#ifndef OPENGLTESTWIDGET_H
#define OPENGLTESTWIDGET_H

#include <QObject>
#include <QDialog>
#include <QOpenGLBuffer>
#include <QLabel>
#include <QMenu>
#include <QCheckBox>
#include <QToolButton>
#include <QToolBar>
#include <QColorDialog>
#include <QDoubleSpinBox>
#include <QOpenGLVertexArrayObject>
#include <QOpenGLShaderProgram>
#include <QComboBox>
#include "molecule.h"
#include "qpeakview.h"
#include <QFileDialog>


#if defined (Q_OS_MAC)
#include <QOpenGLWindow>
class PlayOpenGL: public QOpenGLWindow, protected QOpenGLExtraFunctions { //
#else
#include <QOpenGLWidget>
class PlayOpenGL: public QOpenGLWidget, protected QOpenGLExtraFunctions { //
#endif

Q_OBJECT
  public:
    explicit PlayOpenGL( Molecule *externMole);
    virtual ~PlayOpenGL()  override;
    void setMolecule(Molecule *m);
    int verbosity=5;
    double Ll(){return L;}
    void setMainWindow(QWidget * parent){ chparent=parent; }
    QColor farbverlaufQC(double wrt, double min, double max);
    inline void __RotateCSf( float c, float s, float& X, float& Y );
    inline void __RotateCS( double c, double s, double& X, double& Y );
    void glRotateMat( QMatrix4x4 &mat, const float dang, const float x , const float y, const float z ) ;
    void glRotateMat( Matrix &mat, const float dang, const double x , const double y, const double z );
    void getMVmat(GLdouble mv[16]);
    QMatrix4x4 getMVmat44(){return mvmat;}
    bool canSplitRotate();
    QList<int> lastClicked();
    bool vbyeye(VPoly a,VPoly b);
    int commonFaces(Vert a,Vert b, int f[2]);
    void voronoij(CEnvironment au, int intat);
    void setMapsStuff();
    double getScaleing(){ return L; }
    bool isReady(){ return ready;}
    //void makeBonds(const CEnvironment &atoms, const Connection &bonds, const Connection &lbonds, Connection &hbonds);//, float* atomColors
    void makeBonds(const CEnvironment &atoms, const Connection &bonds, const Connection &lbonds, Connection &hbonds);//, float* atomColors
    void makeDraftBonds(const CEnvironment &atoms, const Connection &bonds, const Connection &lbonds, Connection &hbonds);

    //void makeAtoms(const CEnvironment &atoms);
    void makeAtoms(const CEnvironment &atoms);
    void makeAtomsWOGS(const CEnvironment &atoms);// WOGS with out geometry shader

    //    void makeClassicAtoms(const CEnvironment &atoms);
    void makeSelectedAtoms();
    void makeBackGround();
    void makeEnvi();
    void makeSplitRot();
    void makeSymmetryElements();
    void makeUnitCell();
    void makeContours();
    void contDestruct();
    void voroDestruct();
    void pushRotStat();
    void popRotStat();

    unsigned long long frame_cnt=0;
    bool drawBonds = true;
    bool drawAtoms = true;
    bool drawLabels = true;
    bool isMainWindow = true;
    bool bondsFirst= false;
    bool isOrtho=true;
    bool doNotScale=false;
    QString theIconPath;
    QString GPUInfoString;
    QString GPUInfoString2;
    int ImeanThisAtom;//!< atom index for context menu.
    QMenu *sfacMenu;//!< a sub menu to change the element of a specified atom.
    QMenu *selSfacMenu;
    QList<QVector2D> screenAtoms;
    //CEnvironment theAtoms;
    QList<QImage> altLabels;
     QToolButton *rCenter;
    V3 sumse;
    V3 altemitte;
    int inFocus = -1;
    int rotze = -1;// RotationsZentrum
    double viewAngle = 29.0;
    double previousViewAngle = 29.0;
    float pickradius=256.0f;
    int scrx=0,scry=0;
    int scrx0=0,scry0=0;
    bool rectangle=false;
    double envirange;
    QVector4D rotstat;

    QTimer *idl;//! idle rotation timer
    QTimer *idl_end;//! idle rotation timer


    bool hiddenThings=false;//!< We know that there are hidden objects.
    bool neutrons=false;//!< true if we have neuton scattering factors (negative Fo map is shown if this is true)
    bool atomsClickable=true;
    int SDM_Limit; //!< if the number of atoms in the au exceeds this limit shortest distance is not computed automatically default 350
    bool inRenameMode=false;//!< if true wee are in rename mode
    bool exporting;//!< for exporting bitmaps

    QToolButton *invertMouseZoom=nullptr;//!< inverts the mouse direction that zoom on right click.
    QDoubleSpinBox *pickRadSpin=nullptr;

    bool altpivot=false;
    QComboBox *pivot1cb=nullptr;
    QComboBox *pivot2cb=nullptr;
    QDoubleSpinBox *axx=nullptr;
    QDoubleSpinBox *axy=nullptr;
    QDoubleSpinBox *axz=nullptr;
    QDoubleSpinBox *orgx=nullptr;
    QDoubleSpinBox *orgy=nullptr;
    QDoubleSpinBox *orgz=nullptr;
    QCheckBox *manualAx=nullptr;
    QCheckBox *freeRot=nullptr;
    QLabel *whyNotSplit=nullptr;
    QSpinBox *angle1=nullptr;
    QSpinBox *angle2=nullptr;

    Connection apair;
    bool showPickRad=false;
    QColor labelColor;//!< Color of atom labels.
    QColor selectedLabelColor;
    QColor focusedLabelColor;

    QColor labelBGColor;//!< Color of atom labels.
    QColor selectedLabelBGColor;
    QColor focusedLabelBGColor;
    QColor backGroundColor;//!< Base color of the back groud.
    struct labz{
        int x,y;
        QString l;
        int an;
    };//!< struct labels for exporting bitmaps
    QList<labz> LabelZ;//!< labels for exporting bitmaps


//variables for maps...
    double       lintrans;//!<degree of thansparency for the maps 0-1. 0 is transparent.
    double       linwidth;//!<linewidth of the meshes.
    double       contMin;
    double       contMax;
    QColor       fopc;//!< Fo positive color
    QColor       fomc;//!< Fo negative color
    QColor       dipc;//!< Fo-Fc positive color
    QColor       dimc;//!< Fo-Fc negative color
    QCheckBox   *lighting;//!< enable lighting on the electron density.
    QCheckBox   *niceTrans;//!< draw 6 perspectives for each surface to improve transparent visualization.
    QCheckBox   *fillMap;//!< draw maps in filled style.
    QAction     *foact; //!< QAction toggling Fobs maps.
    QAction     *fofcact;//!< QAction toggling Fobs-Fcalc maps.
    QVector< QVector< float > > fVertexes;
    QVector< QVector< float > > fNormals;

    QList<VPoly> vtriangles;
    float InversionCenterRadius = 0.15;
    float planeTransparence=0.2;
    float glideVectorThickness=0.04;
    float axisThickness=0.04;
    float planeFrameThinkness=0.04;
    double legend_sc = 0.6;
    QAction *h_bonds=nullptr;
    QAction *adp=nullptr;
    QAction *tubes=nullptr;
    QAction *ballStick=nullptr;
    QAction *QPeakLegend=nullptr;
    QAction *hideHydrogen=nullptr;
    QAction *addBond=nullptr;//!< QAction BIND
    QAction *killBond=nullptr;//!< QAction FREE
    QAction *clearSelection=nullptr;//!< QAction deselect
    QAction *centerSelection=nullptr;//!< QAction center selected atoms
    QAction *hideNotSelection=nullptr;//!< QAction hide non selected atoms
    QAction *unhide=nullptr;//!< QAction show all previous hidden objects
    QAction *invhide=nullptr;//!< QAction show all previous hidden and hide visible objects
    QAction *grow=nullptr;//!< QAction grow
    QAction *fuse=nullptr;//!< QAction fuse
    QAction *shpm1=nullptr;//!< QAction show part -N ghost
    QAction *qPeakBonds=nullptr;//!< QAction Q-Peak binds
    QAction *toggleDockWidgets=nullptr;//!< QAction hide text windows
    QAction *delSelAt=nullptr;//!< QAction delete selected atoms
    QAction *highlightParts=nullptr; //!< QAction highlight atoms in parts != 0
    QAction *centroid=nullptr; //!< QAction create a centroid dummy
    QAction *enviNoQ=nullptr;//!< QAction ENVI only for real atoms
    QAction *enviCova=nullptr;//!< QAction ENVI only for covalent contacts
    QAction *invSelection=nullptr;//!< QAction invert selection
    QAction *cntrPlot=nullptr;
    QAction *nocntrPlot=nullptr;
    QAction *rainbowPlot=nullptr;
    QAction *togUnit=nullptr;
    QAction *toggleOrthoView=nullptr;
    QAction *hideBeLo=nullptr;//!< hide/show  BEDE and LONE objects
    QAction *shortLabels=nullptr;//!< Labels without residue number or symmery flag are considered short
    QAction *atomLegend=nullptr;//!< a key legend for atoms used
    QAction *highlightResidues;
    QAction *AMD_Hack_act=nullptr;
    double retinafktr=1.0;
    float qcutoff= -100000.0;
    QFont myFont;//!< there is a Font for Labels of atoms
    QFont nonAtomFont;//!< there is a Font for Labels of Q-Peaks
    bool bggradient = true;
    bool shaderInfo = false;//true;
    Molecule *mol;
    QToolButton *enviButt;//!< Button to clear the envi -ronment.
    QToolButton *enviSelect;//!< Button to select the envi -ronment.
    QList<V3>enviPositions;//!<List of atom postions of an envi-ronment.
    QList<V3> cont;
    QMap<int,GLfloat> contval;
    QList<int>enviKat;//!< List of categorie:qs of contacts in an envi-ronment.
    V3 enviP0;//!< coordinate used for envi.
    QStringList labs;//!< Labels for an envi-ronment list.
    bool burnGPU = false; // true is to measure FPS
    bool AMD_Hack = true;//false;//
    int mk1=0,mk2=0,mk3=0,mk4=0,mk5=0,
        mk6=0,mk7=0,mk8=0,mk9=0,mkA=0,
        mkB=0,mkC=0,mkD=0,mkE=0,mkF=0,mkG=0;

 signals:
    //void message(const QString &);//!< small text messages to the status bar.
    //void qpfoci(double height);//!< when a mouse hovers a Q-Peak the height is emitted to the Q-Peak Legend.

    void no_hw_st();//!< we have no hard ware stereo sorry!
    void message(const QString &);//!< small text messages to the status bar.
    void bigmessage(const QString &);//!< passes HTML strings to the info window.
    void jumpit(int index);//!<An atom has ben left clicked: cursor jump on it.
    void selectionChanged();
    void movedByUser();
    void diffscroll(int numsteps,int diff);//!< to change the iso value of the Fo- or Fo-Fc maps
    void neuemitte(V3 mitte);//!< the rotation center has changet to mitte.
    void inimibas();//!< tells FourXle to initialize the GL display lists.
    void qpfoci(double height);//!< when a mouse hovers a Q-Peak the height is emitted to the Q-Peak Legend.
    void insertDFIX(double value, double esd, QList<MyAtom> selected, QString resiSpec); //!< Insert DFIX restrain to the ins file
    void bindthem();
    void splitable(bool ok);
    void insertDANG(double value, double esd, QList<MyAtom> selected, QString resiSpec); //!< Insert DANG restrain to the ins file
    void insertFLAT(double esd, QList<MyAtom> selected, QString resiSpec); //!< Insert FLAT restrain to the ins file
    void insertEXYZ(QList<MyAtom> selected); //!< Insert EXYZ restrain to the ins file
    void insertEADP(QList<MyAtom> selected); //!< Insert EADP restrain to the ins file
    void insertSADI(int selected); //!< Insert SADI restrain to the ins file
    void insertDELU(double esd1, double esd2, QList<MyAtom> selected, QString resiSpec); //!< Insert DELU restrain to the ins file
    void insertSIMU(double esd1, double st, double dmax, QList<MyAtom> selected, QString resiSpec); //!< Insert SIMU restrain to the ins file
    void insertISOR(double esd1, double esd2, QList<MyAtom> selected, QString resiSpec); //!< Insert ISOR restrain to the ins file
    void insertRIGU(double esd1, double esd2, QList<MyAtom> selected, QString resiSpec); //!< Insert DELU restrain to the ins file
    void insertCHIV(double Vol, double esd1, QList<MyAtom> selected, QString resiSpec); //!< Insert CHIV restrain to the ins file
    void insertANIS(QList<MyAtom> selected); //!< Insert ANIS instruction for selected atoms
    void pivot1Changed(int i);
    void pivot2Changed(int i);
    void initialized();
    void viewAngeleChanged(double);
    void rebuffer(QVector< QVector< float > > &v, QVector< QVector< float > > &n);

public slots:
    void rebuf(){emit rebuffer(fVertexes,fNormals);}
    void addANIS();
    void addCHIV();
    void addDANG();
    void addDELU();
    void addDFIX();
    void addEADP();
    void addEXYZ();
    void addFLAT();
    void addISOR();
    void addRIGU();
    void addSADI();
    void addSIMU();
    void changeEnviRange();
    void makeLegendAtoms();
    void changeBColor();
    void changeTColor();
    double pickRad();
    void setPickRad(double r);
    void changePivot1(int i);
    void changePivot2(int i);
    void hideHydrogens();
    void invertSelection();
    void invertHidden();
    void hidePartMinusOne(bool b){
        //bondsFirst=b;
        mol->nopm1=b;
        update();
    }
    void clearEnvi();
    void envi();
    void lsqpln();
    void growEnvi();
    void showEnvi();
    void expand();
    void getReady(){ready=true;}
    void hideOtherFragments();
    void hideSelected();
    void hideThisFragment();
    void hideNonSelected();
    void selectEnvi();
    void selectThisFragment();
    void selectResiByNr(int nr);
    void setReNaMo(bool b){inRenameMode = b;}
    void showMatrix();
    void saveOrientation();
    void loadOrientation();
    void zoomOut();
    void rotateIdle();
    void fertig();
    void setRotationCenter();
    void setRotationCenter(int rz);
    void setLabelSize(int s);
    void setRotationCenter(V3 center);
    void toogleWithSymmetry(bool b);
    void singleBondDistance();
    void splitRotate();
    void wuff();
    void showHidden();
    void rehide();
    void toggleOrtho(bool b);
    void orthor(){setViewAngle(viewAngle);}
    void makeMaps(QVector< QVector< float > > &v,QVector< QVector< float > > &n);
    void removeAllVAO();
    void highliteQPeak(double co);
    void along001();
    void zoom(double speed);
    void gZoom(double speed);
    void rotZ(double speed=0.01);
    void rotY(double speed=0.01);
    void rotX(double speed=0.01);
    void moveX(double speed);
    void moveY(double speed);
    void makeLabels();
    void renderText(const QVector2D &scpos, const QString &text, int an);    
    void renderText(const QVector2D &scpos, const QString &text, int an, int pts);
    void renderText(const QVector2D &scpos,int i);
    void renderRect();
    void renderText(const V3 &pos, const QString &text, QColor &color, uint bg = Qt::lightGray);
    void renderText(QMatrix4x4 mat, int w, int h, const V3 &pos, const QString &text, QColor &color, uint bg = Qt::lightGray);
    void rotCenter();
    void setAtoms(bool b){drawAtoms = b; update();}
    void setBonds(bool b){drawBonds = b; update();}
    void setBondStrength(double s){ mol->bondStrength = static_cast<float>(s);update();}
    void setLabels(bool b){drawLabels = b; update();}
    void setBGGradient(bool b){bggradient=b; update();}
    void setLOD(int l){
        mol->LOD=l;
        printf("LOD %d\n",mol->LOD);
        if (AMD_Hack) makeAtoms(mol->showatoms);
        update();}
    void homeXY();
    void setViewAngle(double ang);
    void setQpeakCutoff(double cut);
    void toggleAutoUpdate(int b);
    void toggleBondsFirst(int b){
        bondsFirst = (b==Qt::Checked);
        update();
    }
    void updateBondActions(int callLine);
    void updateScreenAtoms();
    void updateAtoms();
    void connectSelection();
    void disConnectSelection(int index);
    void disSelection();
    void reportFPS();
    void geometryShaderAtoms(){
        if (AMD_Hack_act) {
            AMD_Hack=AMD_Hack_act->isChecked();
            makeAtoms(mol->showatoms);
        }
    }

protected:
    void initializeGL() override;
    void resizeGL(int width, int height) override;
    void paintGL() override;
    void mouseMoveEvent(QMouseEvent *event) override;
    void mousePressEvent(QMouseEvent *event) override;
    void wheelEvent(QWheelEvent *event) override;
#if defined (Q_OS_MAC)
    void contextMenuEvent(QMouseEvent *event);
#else
    void contextMenuEvent(QContextMenuEvent *event) override;
#endif
private:
    QWidget *chparent;
    QTimer *tmr = nullptr;
    QMenu *menu = nullptr;
    QMenu *restrMenu = nullptr ;
    QMenu *constrMenu = nullptr;
    QMenu *enviMenu = nullptr;
    bool ready=false;
    int withsymm;

    QList<int> seat;//selected atoms indices for splitrotation
    int pivot;
    int pivot2;
    int ppp=0,pp=0,p=0;//there is no need this should longer GLuint 26.02.2020
    double moux, mouy, L;
    QOpenGLVertexArrayObject bond_vao;
    QOpenGLVertexArrayObject atom_vao;
    QOpenGLVertexArrayObject catom_vao;
    QOpenGLVertexArrayObject latom_vao;
    QOpenGLVertexArrayObject bg_vao;
    QOpenGLVertexArrayObject slct_vao;
    QOpenGLVertexArrayObject map_vao;
    QOpenGLVertexArrayObject voro_vao;
    QOpenGLVertexArrayObject envi_vao;
    QOpenGLVertexArrayObject uc_vao;
    QOpenGLVertexArrayObject cont_vao;
    QOpenGLVertexArrayObject symm_vao0;
    QOpenGLVertexArrayObject symm_vao1;
    QOpenGLVertexArrayObject symm_vao2;
    QOpenGLVertexArrayObject draft_bond_vao;
    QOpenGLVertexArrayObject splitrot_vao;

    QOpenGLBuffer atom_vertexBuffer;
    QOpenGLBuffer atom_indexBuffer;

    QOpenGLBuffer latom_vertexBuffer;
    QOpenGLBuffer satom_vertexBuffer;
    int splitSize = 0;
    int atom_array_size;
    int latom_array_size;
    int axend = 0;
    int invend = 0;
    int planend = 0;
    int glend = 0;
    QOpenGLBuffer backGround_vertexBuffer;
    QOpenGLBuffer bond_vertexBuffer;
    QOpenGLBuffer bond_indexBuffer;
    QOpenGLBuffer map_vertexBuffer;
    QOpenGLBuffer envi_vertexBuffer;
    QOpenGLBuffer voro_vertexBuffer;
    QOpenGLBuffer uc_vertexBuffer;
    QOpenGLBuffer cont_vertexBuffer;
    QOpenGLBuffer symm_vertexBuffer0;
    QOpenGLBuffer symm_vertexBuffer1;
    QOpenGLBuffer symm_vertexBuffer2;
    QOpenGLBuffer splitrot_vertexBuffer;
    //QOpenGLBuffer splitrot_indexBuffer;
    QOpenGLBuffer draft_bond_vertexBuffer;

    QOpenGLShaderProgram      *symm_program0 = nullptr;
    QOpenGLShaderProgram      *symm_program1 = nullptr;
    QOpenGLShaderProgram      *symm_program2 = nullptr;
    QOpenGLShaderProgram   *splitrot_program = nullptr;
    QOpenGLShaderProgram       *cont_program = nullptr;
    QOpenGLShaderProgram         *uc_program = nullptr;
    QOpenGLShaderProgram       *voro_program = nullptr;
    QOpenGLShaderProgram         *bg_program = nullptr;
    QOpenGLShaderProgram        *map_program = nullptr;
    QOpenGLShaderProgram       *bond_program = nullptr;
    QOpenGLShaderProgram       *atom_program = nullptr;// let's nuce the code with this name ;-P
    QOpenGLShaderProgram      *latom_program = nullptr;
    QOpenGLShaderProgram      *satom_program = nullptr;
    QOpenGLShaderProgram       *envi_program = nullptr;
    QOpenGLShaderProgram *draft_bond_program = nullptr;
    QList<int> offsets;
    QMatrix4x4 pmv;
    QMatrix3x3 nm;
    QVector4D eye;
    int nbondx2 = 0;
    int ndrbondx2 = 0;
    int bondStrengthloc;
    int lodloc;
    V3 screenRotx = V3(1,0,0);
    V3 screenRoty = V3(0,1,0);
    V3 SRorg = V3(0,0,0);

    QMatrix4x4 pmat, mvmat;
    Matrix freeRotMat = Matrix(1,0,0 ,0,1,0, 0,0,1);
    double fdx = 0.0;
    double fdy = 0.0;
    GLfloat ang;
    QString glTypeInt2String(int ityp);
    QStringList resis();
    void programInfo(QOpenGLShaderProgram *p, const QString pname);
    void findPivot();
    void paintBonds(const QVector3D org);
    void paintDraftBonds(const QVector3D org);
    void paintAtoms(const QVector3D org);
    void paintLegend();
    void paintMaps(const QVector3D org);
    void paintCell(const V3 org3, const QVector3D org);
    void paintSymmElements(const QVector3D org);
    void paintContours(const QVector3D org);
    void paintSplitrotated(const QVector3D org);
    void paintSelected(const QVector3D org);
    void paintEnvi(const V3 org3, const QVector3D org);
    void paintVoro(const QVector3D org);

    const float fibo=(sqrt(5.0f)+1.0f)*0.5f;
    const float fibom1=fibo-1.0f;
    void pappe(QVector<float> &vert, const QMatrix4x4 &adpmat,         V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi);
    void cube(QVector<float> &vert, const QMatrix4x4 &adpmat,          V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi);
    void icosaeder(QVector<float> &vert, const QMatrix4x4 &adpmat,     V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi);
    void dodecaeder(QVector<float> &vert, const QMatrix4x4 &adpmat,    V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi);
    void triacontaeder(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi);
    void ball(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi);
};

/*
class OpenGLTestWidget : public QDialog {
    Q_OBJECT
public:
    explicit OpenGLTestWidget(Molecule *externMole, QList<QAction*> &acts, QWidget *parent = nullptr);
    PlayOpenGL *ogl;
    QLabel *fpslabel;
    QLabel *messagelabel;
    QPeakView *qpv;
    QToolBar *vito;
    QDoubleSpinBox *fov;
    QWidget *container;
    QSize sizeHint() const override;
    QSize minimumSizeHint() const override;

public slots:
    void printFPS();
    void updateMessage(QString);
    void setViewAngle(double);
signals:

};
*/
#endif // OPENGLTESTWIDGET_H
