#include "shortesdistancematrix.h"

ShortesDistanceMatrix::ShortesDistanceMatrix(Window *parent): QThread{parent}{
    _w = parent;
}

void ShortesDistanceMatrix::finalize(){

    _w->mole.contact.clear();
    _w->mole.sdm.clear();
    _w->mole.contact = contact;
    _w->mole.sdm = sdm;
    _w->mole.packer(brauchSymm);
    _w->mole.showbonds = _w->mole.connecting(_w->mole.showatoms);

    _w->chgl->fuse->setVisible(false);
    _w->chgl->grow->setVisible(true);
    printf("done ShortesDistanceMatrix %lld %lld\n",sdm.size(), _w->mole.knoepfe.size());

}

void ShortesDistanceMatrix::csdm(){
    /*! Calculates the shortest distance matrix.
   * Counts the number of fragmets in the asymmetric unit.
   * Finds possible hydrogen bond contacts.
   * \returns internal symmtry code list for grown structures.
   */
    // George Sheldrick Seminar ideas

    //  QTime zeit; zeit.start();

    double dk,dddd;
    V3 prime,dp,D,floorD;
    contact.clear();
    sdm.clear();
    SdmItem sdmItem;
    sdmItem.a1=0;
    sdmItem.a2=0;
    sdmItem.sn=0;
    SdmItem sdmItem2;
    sdmItem2.a1=0;
    sdmItem2.a2=0;
    sdmItem2.sn=0;

    _w->mole.knoepfe.clear();
    Knopf kn;
    kn.neighbors.clear();
    for (int i=0; i<_w->mole.asymm.size(); i++){
        kn.neighbors.clear();
        for (int j=0; j<_w->mole.asymm.size(); j++ ){
            double min=1000000;
            bool hma=false;
            for (int n=0;n<_w->mole.cell.symmops.size();  n++){
                prime=_w->mole.cell.symmops.at(n) * _w->mole.asymm.at(i).frac + _w->mole.cell.trans.at(n);
                D=prime - _w->mole.asymm.at(j).frac+ V3(0.5,0.5,0.5) ;
                floorD=V3(floor(D.x),floor(D.y),floor(D.z));
                dp=D - floorD - V3(0.5,0.5,0.5);
                dk=_w->mole.fl(dp.x,dp.y,dp.z);
                if (n) dk+=0.0001;
                if ((dk>0.01)&&((min+0.00)>=dk)){
                    min=fmin(dk,min);
                    sdmItem.d=min;
                    sdmItem.floorD=floorD;
                    sdmItem.a1=i;
                    sdmItem.a2=j;
                    sdmItem.sn=n;
                    hma=true;
                }
            }
            if ((_w->mole.asymm[sdmItem.a1].an>-1)&&(_w->mole.asymm[sdmItem.a2].an>-1)&&
                ((_w->mole.asymm[sdmItem.a1].part*_w->mole.asymm[sdmItem.a2].part==0)|| (_w->mole.asymm[sdmItem.a1].part==_w->mole.asymm[sdmItem.a2].part)))
                dddd=(_w->mole.Kovalenz_Radien[_w->mole.asymm[sdmItem.a1].an]+ _w->mole.Kovalenz_Radien[_w->mole.asymm[sdmItem.a2].an])*0.0115;
            else dddd=0;
            if (sdmItem.d<dddd){
                if (hma) kn.neighbors.append(j);
                sdmItem.covalent=true;
            } else {
                sdmItem.covalent=false;
            }
            //if ((sdmItem.sn!=0)&&((_w->mole.asymm[sdmItem.a1].part<0)||(_w->mole.asymm[sdmItem.a2].part<0))) sdmItem.covalent=false;
            if (hma) sdm.append(sdmItem);
        }
        neighborSort(kn);
        _w->mole.knoepfe.append(kn);
    }

    for (int i=0; i<_w->mole.freeatoms.size(); i++){
        MyAtom a1, a2;
        bool bo=false;
        if (_w->mole.freeatoms.at(i).Lab1.contains('_')) a1.resiNr = _w->mole.freeatoms.at(i).Lab1.section('_',1,1).toInt(&bo);
        if (!bo) a1.resiNr=-1;
        if (_w->mole.freeatoms.at(i).Lab2.contains('_'))a2.resiNr = _w->mole.freeatoms.at(i).Lab2.section('_',1,1).toInt(&bo);
        if (!bo)a2.resiNr=-1;
        a1.Label = _w->mole.freeatoms.at(i).Lab1.section('_',0,0);
        a2.Label = _w->mole.freeatoms.at(i).Lab2.section('_',0,0);


        for (int j=0;j<sdm.size();j++){
            if (((a1.resiNr>-1)&& (_w->mole.asymm.at(sdm.at(j).a1).resiNr!=a1.resiNr))||
                ((a2.resiNr>-1)&& (_w->mole.asymm.at(sdm.at(j).a2).resiNr!=a2.resiNr))) continue;
            if (((a1.resiNr>-1)&& (_w->mole.asymm.at(sdm.at(j).a2).resiNr!=a1.resiNr))||
                ((a2.resiNr>-1)&& (_w->mole.asymm.at(sdm.at(j).a1).resiNr!=a2.resiNr))) continue;
            if (((_w->mole.asymm.at(sdm.at(j).a1).Label.section('_',0,0).toUpper()== a1.Label)&&
                 (_w->mole.asymm.at(sdm.at(j).a2).Label.section('_',0,0).toUpper()== a2.Label))||
                ((_w->mole.asymm.at(sdm.at(j).a2).Label.section('_',0,0).toUpper()== a1.Label)&&
                 (_w->mole.asymm.at(sdm.at(j).a1).Label.section('_',0,0).toUpper()== a2.Label))){
                sdm[j].covalent=false;
            }
        }
    }
    for (int i=0; i<_w->mole.bindatoms.size(); i++){
        MyAtom a1, a2;
        bool bo=false;
        bo=false;
        if (_w->mole.bindatoms.at(i).Lab1.contains('_'))a1.resiNr = _w->mole.bindatoms.at(i).Lab1.section('_',1,1).toInt(&bo);
        if (!bo) a1.resiNr=-1;
        if (_w->mole.bindatoms.at(i).Lab2.contains('_'))a2.resiNr = _w->mole.bindatoms.at(i).Lab2.section('_',1,1).toInt(&bo);
        if (!bo)a2.resiNr=-1;
        a1.Label = _w->mole.bindatoms.at(i).Lab1.section('_',0,0);
        a2.Label = _w->mole.bindatoms.at(i).Lab2.section('_',0,0);
        for (int iii=0; iii<_w->mole.asymm.size(); iii++){
            for (int jjj=0; jjj<_w->mole.asymm.size(); jjj++){
                if ((a1.resiNr>-1)&& (_w->mole.asymm.at(iii).resiNr!=a1.resiNr)) continue;
                if ((a2.resiNr>-1)&& (_w->mole.asymm.at(jjj).resiNr!=a2.resiNr)) continue;
                if ((_w->mole.asymm.at(iii).Label.section('_',0,0).section(QString::fromUtf8("»"),0,0).toUpper() == a1.Label.toUpper())&&
                    (_w->mole.asymm.at(jjj).Label.section('_',0,0).section(QString::fromUtf8("»"),0,0).toUpper() == a2.Label.toUpper())){
                    for (int sdmi=0; sdmi < sdm.size(); sdmi++){
                        if ((sdm.at(sdmi).a1==iii)&&(sdm.at(sdmi).a2==jjj)) sdm[sdmi].covalent=true;
                        if ((sdm.at(sdmi).a2==iii)&&(sdm.at(sdmi).a1==jjj)) sdm[sdmi].covalent=true;
                    }
            }
        }
    }
}

for (int i=0; i<_w->mole.asymm.size(); i++){ //wasserstoffbrueckensuche
    for (int j=0; j<_w->mole.asymm.size(); j++ ){
        for (int nn=0;nn<_w->mole.cell.symmops.size();  nn++){
            prime = _w->mole.cell.symmops.at(nn) * _w->mole.asymm.at(i).frac + _w->mole.cell.trans.at(nn);
            D = prime - _w->mole.asymm.at(j).frac+ V3(0.5,0.5,0.5) ;
            floorD = V3(floor(D.x),floor(D.y),floor(D.z));
            dp = D - floorD - V3(0.5,0.5,0.5);
            dk = _w->mole.fl(dp.x,dp.y,dp.z);
            if((1)&&(abs(_w->mole.asymm[i].an-7)<2)&&(abs(_w->mole.asymm[j].an-7)<2)&&(fabs(dk-2.775)<0.325)){//N O F mit abstanden zwischen 2.45 bis 3.1 A koennten H=Bruecken sein
                sdmItem2.d=dk;
                sdmItem2.floorD=floorD;
                sdmItem2.a1=i;
                sdmItem2.a2=j;
                sdmItem2.sn=nn;
                sdmItem2.covalent=true;
                double ang1=361,ang2=361,an;
                for (int oo=0; oo<_w->mole.knoepfe.at(i).neighbors.size(); oo++){
                    an=_w->mole.wl(_w->mole.asymm.at(j).frac,
                            _w->mole.cell.symmops.at(nn) * _w->mole.asymm.at(i).frac + _w->mole.cell.trans.at(nn)-floorD,
                            _w->mole.cell.symmops.at(nn) * _w->mole.asymm.at(_w->mole.knoepfe.at(i).neighbors.at(oo)).frac + _w->mole.cell.trans.at(nn)-floorD);
                    ang1=(an<ang1)?an:ang1;
                }
                for (int oo=0; oo<_w->mole.knoepfe.at(j).neighbors.size(); oo++){
                    an=_w->mole.wl(
                        _w->mole.cell.symmops.at(nn) * _w->mole.asymm.at(i).frac  + _w->mole.cell.trans.at(nn)-floorD,
                        _w->mole.asymm.at(j).frac,
                        _w->mole.asymm.at(_w->mole.knoepfe.at(j).neighbors.at(oo)).frac);
                    ang2=(an<ang2)?an:ang2;
                }
                /*printf("Paarschippen: %s %s %f %f %f %d\n",
            _w->mole.asymm.at(i).Label.toStdString().c_str(),
            _w->mole.asymm.at(j).Label.toStdString().c_str(),
            dk, ang1, ang2,((qMax(ang1,ang2)!=361)&&((qMin(ang1,ang2)<85)||(qMax(ang1,ang2)>180))));// */
                if ((qMax(ang1,ang2)!=361)&&((qMin(ang1,ang2)<85)||(qMax(ang1,ang2)>180))) sdmItem2.covalent=false;
                else {

                    contact.append(sdmItem2);
                }
            }
        }
    }
}

QString bs;
brauchSymm.clear();
int someleft=0,nextmol=0;
int maxmol=1;
for (int i=0; i<_w->mole.asymm.size();i++) _w->mole.asymm[i].molindex=-1;
_w->mole.asymm[0].molindex=1;//starter;
do {
    nextmol=0;
    do {
        someleft=0;
        for  (int i=0; i<sdm.size();i++){
            if ((sdm.at(i).covalent)&&(_w->mole.asymm[sdm.at(i).a1].molindex*_w->mole.asymm[sdm.at(i).a2].molindex<0)) {
                _w->mole.asymm[sdm.at(i).a1].molindex=maxmol;
                _w->mole.asymm[sdm.at(i).a2].molindex=maxmol;
                someleft++;
            }
        }
    }while (someleft);
    for ( int i=0; i<_w->mole.asymm.size();i++) {
        if ((_w->mole.asymm[i].an>-1)&&(_w->mole.asymm[i].molindex<0)) {nextmol=i;break;}
    }
    if (nextmol) {
        _w->mole.asymm[nextmol].molindex=(++maxmol);
    }
}while (nextmol);
_w->mole.Fragments = QString("<b>The _w->mole.asymmetric unit contains %1 fragments.</b><br>").arg(maxmol);
_w->mole.setMaxMols(maxmol);
for (int k =0; k<sdm.size();k++){
    if (sdm.at(k).covalent){
        if (_w->mole.asymm[sdm.at(k).a1].molindex<1) continue;
        if (_w->mole.asymm[sdm.at(k).a1].molindex>6) continue;
        for (int n=0;n<_w->mole.cell.symmops.size();  n++){
            if (((_w->mole.asymm[sdm.at(k).a1].part!=0)&&(_w->mole.asymm[sdm.at(k).a2].part!=0)&&(_w->mole.asymm[sdm.at(k).a1].part!=_w->mole.asymm[sdm.at(k).a2].part)))continue;
            if ((_w->mole.asymm[sdm.at(k).a1].an==_w->mole.asymm[sdm.at(k).a2].an)&&(_w->mole.asymm[sdm.at(k).a1].an==0)) continue;
            prime = _w->mole.cell.symmops.at(n) * _w->mole.asymm[sdm.at(k).a1].frac + _w->mole.cell.trans.at(n);
            D = prime - _w->mole.asymm[sdm.at(k).a2].frac+ V3(0.5,0.5,0.5) ;
            floorD = V3(floor(D.x),floor(D.y),floor(D.z));
            dp = D - floorD - V3(0.5,0.5,0.5);
            if ((n==0)&&(V3(0,0,0)==floorD)) continue;
            dk = _w->mole.fl(dp.x,dp.y,dp.z);
            dddd=(sdm.at(k).d+0.2);
            if ((dk>0.001)&&(dddd>=dk)) {
                bs=QString("%1_%2%3%4:%5,").arg(n+1).arg(5-(int)floorD.x).arg(5-(int)floorD.y).arg(5-(int)floorD.z).arg(_w->mole.asymm[sdm.at(k).a1].molindex);
                if  ((!brauchSymm.contains(bs))) {
                    brauchSymm.append(bs);

                }
            }
        }
    }
}

std::sort(sdm.begin(),sdm.end());
enviSDM(2.5);


QList<int> flags;
for (int i=0; i<_w->mole.asymm.size(); i++) flags.append((_w->mole.asymm.at(i).an<0)?-1:1);
for (int i=0; i<envi_sdm.size(); i++){
    if ((flags.at(envi_sdm.at(i).a1)*flags.at(envi_sdm.at(i).a2))==-1) {
        //if ((qbeforehkl)||((flags.at(sdm.at(i).a1)==-1)&&((_w->mole.asymm.at(sdm.at(i).a2).an>-1)))){
        //   if (((_w->mole.asymm[envi_sdm.at(i).a1].Label=="Q11")||(_w->mole.asymm[envi_sdm.at(i).a2].Label=="Q11"))&&(envi_sdm.at(i).d<2.6))
        //       qDebug()<<_w->mole.asymm[envi_sdm.at(i).a1].Label<<_w->mole.asymm[envi_sdm.at(i).a2].Label<<envi_sdm.at(i).d<<flags[envi_sdm.at(i).a1]<<flags[envi_sdm.at(i).a2];
        if (_w->mole.asymm[envi_sdm.at(i).a1].an > -1) continue;
        if (_w->mole.asymm[envi_sdm.at(i).a1].an == -66) continue;
        if ((envi_sdm.at(i).sn == 0)&&(envi_sdm.at(i).floorD == V3(0,0,0))) {flags[envi_sdm.at(i).a1]=1;continue;}
        if (envi_sdm.at(i).d>2.4) continue;
        _w->mole.asymm[envi_sdm.at(i).a1].frac = _w->mole.cell.symmops.at(envi_sdm.at(i).sn) * _w->mole.asymm[envi_sdm.at(i).a1].frac
                                                 + _w->mole.cell.trans.at(envi_sdm.at(i).sn) - envi_sdm.at(i).floorD;
        flags[envi_sdm.at(i).a1]=1;
        _w->mole.frac2kart(_w->mole.asymm[envi_sdm.at(i).a1].frac,_w->mole.asymm[envi_sdm.at(i).a1].pos);
        //}
    }
}


}

void ShortesDistanceMatrix::neighborSort(Knopf &kn){//sort neighbors by atomic number
// I think bublle sort is fine for such a sort list
for (int i = 0; i < kn.neighbors.size(); i++){
    for (int j = i+1; j < kn.neighbors.size(); j++){
        if ((kn.neighbors.at(i) >= _w->mole.asymm.size())||(kn.neighbors.at(j) >= _w->mole.asymm.size())) continue;
        if (_w->mole.asymm.at(kn.neighbors.at(i)).an < _w->mole.asymm.at(kn.neighbors.at(j)).an){
            int swap = kn.neighbors.at(i);
            kn.neighbors[i] = kn.neighbors.at(j);
            kn.neighbors[j] = swap;
        }
    }
}
}

void ShortesDistanceMatrix::enviSDM(double range){
/*! Calculates the shortes distance matrix for the ENVIronment functionality in the given range.
   * @param range maximal distances around each atom to generate a matrix entry.
   */

// George Sheldrick Seminar ideas
double dk,dddd;
V3 prime,dp,D,floorD;
envi_sdm.clear();
SdmItem sdmItem;
sdmItem.a1=0;
sdmItem.a2=1;
sdmItem.sn=0;
//sdmItem.p1=sdmItem.p2=V3(0,0,0);
for (int i=0; i<_w->mole.asymm.size(); i++){
    for (int j=0; j<_w->mole.asymm.size(); j++ ){
        //  bool hma=false;
        for (int n=0; n < _w->mole.cell.symmops.size();  n++){
            prime = _w->mole.cell.symmops.at(n) * _w->mole.asymm.at(i).frac + _w->mole.cell.trans.at(n);
            D = prime - _w->mole.asymm.at(j).frac+ V3(0.5,0.5,0.5) ;
            floorD = V3(floor(D.x),floor(D.y),floor(D.z));
            for (int h=-1; h<2; h++){
                for (int k=-1; k<2; k++){
                    for (int l=-1; l<2; l++){
                        V3 fD=floorD+V3(h,k,l);
                        dp = D - fD - V3(0.5,0.5,0.5);
                        dk = _w->mole.fl(dp.x,dp.y,dp.z);
                        if ((dk>0.01)&&((range)>=dk)){
                            sdmItem.d=dk;
                            sdmItem.floorD=fD;
                            sdmItem.a1=i;
                            sdmItem.a2=j;
                            sdmItem.sn=n;
                            if ((_w->mole.asymm[i].an>-1)&&(_w->mole.asymm[j].an>-1)&&
                                ((_w->mole.asymm[i].part*_w->mole.asymm[j].part==0)||
                                 (_w->mole.asymm[i].part==_w->mole.asymm[j].part)))
                                dddd=(_w->mole.Kovalenz_Radien[_w->mole.asymm[i].an]+ _w->mole.Kovalenz_Radien[_w->mole.asymm[j].an])*0.012;
                            else dddd=0;
                            if (sdmItem.d<dddd){
                                sdmItem.covalent=true;
                            }else sdmItem.covalent=false;
                            envi_sdm.append(sdmItem);
                        }
                    }
                }
            }
        }
    }
}
std::sort(envi_sdm.begin(),envi_sdm.end());

}
