#include "opengltestwidget.h"
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QDoubleSpinBox>
#include <QSpinBox>
#include <QLineEdit>
#include <QComboBox>
#include <QDialogButtonBox>
#include <QPushButton>
#include <QToolButton>
#include <QApplication>
#include <QInputDialog>
#include <QThread>

///////////////////////////////////////////////////////////////////////
#if defined (Q_OS_MAC)
PlayOpenGL::PlayOpenGL( Molecule *externMole): QOpenGLWindow(){
#else
PlayOpenGL::PlayOpenGL( Molecule *externMole): QOpenGLWidget(){
#endif
    mol = externMole;

    connect(this,SIGNAL(initialized()),this, SLOT(getReady()));
    //QSurfaceFormat glFormat(QSurfaceFormat::DebugContext);
    //myFont = QFont("Arial", 24);
    //nonAtomFont = myFont;
    if (parent()){
        isMainWindow = (parent()->objectName()=="The Main Window");
    }
    //setMouseTracking(true);
    nonAtomFont.setPointSize(qMax(((myFont.pointSize()*2)/3),7));
    QSurfaceFormat glFormat;

    glFormat.setProfile( QSurfaceFormat::CoreProfile ); // Requires >=Qt-4.8.0
    //glFormat.setProfile( QSurfaceFormat::CompatibilityProfile);
//    glFormat.setSwapInterval(0);
#if defined (Q_OS_MAC)
    glFormat.setVersion( 4, 1 );
#else
    glFormat.setVersion( 3, 3 );
#endif
    glFormat.setSamples(4);
    glFormat.setDepthBufferSize(24);
//    glFormat.setSwapBehavior(QSurfaceFormat::TripleBuffer);
    //QSurfaceFormat::setDefaultFormat(glFormat);
    setFormat(glFormat);


    if (externMole->mynewnameis.startsWith("shelXle")){
        theIconPath=QString(":/xle-icons/");
    }else{
        theIconPath=QString(":/bicons/");
    }


    envirange = 3.2;
    //apair.clear();
    enviNoQ = new QAction("Exclude Q-peaks from environment listing", this);
    enviCova = new QAction("List only covalent bonds.",this);

    enviButt = new QToolButton();
    enviButt->setText("clear ENVI");
    //enviButt->setIcon(QIcon(theIconPath+"cancel.svg"));
    enviSelect = new QToolButton();
    enviSelect->setText("Select ENVI");
    enviButt->hide();
    enviSelect->hide();
    //enviSelect->setIcon(QIcon(theIconPath+"cancel.svg"));
    connect(enviButt,SIGNAL(clicked()),this,SLOT(clearEnvi()));
    connect(enviSelect,SIGNAL(clicked()),this,SLOT(selectEnvi()));
    /*wireButt = new QAction("Wiremodel while rotate",this);
    wireButt->setCheckable(true);
    wireButt->setChecked(false);//defaut changed 13.12.2019*/
    rainbowPlot = new QAction("Contour plot in rainbow colors",this);
    rainbowPlot->setCheckable(true);
    rainbowPlot->setChecked(false);
    rainbowPlot->setVisible(false);
    enviNoQ->setCheckable(true);
    enviNoQ->setChecked(false);
    enviCova->setCheckable(true);
    enviCova->setChecked(false);
    //qDebug()<<contextMenuPolicy()<<"contextMenuPolicy" ;

}
PlayOpenGL::~PlayOpenGL(){
    makeCurrent();
    removeAllVAO();


}



void PlayOpenGL::setMolecule(Molecule *m){
    mol = m;
    verbosity = mol->verbosity;


    if (mol->mynewnameis.startsWith("shelXle")){
        theIconPath=QString(":/xle-icons/");
    }else{
        theIconPath=QString(":/bicons/");
    }
    if (mol!=m) {
      if (mol) delete mol;
      mol=m;
    }
    double dim=mol->dimension();
    L=100.0/dim;

}
//QSize PlayOpenGL::sizeHint() const {return QSize(800,800);}
//QSize PlayOpenGL::minimumSizeHint() const {return QSize(60,60);}

void PlayOpenGL::lsqpln(){
    int na=0;
    double xx = 0.0;
    double xy = 0.0;
    double xz = 0.0;
    double yy = 0.0;
    double yz = 0.0;
    double x  = 0.0;
    double y  = 0.0;
    double z  = 0.0;
    QList<V3> k;
    V3 oo=V3(0,0,0);
    for (int i=0; i<mol->showatoms.size(); i++){
      if (mol->showatoms.at(i).an<0) continue;
      oo+=mol->showatoms.at(i).pos;
      na++;
    }
    oo*=1.0/na;
    na=0;
    for (int i=0; i<mol->showatoms.size(); i++){
      if (mol->showatoms.at(i).an<0) continue;
      k.append(mol->showatoms.at(i).pos);
      x  += mol->showatoms.at(i).pos.x - oo.x;
      y  += mol->showatoms.at(i).pos.y - oo.y;
      z  += mol->showatoms.at(i).pos.z - oo.z;
      xx += (mol->showatoms.at(i).pos.x - oo.x) * (mol->showatoms.at(i).pos.x - oo.x);
      xy += (mol->showatoms.at(i).pos.x - oo.x) * (mol->showatoms.at(i).pos.y - oo.y);
      xz += (mol->showatoms.at(i).pos.x - oo.x) * (mol->showatoms.at(i).pos.z - oo.z);
      yy += (mol->showatoms.at(i).pos.y - oo.y) * (mol->showatoms.at(i).pos.y - oo.y);
      yz += (mol->showatoms.at(i).pos.y - oo.y) * (mol->showatoms.at(i).pos.z - oo.z);
      na++;
    }

    Matrix m(xx-xz, xy,    x,
             xy,    yy-yz, y,
             x,     y,     na-z);
    printf("|%g %g %g|\n|%g %g %g|\n|%g %g %g|\n",m.m11,m.m12,m.m13,m.m21,m.m22,m.m23,m.m31,m.m32,m.m33);
    V3 ev;
    double *ro = mol->jacobi(m,ev);
    printf("%f %f %f %f ev (%f %f %f) na=%d %g\n", ro[0], ro[1], ro[2], ro[3], ev.x/na, ev.y/na, ev.z/na, na, Norm(V3(ro[1], ro[2], ro[3])));
    QMatrix4x4 rmatrix;
    rmatrix.setToIdentity();
    glRotateMat(mvmat, ro[0], ro[1], ro[2], ro[3]);
    glRotateMat(rmatrix, ro[0], ro[1], ro[2], ro[3]);
    Matrix rMatrix(rmatrix(0, 0), rmatrix(0, 1), rmatrix(0, 2),
                   rmatrix(1, 0), rmatrix(1, 1), rmatrix(1, 2),
                   rmatrix(2, 0), rmatrix(2, 1), rmatrix(2, 2));
    qDebug()<<rmatrix<<determinant(rMatrix);
    xx = 0.0;
    xy = 0.0;
    xz = 0.0;
    yy = 0.0;
    yz = 0.0;
    x  = 0.0;
    y  = 0.0;
    z  = 0.0;
    printf("ooo %8.4f%8.4f%8.4f ooo\n", oo.x, oo.y, oo.z);
    for (int i=0; i<na; i++){
      V3 bb = (k.at(i)-oo);
      printf("%9.3f%9.3f%9.3f  ", bb.x, bb.y, bb.z);
      k[i] = (bb * rMatrix) ;
      printf("|%9.3f%9.3f%9.3f  \n",k.at(i).x,k.at(i).y,k.at(i).z);

      x  += k.at(i).x;
      y  += k.at(i).y;
      z  += k.at(i).z;
      xx += k.at(i).x * k.at(i).x;
      xy += k.at(i).x * k.at(i).y;
      xz += k.at(i).x * k.at(i).z;
      yy += k.at(i).y * k.at(i).y;
      yz += k.at(i).y * k.at(i).z;
    }
    Matrix m1(xx-xz, xy,    x,
             xy,    yy-yz, y,
             x,     y,     na-z);
    ro = mol->jacobi(m1, ev);
    printf("!! %f %f %f %f ev (%f %f %f) na=%d %g\n", ro[0], ro[1], ro[2], ro[3], ev.x/na, ev.y/na, ev.z/na, na, Norm(V3(ro[1], ro[2], ro[3])));

    pmv = pmat*mvmat;
    nm = mvmat.normalMatrix();
    eye = mvmat.inverted()*QVector4D(0,0,0,1);
}


int PlayOpenGL::commonFaces(Vert a,Vert b, int f[2]){
  int matches=0;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      if (a.faces[i]==b.faces[j])f[matches++]=a.faces[i];
  return matches;
}

void PlayOpenGL::pappe(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi){

  const int HAS_PLAID = 128;
  const GLfloat v[108]={
      -1.f,-1.f, 0.f,
      1.f,-1.f, 0.f,
      -1.f, 1.f, 0.f,

      -1.f, 1.f, 0.f,
      1.f,-1.f, 0.f,
      1.f, 1.f, 0.f,

      -1.f, 0.f,-1.f,
      1.f, 0.f,-1.f,
      -1.f, 0.f, 1.f,

      -1.f, 0.f, 1.f,
      1.f, 0.f,-1.f,
      1.f, 0.f, 1.f,

      0.f,-1.f,-1.f,
      0.f, 1.f,-1.f,
      0.f,-1.f, 1.f,

      0.f,-1.f, 1.f,
      0.f, 1.f,-1.f,
      0.f, 1.f, 1.f,

      1.f,-1.f, 0.f,
      -1.f,-1.f, 0.f,
      -1.f, 1.f, 0.f,

      1.f,-1.f, 0.f,
      -1.f, 1.f, 0.f,
      1.f, 1.f, 0.f,

      1.f, 0.f,-1.f,
      -1.f, 0.f,-1.f,
      -1.f, 0.f, 1.f,

      1.f, 0.f,-1.f,
      -1.f, 0.f, 1.f,
      1.f, 0.f, 1.f,

      0.f, 1.f,-1.f,
      0.f,-1.f,-1.f,
      0.f,-1.f, 1.f,

      0.f, 1.f,-1.f,
      0.f,-1.f, 1.f,
      0.f, 1.f, 1.f
  };
  const GLfloat n[108]={
      0.f, 0.f, 1.f,
      0.f, 0.f, 1.f,
      0.f, 0.f, 1.f,

      0.f, 0.f, 1.f,
      0.f, 0.f, 1.f,
      0.f, 0.f, 1.f,

      0.f,-1.f, 0.f,
      0.f,-1.f, 0.f,
      0.f,-1.f, 0.f,

      0.f,-1.f, 0.f,
      0.f,-1.f, 0.f,
      0.f,-1.f, 0.f,

      1.f, 0.f, 0.f,
      1.f, 0.f, 0.f,
      1.f, 0.f, 0.f,

      1.f, 0.f, 0.f,
      1.f, 0.f, 0.f,
      1.f, 0.f, 0.f,


      0.f, 0.f,-1.f,
      0.f, 0.f,-1.f,
      0.f, 0.f,-1.f,

      0.f, 0.f,-1.f,
      0.f, 0.f,-1.f,
      0.f, 0.f,-1.f,

      0.f, 1.f, 0.f,
      0.f, 1.f, 0.f,
      0.f, 1.f, 0.f,

      0.f, 1.f, 0.f,
      0.f, 1.f, 0.f,
      0.f, 1.f, 0.f,

      -1.f, 0.f, 0.f,
      -1.f, 0.f, 0.f,
      -1.f, 0.f, 0.f,

      -1.f, 0.f, 0.f,
      -1.f, 0.f, 0.f,
      -1.f, 0.f, 0.f
  };

  for (int i=0; i<36; i++){
    QVector4D qv(v[3*i],v[3*i+1],v[3*i+2],1);
    QVector4D qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);
    vert.append(n[3*i]);
    vert.append(n[3*i+1]);
    vert.append(n[3*i+2]);
    vert.append(style|HAS_PLAID);
    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);
    vert.append(_prt);
    vert.append(qv.x());
    vert.append(qv.y());
    vert.append(qv.z());
  }
}

void PlayOpenGL::cube(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi){
  static const GLfloat cubv[108]={
                                    -fibom1, -fibo,  0.000000f, -fibo,  0.000000f, -fibom1,  0.000000f,  fibom1, -fibo,
                                    -fibom1, -fibo,  0.000000f,  0.000000f,  fibom1, -fibo,  1.000000f, -1.000000f, -1.000000f,
                                    -1.000000f,  1.000000f,  1.000000f,  fibom1,  fibo,  0.000000f,  0.000000f,  fibom1, -fibo,
                                    -1.000000f,  1.000000f,  1.000000f,  0.000000f,  fibom1, -fibo, -fibo,  0.000000f, -fibom1,
                                    -1.000000f,  1.000000f,  1.000000f, -fibom1, -fibo,  0.000000f,  0.000000f, -fibom1,  fibo,
                                    -1.000000f,  1.000000f,  1.000000f, -fibo,  0.000000f, -fibom1, -fibom1, -fibo,  0.000000f,
                                    -1.000000f,  1.000000f,  1.000000f,  fibo,  0.000000f,  fibom1,  fibom1,  fibo,  0.000000f,
                                    -1.000000f,  1.000000f,  1.000000f,  0.000000f, -fibom1,  fibo,  fibo,  0.000000f,  fibom1,
                                    0.000000f, -fibom1,  fibo,  1.000000f, -1.000000f, -1.000000f,  fibo,  0.000000f,  fibom1,
                                    0.000000f, -fibom1,  fibo, -fibom1, -fibo,  0.000000f,  1.000000f, -1.000000f, -1.000000f,
                                    fibo,  0.000000f,  fibom1,  0.000000f,  fibom1, -fibo,  fibom1,  fibo,  0.000000f,
                                    fibo,  0.000000f,  fibom1,  1.000000f, -1.000000f, -1.000000f,  0.000000f,  fibom1, -fibo};
  static const GLfloat cubn[108]={
                                    -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f,
                                    -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f,
                                    -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f,
                                    -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f,
                                    -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f,
                                    -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f,
                                    0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,
                                    0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,
                                    0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,
                                    0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,
                                    0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,
                                    0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f};
  //long long int osize=vert.size();
  for (int i=0; i<36; i++){
    QVector4D qv(cubv[3*i], cubv[3*i+1], cubv[3*i+2],1);
    QVector4D qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);
    vert.append(cubn[i*3]);
    vert.append(cubn[i*3+1]);
    vert.append(cubn[i*3+2]);
    vert.append(style);
    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);
    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(0.0f);
  }
  //printf("cube %lld\n", vert.size() - osize);
}

void PlayOpenGL::icosaeder(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi){
  static const GLfloat icov[180]={
                                    1.0f,  0.0f,  fibo,  0.0f,  fibo,  1.0f, -1.0f,  0.0f,  fibo,
                                    0.0f, -fibo,  1.0f,  1.0f,  0.0f,  fibo, -1.0f,  0.0f,  fibo,
                                    1.0f,  0.0f,  fibo,  fibo,  1.0f,  0.0f,  0.0f,  fibo,  1.0f,
                                    1.0f,  0.0f,  fibo,  0.0f, -fibo,  1.0f,  fibo, -1.0f,  0.0f,
                                    1.0f,  0.0f,  fibo,  fibo, -1.0f,  0.0f,  fibo,  1.0f,  0.0f,
                                   -1.0f,  0.0f,  fibo,  0.0f,  fibo,  1.0f, -fibo,  1.0f,  0.0f,
                                   -1.0f,  0.0f,  fibo, -fibo, -1.0f,  0.0f,  0.0f, -fibo,  1.0f,
                                   -1.0f,  0.0f,  fibo, -fibo,  1.0f,  0.0f, -fibo, -1.0f,  0.0f,
                                    0.0f,  fibo,  1.0f,  fibo,  1.0f,  0.0f,  0.0f,  fibo, -1.0f,
                                    0.0f,  fibo,  1.0f,  0.0f,  fibo, -1.0f, -fibo,  1.0f,  0.0f,
                                    0.0f, -fibo,  1.0f, -fibo, -1.0f,  0.0f,  0.0f, -fibo, -1.0f,
                                    0.0f, -fibo,  1.0f,  0.0f, -fibo, -1.0f,  fibo, -1.0f,  0.0f,
                                    fibo,  1.0f,  0.0f,  fibo, -1.0f,  0.0f,  1.0f,  0.0f, -fibo,
                                    fibo,  1.0f,  0.0f,  1.0f,  0.0f, -fibo,  0.0f,  fibo, -1.0f,
                                   -fibo,  1.0f,  0.0f, -1.0f,  0.0f, -fibo, -fibo, -1.0f,  0.0f,
                                   -fibo,  1.0f,  0.0f,  0.0f,  fibo, -1.0f, -1.0f,  0.0f, -fibo,
                                   -fibo, -1.0f,  0.0f, -1.0f,  0.0f, -fibo,  0.0f, -fibo, -1.0f,
                                    fibo, -1.0f,  0.0f,  0.0f, -fibo, -1.0f,  1.0f,  0.0f, -fibo,
                                    0.0f,  fibo, -1.0f,  1.0f,  0.0f, -fibo, -1.0f,  0.0f, -fibo,
                                    0.0f, -fibo, -1.0f, -1.0f,  0.0f, -fibo,  1.0f,  0.0f, -fibo};
  static const GLfloat icon[180]={
                                    0.000000f,  0.356822f,  0.934172f, -0.000000f,  0.356822f,  0.934172f, -0.000000f,  0.356822f,  0.934172f,
                                    0.000000f, -0.356822f,  0.934172f, -0.000000f, -0.356822f,  0.934172f, -0.000000f, -0.356822f,  0.934172f,
                                    0.577350f,  0.577350f,  0.577350f,  0.577350f,  0.577350f,  0.577350f,  0.577350f,  0.577350f,  0.577350f,
                                    0.577350f, -0.577350f,  0.577350f,  0.577350f, -0.577350f,  0.577350f,  0.577350f, -0.577350f,  0.577350f,
                                    0.934172f,  0.000000f,  0.356822f,  0.934172f, -0.000000f,  0.356822f,  0.934172f, -0.000000f,  0.356822f,
                                    -0.577350f,  0.577350f,  0.577350f, -0.577350f,  0.577350f,  0.577350f, -0.577350f,  0.577350f,  0.577350f,
                                    -0.577350f, -0.577350f,  0.577350f, -0.577350f, -0.577350f,  0.577350f, -0.577350f, -0.577350f,  0.577350f,
                                    -0.934172f,  0.000000f,  0.356822f, -0.934172f, -0.000000f,  0.356822f, -0.934172f,  0.000000f,  0.356822f,
                                    0.356822f,  0.934172f, -0.000000f,  0.356822f,  0.934172f, -0.000000f,  0.356822f,  0.934172f,  0.000000f,
                                    -0.356822f,  0.934172f, -0.000000f, -0.356822f,  0.934172f, -0.000000f, -0.356822f,  0.934172f, 0.000000f,
                                    -0.356822f, -0.934172f, -0.000000f, -0.356822f, -0.934172f, -0.000000f, -0.356822f, -0.934172f, 0.000000f,
                                    0.356822f, -0.934172f, -0.000000f,  0.356822f, -0.934172f, -0.000000f,  0.356822f, -0.934172f,  0.000000f,
                                    0.934172f,  0.000000f, -0.356822f,  0.934172f, -0.000000f, -0.356822f,  0.934172f,  0.000000f, -0.356822f,
                                    0.577350f,  0.577350f, -0.577350f,  0.577350f,  0.577350f, -0.577350f,  0.577350f,  0.577350f, -0.577350f,
                                    -0.934172f,  0.000000f, -0.356822f, -0.934172f,  0.000000f, -0.356822f, -0.934172f,  0.000000f, -0.356822f,
                                    -0.577350f,  0.577350f, -0.577350f, -0.577350f,  0.577350f, -0.577350f, -0.577350f,  0.577350f, -0.577350f,
                                    -0.577350f, -0.577350f, -0.577350f, -0.577350f, -0.577350f, -0.577350f, -0.577350f, -0.577350f, -0.577350f,
                                    0.577350f, -0.577350f, -0.577350f,  0.577350f, -0.577350f, -0.577350f,  0.577350f, -0.577350f, -0.577350f,
                                    0.000000f,  0.356822f, -0.934172f,  0.000000f,  0.356822f, -0.934172f,  0.000000f,  0.356822f, -0.934172f,
                                    0.000000f, -0.356822f, -0.934172f,  0.000000f, -0.356822f, -0.934172f,  0.000000f, -0.356822f, -0.934172f};

  for (int i=0; i<60; i++){
    QVector4D qv(icov[3*i], icov[3*i+1], icov[3*i+2], 1);
    QVector4D qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);
    vert.append(icon[3*i]);
    vert.append(icon[3*i+1]);
    vert.append(icon[3*i+2]);
    vert.append(style);
    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);
    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(0.0f);
  }

}

void PlayOpenGL::dodecaeder(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi){
  static const GLfloat dodev[324]={
                                     0.000000f, -fibom1,  fibo,  1.000000f, -1.000000f,  1.000000f,  fibo,  0.000000f,  fibom1,
                                     0.000000f, -fibom1,  fibo,  fibo,  0.000000f,  fibom1,  1.000000f,  1.000000f,  1.000000f,
                                     0.000000f, -fibom1,  fibo,  1.000000f,  1.000000f,  1.000000f,  0.000000f,  fibom1,  fibo,
                                     0.000000f,  fibom1,  fibo, -1.000000f,  1.000000f,  1.000000f, -fibo,  0.000000f,  fibom1,
                                     0.000000f,  fibom1,  fibo, -fibo,  0.000000f,  fibom1, -1.000000f, -1.000000f,  1.000000f,
                                     0.000000f,  fibom1,  fibo, -1.000000f, -1.000000f,  1.000000f,  0.000000f, -fibom1,  fibo,
                                     0.000000f,  fibom1,  fibo,  1.000000f,  1.000000f,  1.000000f,  fibom1,  fibo,  0.000000f,
                                     0.000000f,  fibom1,  fibo,  fibom1,  fibo,  0.000000f, -fibom1,  fibo,  0.000000f,
                                     0.000000f,  fibom1,  fibo, -fibom1,  fibo,  0.000000f, -1.000000f,  1.000000f,  1.000000f,
                                     1.000000f,  1.000000f,  1.000000f,  fibo,  0.000000f,  fibom1,  fibo,  0.000000f, -fibom1,
                                     1.000000f,  1.000000f,  1.000000f,  fibo,  0.000000f, -fibom1,  1.000000f,  1.000000f, -1.000000f,
                                     1.000000f,  1.000000f,  1.000000f,  1.000000f,  1.000000f, -1.000000f,  fibom1,  fibo,  0.000000f,
                                     0.000000f, -fibom1,  fibo, -1.000000f, -1.000000f,  1.000000f, -fibom1, -fibo,  0.000000f,
                                     0.000000f, -fibom1,  fibo, -fibom1, -fibo,  0.000000f,  fibom1, -fibo,  0.000000f,
                                     0.000000f, -fibom1,  fibo,  fibom1, -fibo,  0.000000f,  1.000000f, -1.000000f,  1.000000f,
                                     -1.000000f,  1.000000f,  1.000000f, -fibom1,  fibo,  0.000000f, -1.000000f,  1.000000f, -1.000000f,
                                     -1.000000f,  1.000000f,  1.000000f, -1.000000f,  1.000000f, -1.000000f, -fibo,  0.000000f, -fibom1,
                                     -1.000000f,  1.000000f,  1.000000f, -fibo,  0.000000f, -fibom1, -fibo,  0.000000f,  fibom1,
                                     -1.000000f, -1.000000f,  1.000000f, -fibo,  0.000000f,  fibom1, -fibo,  0.000000f, -fibom1,
                                     -1.000000f, -1.000000f,  1.000000f, -fibo,  0.000000f, -fibom1, -1.000000f, -1.000000f, -1.000000f,
                                     -1.000000f, -1.000000f,  1.000000f, -1.000000f, -1.000000f, -1.000000f, -fibom1, -fibo,  0.000000f,
                                     fibo,  0.000000f,  fibom1,  1.000000f, -1.000000f,  1.000000f,  fibom1, -fibo,  0.000000f,
                                     fibo,  0.000000f,  fibom1,  fibom1, -fibo,  0.000000f,  1.000000f, -1.000000f, -1.000000f,
                                     fibo,  0.000000f,  fibom1,  1.000000f, -1.000000f, -1.000000f,  fibo,  0.000000f, -fibom1,
                                     -fibom1,  fibo,  0.000000f,  fibom1,  fibo,  0.000000f,  1.000000f,  1.000000f, -1.000000f,
                                     -fibom1,  fibo,  0.000000f,  1.000000f,  1.000000f, -1.000000f,  0.000000f,  fibom1, -fibo,
                                     -fibom1,  fibo,  0.000000f,  0.000000f,  fibom1, -fibo, -1.000000f,  1.000000f, -1.000000f,
                                     0.000000f,  fibom1, -fibo,  1.000000f,  1.000000f, -1.000000f,  fibo,  0.000000f, -fibom1,
                                     0.000000f,  fibom1, -fibo,  fibo,  0.000000f, -fibom1,  1.000000f, -1.000000f, -1.000000f,
                                     0.000000f,  fibom1, -fibo,  1.000000f, -1.000000f, -1.000000f,  0.000000f, -fibom1, -fibo,
                                     -fibo,  0.000000f, -fibom1, -1.000000f,  1.000000f, -1.000000f,  0.000000f,  fibom1, -fibo,
                                     -fibo,  0.000000f, -fibom1,  0.000000f,  fibom1, -fibo,  0.000000f, -fibom1, -fibo,
                                     -fibo,  0.000000f, -fibom1,  0.000000f, -fibom1, -fibo, -1.000000f, -1.000000f, -1.000000f,
                                     0.000000f, -fibom1, -fibo,  1.000000f, -1.000000f, -1.000000f,  fibom1, -fibo,  0.000000f,
                                     0.000000f, -fibom1, -fibo,  fibom1, -fibo,  0.000000f, -fibom1, -fibo,  0.000000f,
                                     0.000000f, -fibom1, -fibo, -fibom1, -fibo,  0.000000f, -1.000000f, -1.000000f, -1.000000f};
  static const GLfloat doden[324]={
                                     0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,
                                     0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,
                                     0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,
                                     -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f,
                                     -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f,
                                     -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f,
                                     0.000000f,  0.850651f,  0.525731f,  0.000000f,  0.850651f,  0.525731f,  0.000000f,  0.850651f,  0.525731f,
                                     0.000000f,  0.850651f,  0.525731f,  0.000000f,  0.850651f,  0.525731f,  0.000000f,  0.850651f,  0.525731f,
                                     -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f,
                                     0.850651f,  0.525731f,  0.000000f,  0.850651f,  0.525731f,  0.000000f,  0.850651f,  0.525731f,  0.000000f,
                                     0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,
                                     0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,  0.850651f,  0.525731f, -0.000000f,
                                     -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f,
                                     -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f,
                                     -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f,
                                     -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f,
                                     -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f,
                                     -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f,
                                     -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f,
                                     -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f,
                                     -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f,  0.000000f, -0.850651f, -0.525731f,  0.000000f,
                                     0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f,  0.000000f,  0.850651f, -0.525731f,  0.000000f,
                                     0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,
                                     0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,
                                     -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f,
                                     0.000000f,  0.850651f, -0.525731f,  0.000000f,  0.850651f, -0.525731f,  0.000000f,  0.850651f, -0.525731f,
                                     -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f, -0.000000f,  0.850651f, -0.525731f,
                                     0.525731f,  0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f,  0.000000f, -0.850651f,
                                     0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,
                                     0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,  0.525731f, -0.000000f, -0.850651f,
                                     -0.525731f,  0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f,  0.000000f, -0.850651f,
                                     -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f,
                                     -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f,
                                     0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f,  0.000000f, -0.850651f, -0.525731f,
                                     -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f,
      -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f, -0.000000f, -0.850651f, -0.525731f};

  for (int i=0; i<108; i++){
    QVector4D qv(dodev[3*i], dodev[3*i+1], dodev[3*i+2],1);
    QVector4D qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);

    vert.append(doden[3*i]);
    vert.append(doden[3*i+1]);
    vert.append(doden[3*i+2]);
    vert.append(style)
        ;
    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);

    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(0.0f);
  }
}

void PlayOpenGL::triacontaeder(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi){

  static const GLfloat triv[540]={
                                    1.000000f,  0.000000f,  fibo,  0.000000f,  fibom1,  fibo,  0.000000f, -fibom1,  fibo,  //9
                                    1.000000f,  0.000000f,  fibo,  1.000000f,  1.000000f,  1.000000f,  0.000000f,  fibom1,  fibo,  //18
                                    1.000000f,  0.000000f,  fibo,  0.000000f, -fibom1,  fibo,  1.000000f, -1.000000f,  1.000000f,  //27
                                    1.000000f,  0.000000f,  fibo,  fibo,  0.000000f,  fibom1,  1.000000f,  1.000000f,  1.000000f,  //36
                                    1.000000f,  0.000000f,  fibo,  1.000000f, -1.000000f,  1.000000f,  fibo,  0.000000f,  fibom1,  //45
                                    fibo,  0.000000f,  fibom1,  fibo,  0.000000f, -fibom1,  fibo,  1.000000f,  0.000000f,  //54
                                    fibo,  0.000000f,  fibom1,  fibo, -1.000000f,  0.000000f,  fibo,  0.000000f, -fibom1,  //63
                                    -fibo,  0.000000f,  fibom1, -fibo,  1.000000f,  0.000000f, -fibo,  0.000000f, -fibom1, //72
                                    -fibo,  0.000000f,  fibom1, -fibo,  0.000000f, -fibom1, -fibo, -1.000000f,  0.000000f, //81
                                    fibo,  1.000000f,  0.000000f,  1.000000f,  1.000000f, -1.000000f,  fibom1,  fibo,  0.000000f,  //90
                                    fibo,  1.000000f,  0.000000f,  fibo,  0.000000f, -fibom1,  1.000000f,  1.000000f, -1.000000f,  //99
                                    0.000000f,  fibom1,  fibo, -1.000000f,  0.000000f,  fibo,  0.000000f, -fibom1,  fibo,  //108
                                    0.000000f,  fibom1,  fibo, -1.000000f,  1.000000f,  1.000000f, -1.000000f,  0.000000f,  fibo,  //117
                                    -fibom1,  fibo,  0.000000f, -1.000000f,  1.000000f, -1.000000f, -fibo,  1.000000f,  0.000000f, //126
                                    -fibom1,  fibo,  0.000000f,  0.000000f,  fibo, -1.000000f, -1.000000f,  1.000000f, -1.000000f, //135
                                    -fibo,  1.000000f,  0.000000f, -1.000000f,  1.000000f, -1.000000f, -fibo,  0.000000f, -fibom1, //144
                                    0.000000f,  fibom1,  fibo,  0.000000f,  fibo,  1.000000f, -1.000000f,  1.000000f,  1.000000f,  //153
                                    -fibo, -1.000000f,  0.000000f, -1.000000f, -1.000000f, -1.000000f, -fibom1, -fibo,  0.000000f, //162
                                    -fibo, -1.000000f,  0.000000f, -fibo,  0.000000f, -fibom1, -1.000000f, -1.000000f, -1.000000f, //171
                                    fibom1, -fibo,  0.000000f,  1.000000f, -1.000000f, -1.000000f,  fibo, -1.000000f,  0.000000f,  //180
                                    fibom1, -fibo,  0.000000f,  0.000000f, -fibo, -1.000000f,  1.000000f, -1.000000f, -1.000000f,  //189
                                    fibo, -1.000000f,  0.000000f,  1.000000f, -1.000000f, -1.000000f,  fibo,  0.000000f, -fibom1,  //198
                                    1.000000f,  1.000000f, -1.000000f,  0.000000f,  fibom1, -fibo,  0.000000f,  fibo, -1.000000f,  //207
                                    1.000000f,  1.000000f, -1.000000f,  1.000000f,  0.000000f, -fibo,  0.000000f,  fibom1, -fibo,  //216
                                    0.000000f,  fibo, -1.000000f,  0.000000f,  fibom1, -fibo, -1.000000f,  1.000000f, -1.000000f,  //225
                                    -1.000000f,  0.000000f,  fibo, -1.000000f, -1.000000f,  1.000000f,  0.000000f, -fibom1,  fibo, //234
                                    -1.000000f, -1.000000f, -1.000000f,  0.000000f, -fibom1, -fibo,  0.000000f, -fibo, -1.000000f, //243
                                    -1.000000f, -1.000000f, -1.000000f, -1.000000f,  0.000000f, -fibo,  0.000000f, -fibom1, -fibo, //252
                                    -1.000000f,  0.000000f,  fibo, -1.000000f,  1.000000f,  1.000000f, -fibo,  0.000000f,  fibom1, //261
                                    0.000000f, -fibo, -1.000000f,  0.000000f, -fibom1, -fibo,  1.000000f, -1.000000f, -1.000000f,  //270
                                    -1.000000f,  0.000000f,  fibo, -fibo,  0.000000f,  fibom1, -1.000000f, -1.000000f,  1.000000f, //279
                                    1.000000f, -1.000000f, -1.000000f,  0.000000f, -fibom1, -fibo,  1.000000f,  0.000000f, -fibo,  //288
                                    1.000000f,  0.000000f, -fibo,  0.000000f, -fibom1, -fibo,  0.000000f,  fibom1, -fibo,  //297
                                    0.000000f,  fibom1, -fibo,  0.000000f, -fibom1, -fibo, -1.000000f,  0.000000f, -fibo,  //306
                                    0.000000f, -fibom1,  fibo,  0.000000f, -fibo,  1.000000f,  1.000000f, -1.000000f,  1.000000f,  //315
                                    1.000000f,  1.000000f,  1.000000f,  fibo,  1.000000f,  0.000000f,  fibom1,  fibo,  0.000000f,  //324
                                    1.000000f,  1.000000f,  1.000000f,  fibom1,  fibo,  0.000000f,  0.000000f,  fibo,  1.000000f,  //333
                                    0.000000f,  fibo,  1.000000f,  fibom1,  fibo,  0.000000f, -fibom1,  fibo,  0.000000f,  //342
                                    0.000000f,  fibo,  1.000000f, -fibom1,  fibo,  0.000000f, -1.000000f,  1.000000f,  1.000000f,  //351
                                    -1.000000f, -1.000000f,  1.000000f, -fibo, -1.000000f,  0.000000f, -fibom1, -fibo,  0.000000f, //360
                                    -1.000000f, -1.000000f,  1.000000f, -fibom1, -fibo,  0.000000f,  0.000000f, -fibo,  1.000000f, //369
                                    0.000000f, -fibo,  1.000000f, -fibom1, -fibo,  0.000000f,  fibom1, -fibo,  0.000000f,  //378
                                    0.000000f, -fibo,  1.000000f,  fibom1, -fibo,  0.000000f,  1.000000f, -1.000000f,  1.000000f,  //387
                                    fibom1,  fibo,  0.000000f,  0.000000f,  fibo, -1.000000f, -fibom1,  fibo,  0.000000f,  //396
                                    -fibom1, -fibo,  0.000000f,  0.000000f, -fibo, -1.000000f,  fibom1, -fibo,  0.000000f, //405
                                    fibom1,  fibo,  0.000000f,  1.000000f,  1.000000f, -1.000000f,  0.000000f,  fibo, -1.000000f,  //414
                                    0.000000f,  fibom1,  fibo,  1.000000f,  1.000000f,  1.000000f,  0.000000f,  fibo,  1.000000f,  //423
                                    -fibom1, -fibo,  0.000000f, -1.000000f, -1.000000f, -1.000000f,  0.000000f, -fibo, -1.000000f, //432
                                    fibo,  0.000000f, -fibom1,  1.000000f,  0.000000f, -fibo,  1.000000f,  1.000000f, -1.000000f,  //441
                                    fibo,  0.000000f, -fibom1,  1.000000f, -1.000000f, -1.000000f,  1.000000f,  0.000000f, -fibo,  //450
                                    -fibo,  0.000000f, -fibom1, -1.000000f,  1.000000f, -1.000000f, -1.000000f,  0.000000f, -fibo, //459
                                    -fibo,  0.000000f, -fibom1, -1.000000f,  0.000000f, -fibo, -1.000000f, -1.000000f, -1.000000f, //468
                                    -1.000000f,  1.000000f, -1.000000f,  0.000000f,  fibom1, -fibo, -1.000000f,  0.000000f, -fibo, //477
                                    0.000000f, -fibom1,  fibo, -1.000000f, -1.000000f,  1.000000f,  0.000000f, -fibo,  1.000000f,  //486
                                    1.000000f,  1.000000f,  1.000000f,  fibo,  0.000000f,  fibom1,  fibo,  1.000000f,  0.000000f,  //495
                                    -1.000000f,  1.000000f,  1.000000f, -fibo,  1.000000f,  0.000000f, -fibo,  0.000000f,  fibom1, //504
                                    -1.000000f,  1.000000f,  1.000000f, -fibom1,  fibo,  0.000000f, -fibo,  1.000000f,  0.000000f, //513
                                    -1.000000f, -1.000000f,  1.000000f, -fibo,  0.000000f,  fibom1, -fibo, -1.000000f,  0.000000f, //522
                                    1.000000f, -1.000000f,  1.000000f,  fibo, -1.000000f,  0.000000f,  fibo,  0.000000f,  fibom1,  //531
                                    1.000000f, -1.000000f,  1.000000f,  fibom1, -fibo,  0.000000f,  fibo, -1.000000f,  0.000000f}; //540
  static const GLfloat trin[540]={
                                    -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f,
                                    0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,
                                    0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,
                                    0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,
                                    0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,
                                    1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,
                                    1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,
                                    -1.000000f, -0.000000f, -0.000000f, -1.000000f, -0.000000f, -0.000000f, -1.000000f, -0.000000f, -0.000000f,
                                    -1.000000f, -0.000000f,  0.000000f, -1.000000f, -0.000000f,  0.000000f, -1.000000f, -0.000000f,  0.000000f,
                                    0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,
                                    0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,
                                    -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f,
                                    -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f,
                                    -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f,
                                    -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f,
                                    -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f,
                                    -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f,
                                    -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f,
                                    -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f,
                                    0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,
                                    0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,
                                    0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,
                                    0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,
                                    0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,
                                    -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f,
                                    -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f,
                                    -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f,
                                    -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f,
                                    -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f,
                                    0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,
                                    -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f,
                                    0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,
                                    -0.000000f, -0.000000f, -1.000000f, -0.000000f, -0.000000f, -1.000000f, -0.000000f, -0.000000f, -1.000000f,
                                    -0.000000f,  0.000000f, -1.000000f, -0.000000f,  0.000000f, -1.000000f, -0.000000f,  0.000000f, -1.000000f,
                                    0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,
                                    0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,
                                    0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,
                                    -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f,
                                    -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f,
                                    -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f,
                                    -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f,
                                    -0.000000f, -1.000000f, -0.000000f, -0.000000f, -1.000000f, -0.000000f, -0.000000f, -1.000000f, -0.000000f,
                                    0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,
                                    -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f, -0.000000f,  1.000000f, -0.000000f,
                                    -0.000000f, -1.000000f,  0.000000f, -0.000000f, -1.000000f,  0.000000f, -0.000000f, -1.000000f,  0.000000f,
                                    0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,
                                    0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,
                                    -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f,
                                    0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,
                                    0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,
                                    -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f, -0.809017f,  0.309017f, -0.500000f,
                                    -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f, -0.809017f, -0.309017f, -0.500000f,
                                    -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f,
                                    -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f, -0.309017f, -0.500000f,  0.809017f,
                                    0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,  0.809017f,  0.309017f,  0.500000f,
                                    -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f,
                                    -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f, -0.500000f,  0.809017f,  0.309017f,
                                    -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f, -0.809017f, -0.309017f,  0.500000f,
                                    0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,  0.809017f, -0.309017f,  0.500000f,
                                    0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f,  0.500000f, -0.809017f,  0.309017f};
  for (int i=0; i<180; i++){
    QVector4D qv(triv[3*i], triv[3*i+1], triv[3*i+2],1);
    QVector4D qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);
    vert.append(trin[i*3]);
    vert.append(trin[i*3+1]);
    vert.append(trin[i*3+2]);
    vert.append(style);
    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);
    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(0.0f);
  }
  //printf("ph %f %f %f\n",ph,mol->lmin,mol->lmax);
}

void PlayOpenGL::ball(QVector<float> &vert, const QMatrix4x4 &adpmat, V3 mitte, int an, int _sgi, float _prt, int style, float ph, int resi){
  const int upper_limit = 6;
  static const GLuint idx1[240] = {
      1,     12,     14,//  0 0.546533 0.618034 0.546533
      12,      0,     13,//  1 0.546533 0.546533 0.618034
      13,      5,     14,//  2 0.546533 0.546533 0.618034
      12,     13,     14,//  3 0.618034 0.618034 0.618034
      0,     12,     16,//  4 0.546533 0.618034 0.546533
      12,      1,     15,//  5 0.546533 0.546533 0.618034
      15,      7,     16,//  6 0.546533 0.546533 0.618034
      12,     15,     16,//  7 0.618034 0.618034 0.618034
      10,     17,     18,//  8 0.546533 0.618034 0.546533
      17,      0,     16,//  9 0.546533 0.546533 0.618034
      16,      7,     18,// 10 0.546533 0.546533 0.618034
      17,     16,     18,// 11 0.618034 0.618034 0.618034
      0,     17,     20,// 12 0.546533 0.618034 0.546533
      17,     10,     19,// 13 0.546533 0.546533 0.618034
      19,     11,     20,// 14 0.546533 0.546533 0.618034
      17,     19,     20,// 15 0.618034 0.618034 0.618034
      11,     21,     20,// 16 0.546533 0.618034 0.546533
      21,      5,     13,// 17 0.546533 0.546533 0.618034
      13,      0,     20,// 18 0.546533 0.546533 0.618034
      21,     13,     20,// 19 0.618034 0.618034 0.618034
      5,     21,     23,// 20 0.546533 0.618034 0.546533
      21,     11,     22,// 21 0.546533 0.546533 0.618034
      22,      4,     23,// 22 0.546533 0.546533 0.618034
      21,     22,     23,// 23 0.618034 0.618034 0.618034
      2,     24,     25,// 24 0.546533 0.618034 0.546533
      24,     11,     19,// 25 0.546533 0.546533 0.618034
      19,     10,     25,// 26 0.546533 0.546533 0.618034
      24,     19,     25,// 27 0.618034 0.618034 0.618034
      11,     24,     22,// 28 0.546533 0.618034 0.546533
      24,      2,     26,// 29 0.546533 0.546533 0.618034
      26,      4,     22,// 30 0.546533 0.546533 0.618034
      24,     26,     22,// 31 0.618034 0.618034 0.618034
      10,     27,     25,// 32 0.546533 0.618034 0.546533
      27,      6,     28,// 33 0.546533 0.546533 0.618034
      28,      2,     25,// 34 0.546533 0.546533 0.618034
      27,     28,     25,// 35 0.618034 0.618034 0.618034
      6,     27,     29,// 36 0.546533 0.618034 0.546533
      27,     10,     18,// 37 0.546533 0.546533 0.618034
      18,      7,     29,// 38 0.546533 0.546533 0.618034
      27,     18,     29,// 39 0.618034 0.618034 0.618034
      7,     30,     29,// 40 0.546533 0.618034 0.546533
      30,      8,     31,// 41 0.546533 0.546533 0.618034
      31,      6,     29,// 42 0.546533 0.546533 0.618034
      30,     31,     29,// 43 0.618034 0.618034 0.618034
      8,     30,     32,// 44 0.546533 0.618034 0.546533
      30,      7,     15,// 45 0.546533 0.546533 0.618034
      15,      1,     32,// 46 0.546533 0.546533 0.618034
      30,     15,     32,// 47 0.618034 0.618034 0.618034
      9,     33,     34,// 48 0.546533 0.618034 0.546533
      33,      1,     14,// 49 0.546533 0.546533 0.618034
      14,      5,     34,// 50 0.546533 0.546533 0.618034
      33,     14,     34,// 51 0.618034 0.618034 0.618034
      1,     33,     32,// 52 0.546533 0.618034 0.546533
      33,      9,     35,// 53 0.546533 0.546533 0.618034
      35,      8,     32,// 54 0.546533 0.546533 0.618034
      33,     35,     32,// 55 0.618034 0.618034 0.618034
      4,     36,     23,// 56 0.546533 0.618034 0.546533
      36,      9,     34,// 57 0.546533 0.546533 0.618034
      34,      5,     23,// 58 0.546533 0.546533 0.618034
      36,     34,     23,// 59 0.618034 0.618034 0.618034
      9,     36,     38,// 60 0.546533 0.618034 0.546533
      36,      4,     37,// 61 0.546533 0.546533 0.618034
      37,      3,     38,// 62 0.546533 0.546533 0.618034
      36,     37,     38,// 63 0.618034 0.618034 0.618034
      3,     39,     38,// 64 0.546533 0.618034 0.546533
      39,      8,     35,// 65 0.546533 0.546533 0.618034
      35,      9,     38,// 66 0.546533 0.546533 0.618034
      39,     35,     38,// 67 0.618034 0.618034 0.618034
      8,     39,     31,// 68 0.546533 0.618034 0.546533
      39,      3,     40,// 69 0.546533 0.546533 0.618034
      40,      6,     31,// 70 0.546533 0.546533 0.618034
      39,     40,     31,// 71 0.618034 0.618034 0.618034
      3,     41,     40,// 72 0.546533 0.618034 0.546533
      41,      2,     28,// 73 0.546533 0.546533 0.618034
      28,      6,     40,// 74 0.546533 0.546533 0.618034
      41,     28,     40,// 75 0.618034 0.618034 0.618034
      2,     41,     26,// 76 0.546533 0.618034 0.546533
      41,      3,     37,// 77 0.546533 0.546533 0.618034
      37,      4,     26,// 78 0.546533 0.546533 0.618034
      41,     37,     26 // 79 0.618034 0.618034 0.618034
  };
  static const GLuint idx2[960] = {
      1,     42,     44,//  0 0.275904 0.321244 0.275904
      42,     12,     43,//  1 0.275904 0.312869 0.285473
      43,     14,     44,//  2 0.312869 0.275904 0.285473
      42,     43,     44,//  3 0.285473 0.285473 0.321244
      12,     45,     47,//  4 0.275904 0.285473 0.312869
      45,      0,     46,//  5 0.275904 0.275904 0.321244
      46,     13,     47,//  6 0.275904 0.312869 0.285473
      45,     46,     47,//  7 0.321244 0.285473 0.285473
      13,     48,     50,//  8 0.275904 0.285473 0.312869
      48,      5,     49,//  9 0.275904 0.275904 0.321244
      49,     14,     50,// 10 0.275904 0.312869 0.285473
      48,     49,     50,// 11 0.321244 0.285473 0.285473
      12,     47,     43,// 12 0.312869 0.324920 0.312869
      47,     13,     50,// 13 0.312869 0.312869 0.324920
      50,     14,     43,// 14 0.312869 0.312869 0.324920
      47,     50,     43,// 15 0.324920 0.324920 0.324920
      0,     45,     52,// 16 0.275904 0.321244 0.275904
      45,     12,     51,// 17 0.275904 0.312869 0.285473
      51,     16,     52,// 18 0.312869 0.275904 0.285473
      45,     51,     52,// 19 0.285473 0.285473 0.321244
      12,     42,     54,// 20 0.275904 0.285473 0.312869
      42,      1,     53,// 21 0.275904 0.275904 0.321244
      53,     15,     54,// 22 0.275904 0.312869 0.285473
      42,     53,     54,// 23 0.321244 0.285473 0.285473
      15,     55,     57,// 24 0.275904 0.285473 0.312869
      55,      7,     56,// 25 0.275904 0.275904 0.321244
      56,     16,     57,// 26 0.275904 0.312869 0.285473
      55,     56,     57,// 27 0.321244 0.285473 0.285473
      12,     54,     51,// 28 0.312869 0.324920 0.312869
      54,     15,     57,// 29 0.312869 0.312869 0.324920
      57,     16,     51,// 30 0.312869 0.312869 0.324920
      54,     57,     51,// 31 0.324920 0.324920 0.324920
      10,     58,     60,// 32 0.275904 0.321244 0.275904
      58,     17,     59,// 33 0.275904 0.312869 0.285473
      59,     18,     60,// 34 0.312869 0.275904 0.285473
      58,     59,     60,// 35 0.285473 0.285473 0.321244
      17,     61,     62,// 36 0.275904 0.285473 0.312869
      61,      0,     52,// 37 0.275904 0.275904 0.321244
      52,     16,     62,// 38 0.275904 0.312869 0.285473
      61,     52,     62,// 39 0.321244 0.285473 0.285473
      16,     56,     64,// 40 0.275904 0.285473 0.312869
      56,      7,     63,// 41 0.275904 0.275904 0.321244
      63,     18,     64,// 42 0.275904 0.312869 0.285473
      56,     63,     64,// 43 0.321244 0.285473 0.285473
      17,     62,     59,// 44 0.312869 0.324920 0.312869
      62,     16,     64,// 45 0.312869 0.312869 0.324920
      64,     18,     59,// 46 0.312869 0.312869 0.324920
      62,     64,     59,// 47 0.324920 0.324920 0.324920
      0,     61,     66,// 48 0.275904 0.321244 0.275904
      61,     17,     65,// 49 0.275904 0.312869 0.285473
      65,     20,     66,// 50 0.312869 0.275904 0.285473
      61,     65,     66,// 51 0.285473 0.285473 0.321244
      17,     58,     68,// 52 0.275904 0.285473 0.312869
      58,     10,     67,// 53 0.275904 0.275904 0.321244
      67,     19,     68,// 54 0.275904 0.312869 0.285473
      58,     67,     68,// 55 0.321244 0.285473 0.285473
      19,     69,     71,// 56 0.275904 0.285473 0.312869
      69,     11,     70,// 57 0.275904 0.275904 0.321244
      70,     20,     71,// 58 0.275904 0.312869 0.285473
      69,     70,     71,// 59 0.321244 0.285473 0.285473
      17,     68,     65,// 60 0.312869 0.324920 0.312869
      68,     19,     71,// 61 0.312869 0.312869 0.324920
      71,     20,     65,// 62 0.312869 0.312869 0.324920
      68,     71,     65,// 63 0.324920 0.324920 0.324920
      11,     72,     70,// 64 0.275904 0.321244 0.275904
      72,     21,     73,// 65 0.275904 0.312869 0.285473
      73,     20,     70,// 66 0.312869 0.275904 0.285473
      72,     73,     70,// 67 0.285473 0.285473 0.321244
      21,     74,     75,// 68 0.275904 0.285473 0.312869
      74,      5,     48,// 69 0.275904 0.275904 0.321244
      48,     13,     75,// 70 0.275904 0.312869 0.285473
      74,     48,     75,// 71 0.321244 0.285473 0.285473
      13,     46,     76,// 72 0.275904 0.285473 0.312869
      46,      0,     66,// 73 0.275904 0.275904 0.321244
      66,     20,     76,// 74 0.275904 0.312869 0.285473
      46,     66,     76,// 75 0.321244 0.285473 0.285473
      21,     75,     73,// 76 0.312869 0.324920 0.312869
      75,     13,     76,// 77 0.312869 0.312869 0.324920
      76,     20,     73,// 78 0.312869 0.312869 0.324920
      75,     76,     73,// 79 0.324920 0.324920 0.324920
      5,     74,     78,// 80 0.275904 0.321244 0.275904
      74,     21,     77,// 81 0.275904 0.312869 0.285473
      77,     23,     78,// 82 0.312869 0.275904 0.285473
      74,     77,     78,// 83 0.285473 0.285473 0.321244
      21,     72,     80,// 84 0.275904 0.285473 0.312869
      72,     11,     79,// 85 0.275904 0.275904 0.321244
      79,     22,     80,// 86 0.275904 0.312869 0.285473
      72,     79,     80,// 87 0.321244 0.285473 0.285473
      22,     81,     83,// 88 0.275904 0.285473 0.312869
      81,      4,     82,// 89 0.275904 0.275904 0.321244
      82,     23,     83,// 90 0.275904 0.312869 0.285473
      81,     82,     83,// 91 0.321244 0.285473 0.285473
      21,     80,     77,// 92 0.312869 0.324920 0.312869
      80,     22,     83,// 93 0.312869 0.312869 0.324920
      83,     23,     77,// 94 0.312869 0.312869 0.324920
      80,     83,     77,// 95 0.324920 0.324920 0.324920
      2,     84,     86,// 96 0.275904 0.321244 0.275904
      84,     24,     85,// 97 0.275904 0.312869 0.285473
      85,     25,     86,// 98 0.312869 0.275904 0.285473
      84,     85,     86,// 99 0.285473 0.285473 0.321244
      24,     87,     88,//100 0.275904 0.285473 0.312869
      87,     11,     69,//101 0.275904 0.275904 0.321244
      69,     19,     88,//102 0.275904 0.312869 0.285473
      87,     69,     88,//103 0.321244 0.285473 0.285473
      19,     67,     90,//104 0.275904 0.285473 0.312869
      67,     10,     89,//105 0.275904 0.275904 0.321244
      89,     25,     90,//106 0.275904 0.312869 0.285473
      67,     89,     90,//107 0.321244 0.285473 0.285473
      24,     88,     85,//108 0.312869 0.324920 0.312869
      88,     19,     90,//109 0.312869 0.312869 0.324920
      90,     25,     85,//110 0.312869 0.312869 0.324920
      88,     90,     85,//111 0.324920 0.324920 0.324920
      11,     87,     79,//112 0.275904 0.321244 0.275904
      87,     24,     91,//113 0.275904 0.312869 0.285473
      91,     22,     79,//114 0.312869 0.275904 0.285473
      87,     91,     79,//115 0.285473 0.285473 0.321244
      24,     84,     93,//116 0.275904 0.285473 0.312869
      84,      2,     92,//117 0.275904 0.275904 0.321244
      92,     26,     93,//118 0.275904 0.312869 0.285473
      84,     92,     93,//119 0.321244 0.285473 0.285473
      26,     94,     95,//120 0.275904 0.285473 0.312869
      94,      4,     81,//121 0.275904 0.275904 0.321244
      81,     22,     95,//122 0.275904 0.312869 0.285473
      94,     81,     95,//123 0.321244 0.285473 0.285473
      24,     93,     91,//124 0.312869 0.324920 0.312869
      93,     26,     95,//125 0.312869 0.312869 0.324920
      95,     22,     91,//126 0.312869 0.312869 0.324920
      93,     95,     91,//127 0.324920 0.324920 0.324920
      10,     96,     89,//128 0.275904 0.321244 0.275904
      96,     27,     97,//129 0.275904 0.312869 0.285473
      97,     25,     89,//130 0.312869 0.275904 0.285473
      96,     97,     89,//131 0.285473 0.285473 0.321244
      27,     98,    100,//132 0.275904 0.285473 0.312869
      98,      6,     99,//133 0.275904 0.275904 0.321244
      99,     28,    100,//134 0.275904 0.312869 0.285473
      98,     99,    100,//135 0.321244 0.285473 0.285473
      28,    101,    102,//136 0.275904 0.285473 0.312869
      101,      2,     86,//137 0.275904 0.275904 0.321244
      86,     25,    102,//138 0.275904 0.312869 0.285473
      101,     86,    102,//139 0.321244 0.285473 0.285473
      27,    100,     97,//140 0.312869 0.324920 0.312869
      100,     28,    102,//141 0.312869 0.312869 0.324920
      102,     25,     97,//142 0.312869 0.312869 0.324920
      100,    102,     97,//143 0.324920 0.324920 0.324920
      6,     98,    104,//144 0.275904 0.321244 0.275904
      98,     27,    103,//145 0.275904 0.312869 0.285473
      103,     29,    104,//146 0.312869 0.275904 0.285473
      98,    103,    104,//147 0.285473 0.285473 0.321244
      27,     96,    105,//148 0.275904 0.285473 0.312869
      96,     10,     60,//149 0.275904 0.275904 0.321244
      60,     18,    105,//150 0.275904 0.312869 0.285473
      96,     60,    105,//151 0.321244 0.285473 0.285473
      18,     63,    107,//152 0.275904 0.285473 0.312869
      63,      7,    106,//153 0.275904 0.275904 0.321244
      106,     29,    107,//154 0.275904 0.312869 0.285473
      63,    106,    107,//155 0.321244 0.285473 0.285473
      27,    105,    103,//156 0.312869 0.324920 0.312869
      105,     18,    107,//157 0.312869 0.312869 0.324920
      107,     29,    103,//158 0.312869 0.312869 0.324920
      105,    107,    103,//159 0.324920 0.324920 0.324920
      7,    108,    106,//160 0.275904 0.321244 0.275904
      108,     30,    109,//161 0.275904 0.312869 0.285473
      109,     29,    106,//162 0.312869 0.275904 0.285473
      108,    109,    106,//163 0.285473 0.285473 0.321244
      30,    110,    112,//164 0.275904 0.285473 0.312869
      110,      8,    111,//165 0.275904 0.275904 0.321244
      111,     31,    112,//166 0.275904 0.312869 0.285473
      110,    111,    112,//167 0.321244 0.285473 0.285473
      31,    113,    114,//168 0.275904 0.285473 0.312869
      113,      6,    104,//169 0.275904 0.275904 0.321244
      104,     29,    114,//170 0.275904 0.312869 0.285473
      113,    104,    114,//171 0.321244 0.285473 0.285473
      30,    112,    109,//172 0.312869 0.324920 0.312869
      112,     31,    114,//173 0.312869 0.312869 0.324920
      114,     29,    109,//174 0.312869 0.312869 0.324920
      112,    114,    109,//175 0.324920 0.324920 0.324920
      8,    110,    116,//176 0.275904 0.321244 0.275904
      110,     30,    115,//177 0.275904 0.312869 0.285473
      115,     32,    116,//178 0.312869 0.275904 0.285473
      110,    115,    116,//179 0.285473 0.285473 0.321244
      30,    108,    117,//180 0.275904 0.285473 0.312869
      108,      7,     55,//181 0.275904 0.275904 0.321244
      55,     15,    117,//182 0.275904 0.312869 0.285473
      108,     55,    117,//183 0.321244 0.285473 0.285473
      15,     53,    119,//184 0.275904 0.285473 0.312869
      53,      1,    118,//185 0.275904 0.275904 0.321244
      118,     32,    119,//186 0.275904 0.312869 0.285473
      53,    118,    119,//187 0.321244 0.285473 0.285473
      30,    117,    115,//188 0.312869 0.324920 0.312869
      117,     15,    119,//189 0.312869 0.312869 0.324920
      119,     32,    115,//190 0.312869 0.312869 0.324920
      117,    119,    115,//191 0.324920 0.324920 0.324920
      9,    120,    122,//192 0.275904 0.321244 0.275904
      120,     33,    121,//193 0.275904 0.312869 0.285473
      121,     34,    122,//194 0.312869 0.275904 0.285473
      120,    121,    122,//195 0.285473 0.285473 0.321244
      33,    123,    124,//196 0.275904 0.285473 0.312869
      123,      1,     44,//197 0.275904 0.275904 0.321244
      44,     14,    124,//198 0.275904 0.312869 0.285473
      123,     44,    124,//199 0.321244 0.285473 0.285473
      14,     49,    126,//200 0.275904 0.285473 0.312869
      49,      5,    125,//201 0.275904 0.275904 0.321244
      125,     34,    126,//202 0.275904 0.312869 0.285473
      49,    125,    126,//203 0.321244 0.285473 0.285473
      33,    124,    121,//204 0.312869 0.324920 0.312869
      124,     14,    126,//205 0.312869 0.312869 0.324920
      126,     34,    121,//206 0.312869 0.312869 0.324920
      124,    126,    121,//207 0.324920 0.324920 0.324920
      1,    123,    118,//208 0.275904 0.321244 0.275904
      123,     33,    127,//209 0.275904 0.312869 0.285473
      127,     32,    118,//210 0.312869 0.275904 0.285473
      123,    127,    118,//211 0.285473 0.285473 0.321244
      33,    120,    129,//212 0.275904 0.285473 0.312869
      120,      9,    128,//213 0.275904 0.275904 0.321244
      128,     35,    129,//214 0.275904 0.312869 0.285473
      120,    128,    129,//215 0.321244 0.285473 0.285473
      35,    130,    131,//216 0.275904 0.285473 0.312869
      130,      8,    116,//217 0.275904 0.275904 0.321244
      116,     32,    131,//218 0.275904 0.312869 0.285473
      130,    116,    131,//219 0.321244 0.285473 0.285473
      33,    129,    127,//220 0.312869 0.324920 0.312869
      129,     35,    131,//221 0.312869 0.312869 0.324920
      131,     32,    127,//222 0.312869 0.312869 0.324920
      129,    131,    127,//223 0.324920 0.324920 0.324920
      4,    132,     82,//224 0.275904 0.321244 0.275904
      132,     36,    133,//225 0.275904 0.312869 0.285473
      133,     23,     82,//226 0.312869 0.275904 0.285473
      132,    133,     82,//227 0.285473 0.285473 0.321244
      36,    134,    135,//228 0.275904 0.285473 0.312869
      134,      9,    122,//229 0.275904 0.275904 0.321244
      122,     34,    135,//230 0.275904 0.312869 0.285473
      134,    122,    135,//231 0.321244 0.285473 0.285473
      34,    125,    136,//232 0.275904 0.285473 0.312869
      125,      5,     78,//233 0.275904 0.275904 0.321244
      78,     23,    136,//234 0.275904 0.312869 0.285473
      125,     78,    136,//235 0.321244 0.285473 0.285473
      36,    135,    133,//236 0.312869 0.324920 0.312869
      135,     34,    136,//237 0.312869 0.312869 0.324920
      136,     23,    133,//238 0.312869 0.312869 0.324920
      135,    136,    133,//239 0.324920 0.324920 0.324920
      9,    134,    138,//240 0.275904 0.321244 0.275904
      134,     36,    137,//241 0.275904 0.312869 0.285473
      137,     38,    138,//242 0.312869 0.275904 0.285473
      134,    137,    138,//243 0.285473 0.285473 0.321244
      36,    132,    140,//244 0.275904 0.285473 0.312869
      132,      4,    139,//245 0.275904 0.275904 0.321244
      139,     37,    140,//246 0.275904 0.312869 0.285473
      132,    139,    140,//247 0.321244 0.285473 0.285473
      37,    141,    143,//248 0.275904 0.285473 0.312869
      141,      3,    142,//249 0.275904 0.275904 0.321244
      142,     38,    143,//250 0.275904 0.312869 0.285473
      141,    142,    143,//251 0.321244 0.285473 0.285473
      36,    140,    137,//252 0.312869 0.324920 0.312869
      140,     37,    143,//253 0.312869 0.312869 0.324920
      143,     38,    137,//254 0.312869 0.312869 0.324920
      140,    143,    137,//255 0.324920 0.324920 0.324920
      3,    144,    142,//256 0.275904 0.321244 0.275904
      144,     39,    145,//257 0.275904 0.312869 0.285473
      145,     38,    142,//258 0.312869 0.275904 0.285473
      144,    145,    142,//259 0.285473 0.285473 0.321244
      39,    146,    147,//260 0.275904 0.285473 0.312869
      146,      8,    130,//261 0.275904 0.275904 0.321244
      130,     35,    147,//262 0.275904 0.312869 0.285473
      146,    130,    147,//263 0.321244 0.285473 0.285473
      35,    128,    148,//264 0.275904 0.285473 0.312869
      128,      9,    138,//265 0.275904 0.275904 0.321244
      138,     38,    148,//266 0.275904 0.312869 0.285473
      128,    138,    148,//267 0.321244 0.285473 0.285473
      39,    147,    145,//268 0.312869 0.324920 0.312869
      147,     35,    148,//269 0.312869 0.312869 0.324920
      148,     38,    145,//270 0.312869 0.312869 0.324920
      147,    148,    145,//271 0.324920 0.324920 0.324920
      8,    146,    111,//272 0.275904 0.321244 0.275904
      146,     39,    149,//273 0.275904 0.312869 0.285473
      149,     31,    111,//274 0.312869 0.275904 0.285473
      146,    149,    111,//275 0.285473 0.285473 0.321244
      39,    144,    151,//276 0.275904 0.285473 0.312869
      144,      3,    150,//277 0.275904 0.275904 0.321244
      150,     40,    151,//278 0.275904 0.312869 0.285473
      144,    150,    151,//279 0.321244 0.285473 0.285473
      40,    152,    153,//280 0.275904 0.285473 0.312869
      152,      6,    113,//281 0.275904 0.275904 0.321244
      113,     31,    153,//282 0.275904 0.312869 0.285473
      152,    113,    153,//283 0.321244 0.285473 0.285473
      39,    151,    149,//284 0.312869 0.324920 0.312869
      151,     40,    153,//285 0.312869 0.312869 0.324920
      153,     31,    149,//286 0.312869 0.312869 0.324920
      151,    153,    149,//287 0.324920 0.324920 0.324920
      3,    154,    150,//288 0.275904 0.321244 0.275904
      154,     41,    155,//289 0.275904 0.312869 0.285473
      155,     40,    150,//290 0.312869 0.275904 0.285473
      154,    155,    150,//291 0.285473 0.285473 0.321244
      41,    156,    157,//292 0.275904 0.285473 0.312869
      156,      2,    101,//293 0.275904 0.275904 0.321244
      101,     28,    157,//294 0.275904 0.312869 0.285473
      156,    101,    157,//295 0.321244 0.285473 0.285473
      28,     99,    158,//296 0.275904 0.285473 0.312869
      99,      6,    152,//297 0.275904 0.275904 0.321244
      152,     40,    158,//298 0.275904 0.312869 0.285473
      99,    152,    158,//299 0.321244 0.285473 0.285473
      41,    157,    155,//300 0.312869 0.324920 0.312869
      157,     28,    158,//301 0.312869 0.312869 0.324920
      158,     40,    155,//302 0.312869 0.312869 0.324920
      157,    158,    155,//303 0.324920 0.324920 0.324920
      2,    156,     92,//304 0.275904 0.321244 0.275904
      156,     41,    159,//305 0.275904 0.312869 0.285473
      159,     26,     92,//306 0.312869 0.275904 0.285473
      156,    159,     92,//307 0.285473 0.285473 0.321244
      41,    154,    160,//308 0.275904 0.285473 0.312869
      154,      3,    141,//309 0.275904 0.275904 0.321244
      141,     37,    160,//310 0.275904 0.312869 0.285473
      154,    141,    160,//311 0.321244 0.285473 0.285473
      37,    139,    161,//312 0.275904 0.285473 0.312869
      139,      4,     94,//313 0.275904 0.275904 0.321244
      94,     26,    161,//314 0.275904 0.312869 0.285473
      139,     94,    161,//315 0.321244 0.285473 0.285473
      41,    160,    159,//316 0.312869 0.324920 0.312869
      160,     37,    161,//317 0.312869 0.312869 0.324920
      161,     26,    159,//318 0.312869 0.312869 0.324920
      160,    161,    159 //319 0.324920 0.324920 0.324920
  };
  static const GLuint idx3[3840] = {
      1,    162,    164,//  0 0.138283 0.162173 0.138283
      162,     42,    163,//  1 0.138283 0.161146 0.139482
      163,     44,    164,//  2 0.161146 0.138283 0.139482
      162,    163,    164,//  3 0.139482 0.139482 0.162173
      42,    165,    167,//  4 0.138283 0.157997 0.143103
      165,     12,    166,//  5 0.138283 0.156918 0.144288
      166,     43,    167,//  6 0.156918 0.143103 0.139510
      165,    166,    167,//  7 0.144288 0.139510 0.157997
      43,    168,    170,//  8 0.156918 0.139510 0.143103
      168,     14,    169,//  9 0.156918 0.138283 0.144288
      169,     44,    170,// 10 0.138283 0.143103 0.157997
      168,    169,    170,// 11 0.144288 0.157997 0.139510
      42,    167,    163,// 12 0.143103 0.144387 0.161146
      167,     43,    170,// 13 0.143103 0.143103 0.162284
      170,     44,    163,// 14 0.143103 0.161146 0.144387
      167,    170,    163,// 15 0.162284 0.144387 0.144387
      12,    171,    173,// 16 0.138283 0.144288 0.156918
      171,     45,    172,// 17 0.138283 0.143103 0.157997
      172,     47,    173,// 18 0.143103 0.156918 0.139510
      171,    172,    173,// 19 0.157997 0.139510 0.144288
      45,    174,    176,// 20 0.138283 0.139482 0.161146
      174,      0,    175,// 21 0.138283 0.138283 0.162173
      175,     46,    176,// 22 0.138283 0.161146 0.139482
      174,    175,    176,// 23 0.162173 0.139482 0.139482
      46,    177,    179,// 24 0.138283 0.157997 0.143103
      177,     13,    178,// 25 0.138283 0.156918 0.144288
      178,     47,    179,// 26 0.156918 0.143103 0.139510
      177,    178,    179,// 27 0.144288 0.139510 0.157997
      45,    176,    172,// 28 0.161146 0.144387 0.143103
      176,     46,    179,// 29 0.161146 0.143103 0.144387
      179,     47,    172,// 30 0.143103 0.143103 0.162284
      176,    179,    172,// 31 0.144387 0.162284 0.144387
      13,    180,    182,// 32 0.138283 0.144288 0.156918
      180,     48,    181,// 33 0.138283 0.143103 0.157997
      181,     50,    182,// 34 0.143103 0.156918 0.139510
      180,    181,    182,// 35 0.157997 0.139510 0.144288
      48,    183,    185,// 36 0.138283 0.139482 0.161146
      183,      5,    184,// 37 0.138283 0.138283 0.162173
      184,     49,    185,// 38 0.138283 0.161146 0.139482
      183,    184,    185,// 39 0.162173 0.139482 0.139482
      49,    186,    188,// 40 0.138283 0.157997 0.143103
      186,     14,    187,// 41 0.138283 0.156918 0.144288
      187,     50,    188,// 42 0.156918 0.143103 0.139510
      186,    187,    188,// 43 0.144288 0.139510 0.157997
      48,    185,    181,// 44 0.161146 0.144387 0.143103
      185,     49,    188,// 45 0.161146 0.143103 0.144387
      188,     50,    181,// 46 0.143103 0.143103 0.162284
      185,    188,    181,// 47 0.144387 0.162284 0.144387
      12,    173,    166,// 48 0.156918 0.164485 0.156918
      173,     47,    189,// 49 0.156918 0.163002 0.158459
      189,     43,    166,// 50 0.163002 0.156918 0.158459
      173,    189,    166,// 51 0.158459 0.158459 0.164485
      47,    178,    190,// 52 0.156918 0.158459 0.163002
      178,     13,    182,// 53 0.156918 0.156918 0.164485
      182,     50,    190,// 54 0.156918 0.163002 0.158459
      178,    182,    190,// 55 0.164485 0.158459 0.158459
      50,    187,    191,// 56 0.156918 0.158459 0.163002
      187,     14,    168,// 57 0.156918 0.156918 0.164485
      168,     43,    191,// 58 0.156918 0.163002 0.158459
      187,    168,    191,// 59 0.164485 0.158459 0.158459
      47,    190,    189,// 60 0.163002 0.164647 0.163002
      190,     50,    191,// 61 0.163002 0.163002 0.164647
      191,     43,    189,// 62 0.163002 0.163002 0.164647
      190,    191,    189,// 63 0.164647 0.164647 0.164647
      0,    174,    193,// 64 0.138283 0.162173 0.138283
      174,     45,    192,// 65 0.138283 0.161146 0.139482
      192,     52,    193,// 66 0.161146 0.138283 0.139482
      174,    192,    193,// 67 0.139482 0.139482 0.162173
      45,    171,    195,// 68 0.138283 0.157997 0.143103
      171,     12,    194,// 69 0.138283 0.156918 0.144288
      194,     51,    195,// 70 0.156918 0.143103 0.139510
      171,    194,    195,// 71 0.144288 0.139510 0.157997
      51,    196,    198,// 72 0.156918 0.139510 0.143103
      196,     16,    197,// 73 0.156918 0.138283 0.144288
      197,     52,    198,// 74 0.138283 0.143103 0.157997
      196,    197,    198,// 75 0.144288 0.157997 0.139510
      45,    195,    192,// 76 0.143103 0.144387 0.161146
      195,     51,    198,// 77 0.143103 0.143103 0.162284
      198,     52,    192,// 78 0.143103 0.161146 0.144387
      195,    198,    192,// 79 0.162284 0.144387 0.144387
      12,    165,    200,// 80 0.138283 0.144288 0.156918
      165,     42,    199,// 81 0.138283 0.143103 0.157997
      199,     54,    200,// 82 0.143103 0.156918 0.139510
      165,    199,    200,// 83 0.157997 0.139510 0.144288
      42,    162,    202,// 84 0.138283 0.139482 0.161146
      162,      1,    201,// 85 0.138283 0.138283 0.162173
      201,     53,    202,// 86 0.138283 0.161146 0.139482
      162,    201,    202,// 87 0.162173 0.139482 0.139482
      53,    203,    205,// 88 0.138283 0.157997 0.143103
      203,     15,    204,// 89 0.138283 0.156918 0.144288
      204,     54,    205,// 90 0.156918 0.143103 0.139510
      203,    204,    205,// 91 0.144288 0.139510 0.157997
      42,    202,    199,// 92 0.161146 0.144387 0.143103
      202,     53,    205,// 93 0.161146 0.143103 0.144387
      205,     54,    199,// 94 0.143103 0.143103 0.162284
      202,    205,    199,// 95 0.144387 0.162284 0.144387
      15,    206,    208,// 96 0.138283 0.144288 0.156918
      206,     55,    207,// 97 0.138283 0.143103 0.157997
      207,     57,    208,// 98 0.143103 0.156918 0.139510
      206,    207,    208,// 99 0.157997 0.139510 0.144288
      55,    209,    211,//100 0.138283 0.139482 0.161146
      209,      7,    210,//101 0.138283 0.138283 0.162173
      210,     56,    211,//102 0.138283 0.161146 0.139482
      209,    210,    211,//103 0.162173 0.139482 0.139482
      56,    212,    214,//104 0.138283 0.157997 0.143103
      212,     16,    213,//105 0.138283 0.156918 0.144288
      213,     57,    214,//106 0.156918 0.143103 0.139510
      212,    213,    214,//107 0.144288 0.139510 0.157997
      55,    211,    207,//108 0.161146 0.144387 0.143103
      211,     56,    214,//109 0.161146 0.143103 0.144387
      214,     57,    207,//110 0.143103 0.143103 0.162284
      211,    214,    207,//111 0.144387 0.162284 0.144387
      12,    200,    194,//112 0.156918 0.164485 0.156918
      200,     54,    215,//113 0.156918 0.163002 0.158459
      215,     51,    194,//114 0.163002 0.156918 0.158459
      200,    215,    194,//115 0.158459 0.158459 0.164485
      54,    204,    216,//116 0.156918 0.158459 0.163002
      204,     15,    208,//117 0.156918 0.156918 0.164485
      208,     57,    216,//118 0.156918 0.163002 0.158459
      204,    208,    216,//119 0.164485 0.158459 0.158459
      57,    213,    217,//120 0.156918 0.158459 0.163002
      213,     16,    196,//121 0.156918 0.156918 0.164485
      196,     51,    217,//122 0.156918 0.163002 0.158459
      213,    196,    217,//123 0.164485 0.158459 0.158459
      54,    216,    215,//124 0.163002 0.164647 0.163002
      216,     57,    217,//125 0.163002 0.163002 0.164647
      217,     51,    215,//126 0.163002 0.163002 0.164647
      216,    217,    215,//127 0.164647 0.164647 0.164647
      10,    218,    220,//128 0.138283 0.162173 0.138283
      218,     58,    219,//129 0.138283 0.161146 0.139482
      219,     60,    220,//130 0.161146 0.138283 0.139482
      218,    219,    220,//131 0.139482 0.139482 0.162173
      58,    221,    223,//132 0.138283 0.157997 0.143103
      221,     17,    222,//133 0.138283 0.156918 0.144288
      222,     59,    223,//134 0.156918 0.143103 0.139510
      221,    222,    223,//135 0.144288 0.139510 0.157997
      59,    224,    226,//136 0.156918 0.139510 0.143103
      224,     18,    225,//137 0.156918 0.138283 0.144288
      225,     60,    226,//138 0.138283 0.143103 0.157997
      224,    225,    226,//139 0.144288 0.157997 0.139510
      58,    223,    219,//140 0.143103 0.144387 0.161146
      223,     59,    226,//141 0.143103 0.143103 0.162284
      226,     60,    219,//142 0.143103 0.161146 0.144387
      223,    226,    219,//143 0.162284 0.144387 0.144387
      17,    227,    229,//144 0.138283 0.144288 0.156918
      227,     61,    228,//145 0.138283 0.143103 0.157997
      228,     62,    229,//146 0.143103 0.156918 0.139510
      227,    228,    229,//147 0.157997 0.139510 0.144288
      61,    230,    231,//148 0.138283 0.139482 0.161146
      230,      0,    193,//149 0.138283 0.138283 0.162173
      193,     52,    231,//150 0.138283 0.161146 0.139482
      230,    193,    231,//151 0.162173 0.139482 0.139482
      52,    197,    233,//152 0.138283 0.157997 0.143103
      197,     16,    232,//153 0.138283 0.156918 0.144288
      232,     62,    233,//154 0.156918 0.143103 0.139510
      197,    232,    233,//155 0.144288 0.139510 0.157997
      61,    231,    228,//156 0.161146 0.144387 0.143103
      231,     52,    233,//157 0.161146 0.143103 0.144387
      233,     62,    228,//158 0.143103 0.143103 0.162284
      231,    233,    228,//159 0.144387 0.162284 0.144387
      16,    212,    235,//160 0.138283 0.144288 0.156918
      212,     56,    234,//161 0.138283 0.143103 0.157997
      234,     64,    235,//162 0.143103 0.156918 0.139510
      212,    234,    235,//163 0.157997 0.139510 0.144288
      56,    210,    237,//164 0.138283 0.139482 0.161146
      210,      7,    236,//165 0.138283 0.138283 0.162173
      236,     63,    237,//166 0.138283 0.161146 0.139482
      210,    236,    237,//167 0.162173 0.139482 0.139482
      63,    238,    240,//168 0.138283 0.157997 0.143103
      238,     18,    239,//169 0.138283 0.156918 0.144288
      239,     64,    240,//170 0.156918 0.143103 0.139510
      238,    239,    240,//171 0.144288 0.139510 0.157997
      56,    237,    234,//172 0.161146 0.144387 0.143103
      237,     63,    240,//173 0.161146 0.143103 0.144387
      240,     64,    234,//174 0.143103 0.143103 0.162284
      237,    240,    234,//175 0.144387 0.162284 0.144387
      17,    229,    222,//176 0.156918 0.164485 0.156918
      229,     62,    241,//177 0.156918 0.163002 0.158459
      241,     59,    222,//178 0.163002 0.156918 0.158459
      229,    241,    222,//179 0.158459 0.158459 0.164485
      62,    232,    242,//180 0.156918 0.158459 0.163002
      232,     16,    235,//181 0.156918 0.156918 0.164485
      235,     64,    242,//182 0.156918 0.163002 0.158459
      232,    235,    242,//183 0.164485 0.158459 0.158459
      64,    239,    243,//184 0.156918 0.158459 0.163002
      239,     18,    224,//185 0.156918 0.156918 0.164485
      224,     59,    243,//186 0.156918 0.163002 0.158459
      239,    224,    243,//187 0.164485 0.158459 0.158459
      62,    242,    241,//188 0.163002 0.164647 0.163002
      242,     64,    243,//189 0.163002 0.163002 0.164647
      243,     59,    241,//190 0.163002 0.163002 0.164647
      242,    243,    241,//191 0.164647 0.164647 0.164647
      0,    230,    245,//192 0.138283 0.162173 0.138283
      230,     61,    244,//193 0.138283 0.161146 0.139482
      244,     66,    245,//194 0.161146 0.138283 0.139482
      230,    244,    245,//195 0.139482 0.139482 0.162173
      61,    227,    247,//196 0.138283 0.157997 0.143103
      227,     17,    246,//197 0.138283 0.156918 0.144288
      246,     65,    247,//198 0.156918 0.143103 0.139510
      227,    246,    247,//199 0.144288 0.139510 0.157997
      65,    248,    250,//200 0.156918 0.139510 0.143103
      248,     20,    249,//201 0.156918 0.138283 0.144288
      249,     66,    250,//202 0.138283 0.143103 0.157997
      248,    249,    250,//203 0.144288 0.157997 0.139510
      61,    247,    244,//204 0.143103 0.144387 0.161146
      247,     65,    250,//205 0.143103 0.143103 0.162284
      250,     66,    244,//206 0.143103 0.161146 0.144387
      247,    250,    244,//207 0.162284 0.144387 0.144387
      17,    221,    252,//208 0.138283 0.144288 0.156918
      221,     58,    251,//209 0.138283 0.143103 0.157997
      251,     68,    252,//210 0.143103 0.156918 0.139510
      221,    251,    252,//211 0.157997 0.139510 0.144288
      58,    218,    254,//212 0.138283 0.139482 0.161146
      218,     10,    253,//213 0.138283 0.138283 0.162173
      253,     67,    254,//214 0.138283 0.161146 0.139482
      218,    253,    254,//215 0.162173 0.139482 0.139482
      67,    255,    257,//216 0.138283 0.157997 0.143103
      255,     19,    256,//217 0.138283 0.156918 0.144288
      256,     68,    257,//218 0.156918 0.143103 0.139510
      255,    256,    257,//219 0.144288 0.139510 0.157997
      58,    254,    251,//220 0.161146 0.144387 0.143103
      254,     67,    257,//221 0.161146 0.143103 0.144387
      257,     68,    251,//222 0.143103 0.143103 0.162284
      254,    257,    251,//223 0.144387 0.162284 0.144387
      19,    258,    260,//224 0.138283 0.144288 0.156918
      258,     69,    259,//225 0.138283 0.143103 0.157997
      259,     71,    260,//226 0.143103 0.156918 0.139510
      258,    259,    260,//227 0.157997 0.139510 0.144288
      69,    261,    263,//228 0.138283 0.139482 0.161146
      261,     11,    262,//229 0.138283 0.138283 0.162173
      262,     70,    263,//230 0.138283 0.161146 0.139482
      261,    262,    263,//231 0.162173 0.139482 0.139482
      70,    264,    266,//232 0.138283 0.157997 0.143103
      264,     20,    265,//233 0.138283 0.156918 0.144288
      265,     71,    266,//234 0.156918 0.143103 0.139510
      264,    265,    266,//235 0.144288 0.139510 0.157997
      69,    263,    259,//236 0.161146 0.144387 0.143103
      263,     70,    266,//237 0.161146 0.143103 0.144387
      266,     71,    259,//238 0.143103 0.143103 0.162284
      263,    266,    259,//239 0.144387 0.162284 0.144387
      17,    252,    246,//240 0.156918 0.164485 0.156918
      252,     68,    267,//241 0.156918 0.163002 0.158459
      267,     65,    246,//242 0.163002 0.156918 0.158459
      252,    267,    246,//243 0.158459 0.158459 0.164485
      68,    256,    268,//244 0.156918 0.158459 0.163002
      256,     19,    260,//245 0.156918 0.156918 0.164485
      260,     71,    268,//246 0.156918 0.163002 0.158459
      256,    260,    268,//247 0.164485 0.158459 0.158459
      71,    265,    269,//248 0.156918 0.158459 0.163002
      265,     20,    248,//249 0.156918 0.156918 0.164485
      248,     65,    269,//250 0.156918 0.163002 0.158459
      265,    248,    269,//251 0.164485 0.158459 0.158459
      68,    268,    267,//252 0.163002 0.164647 0.163002
      268,     71,    269,//253 0.163002 0.163002 0.164647
      269,     65,    267,//254 0.163002 0.163002 0.164647
      268,    269,    267,//255 0.164647 0.164647 0.164647
      11,    270,    262,//256 0.138283 0.162173 0.138283
      270,     72,    271,//257 0.138283 0.161146 0.139482
      271,     70,    262,//258 0.161146 0.138283 0.139482
      270,    271,    262,//259 0.139482 0.139482 0.162173
      72,    272,    274,//260 0.138283 0.157997 0.143103
      272,     21,    273,//261 0.138283 0.156918 0.144288
      273,     73,    274,//262 0.156918 0.143103 0.139510
      272,    273,    274,//263 0.144288 0.139510 0.157997
      73,    275,    276,//264 0.156918 0.139510 0.143103
      275,     20,    264,//265 0.156918 0.138283 0.144288
      264,     70,    276,//266 0.138283 0.143103 0.157997
      275,    264,    276,//267 0.144288 0.157997 0.139510
      72,    274,    271,//268 0.143103 0.144387 0.161146
      274,     73,    276,//269 0.143103 0.143103 0.162284
      276,     70,    271,//270 0.143103 0.161146 0.144387
      274,    276,    271,//271 0.162284 0.144387 0.144387
      21,    277,    279,//272 0.138283 0.144288 0.156918
      277,     74,    278,//273 0.138283 0.143103 0.157997
      278,     75,    279,//274 0.143103 0.156918 0.139510
      277,    278,    279,//275 0.157997 0.139510 0.144288
      74,    280,    281,//276 0.138283 0.139482 0.161146
      280,      5,    183,//277 0.138283 0.138283 0.162173
      183,     48,    281,//278 0.138283 0.161146 0.139482
      280,    183,    281,//279 0.162173 0.139482 0.139482
      48,    180,    283,//280 0.138283 0.157997 0.143103
      180,     13,    282,//281 0.138283 0.156918 0.144288
      282,     75,    283,//282 0.156918 0.143103 0.139510
      180,    282,    283,//283 0.144288 0.139510 0.157997
      74,    281,    278,//284 0.161146 0.144387 0.143103
      281,     48,    283,//285 0.161146 0.143103 0.144387
      283,     75,    278,//286 0.143103 0.143103 0.162284
      281,    283,    278,//287 0.144387 0.162284 0.144387
      13,    177,    285,//288 0.138283 0.144288 0.156918
      177,     46,    284,//289 0.138283 0.143103 0.157997
      284,     76,    285,//290 0.143103 0.156918 0.139510
      177,    284,    285,//291 0.157997 0.139510 0.144288
      46,    175,    286,//292 0.138283 0.139482 0.161146
      175,      0,    245,//293 0.138283 0.138283 0.162173
      245,     66,    286,//294 0.138283 0.161146 0.139482
      175,    245,    286,//295 0.162173 0.139482 0.139482
      66,    249,    288,//296 0.138283 0.157997 0.143103
      249,     20,    287,//297 0.138283 0.156918 0.144288
      287,     76,    288,//298 0.156918 0.143103 0.139510
      249,    287,    288,//299 0.144288 0.139510 0.157997
      46,    286,    284,//300 0.161146 0.144387 0.143103
      286,     66,    288,//301 0.161146 0.143103 0.144387
      288,     76,    284,//302 0.143103 0.143103 0.162284
      286,    288,    284,//303 0.144387 0.162284 0.144387
      21,    279,    273,//304 0.156918 0.164485 0.156918
      279,     75,    289,//305 0.156918 0.163002 0.158459
      289,     73,    273,//306 0.163002 0.156918 0.158459
      279,    289,    273,//307 0.158459 0.158459 0.164485
      75,    282,    290,//308 0.156918 0.158459 0.163002
      282,     13,    285,//309 0.156918 0.156918 0.164485
      285,     76,    290,//310 0.156918 0.163002 0.158459
      282,    285,    290,//311 0.164485 0.158459 0.158459
      76,    287,    291,//312 0.156918 0.158459 0.163002
      287,     20,    275,//313 0.156918 0.156918 0.164485
      275,     73,    291,//314 0.156918 0.163002 0.158459
      287,    275,    291,//315 0.164485 0.158459 0.158459
      75,    290,    289,//316 0.163002 0.164647 0.163002
      290,     76,    291,//317 0.163002 0.163002 0.164647
      291,     73,    289,//318 0.163002 0.163002 0.164647
      290,    291,    289,//319 0.164647 0.164647 0.164647
      5,    280,    293,//320 0.138283 0.162173 0.138283
      280,     74,    292,//321 0.138283 0.161146 0.139482
      292,     78,    293,//322 0.161146 0.138283 0.139482
      280,    292,    293,//323 0.139482 0.139482 0.162173
      74,    277,    295,//324 0.138283 0.157997 0.143103
      277,     21,    294,//325 0.138283 0.156918 0.144288
      294,     77,    295,//326 0.156918 0.143103 0.139510
      277,    294,    295,//327 0.144288 0.139510 0.157997
      77,    296,    298,//328 0.156918 0.139510 0.143103
      296,     23,    297,//329 0.156918 0.138283 0.144288
      297,     78,    298,//330 0.138283 0.143103 0.157997
      296,    297,    298,//331 0.144288 0.157997 0.139510
      74,    295,    292,//332 0.143103 0.144387 0.161146
      295,     77,    298,//333 0.143103 0.143103 0.162284
      298,     78,    292,//334 0.143103 0.161146 0.144387
      295,    298,    292,//335 0.162284 0.144387 0.144387
      21,    272,    300,//336 0.138283 0.144288 0.156918
      272,     72,    299,//337 0.138283 0.143103 0.157997
      299,     80,    300,//338 0.143103 0.156918 0.139510
      272,    299,    300,//339 0.157997 0.139510 0.144288
      72,    270,    302,//340 0.138283 0.139482 0.161146
      270,     11,    301,//341 0.138283 0.138283 0.162173
      301,     79,    302,//342 0.138283 0.161146 0.139482
      270,    301,    302,//343 0.162173 0.139482 0.139482
      79,    303,    305,//344 0.138283 0.157997 0.143103
      303,     22,    304,//345 0.138283 0.156918 0.144288
      304,     80,    305,//346 0.156918 0.143103 0.139510
      303,    304,    305,//347 0.144288 0.139510 0.157997
      72,    302,    299,//348 0.161146 0.144387 0.143103
      302,     79,    305,//349 0.161146 0.143103 0.144387
      305,     80,    299,//350 0.143103 0.143103 0.162284
      302,    305,    299,//351 0.144387 0.162284 0.144387
      22,    306,    308,//352 0.138283 0.144288 0.156918
      306,     81,    307,//353 0.138283 0.143103 0.157997
      307,     83,    308,//354 0.143103 0.156918 0.139510
      306,    307,    308,//355 0.157997 0.139510 0.144288
      81,    309,    311,//356 0.138283 0.139482 0.161146
      309,      4,    310,//357 0.138283 0.138283 0.162173
      310,     82,    311,//358 0.138283 0.161146 0.139482
      309,    310,    311,//359 0.162173 0.139482 0.139482
      82,    312,    314,//360 0.138283 0.157997 0.143103
      312,     23,    313,//361 0.138283 0.156918 0.144288
      313,     83,    314,//362 0.156918 0.143103 0.139510
      312,    313,    314,//363 0.144288 0.139510 0.157997
      81,    311,    307,//364 0.161146 0.144387 0.143103
      311,     82,    314,//365 0.161146 0.143103 0.144387
      314,     83,    307,//366 0.143103 0.143103 0.162284
      311,    314,    307,//367 0.144387 0.162284 0.144387
      21,    300,    294,//368 0.156918 0.164485 0.156918
      300,     80,    315,//369 0.156918 0.163002 0.158459
      315,     77,    294,//370 0.163002 0.156918 0.158459
      300,    315,    294,//371 0.158459 0.158459 0.164485
      80,    304,    316,//372 0.156918 0.158459 0.163002
      304,     22,    308,//373 0.156918 0.156918 0.164485
      308,     83,    316,//374 0.156918 0.163002 0.158459
      304,    308,    316,//375 0.164485 0.158459 0.158459
      83,    313,    317,//376 0.156918 0.158459 0.163002
      313,     23,    296,//377 0.156918 0.156918 0.164485
      296,     77,    317,//378 0.156918 0.163002 0.158459
      313,    296,    317,//379 0.164485 0.158459 0.158459
      80,    316,    315,//380 0.163002 0.164647 0.163002
      316,     83,    317,//381 0.163002 0.163002 0.164647
      317,     77,    315,//382 0.163002 0.163002 0.164647
      316,    317,    315,//383 0.164647 0.164647 0.164647
      2,    318,    320,//384 0.138283 0.162173 0.138283
      318,     84,    319,//385 0.138283 0.161146 0.139482
      319,     86,    320,//386 0.161146 0.138283 0.139482
      318,    319,    320,//387 0.139482 0.139482 0.162173
      84,    321,    323,//388 0.138283 0.157997 0.143103
      321,     24,    322,//389 0.138283 0.156918 0.144288
      322,     85,    323,//390 0.156918 0.143103 0.139510
      321,    322,    323,//391 0.144288 0.139510 0.157997
      85,    324,    326,//392 0.156918 0.139510 0.143103
      324,     25,    325,//393 0.156918 0.138283 0.144288
      325,     86,    326,//394 0.138283 0.143103 0.157997
      324,    325,    326,//395 0.144288 0.157997 0.139510
      84,    323,    319,//396 0.143103 0.144387 0.161146
      323,     85,    326,//397 0.143103 0.143103 0.162284
      326,     86,    319,//398 0.143103 0.161146 0.144387
      323,    326,    319,//399 0.162284 0.144387 0.144387
      24,    327,    329,//400 0.138283 0.144288 0.156918
      327,     87,    328,//401 0.138283 0.143103 0.157997
      328,     88,    329,//402 0.143103 0.156918 0.139510
      327,    328,    329,//403 0.157997 0.139510 0.144288
      87,    330,    331,//404 0.138283 0.139482 0.161146
      330,     11,    261,//405 0.138283 0.138283 0.162173
      261,     69,    331,//406 0.138283 0.161146 0.139482
      330,    261,    331,//407 0.162173 0.139482 0.139482
      69,    258,    333,//408 0.138283 0.157997 0.143103
      258,     19,    332,//409 0.138283 0.156918 0.144288
      332,     88,    333,//410 0.156918 0.143103 0.139510
      258,    332,    333,//411 0.144288 0.139510 0.157997
      87,    331,    328,//412 0.161146 0.144387 0.143103
      331,     69,    333,//413 0.161146 0.143103 0.144387
      333,     88,    328,//414 0.143103 0.143103 0.162284
      331,    333,    328,//415 0.144387 0.162284 0.144387
      19,    255,    335,//416 0.138283 0.144288 0.156918
      255,     67,    334,//417 0.138283 0.143103 0.157997
      334,     90,    335,//418 0.143103 0.156918 0.139510
      255,    334,    335,//419 0.157997 0.139510 0.144288
      67,    253,    337,//420 0.138283 0.139482 0.161146
      253,     10,    336,//421 0.138283 0.138283 0.162173
      336,     89,    337,//422 0.138283 0.161146 0.139482
      253,    336,    337,//423 0.162173 0.139482 0.139482
      89,    338,    340,//424 0.138283 0.157997 0.143103
      338,     25,    339,//425 0.138283 0.156918 0.144288
      339,     90,    340,//426 0.156918 0.143103 0.139510
      338,    339,    340,//427 0.144288 0.139510 0.157997
      67,    337,    334,//428 0.161146 0.144387 0.143103
      337,     89,    340,//429 0.161146 0.143103 0.144387
      340,     90,    334,//430 0.143103 0.143103 0.162284
      337,    340,    334,//431 0.144387 0.162284 0.144387
      24,    329,    322,//432 0.156918 0.164485 0.156918
      329,     88,    341,//433 0.156918 0.163002 0.158459
      341,     85,    322,//434 0.163002 0.156918 0.158459
      329,    341,    322,//435 0.158459 0.158459 0.164485
      88,    332,    342,//436 0.156918 0.158459 0.163002
      332,     19,    335,//437 0.156918 0.156918 0.164485
      335,     90,    342,//438 0.156918 0.163002 0.158459
      332,    335,    342,//439 0.164485 0.158459 0.158459
      90,    339,    343,//440 0.156918 0.158459 0.163002
      339,     25,    324,//441 0.156918 0.156918 0.164485
      324,     85,    343,//442 0.156918 0.163002 0.158459
      339,    324,    343,//443 0.164485 0.158459 0.158459
      88,    342,    341,//444 0.163002 0.164647 0.163002
      342,     90,    343,//445 0.163002 0.163002 0.164647
      343,     85,    341,//446 0.163002 0.163002 0.164647
      342,    343,    341,//447 0.164647 0.164647 0.164647
      11,    330,    301,//448 0.138283 0.162173 0.138283
      330,     87,    344,//449 0.138283 0.161146 0.139482
      344,     79,    301,//450 0.161146 0.138283 0.139482
      330,    344,    301,//451 0.139482 0.139482 0.162173
      87,    327,    346,//452 0.138283 0.157997 0.143103
      327,     24,    345,//453 0.138283 0.156918 0.144288
      345,     91,    346,//454 0.156918 0.143103 0.139510
      327,    345,    346,//455 0.144288 0.139510 0.157997
      91,    347,    348,//456 0.156918 0.139510 0.143103
      347,     22,    303,//457 0.156918 0.138283 0.144288
      303,     79,    348,//458 0.138283 0.143103 0.157997
      347,    303,    348,//459 0.144288 0.157997 0.139510
      87,    346,    344,//460 0.143103 0.144387 0.161146
      346,     91,    348,//461 0.143103 0.143103 0.162284
      348,     79,    344,//462 0.143103 0.161146 0.144387
      346,    348,    344,//463 0.162284 0.144387 0.144387
      24,    321,    350,//464 0.138283 0.144288 0.156918
      321,     84,    349,//465 0.138283 0.143103 0.157997
      349,     93,    350,//466 0.143103 0.156918 0.139510
      321,    349,    350,//467 0.157997 0.139510 0.144288
      84,    318,    352,//468 0.138283 0.139482 0.161146
      318,      2,    351,//469 0.138283 0.138283 0.162173
      351,     92,    352,//470 0.138283 0.161146 0.139482
      318,    351,    352,//471 0.162173 0.139482 0.139482
      92,    353,    355,//472 0.138283 0.157997 0.143103
      353,     26,    354,//473 0.138283 0.156918 0.144288
      354,     93,    355,//474 0.156918 0.143103 0.139510
      353,    354,    355,//475 0.144288 0.139510 0.157997
      84,    352,    349,//476 0.161146 0.144387 0.143103
      352,     92,    355,//477 0.161146 0.143103 0.144387
      355,     93,    349,//478 0.143103 0.143103 0.162284
      352,    355,    349,//479 0.144387 0.162284 0.144387
      26,    356,    358,//480 0.138283 0.144288 0.156918
      356,     94,    357,//481 0.138283 0.143103 0.157997
      357,     95,    358,//482 0.143103 0.156918 0.139510
      356,    357,    358,//483 0.157997 0.139510 0.144288
      94,    359,    360,//484 0.138283 0.139482 0.161146
      359,      4,    309,//485 0.138283 0.138283 0.162173
      309,     81,    360,//486 0.138283 0.161146 0.139482
      359,    309,    360,//487 0.162173 0.139482 0.139482
      81,    306,    362,//488 0.138283 0.157997 0.143103
      306,     22,    361,//489 0.138283 0.156918 0.144288
      361,     95,    362,//490 0.156918 0.143103 0.139510
      306,    361,    362,//491 0.144288 0.139510 0.157997
      94,    360,    357,//492 0.161146 0.144387 0.143103
      360,     81,    362,//493 0.161146 0.143103 0.144387
      362,     95,    357,//494 0.143103 0.143103 0.162284
      360,    362,    357,//495 0.144387 0.162284 0.144387
      24,    350,    345,//496 0.156918 0.164485 0.156918
      350,     93,    363,//497 0.156918 0.163002 0.158459
      363,     91,    345,//498 0.163002 0.156918 0.158459
      350,    363,    345,//499 0.158459 0.158459 0.164485
      93,    354,    364,//500 0.156918 0.158459 0.163002
      354,     26,    358,//501 0.156918 0.156918 0.164485
      358,     95,    364,//502 0.156918 0.163002 0.158459
      354,    358,    364,//503 0.164485 0.158459 0.158459
      95,    361,    365,//504 0.156918 0.158459 0.163002
      361,     22,    347,//505 0.156918 0.156918 0.164485
      347,     91,    365,//506 0.156918 0.163002 0.158459
      361,    347,    365,//507 0.164485 0.158459 0.158459
      93,    364,    363,//508 0.163002 0.164647 0.163002
      364,     95,    365,//509 0.163002 0.163002 0.164647
      365,     91,    363,//510 0.163002 0.163002 0.164647
      364,    365,    363,//511 0.164647 0.164647 0.164647
      10,    366,    336,//512 0.138283 0.162173 0.138283
      366,     96,    367,//513 0.138283 0.161146 0.139482
      367,     89,    336,//514 0.161146 0.138283 0.139482
      366,    367,    336,//515 0.139482 0.139482 0.162173
      96,    368,    370,//516 0.138283 0.157997 0.143103
      368,     27,    369,//517 0.138283 0.156918 0.144288
      369,     97,    370,//518 0.156918 0.143103 0.139510
      368,    369,    370,//519 0.144288 0.139510 0.157997
      97,    371,    372,//520 0.156918 0.139510 0.143103
      371,     25,    338,//521 0.156918 0.138283 0.144288
      338,     89,    372,//522 0.138283 0.143103 0.157997
      371,    338,    372,//523 0.144288 0.157997 0.139510
      96,    370,    367,//524 0.143103 0.144387 0.161146
      370,     97,    372,//525 0.143103 0.143103 0.162284
      372,     89,    367,//526 0.143103 0.161146 0.144387
      370,    372,    367,//527 0.162284 0.144387 0.144387
      27,    373,    375,//528 0.138283 0.144288 0.156918
      373,     98,    374,//529 0.138283 0.143103 0.157997
      374,    100,    375,//530 0.143103 0.156918 0.139510
      373,    374,    375,//531 0.157997 0.139510 0.144288
      98,    376,    378,//532 0.138283 0.139482 0.161146
      376,      6,    377,//533 0.138283 0.138283 0.162173
      377,     99,    378,//534 0.138283 0.161146 0.139482
      376,    377,    378,//535 0.162173 0.139482 0.139482
      99,    379,    381,//536 0.138283 0.157997 0.143103
      379,     28,    380,//537 0.138283 0.156918 0.144288
      380,    100,    381,//538 0.156918 0.143103 0.139510
      379,    380,    381,//539 0.144288 0.139510 0.157997
      98,    378,    374,//540 0.161146 0.144387 0.143103
      378,     99,    381,//541 0.161146 0.143103 0.144387
      381,    100,    374,//542 0.143103 0.143103 0.162284
      378,    381,    374,//543 0.144387 0.162284 0.144387
      28,    382,    384,//544 0.138283 0.144288 0.156918
      382,    101,    383,//545 0.138283 0.143103 0.157997
      383,    102,    384,//546 0.143103 0.156918 0.139510
      382,    383,    384,//547 0.157997 0.139510 0.144288
      101,    385,    386,//548 0.138283 0.139482 0.161146
      385,      2,    320,//549 0.138283 0.138283 0.162173
      320,     86,    386,//550 0.138283 0.161146 0.139482
      385,    320,    386,//551 0.162173 0.139482 0.139482
      86,    325,    388,//552 0.138283 0.157997 0.143103
      325,     25,    387,//553 0.138283 0.156918 0.144288
      387,    102,    388,//554 0.156918 0.143103 0.139510
      325,    387,    388,//555 0.144288 0.139510 0.157997
      101,    386,    383,//556 0.161146 0.144387 0.143103
      386,     86,    388,//557 0.161146 0.143103 0.144387
      388,    102,    383,//558 0.143103 0.143103 0.162284
      386,    388,    383,//559 0.144387 0.162284 0.144387
      27,    375,    369,//560 0.156918 0.164485 0.156918
      375,    100,    389,//561 0.156918 0.163002 0.158459
      389,     97,    369,//562 0.163002 0.156918 0.158459
      375,    389,    369,//563 0.158459 0.158459 0.164485
      100,    380,    390,//564 0.156918 0.158459 0.163002
      380,     28,    384,//565 0.156918 0.156918 0.164485
      384,    102,    390,//566 0.156918 0.163002 0.158459
      380,    384,    390,//567 0.164485 0.158459 0.158459
      102,    387,    391,//568 0.156918 0.158459 0.163002
      387,     25,    371,//569 0.156918 0.156918 0.164485
      371,     97,    391,//570 0.156918 0.163002 0.158459
      387,    371,    391,//571 0.164485 0.158459 0.158459
      100,    390,    389,//572 0.163002 0.164647 0.163002
      390,    102,    391,//573 0.163002 0.163002 0.164647
      391,     97,    389,//574 0.163002 0.163002 0.164647
      390,    391,    389,//575 0.164647 0.164647 0.164647
      6,    376,    393,//576 0.138283 0.162173 0.138283
      376,     98,    392,//577 0.138283 0.161146 0.139482
      392,    104,    393,//578 0.161146 0.138283 0.139482
      376,    392,    393,//579 0.139482 0.139482 0.162173
      98,    373,    395,//580 0.138283 0.157997 0.143103
      373,     27,    394,//581 0.138283 0.156918 0.144288
      394,    103,    395,//582 0.156918 0.143103 0.139510
      373,    394,    395,//583 0.144288 0.139510 0.157997
      103,    396,    398,//584 0.156918 0.139510 0.143103
      396,     29,    397,//585 0.156918 0.138283 0.144288
      397,    104,    398,//586 0.138283 0.143103 0.157997
      396,    397,    398,//587 0.144288 0.157997 0.139510
      98,    395,    392,//588 0.143103 0.144387 0.161146
      395,    103,    398,//589 0.143103 0.143103 0.162284
      398,    104,    392,//590 0.143103 0.161146 0.144387
      395,    398,    392,//591 0.162284 0.144387 0.144387
      27,    368,    400,//592 0.138283 0.144288 0.156918
      368,     96,    399,//593 0.138283 0.143103 0.157997
      399,    105,    400,//594 0.143103 0.156918 0.139510
      368,    399,    400,//595 0.157997 0.139510 0.144288
      96,    366,    401,//596 0.138283 0.139482 0.161146
      366,     10,    220,//597 0.138283 0.138283 0.162173
      220,     60,    401,//598 0.138283 0.161146 0.139482
      366,    220,    401,//599 0.162173 0.139482 0.139482
      60,    225,    403,//600 0.138283 0.157997 0.143103
      225,     18,    402,//601 0.138283 0.156918 0.144288
      402,    105,    403,//602 0.156918 0.143103 0.139510
      225,    402,    403,//603 0.144288 0.139510 0.157997
      96,    401,    399,//604 0.161146 0.144387 0.143103
      401,     60,    403,//605 0.161146 0.143103 0.144387
      403,    105,    399,//606 0.143103 0.143103 0.162284
      401,    403,    399,//607 0.144387 0.162284 0.144387
      18,    238,    405,//608 0.138283 0.144288 0.156918
      238,     63,    404,//609 0.138283 0.143103 0.157997
      404,    107,    405,//610 0.143103 0.156918 0.139510
      238,    404,    405,//611 0.157997 0.139510 0.144288
      63,    236,    407,//612 0.138283 0.139482 0.161146
      236,      7,    406,//613 0.138283 0.138283 0.162173
      406,    106,    407,//614 0.138283 0.161146 0.139482
      236,    406,    407,//615 0.162173 0.139482 0.139482
      106,    408,    410,//616 0.138283 0.157997 0.143103
      408,     29,    409,//617 0.138283 0.156918 0.144288
      409,    107,    410,//618 0.156918 0.143103 0.139510
      408,    409,    410,//619 0.144288 0.139510 0.157997
      63,    407,    404,//620 0.161146 0.144387 0.143103
      407,    106,    410,//621 0.161146 0.143103 0.144387
      410,    107,    404,//622 0.143103 0.143103 0.162284
      407,    410,    404,//623 0.144387 0.162284 0.144387
      27,    400,    394,//624 0.156918 0.164485 0.156918
      400,    105,    411,//625 0.156918 0.163002 0.158459
      411,    103,    394,//626 0.163002 0.156918 0.158459
      400,    411,    394,//627 0.158459 0.158459 0.164485
      105,    402,    412,//628 0.156918 0.158459 0.163002
      402,     18,    405,//629 0.156918 0.156918 0.164485
      405,    107,    412,//630 0.156918 0.163002 0.158459
      402,    405,    412,//631 0.164485 0.158459 0.158459
      107,    409,    413,//632 0.156918 0.158459 0.163002
      409,     29,    396,//633 0.156918 0.156918 0.164485
      396,    103,    413,//634 0.156918 0.163002 0.158459
      409,    396,    413,//635 0.164485 0.158459 0.158459
      105,    412,    411,//636 0.163002 0.164647 0.163002
      412,    107,    413,//637 0.163002 0.163002 0.164647
      413,    103,    411,//638 0.163002 0.163002 0.164647
      412,    413,    411,//639 0.164647 0.164647 0.164647
      7,    414,    406,//640 0.138283 0.162173 0.138283
      414,    108,    415,//641 0.138283 0.161146 0.139482
      415,    106,    406,//642 0.161146 0.138283 0.139482
      414,    415,    406,//643 0.139482 0.139482 0.162173
      108,    416,    418,//644 0.138283 0.157997 0.143103
      416,     30,    417,//645 0.138283 0.156918 0.144288
      417,    109,    418,//646 0.156918 0.143103 0.139510
      416,    417,    418,//647 0.144288 0.139510 0.157997
      109,    419,    420,//648 0.156918 0.139510 0.143103
      419,     29,    408,//649 0.156918 0.138283 0.144288
      408,    106,    420,//650 0.138283 0.143103 0.157997
      419,    408,    420,//651 0.144288 0.157997 0.139510
      108,    418,    415,//652 0.143103 0.144387 0.161146
      418,    109,    420,//653 0.143103 0.143103 0.162284
      420,    106,    415,//654 0.143103 0.161146 0.144387
      418,    420,    415,//655 0.162284 0.144387 0.144387
      30,    421,    423,//656 0.138283 0.144288 0.156918
      421,    110,    422,//657 0.138283 0.143103 0.157997
      422,    112,    423,//658 0.143103 0.156918 0.139510
      421,    422,    423,//659 0.157997 0.139510 0.144288
      110,    424,    426,//660 0.138283 0.139482 0.161146
      424,      8,    425,//661 0.138283 0.138283 0.162173
      425,    111,    426,//662 0.138283 0.161146 0.139482
      424,    425,    426,//663 0.162173 0.139482 0.139482
      111,    427,    429,//664 0.138283 0.157997 0.143103
      427,     31,    428,//665 0.138283 0.156918 0.144288
      428,    112,    429,//666 0.156918 0.143103 0.139510
      427,    428,    429,//667 0.144288 0.139510 0.157997
      110,    426,    422,//668 0.161146 0.144387 0.143103
      426,    111,    429,//669 0.161146 0.143103 0.144387
      429,    112,    422,//670 0.143103 0.143103 0.162284
      426,    429,    422,//671 0.144387 0.162284 0.144387
      31,    430,    432,//672 0.138283 0.144288 0.156918
      430,    113,    431,//673 0.138283 0.143103 0.157997
      431,    114,    432,//674 0.143103 0.156918 0.139510
      430,    431,    432,//675 0.157997 0.139510 0.144288
      113,    433,    434,//676 0.138283 0.139482 0.161146
      433,      6,    393,//677 0.138283 0.138283 0.162173
      393,    104,    434,//678 0.138283 0.161146 0.139482
      433,    393,    434,//679 0.162173 0.139482 0.139482
      104,    397,    436,//680 0.138283 0.157997 0.143103
      397,     29,    435,//681 0.138283 0.156918 0.144288
      435,    114,    436,//682 0.156918 0.143103 0.139510
      397,    435,    436,//683 0.144288 0.139510 0.157997
      113,    434,    431,//684 0.161146 0.144387 0.143103
      434,    104,    436,//685 0.161146 0.143103 0.144387
      436,    114,    431,//686 0.143103 0.143103 0.162284
      434,    436,    431,//687 0.144387 0.162284 0.144387
      30,    423,    417,//688 0.156918 0.164485 0.156918
      423,    112,    437,//689 0.156918 0.163002 0.158459
      437,    109,    417,//690 0.163002 0.156918 0.158459
      423,    437,    417,//691 0.158459 0.158459 0.164485
      112,    428,    438,//692 0.156918 0.158459 0.163002
      428,     31,    432,//693 0.156918 0.156918 0.164485
      432,    114,    438,//694 0.156918 0.163002 0.158459
      428,    432,    438,//695 0.164485 0.158459 0.158459
      114,    435,    439,//696 0.156918 0.158459 0.163002
      435,     29,    419,//697 0.156918 0.156918 0.164485
      419,    109,    439,//698 0.156918 0.163002 0.158459
      435,    419,    439,//699 0.164485 0.158459 0.158459
      112,    438,    437,//700 0.163002 0.164647 0.163002
      438,    114,    439,//701 0.163002 0.163002 0.164647
      439,    109,    437,//702 0.163002 0.163002 0.164647
      438,    439,    437,//703 0.164647 0.164647 0.164647
      8,    424,    441,//704 0.138283 0.162173 0.138283
      424,    110,    440,//705 0.138283 0.161146 0.139482
      440,    116,    441,//706 0.161146 0.138283 0.139482
      424,    440,    441,//707 0.139482 0.139482 0.162173
      110,    421,    443,//708 0.138283 0.157997 0.143103
      421,     30,    442,//709 0.138283 0.156918 0.144288
      442,    115,    443,//710 0.156918 0.143103 0.139510
      421,    442,    443,//711 0.144288 0.139510 0.157997
      115,    444,    446,//712 0.156918 0.139510 0.143103
      444,     32,    445,//713 0.156918 0.138283 0.144288
      445,    116,    446,//714 0.138283 0.143103 0.157997
      444,    445,    446,//715 0.144288 0.157997 0.139510
      110,    443,    440,//716 0.143103 0.144387 0.161146
      443,    115,    446,//717 0.143103 0.143103 0.162284
      446,    116,    440,//718 0.143103 0.161146 0.144387
      443,    446,    440,//719 0.162284 0.144387 0.144387
      30,    416,    448,//720 0.138283 0.144288 0.156918
      416,    108,    447,//721 0.138283 0.143103 0.157997
      447,    117,    448,//722 0.143103 0.156918 0.139510
      416,    447,    448,//723 0.157997 0.139510 0.144288
      108,    414,    449,//724 0.138283 0.139482 0.161146
      414,      7,    209,//725 0.138283 0.138283 0.162173
      209,     55,    449,//726 0.138283 0.161146 0.139482
      414,    209,    449,//727 0.162173 0.139482 0.139482
      55,    206,    451,//728 0.138283 0.157997 0.143103
      206,     15,    450,//729 0.138283 0.156918 0.144288
      450,    117,    451,//730 0.156918 0.143103 0.139510
      206,    450,    451,//731 0.144288 0.139510 0.157997
      108,    449,    447,//732 0.161146 0.144387 0.143103
      449,     55,    451,//733 0.161146 0.143103 0.144387
      451,    117,    447,//734 0.143103 0.143103 0.162284
      449,    451,    447,//735 0.144387 0.162284 0.144387
      15,    203,    453,//736 0.138283 0.144288 0.156918
      203,     53,    452,//737 0.138283 0.143103 0.157997
      452,    119,    453,//738 0.143103 0.156918 0.139510
      203,    452,    453,//739 0.157997 0.139510 0.144288
      53,    201,    455,//740 0.138283 0.139482 0.161146
      201,      1,    454,//741 0.138283 0.138283 0.162173
      454,    118,    455,//742 0.138283 0.161146 0.139482
      201,    454,    455,//743 0.162173 0.139482 0.139482
      118,    456,    458,//744 0.138283 0.157997 0.143103
      456,     32,    457,//745 0.138283 0.156918 0.144288
      457,    119,    458,//746 0.156918 0.143103 0.139510
      456,    457,    458,//747 0.144288 0.139510 0.157997
      53,    455,    452,//748 0.161146 0.144387 0.143103
      455,    118,    458,//749 0.161146 0.143103 0.144387
      458,    119,    452,//750 0.143103 0.143103 0.162284
      455,    458,    452,//751 0.144387 0.162284 0.144387
      30,    448,    442,//752 0.156918 0.164485 0.156918
      448,    117,    459,//753 0.156918 0.163002 0.158459
      459,    115,    442,//754 0.163002 0.156918 0.158459
      448,    459,    442,//755 0.158459 0.158459 0.164485
      117,    450,    460,//756 0.156918 0.158459 0.163002
      450,     15,    453,//757 0.156918 0.156918 0.164485
      453,    119,    460,//758 0.156918 0.163002 0.158459
      450,    453,    460,//759 0.164485 0.158459 0.158459
      119,    457,    461,//760 0.156918 0.158459 0.163002
      457,     32,    444,//761 0.156918 0.156918 0.164485
      444,    115,    461,//762 0.156918 0.163002 0.158459
      457,    444,    461,//763 0.164485 0.158459 0.158459
      117,    460,    459,//764 0.163002 0.164647 0.163002
      460,    119,    461,//765 0.163002 0.163002 0.164647
      461,    115,    459,//766 0.163002 0.163002 0.164647
      460,    461,    459,//767 0.164647 0.164647 0.164647
      9,    462,    464,//768 0.138283 0.162173 0.138283
      462,    120,    463,//769 0.138283 0.161146 0.139482
      463,    122,    464,//770 0.161146 0.138283 0.139482
      462,    463,    464,//771 0.139482 0.139482 0.162173
      120,    465,    467,//772 0.138283 0.157997 0.143103
      465,     33,    466,//773 0.138283 0.156918 0.144288
      466,    121,    467,//774 0.156918 0.143103 0.139510
      465,    466,    467,//775 0.144288 0.139510 0.157997
      121,    468,    470,//776 0.156918 0.139510 0.143103
      468,     34,    469,//777 0.156918 0.138283 0.144288
      469,    122,    470,//778 0.138283 0.143103 0.157997
      468,    469,    470,//779 0.144288 0.157997 0.139510
      120,    467,    463,//780 0.143103 0.144387 0.161146
      467,    121,    470,//781 0.143103 0.143103 0.162284
      470,    122,    463,//782 0.143103 0.161146 0.144387
      467,    470,    463,//783 0.162284 0.144387 0.144387
      33,    471,    473,//784 0.138283 0.144288 0.156918
      471,    123,    472,//785 0.138283 0.143103 0.157997
      472,    124,    473,//786 0.143103 0.156918 0.139510
      471,    472,    473,//787 0.157997 0.139510 0.144288
      123,    474,    475,//788 0.138283 0.139482 0.161146
      474,      1,    164,//789 0.138283 0.138283 0.162173
      164,     44,    475,//790 0.138283 0.161146 0.139482
      474,    164,    475,//791 0.162173 0.139482 0.139482
      44,    169,    477,//792 0.138283 0.157997 0.143103
      169,     14,    476,//793 0.138283 0.156918 0.144288
      476,    124,    477,//794 0.156918 0.143103 0.139510
      169,    476,    477,//795 0.144288 0.139510 0.157997
      123,    475,    472,//796 0.161146 0.144387 0.143103
      475,     44,    477,//797 0.161146 0.143103 0.144387
      477,    124,    472,//798 0.143103 0.143103 0.162284
      475,    477,    472,//799 0.144387 0.162284 0.144387
      14,    186,    479,//800 0.138283 0.144288 0.156918
      186,     49,    478,//801 0.138283 0.143103 0.157997
      478,    126,    479,//802 0.143103 0.156918 0.139510
      186,    478,    479,//803 0.157997 0.139510 0.144288
      49,    184,    481,//804 0.138283 0.139482 0.161146
      184,      5,    480,//805 0.138283 0.138283 0.162173
      480,    125,    481,//806 0.138283 0.161146 0.139482
      184,    480,    481,//807 0.162173 0.139482 0.139482
      125,    482,    484,//808 0.138283 0.157997 0.143103
      482,     34,    483,//809 0.138283 0.156918 0.144288
      483,    126,    484,//810 0.156918 0.143103 0.139510
      482,    483,    484,//811 0.144288 0.139510 0.157997
      49,    481,    478,//812 0.161146 0.144387 0.143103
      481,    125,    484,//813 0.161146 0.143103 0.144387
      484,    126,    478,//814 0.143103 0.143103 0.162284
      481,    484,    478,//815 0.144387 0.162284 0.144387
      33,    473,    466,//816 0.156918 0.164485 0.156918
      473,    124,    485,//817 0.156918 0.163002 0.158459
      485,    121,    466,//818 0.163002 0.156918 0.158459
      473,    485,    466,//819 0.158459 0.158459 0.164485
      124,    476,    486,//820 0.156918 0.158459 0.163002
      476,     14,    479,//821 0.156918 0.156918 0.164485
      479,    126,    486,//822 0.156918 0.163002 0.158459
      476,    479,    486,//823 0.164485 0.158459 0.158459
      126,    483,    487,//824 0.156918 0.158459 0.163002
      483,     34,    468,//825 0.156918 0.156918 0.164485
      468,    121,    487,//826 0.156918 0.163002 0.158459
      483,    468,    487,//827 0.164485 0.158459 0.158459
      124,    486,    485,//828 0.163002 0.164647 0.163002
      486,    126,    487,//829 0.163002 0.163002 0.164647
      487,    121,    485,//830 0.163002 0.163002 0.164647
      486,    487,    485,//831 0.164647 0.164647 0.164647
      1,    474,    454,//832 0.138283 0.162173 0.138283
      474,    123,    488,//833 0.138283 0.161146 0.139482
      488,    118,    454,//834 0.161146 0.138283 0.139482
      474,    488,    454,//835 0.139482 0.139482 0.162173
      123,    471,    490,//836 0.138283 0.157997 0.143103
      471,     33,    489,//837 0.138283 0.156918 0.144288
      489,    127,    490,//838 0.156918 0.143103 0.139510
      471,    489,    490,//839 0.144288 0.139510 0.157997
      127,    491,    492,//840 0.156918 0.139510 0.143103
      491,     32,    456,//841 0.156918 0.138283 0.144288
      456,    118,    492,//842 0.138283 0.143103 0.157997
      491,    456,    492,//843 0.144288 0.157997 0.139510
      123,    490,    488,//844 0.143103 0.144387 0.161146
      490,    127,    492,//845 0.143103 0.143103 0.162284
      492,    118,    488,//846 0.143103 0.161146 0.144387
      490,    492,    488,//847 0.162284 0.144387 0.144387
      33,    465,    494,//848 0.138283 0.144288 0.156918
      465,    120,    493,//849 0.138283 0.143103 0.157997
      493,    129,    494,//850 0.143103 0.156918 0.139510
      465,    493,    494,//851 0.157997 0.139510 0.144288
      120,    462,    496,//852 0.138283 0.139482 0.161146
      462,      9,    495,//853 0.138283 0.138283 0.162173
      495,    128,    496,//854 0.138283 0.161146 0.139482
      462,    495,    496,//855 0.162173 0.139482 0.139482
      128,    497,    499,//856 0.138283 0.157997 0.143103
      497,     35,    498,//857 0.138283 0.156918 0.144288
      498,    129,    499,//858 0.156918 0.143103 0.139510
      497,    498,    499,//859 0.144288 0.139510 0.157997
      120,    496,    493,//860 0.161146 0.144387 0.143103
      496,    128,    499,//861 0.161146 0.143103 0.144387
      499,    129,    493,//862 0.143103 0.143103 0.162284
      496,    499,    493,//863 0.144387 0.162284 0.144387
      35,    500,    502,//864 0.138283 0.144288 0.156918
      500,    130,    501,//865 0.138283 0.143103 0.157997
      501,    131,    502,//866 0.143103 0.156918 0.139510
      500,    501,    502,//867 0.157997 0.139510 0.144288
      130,    503,    504,//868 0.138283 0.139482 0.161146
      503,      8,    441,//869 0.138283 0.138283 0.162173
      441,    116,    504,//870 0.138283 0.161146 0.139482
      503,    441,    504,//871 0.162173 0.139482 0.139482
      116,    445,    506,//872 0.138283 0.157997 0.143103
      445,     32,    505,//873 0.138283 0.156918 0.144288
      505,    131,    506,//874 0.156918 0.143103 0.139510
      445,    505,    506,//875 0.144288 0.139510 0.157997
      130,    504,    501,//876 0.161146 0.144387 0.143103
      504,    116,    506,//877 0.161146 0.143103 0.144387
      506,    131,    501,//878 0.143103 0.143103 0.162284
      504,    506,    501,//879 0.144387 0.162284 0.144387
      33,    494,    489,//880 0.156918 0.164485 0.156918
      494,    129,    507,//881 0.156918 0.163002 0.158459
      507,    127,    489,//882 0.163002 0.156918 0.158459
      494,    507,    489,//883 0.158459 0.158459 0.164485
      129,    498,    508,//884 0.156918 0.158459 0.163002
      498,     35,    502,//885 0.156918 0.156918 0.164485
      502,    131,    508,//886 0.156918 0.163002 0.158459
      498,    502,    508,//887 0.164485 0.158459 0.158459
      131,    505,    509,//888 0.156918 0.158459 0.163002
      505,     32,    491,//889 0.156918 0.156918 0.164485
      491,    127,    509,//890 0.156918 0.163002 0.158459
      505,    491,    509,//891 0.164485 0.158459 0.158459
      129,    508,    507,//892 0.163002 0.164647 0.163002
      508,    131,    509,//893 0.163002 0.163002 0.164647
      509,    127,    507,//894 0.163002 0.163002 0.164647
      508,    509,    507,//895 0.164647 0.164647 0.164647
      4,    510,    310,//896 0.138283 0.162173 0.138283
      510,    132,    511,//897 0.138283 0.161146 0.139482
      511,     82,    310,//898 0.161146 0.138283 0.139482
      510,    511,    310,//899 0.139482 0.139482 0.162173
      132,    512,    514,//900 0.138283 0.157997 0.143103
      512,     36,    513,//901 0.138283 0.156918 0.144288
      513,    133,    514,//902 0.156918 0.143103 0.139510
      512,    513,    514,//903 0.144288 0.139510 0.157997
      133,    515,    516,//904 0.156918 0.139510 0.143103
      515,     23,    312,//905 0.156918 0.138283 0.144288
      312,     82,    516,//906 0.138283 0.143103 0.157997
      515,    312,    516,//907 0.144288 0.157997 0.139510
      132,    514,    511,//908 0.143103 0.144387 0.161146
      514,    133,    516,//909 0.143103 0.143103 0.162284
      516,     82,    511,//910 0.143103 0.161146 0.144387
      514,    516,    511,//911 0.162284 0.144387 0.144387
      36,    517,    519,//912 0.138283 0.144288 0.156918
      517,    134,    518,//913 0.138283 0.143103 0.157997
      518,    135,    519,//914 0.143103 0.156918 0.139510
      517,    518,    519,//915 0.157997 0.139510 0.144288
      134,    520,    521,//916 0.138283 0.139482 0.161146
      520,      9,    464,//917 0.138283 0.138283 0.162173
      464,    122,    521,//918 0.138283 0.161146 0.139482
      520,    464,    521,//919 0.162173 0.139482 0.139482
      122,    469,    523,//920 0.138283 0.157997 0.143103
      469,     34,    522,//921 0.138283 0.156918 0.144288
      522,    135,    523,//922 0.156918 0.143103 0.139510
      469,    522,    523,//923 0.144288 0.139510 0.157997
      134,    521,    518,//924 0.161146 0.144387 0.143103
      521,    122,    523,//925 0.161146 0.143103 0.144387
      523,    135,    518,//926 0.143103 0.143103 0.162284
      521,    523,    518,//927 0.144387 0.162284 0.144387
      34,    482,    525,//928 0.138283 0.144288 0.156918
      482,    125,    524,//929 0.138283 0.143103 0.157997
      524,    136,    525,//930 0.143103 0.156918 0.139510
      482,    524,    525,//931 0.157997 0.139510 0.144288
      125,    480,    526,//932 0.138283 0.139482 0.161146
      480,      5,    293,//933 0.138283 0.138283 0.162173
      293,     78,    526,//934 0.138283 0.161146 0.139482
      480,    293,    526,//935 0.162173 0.139482 0.139482
      78,    297,    528,//936 0.138283 0.157997 0.143103
      297,     23,    527,//937 0.138283 0.156918 0.144288
      527,    136,    528,//938 0.156918 0.143103 0.139510
      297,    527,    528,//939 0.144288 0.139510 0.157997
      125,    526,    524,//940 0.161146 0.144387 0.143103
      526,     78,    528,//941 0.161146 0.143103 0.144387
      528,    136,    524,//942 0.143103 0.143103 0.162284
      526,    528,    524,//943 0.144387 0.162284 0.144387
      36,    519,    513,//944 0.156918 0.164485 0.156918
      519,    135,    529,//945 0.156918 0.163002 0.158459
      529,    133,    513,//946 0.163002 0.156918 0.158459
      519,    529,    513,//947 0.158459 0.158459 0.164485
      135,    522,    530,//948 0.156918 0.158459 0.163002
      522,     34,    525,//949 0.156918 0.156918 0.164485
      525,    136,    530,//950 0.156918 0.163002 0.158459
      522,    525,    530,//951 0.164485 0.158459 0.158459
      136,    527,    531,//952 0.156918 0.158459 0.163002
      527,     23,    515,//953 0.156918 0.156918 0.164485
      515,    133,    531,//954 0.156918 0.163002 0.158459
      527,    515,    531,//955 0.164485 0.158459 0.158459
      135,    530,    529,//956 0.163002 0.164647 0.163002
      530,    136,    531,//957 0.163002 0.163002 0.164647
      531,    133,    529,//958 0.163002 0.163002 0.164647
      530,    531,    529,//959 0.164647 0.164647 0.164647
      9,    520,    533,//960 0.138283 0.162173 0.138283
      520,    134,    532,//961 0.138283 0.161146 0.139482
      532,    138,    533,//962 0.161146 0.138283 0.139482
      520,    532,    533,//963 0.139482 0.139482 0.162173
      134,    517,    535,//964 0.138283 0.157997 0.143103
      517,     36,    534,//965 0.138283 0.156918 0.144288
      534,    137,    535,//966 0.156918 0.143103 0.139510
      517,    534,    535,//967 0.144288 0.139510 0.157997
      137,    536,    538,//968 0.156918 0.139510 0.143103
      536,     38,    537,//969 0.156918 0.138283 0.144288
      537,    138,    538,//970 0.138283 0.143103 0.157997
      536,    537,    538,//971 0.144288 0.157997 0.139510
      134,    535,    532,//972 0.143103 0.144387 0.161146
      535,    137,    538,//973 0.143103 0.143103 0.162284
      538,    138,    532,//974 0.143103 0.161146 0.144387
      535,    538,    532,//975 0.162284 0.144387 0.144387
      36,    512,    540,//976 0.138283 0.144288 0.156918
      512,    132,    539,//977 0.138283 0.143103 0.157997
      539,    140,    540,//978 0.143103 0.156918 0.139510
      512,    539,    540,//979 0.157997 0.139510 0.144288
      132,    510,    542,//980 0.138283 0.139482 0.161146
      510,      4,    541,//981 0.138283 0.138283 0.162173
      541,    139,    542,//982 0.138283 0.161146 0.139482
      510,    541,    542,//983 0.162173 0.139482 0.139482
      139,    543,    545,//984 0.138283 0.157997 0.143103
      543,     37,    544,//985 0.138283 0.156918 0.144288
      544,    140,    545,//986 0.156918 0.143103 0.139510
      543,    544,    545,//987 0.144288 0.139510 0.157997
      132,    542,    539,//988 0.161146 0.144387 0.143103
      542,    139,    545,//989 0.161146 0.143103 0.144387
      545,    140,    539,//990 0.143103 0.143103 0.162284
      542,    545,    539,//991 0.144387 0.162284 0.144387
      37,    546,    548,//992 0.138283 0.144288 0.156918
      546,    141,    547,//993 0.138283 0.143103 0.157997
      547,    143,    548,//994 0.143103 0.156918 0.139510
      546,    547,    548,//995 0.157997 0.139510 0.144288
      141,    549,    551,//996 0.138283 0.139482 0.161146
      549,      3,    550,//997 0.138283 0.138283 0.162173
      550,    142,    551,//998 0.138283 0.161146 0.139482
      549,    550,    551,//999 0.162173 0.139482 0.139482
      142,    552,    554,//1000 0.138283 0.157997 0.143103
      552,     38,    553,//1001 0.138283 0.156918 0.144288
      553,    143,    554,//1002 0.156918 0.143103 0.139510
      552,    553,    554,//1003 0.144288 0.139510 0.157997
      141,    551,    547,//1004 0.161146 0.144387 0.143103
      551,    142,    554,//1005 0.161146 0.143103 0.144387
      554,    143,    547,//1006 0.143103 0.143103 0.162284
      551,    554,    547,//1007 0.144387 0.162284 0.144387
      36,    540,    534,//1008 0.156918 0.164485 0.156918
      540,    140,    555,//1009 0.156918 0.163002 0.158459
      555,    137,    534,//1010 0.163002 0.156918 0.158459
      540,    555,    534,//1011 0.158459 0.158459 0.164485
      140,    544,    556,//1012 0.156918 0.158459 0.163002
      544,     37,    548,//1013 0.156918 0.156918 0.164485
      548,    143,    556,//1014 0.156918 0.163002 0.158459
      544,    548,    556,//1015 0.164485 0.158459 0.158459
      143,    553,    557,//1016 0.156918 0.158459 0.163002
      553,     38,    536,//1017 0.156918 0.156918 0.164485
      536,    137,    557,//1018 0.156918 0.163002 0.158459
      553,    536,    557,//1019 0.164485 0.158459 0.158459
      140,    556,    555,//1020 0.163002 0.164647 0.163002
      556,    143,    557,//1021 0.163002 0.163002 0.164647
      557,    137,    555,//1022 0.163002 0.163002 0.164647
      556,    557,    555,//1023 0.164647 0.164647 0.164647
      3,    558,    550,//1024 0.138283 0.162173 0.138283
      558,    144,    559,//1025 0.138283 0.161146 0.139482
      559,    142,    550,//1026 0.161146 0.138283 0.139482
      558,    559,    550,//1027 0.139482 0.139482 0.162173
      144,    560,    562,//1028 0.138283 0.157997 0.143103
      560,     39,    561,//1029 0.138283 0.156918 0.144288
      561,    145,    562,//1030 0.156918 0.143103 0.139510
      560,    561,    562,//1031 0.144288 0.139510 0.157997
      145,    563,    564,//1032 0.156918 0.139510 0.143103
      563,     38,    552,//1033 0.156918 0.138283 0.144288
      552,    142,    564,//1034 0.138283 0.143103 0.157997
      563,    552,    564,//1035 0.144288 0.157997 0.139510
      144,    562,    559,//1036 0.143103 0.144387 0.161146
      562,    145,    564,//1037 0.143103 0.143103 0.162284
      564,    142,    559,//1038 0.143103 0.161146 0.144387
      562,    564,    559,//1039 0.162284 0.144387 0.144387
      39,    565,    567,//1040 0.138283 0.144288 0.156918
      565,    146,    566,//1041 0.138283 0.143103 0.157997
      566,    147,    567,//1042 0.143103 0.156918 0.139510
      565,    566,    567,//1043 0.157997 0.139510 0.144288
      146,    568,    569,//1044 0.138283 0.139482 0.161146
      568,      8,    503,//1045 0.138283 0.138283 0.162173
      503,    130,    569,//1046 0.138283 0.161146 0.139482
      568,    503,    569,//1047 0.162173 0.139482 0.139482
      130,    500,    571,//1048 0.138283 0.157997 0.143103
      500,     35,    570,//1049 0.138283 0.156918 0.144288
      570,    147,    571,//1050 0.156918 0.143103 0.139510
      500,    570,    571,//1051 0.144288 0.139510 0.157997
      146,    569,    566,//1052 0.161146 0.144387 0.143103
      569,    130,    571,//1053 0.161146 0.143103 0.144387
      571,    147,    566,//1054 0.143103 0.143103 0.162284
      569,    571,    566,//1055 0.144387 0.162284 0.144387
      35,    497,    573,//1056 0.138283 0.144288 0.156918
      497,    128,    572,//1057 0.138283 0.143103 0.157997
      572,    148,    573,//1058 0.143103 0.156918 0.139510
      497,    572,    573,//1059 0.157997 0.139510 0.144288
      128,    495,    574,//1060 0.138283 0.139482 0.161146
      495,      9,    533,//1061 0.138283 0.138283 0.162173
      533,    138,    574,//1062 0.138283 0.161146 0.139482
      495,    533,    574,//1063 0.162173 0.139482 0.139482
      138,    537,    576,//1064 0.138283 0.157997 0.143103
      537,     38,    575,//1065 0.138283 0.156918 0.144288
      575,    148,    576,//1066 0.156918 0.143103 0.139510
      537,    575,    576,//1067 0.144288 0.139510 0.157997
      128,    574,    572,//1068 0.161146 0.144387 0.143103
      574,    138,    576,//1069 0.161146 0.143103 0.144387
      576,    148,    572,//1070 0.143103 0.143103 0.162284
      574,    576,    572,//1071 0.144387 0.162284 0.144387
      39,    567,    561,//1072 0.156918 0.164485 0.156918
      567,    147,    577,//1073 0.156918 0.163002 0.158459
      577,    145,    561,//1074 0.163002 0.156918 0.158459
      567,    577,    561,//1075 0.158459 0.158459 0.164485
      147,    570,    578,//1076 0.156918 0.158459 0.163002
      570,     35,    573,//1077 0.156918 0.156918 0.164485
      573,    148,    578,//1078 0.156918 0.163002 0.158459
      570,    573,    578,//1079 0.164485 0.158459 0.158459
      148,    575,    579,//1080 0.156918 0.158459 0.163002
      575,     38,    563,//1081 0.156918 0.156918 0.164485
      563,    145,    579,//1082 0.156918 0.163002 0.158459
      575,    563,    579,//1083 0.164485 0.158459 0.158459
      147,    578,    577,//1084 0.163002 0.164647 0.163002
      578,    148,    579,//1085 0.163002 0.163002 0.164647
      579,    145,    577,//1086 0.163002 0.163002 0.164647
      578,    579,    577,//1087 0.164647 0.164647 0.164647
      8,    568,    425,//1088 0.138283 0.162173 0.138283
      568,    146,    580,//1089 0.138283 0.161146 0.139482
      580,    111,    425,//1090 0.161146 0.138283 0.139482
      568,    580,    425,//1091 0.139482 0.139482 0.162173
      146,    565,    582,//1092 0.138283 0.157997 0.143103
      565,     39,    581,//1093 0.138283 0.156918 0.144288
      581,    149,    582,//1094 0.156918 0.143103 0.139510
      565,    581,    582,//1095 0.144288 0.139510 0.157997
      149,    583,    584,//1096 0.156918 0.139510 0.143103
      583,     31,    427,//1097 0.156918 0.138283 0.144288
      427,    111,    584,//1098 0.138283 0.143103 0.157997
      583,    427,    584,//1099 0.144288 0.157997 0.139510
      146,    582,    580,//1100 0.143103 0.144387 0.161146
      582,    149,    584,//1101 0.143103 0.143103 0.162284
      584,    111,    580,//1102 0.143103 0.161146 0.144387
      582,    584,    580,//1103 0.162284 0.144387 0.144387
      39,    560,    586,//1104 0.138283 0.144288 0.156918
      560,    144,    585,//1105 0.138283 0.143103 0.157997
      585,    151,    586,//1106 0.143103 0.156918 0.139510
      560,    585,    586,//1107 0.157997 0.139510 0.144288
      144,    558,    588,//1108 0.138283 0.139482 0.161146
      558,      3,    587,//1109 0.138283 0.138283 0.162173
      587,    150,    588,//1110 0.138283 0.161146 0.139482
      558,    587,    588,//1111 0.162173 0.139482 0.139482
      150,    589,    591,//1112 0.138283 0.157997 0.143103
      589,     40,    590,//1113 0.138283 0.156918 0.144288
      590,    151,    591,//1114 0.156918 0.143103 0.139510
      589,    590,    591,//1115 0.144288 0.139510 0.157997
      144,    588,    585,//1116 0.161146 0.144387 0.143103
      588,    150,    591,//1117 0.161146 0.143103 0.144387
      591,    151,    585,//1118 0.143103 0.143103 0.162284
      588,    591,    585,//1119 0.144387 0.162284 0.144387
      40,    592,    594,//1120 0.138283 0.144288 0.156918
      592,    152,    593,//1121 0.138283 0.143103 0.157997
      593,    153,    594,//1122 0.143103 0.156918 0.139510
      592,    593,    594,//1123 0.157997 0.139510 0.144288
      152,    595,    596,//1124 0.138283 0.139482 0.161146
      595,      6,    433,//1125 0.138283 0.138283 0.162173
      433,    113,    596,//1126 0.138283 0.161146 0.139482
      595,    433,    596,//1127 0.162173 0.139482 0.139482
      113,    430,    598,//1128 0.138283 0.157997 0.143103
      430,     31,    597,//1129 0.138283 0.156918 0.144288
      597,    153,    598,//1130 0.156918 0.143103 0.139510
      430,    597,    598,//1131 0.144288 0.139510 0.157997
      152,    596,    593,//1132 0.161146 0.144387 0.143103
      596,    113,    598,//1133 0.161146 0.143103 0.144387
      598,    153,    593,//1134 0.143103 0.143103 0.162284
      596,    598,    593,//1135 0.144387 0.162284 0.144387
      39,    586,    581,//1136 0.156918 0.164485 0.156918
      586,    151,    599,//1137 0.156918 0.163002 0.158459
      599,    149,    581,//1138 0.163002 0.156918 0.158459
      586,    599,    581,//1139 0.158459 0.158459 0.164485
      151,    590,    600,//1140 0.156918 0.158459 0.163002
      590,     40,    594,//1141 0.156918 0.156918 0.164485
      594,    153,    600,//1142 0.156918 0.163002 0.158459
      590,    594,    600,//1143 0.164485 0.158459 0.158459
      153,    597,    601,//1144 0.156918 0.158459 0.163002
      597,     31,    583,//1145 0.156918 0.156918 0.164485
      583,    149,    601,//1146 0.156918 0.163002 0.158459
      597,    583,    601,//1147 0.164485 0.158459 0.158459
      151,    600,    599,//1148 0.163002 0.164647 0.163002
      600,    153,    601,//1149 0.163002 0.163002 0.164647
      601,    149,    599,//1150 0.163002 0.163002 0.164647
      600,    601,    599,//1151 0.164647 0.164647 0.164647
      3,    602,    587,//1152 0.138283 0.162173 0.138283
      602,    154,    603,//1153 0.138283 0.161146 0.139482
      603,    150,    587,//1154 0.161146 0.138283 0.139482
      602,    603,    587,//1155 0.139482 0.139482 0.162173
      154,    604,    606,//1156 0.138283 0.157997 0.143103
      604,     41,    605,//1157 0.138283 0.156918 0.144288
      605,    155,    606,//1158 0.156918 0.143103 0.139510
      604,    605,    606,//1159 0.144288 0.139510 0.157997
      155,    607,    608,//1160 0.156918 0.139510 0.143103
      607,     40,    589,//1161 0.156918 0.138283 0.144288
      589,    150,    608,//1162 0.138283 0.143103 0.157997
      607,    589,    608,//1163 0.144288 0.157997 0.139510
      154,    606,    603,//1164 0.143103 0.144387 0.161146
      606,    155,    608,//1165 0.143103 0.143103 0.162284
      608,    150,    603,//1166 0.143103 0.161146 0.144387
      606,    608,    603,//1167 0.162284 0.144387 0.144387
      41,    609,    611,//1168 0.138283 0.144288 0.156918
      609,    156,    610,//1169 0.138283 0.143103 0.157997
      610,    157,    611,//1170 0.143103 0.156918 0.139510
      609,    610,    611,//1171 0.157997 0.139510 0.144288
      156,    612,    613,//1172 0.138283 0.139482 0.161146
      612,      2,    385,//1173 0.138283 0.138283 0.162173
      385,    101,    613,//1174 0.138283 0.161146 0.139482
      612,    385,    613,//1175 0.162173 0.139482 0.139482
      101,    382,    615,//1176 0.138283 0.157997 0.143103
      382,     28,    614,//1177 0.138283 0.156918 0.144288
      614,    157,    615,//1178 0.156918 0.143103 0.139510
      382,    614,    615,//1179 0.144288 0.139510 0.157997
      156,    613,    610,//1180 0.161146 0.144387 0.143103
      613,    101,    615,//1181 0.161146 0.143103 0.144387
      615,    157,    610,//1182 0.143103 0.143103 0.162284
      613,    615,    610,//1183 0.144387 0.162284 0.144387
      28,    379,    617,//1184 0.138283 0.144288 0.156918
      379,     99,    616,//1185 0.138283 0.143103 0.157997
      616,    158,    617,//1186 0.143103 0.156918 0.139510
      379,    616,    617,//1187 0.157997 0.139510 0.144288
      99,    377,    618,//1188 0.138283 0.139482 0.161146
      377,      6,    595,//1189 0.138283 0.138283 0.162173
      595,    152,    618,//1190 0.138283 0.161146 0.139482
      377,    595,    618,//1191 0.162173 0.139482 0.139482
      152,    592,    620,//1192 0.138283 0.157997 0.143103
      592,     40,    619,//1193 0.138283 0.156918 0.144288
      619,    158,    620,//1194 0.156918 0.143103 0.139510
      592,    619,    620,//1195 0.144288 0.139510 0.157997
      99,    618,    616,//1196 0.161146 0.144387 0.143103
      618,    152,    620,//1197 0.161146 0.143103 0.144387
      620,    158,    616,//1198 0.143103 0.143103 0.162284
      618,    620,    616,//1199 0.144387 0.162284 0.144387
      41,    611,    605,//1200 0.156918 0.164485 0.156918
      611,    157,    621,//1201 0.156918 0.163002 0.158459
      621,    155,    605,//1202 0.163002 0.156918 0.158459
      611,    621,    605,//1203 0.158459 0.158459 0.164485
      157,    614,    622,//1204 0.156918 0.158459 0.163002
      614,     28,    617,//1205 0.156918 0.156918 0.164485
      617,    158,    622,//1206 0.156918 0.163002 0.158459
      614,    617,    622,//1207 0.164485 0.158459 0.158459
      158,    619,    623,//1208 0.156918 0.158459 0.163002
      619,     40,    607,//1209 0.156918 0.156918 0.164485
      607,    155,    623,//1210 0.156918 0.163002 0.158459
      619,    607,    623,//1211 0.164485 0.158459 0.158459
      157,    622,    621,//1212 0.163002 0.164647 0.163002
      622,    158,    623,//1213 0.163002 0.163002 0.164647
      623,    155,    621,//1214 0.163002 0.163002 0.164647
      622,    623,    621,//1215 0.164647 0.164647 0.164647
      2,    612,    351,//1216 0.138283 0.162173 0.138283
      612,    156,    624,//1217 0.138283 0.161146 0.139482
      624,     92,    351,//1218 0.161146 0.138283 0.139482
      612,    624,    351,//1219 0.139482 0.139482 0.162173
      156,    609,    626,//1220 0.138283 0.157997 0.143103
      609,     41,    625,//1221 0.138283 0.156918 0.144288
      625,    159,    626,//1222 0.156918 0.143103 0.139510
      609,    625,    626,//1223 0.144288 0.139510 0.157997
      159,    627,    628,//1224 0.156918 0.139510 0.143103
      627,     26,    353,//1225 0.156918 0.138283 0.144288
      353,     92,    628,//1226 0.138283 0.143103 0.157997
      627,    353,    628,//1227 0.144288 0.157997 0.139510
      156,    626,    624,//1228 0.143103 0.144387 0.161146
      626,    159,    628,//1229 0.143103 0.143103 0.162284
      628,     92,    624,//1230 0.143103 0.161146 0.144387
      626,    628,    624,//1231 0.162284 0.144387 0.144387
      41,    604,    630,//1232 0.138283 0.144288 0.156918
      604,    154,    629,//1233 0.138283 0.143103 0.157997
      629,    160,    630,//1234 0.143103 0.156918 0.139510
      604,    629,    630,//1235 0.157997 0.139510 0.144288
      154,    602,    631,//1236 0.138283 0.139482 0.161146
      602,      3,    549,//1237 0.138283 0.138283 0.162173
      549,    141,    631,//1238 0.138283 0.161146 0.139482
      602,    549,    631,//1239 0.162173 0.139482 0.139482
      141,    546,    633,//1240 0.138283 0.157997 0.143103
      546,     37,    632,//1241 0.138283 0.156918 0.144288
      632,    160,    633,//1242 0.156918 0.143103 0.139510
      546,    632,    633,//1243 0.144288 0.139510 0.157997
      154,    631,    629,//1244 0.161146 0.144387 0.143103
      631,    141,    633,//1245 0.161146 0.143103 0.144387
      633,    160,    629,//1246 0.143103 0.143103 0.162284
      631,    633,    629,//1247 0.144387 0.162284 0.144387
      37,    543,    635,//1248 0.138283 0.144288 0.156918
      543,    139,    634,//1249 0.138283 0.143103 0.157997
      634,    161,    635,//1250 0.143103 0.156918 0.139510
      543,    634,    635,//1251 0.157997 0.139510 0.144288
      139,    541,    636,//1252 0.138283 0.139482 0.161146
      541,      4,    359,//1253 0.138283 0.138283 0.162173
      359,     94,    636,//1254 0.138283 0.161146 0.139482
      541,    359,    636,//1255 0.162173 0.139482 0.139482
      94,    356,    638,//1256 0.138283 0.157997 0.143103
      356,     26,    637,//1257 0.138283 0.156918 0.144288
      637,    161,    638,//1258 0.156918 0.143103 0.139510
      356,    637,    638,//1259 0.144288 0.139510 0.157997
      139,    636,    634,//1260 0.161146 0.144387 0.143103
      636,     94,    638,//1261 0.161146 0.143103 0.144387
      638,    161,    634,//1262 0.143103 0.143103 0.162284
      636,    638,    634,//1263 0.144387 0.162284 0.144387
      41,    630,    625,//1264 0.156918 0.164485 0.156918
      630,    160,    639,//1265 0.156918 0.163002 0.158459
      639,    159,    625,//1266 0.163002 0.156918 0.158459
      630,    639,    625,//1267 0.158459 0.158459 0.164485
      160,    632,    640,//1268 0.156918 0.158459 0.163002
      632,     37,    635,//1269 0.156918 0.156918 0.164485
      635,    161,    640,//1270 0.156918 0.163002 0.158459
      632,    635,    640,//1271 0.164485 0.158459 0.158459
      161,    637,    641,//1272 0.156918 0.158459 0.163002
      637,     26,    627,//1273 0.156918 0.156918 0.164485
      627,    159,    641,//1274 0.156918 0.163002 0.158459
      637,    627,    641,//1275 0.164485 0.158459 0.158459
      160,    640,    639,//1276 0.163002 0.164647 0.163002
      640,    161,    641,//1277 0.163002 0.163002 0.164647
      641,    159,    639,//1278 0.163002 0.163002 0.164647
      640,    641,    639 //1279 0.164647 0.164647 0.164647
  };
  static const GLfloat v[1926] = {
      -0.525731f,   0.850651f,   0.000000f,//0
      0.525731f,   0.850651f,   0.000000f,//1
      -0.525731f,  -0.850651f,   0.000000f,//2
      0.525731f,  -0.850651f,   0.000000f,//3
      0.000000f,  -0.525731f,   0.850651f,//4
      0.000000f,   0.525731f,   0.850651f,//5
      0.000000f,  -0.525731f,  -0.850651f,//6
      0.000000f,   0.525731f,  -0.850651f,//7
      0.850651f,   0.000000f,  -0.525731f,//8
      0.850651f,   0.000000f,   0.525731f,//9
      -0.850651f,   0.000000f,  -0.525731f,//10
      -0.850651f,   0.000000f,   0.525731f,//11
      0.000000f,   1.000000f,   0.000000f,//12
      -0.309017f,   0.809017f,   0.500000f,//13
      0.309017f,   0.809017f,   0.500000f,//14
      0.309017f,   0.809017f,  -0.500000f,//15
      -0.309017f,   0.809017f,  -0.500000f,//16
      -0.809017f,   0.500000f,  -0.309017f,//17
      -0.500000f,   0.309017f,  -0.809017f,//18
      -1.000000f,   0.000000f,   0.000000f,//19
      -0.809017f,   0.500000f,   0.309017f,//20
      -0.500000f,   0.309017f,   0.809017f,//21
      -0.500000f,  -0.309017f,   0.809017f,//22
      0.000000f,   0.000000f,   1.000000f,//23
      -0.809017f,  -0.500000f,   0.309017f,//24
      -0.809017f,  -0.500000f,  -0.309017f,//25
      -0.309017f,  -0.809017f,   0.500000f,//26
      -0.500000f,  -0.309017f,  -0.809017f,//27
      -0.309017f,  -0.809017f,  -0.500000f,//28
      0.000000f,   0.000000f,  -1.000000f,//29
      0.500000f,   0.309017f,  -0.809017f,//30
      0.500000f,  -0.309017f,  -0.809017f,//31
      0.809017f,   0.500000f,  -0.309017f,//32
      0.809017f,   0.500000f,   0.309017f,//33
      0.500000f,   0.309017f,   0.809017f,//34
      1.000000f,   0.000000f,   0.000000f,//35
      0.500000f,  -0.309017f,   0.809017f,//36
      0.309017f,  -0.809017f,   0.500000f,//37
      0.809017f,  -0.500000f,   0.309017f,//38
      0.809017f,  -0.500000f,  -0.309017f,//39
      0.309017f,  -0.809017f,  -0.500000f,//40
      0.000000f,  -1.000000f,   0.000000f,//41
      0.273267f,   0.961938f,   0.000000f,//42
      0.162460f,   0.951057f,   0.262866f,//43
      0.433889f,   0.862668f,   0.259892f,//44
      -0.273267f,   0.961938f,   0.000000f,//45
      -0.433889f,   0.862668f,   0.259892f,//46
      -0.162460f,   0.951057f,   0.262866f,//47
      -0.160622f,   0.693780f,   0.702046f,//48
      0.160622f,   0.693780f,   0.702046f,//49
      0.000000f,   0.850651f,   0.525731f,//50
      -0.162460f,   0.951057f,  -0.262866f,//51
      -0.433889f,   0.862668f,  -0.259892f,//52
      0.433889f,   0.862668f,  -0.259892f,//53
      0.162460f,   0.951057f,  -0.262866f,//54
      0.160622f,   0.693780f,  -0.702046f,//55
      -0.160622f,   0.693780f,  -0.702046f,//56
      0.000000f,   0.850651f,  -0.525731f,//57
      -0.862668f,   0.259892f,  -0.433889f,//58
      -0.688191f,   0.425325f,  -0.587785f,//59
      -0.702046f,   0.160622f,  -0.693780f,//60
      -0.693780f,   0.702046f,  -0.160622f,//61
      -0.587785f,   0.688191f,  -0.425325f,//62
      -0.259892f,   0.433889f,  -0.862668f,//63
      -0.425325f,   0.587785f,  -0.688191f,//64
      -0.850651f,   0.525731f,   0.000000f,//65
      -0.693780f,   0.702046f,   0.160622f,//66
      -0.961938f,   0.000000f,  -0.273267f,//67
      -0.951057f,   0.262866f,  -0.162460f,//68
      -0.961938f,   0.000000f,   0.273267f,//69
      -0.862668f,   0.259892f,   0.433889f,//70
      -0.951057f,   0.262866f,   0.162460f,//71
      -0.702046f,   0.160622f,   0.693780f,//72
      -0.688191f,   0.425325f,   0.587785f,//73
      -0.259892f,   0.433889f,   0.862668f,//74
      -0.425325f,   0.587785f,   0.688191f,//75
      -0.587785f,   0.688191f,   0.425325f,//76
      -0.262866f,   0.162460f,   0.951057f,//77
      0.000000f,   0.273267f,   0.961938f,//78
      -0.702046f,  -0.160622f,   0.693780f,//79
      -0.525731f,   0.000000f,   0.850651f,//80
      -0.259892f,  -0.433889f,   0.862668f,//81
      0.000000f,  -0.273267f,   0.961938f,//82
      -0.262866f,  -0.162460f,   0.951057f,//83
      -0.693780f,  -0.702046f,   0.160622f,//84
      -0.850651f,  -0.525731f,   0.000000f,//85
      -0.693780f,  -0.702046f,  -0.160622f,//86
      -0.862668f,  -0.259892f,   0.433889f,//87
      -0.951057f,  -0.262866f,   0.162460f,//88
      -0.862668f,  -0.259892f,  -0.433889f,//89
      -0.951057f,  -0.262866f,  -0.162460f,//90
      -0.688191f,  -0.425325f,   0.587785f,//91
      -0.433889f,  -0.862668f,   0.259892f,//92
      -0.587785f,  -0.688191f,   0.425325f,//93
      -0.160622f,  -0.693780f,   0.702046f,//94
      -0.425325f,  -0.587785f,   0.688191f,//95
      -0.702046f,  -0.160622f,  -0.693780f,//96
      -0.688191f,  -0.425325f,  -0.587785f,//97
      -0.259892f,  -0.433889f,  -0.862668f,//98
      -0.160622f,  -0.693780f,  -0.702046f,//99
      -0.425325f,  -0.587785f,  -0.688191f,//100
      -0.433889f,  -0.862668f,  -0.259892f,//101
      -0.587785f,  -0.688191f,  -0.425325f,//102
      -0.262866f,  -0.162460f,  -0.951057f,//103
      0.000000f,  -0.273267f,  -0.961938f,//104
      -0.525731f,   0.000000f,  -0.850651f,//105
      0.000000f,   0.273267f,  -0.961938f,//106
      -0.262866f,   0.162460f,  -0.951057f,//107
      0.259892f,   0.433889f,  -0.862668f,//108
      0.262866f,   0.162460f,  -0.951057f,//109
      0.702046f,   0.160622f,  -0.693780f,//110
      0.702046f,  -0.160622f,  -0.693780f,//111
      0.525731f,   0.000000f,  -0.850651f,//112
      0.259892f,  -0.433889f,  -0.862668f,//113
      0.262866f,  -0.162460f,  -0.951057f,//114
      0.688191f,   0.425325f,  -0.587785f,//115
      0.862668f,   0.259892f,  -0.433889f,//116
      0.425325f,   0.587785f,  -0.688191f,//117
      0.693780f,   0.702046f,  -0.160622f,//118
      0.587785f,   0.688191f,  -0.425325f,//119
      0.862668f,   0.259892f,   0.433889f,//120
      0.688191f,   0.425325f,   0.587785f,//121
      0.702046f,   0.160622f,   0.693780f,//122
      0.693780f,   0.702046f,   0.160622f,//123
      0.587785f,   0.688191f,   0.425325f,//124
      0.259892f,   0.433889f,   0.862668f,//125
      0.425325f,   0.587785f,   0.688191f,//126
      0.850651f,   0.525731f,   0.000000f,//127
      0.961938f,   0.000000f,   0.273267f,//128
      0.951057f,   0.262866f,   0.162460f,//129
      0.961938f,   0.000000f,  -0.273267f,//130
      0.951057f,   0.262866f,  -0.162460f,//131
      0.259892f,  -0.433889f,   0.862668f,//132
      0.262866f,  -0.162460f,   0.951057f,//133
      0.702046f,  -0.160622f,   0.693780f,//134
      0.525731f,   0.000000f,   0.850651f,//135
      0.262866f,   0.162460f,   0.951057f,//136
      0.688191f,  -0.425325f,   0.587785f,//137
      0.862668f,  -0.259892f,   0.433889f,//138
      0.160622f,  -0.693780f,   0.702046f,//139
      0.425325f,  -0.587785f,   0.688191f,//140
      0.433889f,  -0.862668f,   0.259892f,//141
      0.693780f,  -0.702046f,   0.160622f,//142
      0.587785f,  -0.688191f,   0.425325f,//143
      0.693780f,  -0.702046f,  -0.160622f,//144
      0.850651f,  -0.525731f,   0.000000f,//145
      0.862668f,  -0.259892f,  -0.433889f,//146
      0.951057f,  -0.262866f,  -0.162460f,//147
      0.951057f,  -0.262866f,   0.162460f,//148
      0.688191f,  -0.425325f,  -0.587785f,//149
      0.433889f,  -0.862668f,  -0.259892f,//150
      0.587785f,  -0.688191f,  -0.425325f,//151
      0.160622f,  -0.693780f,  -0.702046f,//152
      0.425325f,  -0.587785f,  -0.688191f,//153
      0.273267f,  -0.961938f,   0.000000f,//154
      0.162460f,  -0.951057f,  -0.262866f,//155
      -0.273267f,  -0.961938f,   0.000000f,//156
      -0.162460f,  -0.951057f,  -0.262866f,//157
      0.000000f,  -0.850651f,  -0.525731f,//158
      -0.162460f,  -0.951057f,   0.262866f,//159
      0.162460f,  -0.951057f,   0.262866f,//160
      0.000000f,  -0.850651f,   0.525731f,//161
      0.403355f,   0.915043f,   0.000000f,//162
      0.358229f,   0.924305f,   0.131655f,//163
      0.484442f,   0.864929f,   0.131200f,//164
      0.137952f,   0.990439f,   0.000000f,//165
      0.082242f,   0.987688f,   0.133071f,//166
      0.220117f,   0.966393f,   0.132792f,//167
      0.238677f,   0.891007f,   0.386187f,//168
      0.375039f,   0.843911f,   0.383614f,//169
      0.301259f,   0.916244f,   0.264083f,//170
      -0.137952f,   0.990439f,   0.000000f,//171
      -0.220117f,   0.966393f,   0.132792f,//172
      -0.082242f,   0.987688f,   0.133071f,//173
      -0.403355f,   0.915043f,   0.000000f,//174
      -0.484442f,   0.864929f,   0.131200f,//175
      -0.358229f,   0.924305f,   0.131655f,//176
      -0.375039f,   0.843911f,   0.383614f,//177
      -0.238677f,   0.891007f,   0.386187f,//178
      -0.301259f,   0.916244f,   0.264083f,//179
      -0.237086f,   0.758652f,   0.606825f,//180
      -0.081142f,   0.780204f,   0.620240f,//181
      -0.156434f,   0.840178f,   0.519258f,//182
      -0.081086f,   0.615642f,   0.783843f,//183
      0.081086f,   0.615642f,   0.783843f,//184
      0.000000f,   0.702907f,   0.711282f,//185
      0.237086f,   0.758652f,   0.606825f,//186
      0.156434f,   0.840178f,   0.519258f,//187
      0.081142f,   0.780204f,   0.620240f,//188
      0.000000f,   0.963861f,   0.266405f,//189
      -0.082324f,   0.912983f,   0.399607f,//190
      0.082324f,   0.912983f,   0.399607f,//191
      -0.358229f,   0.924305f,  -0.131655f,//192
      -0.484442f,   0.864929f,  -0.131200f,//193
      -0.082242f,   0.987688f,  -0.133071f,//194
      -0.220117f,   0.966393f,  -0.132792f,//195
      -0.238677f,   0.891007f,  -0.386187f,//196
      -0.375039f,   0.843911f,  -0.383614f,//197
      -0.301259f,   0.916244f,  -0.264083f,//198
      0.220117f,   0.966393f,  -0.132792f,//199
      0.082242f,   0.987688f,  -0.133071f,//200
      0.484442f,   0.864929f,  -0.131200f,//201
      0.358229f,   0.924305f,  -0.131655f,//202
      0.375039f,   0.843911f,  -0.383614f,//203
      0.238677f,   0.891007f,  -0.386187f,//204
      0.301259f,   0.916244f,  -0.264083f,//205
      0.237086f,   0.758652f,  -0.606825f,//206
      0.081142f,   0.780204f,  -0.620240f,//207
      0.156434f,   0.840178f,  -0.519258f,//208
      0.081086f,   0.615642f,  -0.783843f,//209
      -0.081086f,   0.615642f,  -0.783843f,//210
      0.000000f,   0.702907f,  -0.711282f,//211
      -0.237086f,   0.758652f,  -0.606825f,//212
      -0.156434f,   0.840178f,  -0.519258f,//213
      -0.081142f,   0.780204f,  -0.620240f,//214
      0.000000f,   0.963861f,  -0.266405f,//215
      0.082324f,   0.912983f,  -0.399607f,//216
      -0.082324f,   0.912983f,  -0.399607f,//217
      -0.864929f,   0.131200f,  -0.484442f,//218
      -0.792649f,   0.213023f,  -0.571252f,//219
      -0.783843f,   0.081086f,  -0.615642f,//220
      -0.843911f,   0.383614f,  -0.375039f,//221
      -0.757935f,   0.468430f,  -0.453991f,//222
      -0.783452f,   0.346153f,  -0.516122f,//223
      -0.601501f,   0.371748f,  -0.707107f,//224
      -0.606825f,   0.237086f,  -0.758652f,//225
      -0.702310f,   0.296005f,  -0.647412f,//226
      -0.758652f,   0.606825f,  -0.237086f,//227
      -0.647412f,   0.702310f,  -0.296005f,//228
      -0.707107f,   0.601501f,  -0.371748f,//229
      -0.615642f,   0.783843f,  -0.081086f,//230
      -0.571252f,   0.792649f,  -0.213023f,//231
      -0.453991f,   0.757935f,  -0.468430f,//232
      -0.516122f,   0.783452f,  -0.346153f,//233
      -0.296005f,   0.647412f,  -0.702310f,//234
      -0.371748f,   0.707107f,  -0.601501f,//235
      -0.131200f,   0.484442f,  -0.864929f,//236
      -0.213023f,   0.571252f,  -0.792649f,//237
      -0.383614f,   0.375039f,  -0.843911f,//238
      -0.468430f,   0.453991f,  -0.757935f,//239
      -0.346153f,   0.516122f,  -0.783452f,//240
      -0.646578f,   0.564254f,  -0.513375f,//241
      -0.513375f,   0.646578f,  -0.564254f,//242
      -0.564254f,   0.513375f,  -0.646578f,//243
      -0.702907f,   0.711282f,   0.000000f,//244
      -0.615642f,   0.783843f,   0.081086f,//245
      -0.840178f,   0.519258f,  -0.156434f,//246
      -0.780204f,   0.620240f,  -0.081142f,//247
      -0.840178f,   0.519258f,   0.156434f,//248
      -0.758652f,   0.606825f,   0.237086f,//249
      -0.780204f,   0.620240f,   0.081142f,//250
      -0.916244f,   0.264083f,  -0.301259f,//251
      -0.891007f,   0.386187f,  -0.238677f,//252
      -0.915043f,   0.000000f,  -0.403355f,//253
      -0.924305f,   0.131655f,  -0.358229f,//254
      -0.990439f,   0.000000f,  -0.137952f,//255
      -0.987688f,   0.133071f,  -0.082242f,//256
      -0.966393f,   0.132792f,  -0.220117f,//257
      -0.990439f,   0.000000f,   0.137952f,//258
      -0.966393f,   0.132792f,   0.220117f,//259
      -0.987688f,   0.133071f,   0.082242f,//260
      -0.915043f,   0.000000f,   0.403355f,//261
      -0.864929f,   0.131200f,   0.484442f,//262
      -0.924305f,   0.131655f,   0.358229f,//263
      -0.843911f,   0.383614f,   0.375039f,//264
      -0.891007f,   0.386187f,   0.238677f,//265
      -0.916244f,   0.264083f,   0.301259f,//266
      -0.912983f,   0.399607f,  -0.082324f,//267
      -0.963861f,   0.266405f,   0.000000f,//268
      -0.912983f,   0.399607f,   0.082324f,//269
      -0.783843f,   0.081086f,   0.615642f,//270
      -0.792649f,   0.213023f,   0.571252f,//271
      -0.606825f,   0.237086f,   0.758652f,//272
      -0.601501f,   0.371748f,   0.707107f,//273
      -0.702310f,   0.296005f,   0.647412f,//274
      -0.757935f,   0.468430f,   0.453991f,//275
      -0.783452f,   0.346153f,   0.516122f,//276
      -0.383614f,   0.375039f,   0.843911f,//277
      -0.346153f,   0.516122f,   0.783452f,//278
      -0.468430f,   0.453991f,   0.757935f,//279
      -0.131200f,   0.484442f,   0.864929f,//280
      -0.213023f,   0.571252f,   0.792649f,//281
      -0.371748f,   0.707107f,   0.601501f,//282
      -0.296005f,   0.647412f,   0.702310f,//283
      -0.516122f,   0.783452f,   0.346153f,//284
      -0.453991f,   0.757935f,   0.468430f,//285
      -0.571252f,   0.792649f,   0.213023f,//286
      -0.707107f,   0.601501f,   0.371748f,//287
      -0.647412f,   0.702310f,   0.296005f,//288
      -0.564254f,   0.513375f,   0.646578f,//289
      -0.513375f,   0.646578f,   0.564254f,//290
      -0.646578f,   0.564254f,   0.513375f,//291
      -0.131655f,   0.358229f,   0.924305f,//292
      0.000000f,   0.403355f,   0.915043f,//293
      -0.386187f,   0.238677f,   0.891007f,//294
      -0.264083f,   0.301259f,   0.916244f,//295
      -0.133071f,   0.082242f,   0.987688f,//296
      0.000000f,   0.137952f,   0.990439f,//297
      -0.132792f,   0.220117f,   0.966393f,//298
      -0.620240f,   0.081142f,   0.780204f,//299
      -0.519258f,   0.156434f,   0.840178f,//300
      -0.783843f,  -0.081086f,   0.615642f,//301
      -0.711282f,   0.000000f,   0.702907f,//302
      -0.606825f,  -0.237086f,   0.758652f,//303
      -0.519258f,  -0.156434f,   0.840178f,//304
      -0.620240f,  -0.081142f,   0.780204f,//305
      -0.383614f,  -0.375039f,   0.843911f,//306
      -0.264083f,  -0.301259f,   0.916244f,//307
      -0.386187f,  -0.238677f,   0.891007f,//308
      -0.131200f,  -0.484442f,   0.864929f,//309
      0.000000f,  -0.403355f,   0.915043f,//310
      -0.131655f,  -0.358229f,   0.924305f,//311
      0.000000f,  -0.137952f,   0.990439f,//312
      -0.133071f,  -0.082242f,   0.987688f,//313
      -0.132792f,  -0.220117f,   0.966393f,//314
      -0.399607f,   0.082324f,   0.912983f,//315
      -0.399607f,  -0.082324f,   0.912983f,//316
      -0.266405f,   0.000000f,   0.963861f,//317
      -0.615642f,  -0.783843f,   0.081086f,//318
      -0.702907f,  -0.711282f,   0.000000f,//319
      -0.615642f,  -0.783843f,  -0.081086f,//320
      -0.758652f,  -0.606825f,   0.237086f,//321
      -0.840178f,  -0.519258f,   0.156434f,//322
      -0.780204f,  -0.620240f,   0.081142f,//323
      -0.840178f,  -0.519258f,  -0.156434f,//324
      -0.758652f,  -0.606825f,  -0.237086f,//325
      -0.780204f,  -0.620240f,  -0.081142f,//326
      -0.843911f,  -0.383614f,   0.375039f,//327
      -0.916244f,  -0.264083f,   0.301259f,//328
      -0.891007f,  -0.386187f,   0.238677f,//329
      -0.864929f,  -0.131200f,   0.484442f,//330
      -0.924305f,  -0.131655f,   0.358229f,//331
      -0.987688f,  -0.133071f,   0.082242f,//332
      -0.966393f,  -0.132792f,   0.220117f,//333
      -0.966393f,  -0.132792f,  -0.220117f,//334
      -0.987688f,  -0.133071f,  -0.082242f,//335
      -0.864929f,  -0.131200f,  -0.484442f,//336
      -0.924305f,  -0.131655f,  -0.358229f,//337
      -0.843911f,  -0.383614f,  -0.375039f,//338
      -0.891007f,  -0.386187f,  -0.238677f,//339
      -0.916244f,  -0.264083f,  -0.301259f,//340
      -0.912983f,  -0.399607f,   0.082324f,//341
      -0.963861f,  -0.266405f,   0.000000f,//342
      -0.912983f,  -0.399607f,  -0.082324f,//343
      -0.792649f,  -0.213023f,   0.571252f,//344
      -0.757935f,  -0.468430f,   0.453991f,//345
      -0.783452f,  -0.346153f,   0.516122f,//346
      -0.601501f,  -0.371748f,   0.707107f,//347
      -0.702310f,  -0.296005f,   0.647412f,//348
      -0.647412f,  -0.702310f,   0.296005f,//349
      -0.707107f,  -0.601501f,   0.371748f,//350
      -0.484442f,  -0.864929f,   0.131200f,//351
      -0.571252f,  -0.792649f,   0.213023f,//352
      -0.375039f,  -0.843911f,   0.383614f,//353
      -0.453991f,  -0.757935f,   0.468430f,//354
      -0.516122f,  -0.783452f,   0.346153f,//355
      -0.237086f,  -0.758652f,   0.606825f,//356
      -0.296005f,  -0.647412f,   0.702310f,//357
      -0.371748f,  -0.707107f,   0.601501f,//358
      -0.081086f,  -0.615642f,   0.783843f,//359
      -0.213023f,  -0.571252f,   0.792649f,//360
      -0.468430f,  -0.453991f,   0.757935f,//361
      -0.346153f,  -0.516122f,   0.783452f,//362
      -0.646578f,  -0.564254f,   0.513375f,//363
      -0.513375f,  -0.646578f,   0.564254f,//364
      -0.564254f,  -0.513375f,   0.646578f,//365
      -0.783843f,  -0.081086f,  -0.615642f,//366
      -0.792649f,  -0.213023f,  -0.571252f,//367
      -0.606825f,  -0.237086f,  -0.758652f,//368
      -0.601501f,  -0.371748f,  -0.707107f,//369
      -0.702310f,  -0.296005f,  -0.647412f,//370
      -0.757935f,  -0.468430f,  -0.453991f,//371
      -0.783452f,  -0.346153f,  -0.516122f,//372
      -0.383614f,  -0.375039f,  -0.843911f,//373
      -0.346153f,  -0.516122f,  -0.783452f,//374
      -0.468430f,  -0.453991f,  -0.757935f,//375
      -0.131200f,  -0.484442f,  -0.864929f,//376
      -0.081086f,  -0.615642f,  -0.783843f,//377
      -0.213023f,  -0.571252f,  -0.792649f,//378
      -0.237086f,  -0.758652f,  -0.606825f,//379
      -0.371748f,  -0.707107f,  -0.601501f,//380
      -0.296005f,  -0.647412f,  -0.702310f,//381
      -0.375039f,  -0.843911f,  -0.383614f,//382
      -0.516122f,  -0.783452f,  -0.346153f,//383
      -0.453991f,  -0.757935f,  -0.468430f,//384
      -0.484442f,  -0.864929f,  -0.131200f,//385
      -0.571252f,  -0.792649f,  -0.213023f,//386
      -0.707107f,  -0.601501f,  -0.371748f,//387
      -0.647412f,  -0.702310f,  -0.296005f,//388
      -0.564254f,  -0.513375f,  -0.646578f,//389
      -0.513375f,  -0.646578f,  -0.564254f,//390
      -0.646578f,  -0.564254f,  -0.513375f,//391
      -0.131655f,  -0.358229f,  -0.924305f,//392
      0.000000f,  -0.403355f,  -0.915043f,//393
      -0.386187f,  -0.238677f,  -0.891007f,//394
      -0.264083f,  -0.301259f,  -0.916244f,//395
      -0.133071f,  -0.082242f,  -0.987688f,//396
      0.000000f,  -0.137952f,  -0.990439f,//397
      -0.132792f,  -0.220117f,  -0.966393f,//398
      -0.620240f,  -0.081142f,  -0.780204f,//399
      -0.519258f,  -0.156434f,  -0.840178f,//400
      -0.711282f,   0.000000f,  -0.702907f,//401
      -0.519258f,   0.156434f,  -0.840178f,//402
      -0.620240f,   0.081142f,  -0.780204f,//403
      -0.264083f,   0.301259f,  -0.916244f,//404
      -0.386187f,   0.238677f,  -0.891007f,//405
      0.000000f,   0.403355f,  -0.915043f,//406
      -0.131655f,   0.358229f,  -0.924305f,//407
      0.000000f,   0.137952f,  -0.990439f,//408
      -0.133071f,   0.082242f,  -0.987688f,//409
      -0.132792f,   0.220117f,  -0.966393f,//410
      -0.399607f,  -0.082324f,  -0.912983f,//411
      -0.399607f,   0.082324f,  -0.912983f,//412
      -0.266405f,   0.000000f,  -0.963861f,//413
      0.131200f,   0.484442f,  -0.864929f,//414
      0.131655f,   0.358229f,  -0.924305f,//415
      0.383614f,   0.375039f,  -0.843911f,//416
      0.386187f,   0.238677f,  -0.891007f,//417
      0.264083f,   0.301259f,  -0.916244f,//418
      0.133071f,   0.082242f,  -0.987688f,//419
      0.132792f,   0.220117f,  -0.966393f,//420
      0.606825f,   0.237086f,  -0.758652f,//421
      0.620240f,   0.081142f,  -0.780204f,//422
      0.519258f,   0.156434f,  -0.840178f,//423
      0.783843f,   0.081086f,  -0.615642f,//424
      0.783843f,  -0.081086f,  -0.615642f,//425
      0.711282f,   0.000000f,  -0.702907f,//426
      0.606825f,  -0.237086f,  -0.758652f,//427
      0.519258f,  -0.156434f,  -0.840178f,//428
      0.620240f,  -0.081142f,  -0.780204f,//429
      0.383614f,  -0.375039f,  -0.843911f,//430
      0.264083f,  -0.301259f,  -0.916244f,//431
      0.386187f,  -0.238677f,  -0.891007f,//432
      0.131200f,  -0.484442f,  -0.864929f,//433
      0.131655f,  -0.358229f,  -0.924305f,//434
      0.133071f,  -0.082242f,  -0.987688f,//435
      0.132792f,  -0.220117f,  -0.966393f,//436
      0.399607f,   0.082324f,  -0.912983f,//437
      0.399607f,  -0.082324f,  -0.912983f,//438
      0.266405f,   0.000000f,  -0.963861f,//439
      0.792649f,   0.213023f,  -0.571252f,//440
      0.864929f,   0.131200f,  -0.484442f,//441
      0.601501f,   0.371748f,  -0.707107f,//442
      0.702310f,   0.296005f,  -0.647412f,//443
      0.757935f,   0.468430f,  -0.453991f,//444
      0.843911f,   0.383614f,  -0.375039f,//445
      0.783452f,   0.346153f,  -0.516122f,//446
      0.346153f,   0.516122f,  -0.783452f,//447
      0.468430f,   0.453991f,  -0.757935f,//448
      0.213023f,   0.571252f,  -0.792649f,//449
      0.371748f,   0.707107f,  -0.601501f,//450
      0.296005f,   0.647412f,  -0.702310f,//451
      0.516122f,   0.783452f,  -0.346153f,//452
      0.453991f,   0.757935f,  -0.468430f,//453
      0.615642f,   0.783843f,  -0.081086f,//454
      0.571252f,   0.792649f,  -0.213023f,//455
      0.758652f,   0.606825f,  -0.237086f,//456
      0.707107f,   0.601501f,  -0.371748f,//457
      0.647412f,   0.702310f,  -0.296005f,//458
      0.564254f,   0.513375f,  -0.646578f,//459
      0.513375f,   0.646578f,  -0.564254f,//460
      0.646578f,   0.564254f,  -0.513375f,//461
      0.864929f,   0.131200f,   0.484442f,//462
      0.792649f,   0.213023f,   0.571252f,//463
      0.783843f,   0.081086f,   0.615642f,//464
      0.843911f,   0.383614f,   0.375039f,//465
      0.757935f,   0.468430f,   0.453991f,//466
      0.783452f,   0.346153f,   0.516122f,//467
      0.601501f,   0.371748f,   0.707107f,//468
      0.606825f,   0.237086f,   0.758652f,//469
      0.702310f,   0.296005f,   0.647412f,//470
      0.758652f,   0.606825f,   0.237086f,//471
      0.647412f,   0.702310f,   0.296005f,//472
      0.707107f,   0.601501f,   0.371748f,//473
      0.615642f,   0.783843f,   0.081086f,//474
      0.571252f,   0.792649f,   0.213023f,//475
      0.453991f,   0.757935f,   0.468430f,//476
      0.516122f,   0.783452f,   0.346153f,//477
      0.296005f,   0.647412f,   0.702310f,//478
      0.371748f,   0.707107f,   0.601501f,//479
      0.131200f,   0.484442f,   0.864929f,//480
      0.213023f,   0.571252f,   0.792649f,//481
      0.383614f,   0.375039f,   0.843911f,//482
      0.468430f,   0.453991f,   0.757935f,//483
      0.346153f,   0.516122f,   0.783452f,//484
      0.646578f,   0.564254f,   0.513375f,//485
      0.513375f,   0.646578f,   0.564254f,//486
      0.564254f,   0.513375f,   0.646578f,//487
      0.702907f,   0.711282f,   0.000000f,//488
      0.840178f,   0.519258f,   0.156434f,//489
      0.780204f,   0.620240f,   0.081142f,//490
      0.840178f,   0.519258f,  -0.156434f,//491
      0.780204f,   0.620240f,  -0.081142f,//492
      0.916244f,   0.264083f,   0.301259f,//493
      0.891007f,   0.386187f,   0.238677f,//494
      0.915043f,   0.000000f,   0.403355f,//495
      0.924305f,   0.131655f,   0.358229f,//496
      0.990439f,   0.000000f,   0.137952f,//497
      0.987688f,   0.133071f,   0.082242f,//498
      0.966393f,   0.132792f,   0.220117f,//499
      0.990439f,   0.000000f,  -0.137952f,//500
      0.966393f,   0.132792f,  -0.220117f,//501
      0.987688f,   0.133071f,  -0.082242f,//502
      0.915043f,   0.000000f,  -0.403355f,//503
      0.924305f,   0.131655f,  -0.358229f,//504
      0.891007f,   0.386187f,  -0.238677f,//505
      0.916244f,   0.264083f,  -0.301259f,//506
      0.912983f,   0.399607f,   0.082324f,//507
      0.963861f,   0.266405f,   0.000000f,//508
      0.912983f,   0.399607f,  -0.082324f,//509
      0.131200f,  -0.484442f,   0.864929f,//510
      0.131655f,  -0.358229f,   0.924305f,//511
      0.383614f,  -0.375039f,   0.843911f,//512
      0.386187f,  -0.238677f,   0.891007f,//513
      0.264083f,  -0.301259f,   0.916244f,//514
      0.133071f,  -0.082242f,   0.987688f,//515
      0.132792f,  -0.220117f,   0.966393f,//516
      0.606825f,  -0.237086f,   0.758652f,//517
      0.620240f,  -0.081142f,   0.780204f,//518
      0.519258f,  -0.156434f,   0.840178f,//519
      0.783843f,  -0.081086f,   0.615642f,//520
      0.711282f,   0.000000f,   0.702907f,//521
      0.519258f,   0.156434f,   0.840178f,//522
      0.620240f,   0.081142f,   0.780204f,//523
      0.264083f,   0.301259f,   0.916244f,//524
      0.386187f,   0.238677f,   0.891007f,//525
      0.131655f,   0.358229f,   0.924305f,//526
      0.133071f,   0.082242f,   0.987688f,//527
      0.132792f,   0.220117f,   0.966393f,//528
      0.399607f,  -0.082324f,   0.912983f,//529
      0.399607f,   0.082324f,   0.912983f,//530
      0.266405f,   0.000000f,   0.963861f,//531
      0.792649f,  -0.213023f,   0.571252f,//532
      0.864929f,  -0.131200f,   0.484442f,//533
      0.601501f,  -0.371748f,   0.707107f,//534
      0.702310f,  -0.296005f,   0.647412f,//535
      0.757935f,  -0.468430f,   0.453991f,//536
      0.843911f,  -0.383614f,   0.375039f,//537
      0.783452f,  -0.346153f,   0.516122f,//538
      0.346153f,  -0.516122f,   0.783452f,//539
      0.468430f,  -0.453991f,   0.757935f,//540
      0.081086f,  -0.615642f,   0.783843f,//541
      0.213023f,  -0.571252f,   0.792649f,//542
      0.237086f,  -0.758652f,   0.606825f,//543
      0.371748f,  -0.707107f,   0.601501f,//544
      0.296005f,  -0.647412f,   0.702310f,//545
      0.375039f,  -0.843911f,   0.383614f,//546
      0.516122f,  -0.783452f,   0.346153f,//547
      0.453991f,  -0.757935f,   0.468430f,//548
      0.484442f,  -0.864929f,   0.131200f,//549
      0.615642f,  -0.783843f,   0.081086f,//550
      0.571252f,  -0.792649f,   0.213023f,//551
      0.758652f,  -0.606825f,   0.237086f,//552
      0.707107f,  -0.601501f,   0.371748f,//553
      0.647412f,  -0.702310f,   0.296005f,//554
      0.564254f,  -0.513375f,   0.646578f,//555
      0.513375f,  -0.646578f,   0.564254f,//556
      0.646578f,  -0.564254f,   0.513375f,//557
      0.615642f,  -0.783843f,  -0.081086f,//558
      0.702907f,  -0.711282f,   0.000000f,//559
      0.758652f,  -0.606825f,  -0.237086f,//560
      0.840178f,  -0.519258f,  -0.156434f,//561
      0.780204f,  -0.620240f,  -0.081142f,//562
      0.840178f,  -0.519258f,   0.156434f,//563
      0.780204f,  -0.620240f,   0.081142f,//564
      0.843911f,  -0.383614f,  -0.375039f,//565
      0.916244f,  -0.264083f,  -0.301259f,//566
      0.891007f,  -0.386187f,  -0.238677f,//567
      0.864929f,  -0.131200f,  -0.484442f,//568
      0.924305f,  -0.131655f,  -0.358229f,//569
      0.987688f,  -0.133071f,  -0.082242f,//570
      0.966393f,  -0.132792f,  -0.220117f,//571
      0.966393f,  -0.132792f,   0.220117f,//572
      0.987688f,  -0.133071f,   0.082242f,//573
      0.924305f,  -0.131655f,   0.358229f,//574
      0.891007f,  -0.386187f,   0.238677f,//575
      0.916244f,  -0.264083f,   0.301259f,//576
      0.912983f,  -0.399607f,  -0.082324f,//577
      0.963861f,  -0.266405f,   0.000000f,//578
      0.912983f,  -0.399607f,   0.082324f,//579
      0.792649f,  -0.213023f,  -0.571252f,//580
      0.757935f,  -0.468430f,  -0.453991f,//581
      0.783452f,  -0.346153f,  -0.516122f,//582
      0.601501f,  -0.371748f,  -0.707107f,//583
      0.702310f,  -0.296005f,  -0.647412f,//584
      0.647412f,  -0.702310f,  -0.296005f,//585
      0.707107f,  -0.601501f,  -0.371748f,//586
      0.484442f,  -0.864929f,  -0.131200f,//587
      0.571252f,  -0.792649f,  -0.213023f,//588
      0.375039f,  -0.843911f,  -0.383614f,//589
      0.453991f,  -0.757935f,  -0.468430f,//590
      0.516122f,  -0.783452f,  -0.346153f,//591
      0.237086f,  -0.758652f,  -0.606825f,//592
      0.296005f,  -0.647412f,  -0.702310f,//593
      0.371748f,  -0.707107f,  -0.601501f,//594
      0.081086f,  -0.615642f,  -0.783843f,//595
      0.213023f,  -0.571252f,  -0.792649f,//596
      0.468430f,  -0.453991f,  -0.757935f,//597
      0.346153f,  -0.516122f,  -0.783452f,//598
      0.646578f,  -0.564254f,  -0.513375f,//599
      0.513375f,  -0.646578f,  -0.564254f,//600
      0.564254f,  -0.513375f,  -0.646578f,//601
      0.403355f,  -0.915043f,   0.000000f,//602
      0.358229f,  -0.924305f,  -0.131655f,//603
      0.137952f,  -0.990439f,   0.000000f,//604
      0.082242f,  -0.987688f,  -0.133071f,//605
      0.220117f,  -0.966393f,  -0.132792f,//606
      0.238677f,  -0.891007f,  -0.386187f,//607
      0.301259f,  -0.916244f,  -0.264083f,//608
      -0.137952f,  -0.990439f,   0.000000f,//609
      -0.220117f,  -0.966393f,  -0.132792f,//610
      -0.082242f,  -0.987688f,  -0.133071f,//611
      -0.403355f,  -0.915043f,   0.000000f,//612
      -0.358229f,  -0.924305f,  -0.131655f,//613
      -0.238677f,  -0.891007f,  -0.386187f,//614
      -0.301259f,  -0.916244f,  -0.264083f,//615
      -0.081142f,  -0.780204f,  -0.620240f,//616
      -0.156434f,  -0.840178f,  -0.519258f,//617
      0.000000f,  -0.702907f,  -0.711282f,//618
      0.156434f,  -0.840178f,  -0.519258f,//619
      0.081142f,  -0.780204f,  -0.620240f,//620
      0.000000f,  -0.963861f,  -0.266405f,//621
      -0.082324f,  -0.912983f,  -0.399607f,//622
      0.082324f,  -0.912983f,  -0.399607f,//623
      -0.358229f,  -0.924305f,   0.131655f,//624
      -0.082242f,  -0.987688f,   0.133071f,//625
      -0.220117f,  -0.966393f,   0.132792f,//626
      -0.238677f,  -0.891007f,   0.386187f,//627
      -0.301259f,  -0.916244f,   0.264083f,//628
      0.220117f,  -0.966393f,   0.132792f,//629
      0.082242f,  -0.987688f,   0.133071f,//630
      0.358229f,  -0.924305f,   0.131655f,//631
      0.238677f,  -0.891007f,   0.386187f,//632
      0.301259f,  -0.916244f,   0.264083f,//633
      0.081142f,  -0.780204f,   0.620240f,//634
      0.156434f,  -0.840178f,   0.519258f,//635
      0.000000f,  -0.702907f,   0.711282f,//636
      -0.156434f,  -0.840178f,   0.519258f,//637
      -0.081142f,  -0.780204f,   0.620240f,//638
      0.000000f,  -0.963861f,   0.266405f,//639
      0.082324f,  -0.912983f,   0.399607f,//640
      -0.082324f,  -0.912983f,   0.399607f //641
  };
  //long long int osize=vert.size();

  for (int i=0; i<1280; i++){
    if ((mol->LOD==2)&&(i>=80)) break;
    if ((mol->LOD<upper_limit)&&(i>=320)) break;
    int i1 = (mol->LOD==2)?idx1[i*3]  :(mol->LOD<upper_limit)?idx2[i*3]  :idx3[i*3];
    int i2 = (mol->LOD==2)?idx1[i*3+1]:(mol->LOD<upper_limit)?idx2[i*3+1]:idx3[i*3+1];
    int i3 = (mol->LOD==2)?idx1[i*3+2]:(mol->LOD<upper_limit)?idx2[i*3+2]:idx3[i*3+2];
    QVector4D qv(v[i1*3],v[i1*3+1],v[i1*3+2],1);
    QVector4D qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);

    vert.append(v[i1*3]);
    vert.append(v[i1*3+1]);
    vert.append(v[i1*3+2]);
    vert.append(style);

    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);

    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(i1);

    qv = QVector4D(v[i2*3],v[i2*3+1],v[i2*3+2],1);
    qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);

    vert.append(v[i2*3]);
    vert.append(v[i2*3+1]);
    vert.append(v[i2*3+2]);
    vert.append(style);

    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);

    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(i1);

    qv = QVector4D(v[i3*3],v[i3*3+1],v[i3*3+2],1);
    qvp = adpmat * qv;
    vert.append(qvp.x());
    vert.append(qvp.y());
    vert.append(qvp.z());
    vert.append(an);

    vert.append(v[i3*3]);
    vert.append(v[i3*3+1]);
    vert.append(v[i3*3+2]);
    vert.append(style);

    vert.append(mitte.x);
    vert.append(mitte.y);
    vert.append(mitte.z);
    vert.append(_sgi);

    vert.append(_prt);
    vert.append(ph);
    vert.append(resi);
    vert.append(i1);
  }
  //printf("ball %lld\n", vert.size() - osize);
}

void PlayOpenGL::updateAtoms(){
  if (AMD_Hack)makeAtoms(mol->showatoms);
  update();

}

void PlayOpenGL::makeAtomsWOGS(const CEnvironment &atoms){
  if (verbosity >2 ) printf("makeAtomsWOGS\n");

  const int HAS_WALLS = 1;
  const int HAS_RINGS = 2;
  const int HAS_SPHERE = 4;
  const int IS_SOLID = 8;
  const int HAS_WHITERING = 16;
  const int HAS_NOLABEL = 32;
  const int HAS_NOADP = 64;
  const int HAS_PLAID = 128;
  const int IS_METAL = 256;
  const int IS_HIDDEN = 512;
  const int IS_DEUTERIUM = 1024;
  const int IS_ISO = 2048;
  QVector<float> vert;

  rotze = (!isMainWindow)?0:(rotze>=atoms.size())?-1:rotze;

  makeCurrent();
  GLenum err = GL_NO_ERROR;
  while((err = glGetError()) != GL_NO_ERROR){
    printf("before makeAtoms %s error %X\n", parent()->objectName().toStdString().c_str(),err);

  }
  atom_vao.release();
  if (atom_vao.isCreated()) {
    atom_vao.release();
    atom_vao.destroy();
    atom_vertexBuffer.destroy();
    //atom_indexBuffer.destroy();
    atom_program->removeAllShaders();
    atom_program->release();
    glFlush();
  }
  makeLabels();
  int zz=0;
  sumse=V3(0,0,0);
  for (int i = 0; i < atoms.size(); i++){
    if ((!atoms.at(i).hidden)) {//(mol->showatoms.at(i).an>-1)&&
      sumse += atoms[i].pos;
      zz++;
    }
  }
  if (zz) sumse*=1.0/zz;
  if (false){
    pushRotStat();
    popRotStat();
    makeSelectedAtoms();
    updateScreenAtoms();
    makeLegendAtoms();
    //setViewAngle(va);
    update();
    while((err = glGetError()) != GL_NO_ERROR){
      printf("after2 makeAtoms %s error %X\n", parent()->objectName().toStdString().c_str(),err);
    }
    return;
  }
  bool wehaveh=false;
  for (int i = 0; i< atoms.size(); i++){
    if (atoms.at(i).hidden) continue;
    if ((atoms.at(i).an==-1)&&(atoms.at(i).peakHeight<qcutoff)) continue;
    if ((mol->nopm1)&&(atoms.at(i).symmGroup)&&(atoms.at(i).part<0))continue;
    if ((mol->noQPeaksPlease)&&(atoms.at(i).an==-1)) continue;
    if ((atoms.at(i).an==0)) wehaveh=true;
    //ball(QVector<float> &vert, int LOD, V3 mitte, int an, int _sgi, float _prt, int style)
    bool isd=((atoms.at(i).an==0)&&(atoms.at(i).Label[0]=='D'));
    int style = (atoms.at(i).an<0)? (atoms.at(i).hidden)?512:0:
                    mol->AtomStyle[atoms.at(i).an]
                        |((atoms.at(i).hidden)?512:0)
                        |((isd)?1024:0)
                        |(((atoms.at(i).isIso))?2048:0);
    if (atoms.at(i).Label.contains("noADP"))style |=ATOM_STYLE_NOADP;
    V3 ev = V3(0,0,0);
    bool npd = false;
    QMatrix4x4 mtest;
    mtest.setToIdentity();
    QMatrix4x4 r,scale;//(ev.x * fiftypct,0,0,0, 0,ev.y * fiftypct,0,0, 0,0,ev.z * fiftypct,0, 0,0,0,1);
    scale.setToIdentity();
    if ((atoms.at(i).an<0)||(!adp->isChecked())||((style &HAS_NOADP) == HAS_NOADP)) {
      float radi= (atoms.at(i).an<-65)?mol->qPeakRad*2:mol->qPeakRad;
      if (atoms.at(i).an<0) scale = QMatrix4x4(1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1)*radi;
      else if ((!adp->isChecked())||((style &HAS_NOADP)==HAS_NOADP)) {//||((style & IS_ISO) == IS_ISO)
          radi = (tubes->isChecked())?mol->bondStrength:mol->arad[int(atoms.at(i).an)];
          scale = QMatrix4x4(1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1)*radi;
      }
    }else {
    double *arr = mol->jacobi(atoms.at(i).uc, ev);
    npd = ((ev.x<0)||(ev.y<0)||(ev.z<0));
    mtest.rotate(arr[0], arr[1], arr[2], arr[3]);
    float fiftypct =1.54f;
    switch (mol->proba ) {
    case 10 :{ fiftypct = 0.76; break;}   //Hauptachsen der Eliipsoide 10% Wahrscheinlichkeit
    case 30 :{ fiftypct = 1.19; break;}   //Hauptachsen der Eliipsoide 30% Wahrscheinlichkeit
    case 50 :{ fiftypct = 1.54; break;}   //Hauptachsen der Eliipsoide 50% Wahrscheinlichkeit
    case 70 :{ fiftypct = 1.91; break;}   //Hauptachsen der Eliipsoide 70% Wahrscheinlichkeit
    case 90 :{ fiftypct = 2.50; break;}   //Hauptachsen der Eliipsoide 90% Wahrscheinlichkeit
    default: ;
    }
    scale.setToIdentity();
    scale.scale(sqrt(ev.x) * fiftypct,
                sqrt(ev.y) * fiftypct,
                sqrt(ev.z) * fiftypct);
    }
    if (!adp->isChecked()) style|=HAS_NOADP;
    QMatrix4x4 m = mtest * scale;
    if ((atoms.at(i).an >= 0) && (npd)){
      r = QMatrix4x4(0.202072,0,0,0, 0,0.202072,0,0, 0,0,0.202072,0, 0,0,0,1);
      style |= IS_SOLID;
      cube(vert, r, atoms.at(i).pos, atoms.at(i).an, atoms.at(i).sg, atoms.at(i).part, style, atoms.at(i).peakHeight, atoms.at(i).resiNr);
    } else {
      if (atoms.at(i).an==-1) icosaeder(vert, m, atoms.at(i).pos, atoms.at(i).an, atoms.at(i).sg, atoms.at(i).part, style, atoms.at(i).peakHeight, atoms.at(i).resiNr);//qpeaks
      else if (atoms.at(i).an==-66) dodecaeder(vert, m, atoms.at(i).pos, atoms.at(i).an, atoms.at(i).sg, atoms.at(i).part, style, atoms.at(i).peakHeight, atoms.at(i).resiNr);//centroids
      else if (atoms.at(i).an==-42) triacontaeder(vert, m, atoms.at(i).pos, atoms.at(i).an, atoms.at(i).sg, atoms.at(i).part, style, atoms.at(i).sof, atoms.at(i).resiNr);//belo
      else {
          if ((adp->isChecked())&&(HAS_WALLS==(HAS_WALLS&style))&&((IS_SOLID&style)==0)) pappe(vert, m, atoms.at(i).pos, atoms.at(i).an, atoms.at(i).sg, atoms.at(i).part, style, atoms.at(i).peakHeight, atoms.at(i).resiNr);
          if ((HAS_SPHERE&style)==HAS_SPHERE)
          {ball(vert, m, atoms.at(i).pos, atoms.at(i).an, atoms.at(i).sg, atoms.at(i).part, style, atoms.at(i).peakHeight, atoms.at(i).resiNr);}
      }//else qpeak centroids
    }//npd
  }
  hideHydrogen->setVisible(wehaveh);
  atom_array_size = vert.size() / 16; // one chunk has 16 floats

  atom_vao.create();
  atom_vao.bind();
  //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
  atom_vertexBuffer = QOpenGLBuffer(QOpenGLBuffer::VertexBuffer);
  atom_vertexBuffer.create();
  atom_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
  if (!atom_vertexBuffer.bind()) {printf("I was unable to bind atom_vertexBuffer.\n");}

  atom_vertexBuffer.allocate(vert.data(), vert.size() * sizeof(float));

  if (!atom_program) atom_program = new QOpenGLShaderProgram(this);
  else atom_program->removeAllShaders();

  bool result = atom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/atomwog.vert");
  //bool result = atom_program->addShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/atom.vert");

  if ( !result ) qWarning() << atom_program->log();

  result = atom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/atom.frag");
  //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/atom.frag");

  if ( !result ) qWarning() << atom_program->log();
  result = atom_program->link();
  if ( !result ) qWarning() << atom_program->log();
  \
      if ( !atom_program->bind() )    {
    qWarning() << "Could not bind shader program to context";
    return;
  }

  programInfo(atom_program,"without GS atom");
  GLfloat acv[110*3];
  for (int i=0;i<109;i++) {
    acv[3*i]=mol->atomColor(i,false).redF();
    acv[3*i+1]=mol->atomColor(i,false).greenF();
    acv[3*i+2]=mol->atomColor(i,false).blueF();
  }

  atom_program->setUniformValueArray("atomColors",acv, 109,3);
  atom_program->enableAttributeArray(0);
  atom_program->enableAttributeArray(1);
  atom_program->enableAttributeArray(2);
  atom_program->enableAttributeArray(3);

  int stride = 4 * 4 * sizeof(float);
  atom_program->setAttributeBuffer(0, GL_FLOAT, 0, 4, stride);
  atom_program->setAttributeBuffer(1, GL_FLOAT, sizeof(float)*4  , 4, stride);
  atom_program->setAttributeBuffer(2, GL_FLOAT, sizeof(float)*8  , 4, stride);
  atom_program->setAttributeBuffer(3, GL_FLOAT, sizeof(float)*12 , 4, stride);

  atom_vao.release();

  //atom_program->disableAttributeArray(0);
  //atom_program->disableAttributeArray(1);
  //atom_program->disableAttributeArray(2);
  //atom_program->disableAttributeArray(3);
  mk2++;
  atom_program->release();
  atom_vertexBuffer.release();
  atom_vertexBuffer.destroy();
  while((err = glGetError()) != GL_NO_ERROR){
    printf("after1 makeAtomsWOGS %s error %X\n", parent()->objectName().toStdString().c_str(),err);
  }
  pushRotStat();
  popRotStat();
  makeSelectedAtoms();
  updateScreenAtoms();
  makeLegendAtoms();
  update();
  while((err = glGetError()) != GL_NO_ERROR){
    printf("after2 makeAtomsWOGS %s error %X\n", parent()->objectName().toStdString().c_str(),err);
  }
  //lsqpln();
}

void PlayOpenGL::voronoij(CEnvironment au, int intat){
  double dk,range=5.0;
  QString voroMsg;
  voroMsg.clear();
  voroMsg.append("<b>Voronoi:</b><br>");
  //QTime speedTest;
  //speedTest.start();
  vtriangles.clear();
  VPoly triangle;
  //vorobas=1;
  V3 prime,dp,D,floorD;
  QList<SdmItem> sdm;
  SdmItem sdmItem;
  sdmItem.a1=0;
  sdmItem.a2=1;
  sdmItem.sn=0;
  //sdmItem.p1=sdmItem.p2=V3(0,0,0);
  for (int i=0; i<au.size(); i++){
    for (int j=0; j<au.size(); j++ ){
      if ((au.at(i).sg)||(au.at(j).sg)) continue;
      //  bool hma=false;
      for (int n=0;n<mol->cell.symmops.size();  n++){
        prime=mol->cell.symmops.at(n) * au.at(i).frac + mol->cell.trans.at(n);
        D=prime - au.at(j).frac+ V3(0.5,0.5,0.5) ;
        floorD=V3(floor(D.x),floor(D.y),floor(D.z));
        for (int h=-1; h<2; h++){
          for (int k=-1; k<2; k++){
            for (int l=-1; l<2; l++){
              V3 fD=floorD+V3(h,k,l);
              dp=D - fD - V3(0.5,0.5,0.5);
              dk = mol->fl(dp.x,dp.y,dp.z);
              if ((dk>0.01)&&((range)>=dk)){
                sdmItem.d=dk;
                sdmItem.floorD=fD;
                sdmItem.a1=i;
                sdmItem.a2=j;
                sdmItem.sn=n;
                sdm.append(sdmItem);
              }
            }
          }
        }
      }
    }
  }
#if (QT_VERSION >= 0x050000)
  std::sort(sdm.begin(),sdm.end());
#else
  qSort(sdm.begin(),sdm.end());
#endif
  //SDMprint(sdm,au);
  QList<V3> n,m,doneLine;
  QList<int> intra, altcol;
  QList<Vert> v;
  Vert vert;
  V3 pf,pc,mx,nx,of,oc;
  Matrix mat;
  double vol=0.0,avol,tvol;
  for (int i=(intat>-1)?intat:0; i<au.size(); i++){
    if (au.at(i).an<0) continue;
    //find faces
    n.clear();
    m.clear();
    v.clear();
    doneLine.clear();
    intra.clear();
    avol=0.0;
    int faci=0;
    //int vcnt=0;
    mol->frac2kart(au.at(i).frac,oc);
    int tris=0;
    for (int j=0; j<sdm.size();j++){
      if(sdm.at(j).d<0.1) continue;
      if ((sdm.at(j).a2==i)&&(au[sdm.at(j).a1].an>-1)) {
        pf = mol->cell.symmops.at(sdm.at(j).sn) * au[sdm.at(j).a1].frac + mol->cell.trans.at(sdm.at(j).sn) - sdm.at(j).floorD;
        mol->frac2kart(pf,pc);
        mx =0.5*(oc+pc);
        nx = Normalize(pc-oc);
        bool exists = false;
        double dpl;
        for (int tp=0; tp<m.size(); tp++){
          dpl=n.at(tp)*(mx-m.at(tp));
          if (fabs(dpl)<0.001){exists=true;break;}
        }
        if (!exists){
          m.append(mx);
          n.append(nx);
          intra.append((sdm.at(j).sn==0)&&(Norm(sdm.at(j).floorD)==0.0));
          if (intat>-1)altcol.append(au[sdm.at(j).a1].an);
        }
      }//is neighbour of i
    }// j sdm
    for (int h=0; h<n.size()-2; h++)
      for (int k=h+1; k<n.size()-1; k++)
        for (int l=k+1; l<n.size(); l++){
          mat=Matrix(n.at(h),n.at(k),n.at(l));
          double d=determinant(mat),dpl;

          if (fabs(d)<0.0002) continue;

          //p = (dot(p1,n1)*cross(n2,n3)-dot(p2,n2)*cross(n1,n3)+dot(p3,n3)*cross(n1,n2))/d
          bool out=false;
          vert.pos=(
              (m.at(h)*n.at(h))*(n.at(k)%n.at(l))+
              (m.at(k)*n.at(k))*(n.at(l)%n.at(h))+
              (m.at(l)*n.at(l))*(n.at(h)%n.at(k)))*(1.0/d);
          vert.faces[0]=h;
          vert.faces[1]=k;
          vert.faces[2]=l;
          for (int pli=0; pli<n.size(); pli++){//teste lage zu allen ebenen
            dpl=n.at(pli)*(vert.pos-m.at(pli));
            if (dpl>0.00001) {
              out=true;
              break;
            }
          }
          if (!out) {
            v.append(vert);

          }
        }
    for (int mi=0; mi<m.size();mi++){// this is to get the midpoint of a face in its actual center
      V3 mnew=V3(0,0,0);
      int poli=0;
      for (int vi=0; vi<v.size(); vi++){
        if ((v.at(vi).faces[0]==mi)||(v.at(vi).faces[1]==mi)||(v.at(vi).faces[2]==mi)){
          mnew+=v.at(vi).pos;
          poli++;
        }
      }
      mnew*=1.0/poli;

      if (poli) {m[mi]=mnew;faci++;}
    }

    for(int vi=0; vi<v.size()-1; vi++)
      for(int vj=vi+1; vj<v.size(); vj++){
        int fc[2];
        if (commonFaces(v.at(vi),v.at(vj),fc)==2){
          V3 lin=v.at(vi).pos+v.at(vj).pos;
          double dis;
          bool done=false;
          for (int di=0; di<doneLine.size(); di++){
            dis=Distance(lin,doneLine.at(di));
            if (dis<0.000001) {done=true; break;}
          }
          if (done) continue;
          doneLine.append(lin);
          tvol=determinant(Matrix(m.at(fc[0])-oc, v.at(vi).pos-oc,v.at(vj).pos-oc));
          int v1=(tvol<0)?vj:vi,
              v2=(tvol<0)?vi:vj;
          triangle.mid=m.at(fc[0]);
          triangle.nor=n.at(fc[0]);
          triangle.acol=(intat>-1)?altcol.at(fc[0]):au[i].an;
          triangle.intra=(intat>-1)?0:intra.at(fc[0]);
          triangle.verts0=v[v1].pos;
          triangle.verts1=v[v2].pos;

          vtriangles.append(triangle);

          tris++;
          double tv=fabs(tvol)/6.0;
          avol+=tv;
          vol+=tv*au[i].sof;
          tvol=determinant(Matrix(m.at(fc[1])-oc, v.at(vi).pos-oc,v.at(vj).pos-oc));
          v1=(tvol<0)?vj:vi;
          v2=(tvol<0)?vi:vj;
          triangle.mid=m.at(fc[1]);
          triangle.nor=n.at(fc[1]);

          triangle.acol=(intat>-1)?altcol.at(fc[1]):au[i].an;
          triangle.intra=(intat>-1)?0:intra.at(fc[1]);
          //         triangle.col[3]=((intat>-1)&&(intra.at(fc[1])))?0.8:0.4;
          triangle.verts0=v[v1].pos;
          triangle.verts1=v[v2].pos;
          vtriangles.append(triangle);
          tris++;
          tv=fabs(tvol)/6.0;
          avol+=tv;
          vol+=tv*au[i].sof;
        }
      }
    if (verbosity >2 ) printf("%-12s: Triangles %6d Volume %18.5f\n", au.at(i).Label.toStdString().c_str(), tris, avol);
    voroMsg.append(QString("%1 : EulerTest: %2 Volume %3 &Aring;<sup>3</sup>.<br>").arg(au.at(i).Label).arg(-tris/2+faci+v.size()).arg(avol));

    if (intat>-1) break;
  }//i atoms au
  if (intat==-1) {
    if (verbosity >2 ) printf("Total Volume= %18.5f Total Volume * z = %18.5f cell volume %18.5f delta =%f%%\n",
           vol,
           vol*mol->cell.symmops.size(),
           mol->cell.V,
           (vol*mol->cell.symmops.size()-mol->cell.V)/mol->cell.V*100.0);
    voroMsg.append(QString("Total Volume= %1 &Aring;<sup>3</sup> cell volume %2 &Aring;<sup>3</sup> &Delta;V %3%<br>")
                   .arg(vol*mol->cell.symmops.size())
                   .arg(mol->cell.V)
                   .arg((vol*mol->cell.symmops.size()-mol->cell.V)/mol->cell.V*100.0));
  }
  emit bigmessage(voroMsg);

 // eye=auge;
  /*
#if (QT_VERSION >= 0x050000)
  std::sort(vtriangles.begin(),vtriangles.end(),vbyeye);
#else
  qSort(vtriangles.begin(),vtriangles.end(),vbyeye);
#endif
*/
  QVector<float> vp;
  for (int ti=0; ti<vtriangles.size(); ti++){
      QColor c=mol->atomColor(vtriangles.at(ti).acol);
      vp.append(vtriangles.at(ti).mid.x);
      vp.append(vtriangles.at(ti).mid.y);
      vp.append(vtriangles.at(ti).mid.z);
      vp.append(static_cast<float>(vtriangles.at(ti).acol)+0.99*c.blueF());

      vp.append(vtriangles.at(ti).nor.x);
      vp.append(vtriangles.at(ti).nor.y);
      vp.append(vtriangles.at(ti).nor.z);
      vp.append(static_cast<float>(vtriangles.at(ti).intra));

      vp.append(vtriangles.at(ti).verts0.x);
      vp.append(vtriangles.at(ti).verts0.y);
      vp.append(vtriangles.at(ti).verts0.z);
      vp.append(c.redF());

      vp.append(vtriangles.at(ti).verts1.x);
      vp.append(vtriangles.at(ti).verts1.y);
      vp.append(vtriangles.at(ti).verts1.z);
      vp.append(c.greenF());

  }

  makeCurrent();
  voro_vao.release();
  if (voro_vao.isCreated()) {
      voro_vao.release();
      voro_vao.destroy();
      voro_vertexBuffer.destroy();
      //voro_indexBuffer.destroy();
  }
  voro_vao.create();
  voro_vao.bind();
  voro_vertexBuffer.create();
  voro_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
  //glDisable(GL_BLEND);
  if (!voro_vertexBuffer.bind()) {printf("voro_vertexBuffer: I was unable to bind!\n");}
  voro_vertexBuffer.allocate(vp.data(), vp.size()*sizeof(float));
  if (!voro_program) voro_program = new QOpenGLShaderProgram(this);
  else voro_program->removeAllShaders();


  bool result = voro_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/voro.vert");

  if ( !result ) qWarning() << voro_program->log();
  result = voro_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/voro.geom");

  if ( !result ) qWarning() << voro_program->log();
  result = voro_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/voro.frag");

  if ( !result ) qWarning() << voro_program->log();
  result = voro_program->link();
 if ( !result ) qWarning() << voro_program->log();
  if ( !voro_program->bind() )    {
      qWarning() << "Could not bind shader program to context";
      return;
  }


  programInfo(voro_program,"voro");

  voro_program->enableAttributeArray(0);//mid
  voro_program->enableAttributeArray(1);//nor
  voro_program->enableAttributeArray(2);//vert0
  voro_program->enableAttributeArray(3);//vert1

  int stride = 4 * 4 * sizeof(float);
  voro_program->setAttributeBuffer(0, GL_FLOAT, 0, 4, stride);
  voro_program->setAttributeBuffer(1, GL_FLOAT, sizeof(float)*4  , 4, stride);
  voro_program->setAttributeBuffer(2, GL_FLOAT, sizeof(float)*8  , 4, stride);
  voro_program->setAttributeBuffer(3, GL_FLOAT, sizeof(float)*12  , 4, stride);


  voro_program->release();
  voro_vao.release();
  voro_vertexBuffer.release();
  voro_vertexBuffer.destroy();
  update();

}//voronoij
/*V3 eye;
bool PlayOpenGL::vbyeye(VPoly a,VPoly b){
  //return (Distance((a.mid+a.verts0+a.verts1)*0.33333,eye)>Distance((b.mid+b.verts0+b.verts1)*0.33333,eye));
  return (mol->Distance(a.mid,eye)>Distance(b.mid,eye));
}*/



void PlayOpenGL::connectSelection(){//!creates a bond betwen two selected atoms.
  if (mol->selectedatoms.size()!=2) return;
  MyBond b;
  b.a1=mol->selectedatoms.at(0).style;
  b.a2=mol->selectedatoms.at(1).style;
  b.ato1=&mol->showatoms[mol->selectedatoms.at(0).style];
  b.ato2=&mol->showatoms[mol->selectedatoms.at(1).style];
  b.length=sqrt(Distance(b.ato1->pos,b.ato2->pos));
  mol->showbonds.append(b);
  mol->selectedatoms.clear();
  updateBondActions(__LINE__);
  update();
}

void PlayOpenGL::disConnectSelection(int index){//! destroyes the bond between two selected atoms.
  mol->showbonds.removeAt(index);
  mol->selectedatoms.clear();
  updateBondActions(__LINE__);
  update();
}

void PlayOpenGL::disSelection(){//! Un select everything.
  mol->selectedatoms.clear();
  //apair.clear();
  updateBondActions(__LINE__);
  update();
}

void PlayOpenGL::addANIS(){
  emit insertANIS(mol->selectedatoms);
  mol->selectedatoms.clear();
}

void PlayOpenGL::addCHIV(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("CHIV: Geometrical restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *volline = new QLineEdit(dia);
  QLineEdit *esd1line = new QLineEdit(dia);
  QLabel *vol = new QLabel("Volume:", dia);
  QLabel *esd1 = new QLabel("ESD:", dia);
  // Set default value for distance and esd
  volline->setText("0.0");
  esd1line->setText("0.1");
  int i(0); // index for position of widgets in the container
  grid->addWidget(vol,i,0);
  grid->addWidget(volline,i++,1);
  grid->addWidget(esd1,i,0);
  grid->addWidget(esd1line,i++,1);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,i,0);
    grid->addWidget(resiCom,i++,1);
  }

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,i,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertCHIV(volline->text().toDouble(),esd1line->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }
}

void PlayOpenGL::addDANG(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("DANG: Geometrical restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *valueline = new QLineEdit(dia);
  QLineEdit *esdline = new QLineEdit(dia);
  QLabel *value = new QLabel("Value:", dia);
  QLabel *esd = new QLabel("ESD:", dia);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,2,0);
    grid->addWidget(resiCom,2,1);
  }
  // Set default value for distance and esd
  // Get value for DANG. Current bond distance and default esd - 0.04
  valueline->setText(QString::number(sqrt(Distance(mol->selectedatoms.at(0).pos,mol->selectedatoms.at(1).pos)),10,3));
  esdline->setText("0.04");

  grid->addWidget(value,0,0);
  grid->addWidget(valueline,0,1);
  grid->addWidget(esd,1,0);
  grid->addWidget(esdline,1,1);

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,4,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertDANG(valueline->text().toDouble(),esdline->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }

}

void PlayOpenGL::addDELU(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("DELU: ADP restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *esd1line = new QLineEdit(dia);
  QLineEdit *esd2line = new QLineEdit(dia);
  QLabel *esd1 = new QLabel("ESD1:", dia);
  QLabel *esd2 = new QLabel("ESD2:", dia);
  // Set default value for distance and esd
  esd1line->setText("0.01");
  esd2line->setText("0.01");

  grid->addWidget(esd1,0,0);
  grid->addWidget(esd1line,0,1);
  grid->addWidget(esd2,1,0);
  grid->addWidget(esd2line,1,1);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,3,0);
    grid->addWidget(resiCom,3,1);
  }

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,4,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertDELU(esd1line->text().toDouble(),esd2line->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }
}

void PlayOpenGL::addDFIX(){
    if (mol->selectedatoms.size()!=2)return;
  QDialog *dia = new QDialog();
  dia->setWindowTitle("DFIX: Geometrical restraint");
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *valueline = new QLineEdit(dia);
  valueline->setObjectName("DFIX_DIST");
  QLineEdit *esdline = new QLineEdit(dia);
  QLabel *value = new QLabel("Value:", dia);
  QLabel *esd = new QLabel("ESD:", dia);
  QLabel *atoms = new QLabel(QString("%1==%2  : %3 A")
                             .arg(mol->selectedatoms.at(0).Label)
                             .arg(mol->selectedatoms.at(1).Label)
                             .arg( sqrt(Distance(mol->selectedatoms.at(0).pos,mol->selectedatoms.at(1).pos)),10,'f',3));
  QCheckBox *bindit = new QCheckBox("Insert BIND instruction",dia);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    // Usually you want to define residue restraints with numbers at the atom:
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,3,0);
    grid->addWidget(resiCom,3,1);
  }
  // Set default value for distance and esd
  // Get value for DFIX. Current bond distance and default esd - 0.02
  valueline->setText(QString::number(sqrt(Distance(mol->selectedatoms.at(0).pos,mol->selectedatoms.at(1).pos)),10,3));
  esdline->setText("0.02");
  grid->addWidget(atoms,0,0,1,2);
  grid->addWidget(value,1,0);
  grid->addWidget(valueline,1,1);
  grid->addWidget(esd,2,0);
  grid->addWidget(esdline,2,1);
  grid->addWidget(bindit,4,0);
  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;
  QPushButton *singleBondButton = new QPushButton(QString("%1 A").arg(0.01*(mol->Kovalenz_Radien[mol->selectedatoms.at(0).an]+mol->Kovalenz_Radien[mol->selectedatoms.at(1).an])));

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);
  buttonBox->addButton(singleBondButton,QDialogButtonBox::ActionRole);


  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,10,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));
  connect(singleBondButton, SIGNAL(clicked()), this, SLOT(singleBondDistance()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertDFIX(valueline->text().toDouble(),esdline->text().toDouble(),mol->selectedatoms,s);
    if (bindit->isChecked()) emit bindthem();
    mol->selectedatoms.clear();
  }
  else  {
    if(a == QDialog::Rejected)
      dia->close();
  }
  //    updateBondActions(__LINE__);
}

void PlayOpenGL::addEADP(){
  emit insertEADP(mol->selectedatoms);
  mol->selectedatoms.clear();
}

void PlayOpenGL::addEXYZ(){
  emit insertEXYZ(mol->selectedatoms);
  mol->selectedatoms.clear();
}

void PlayOpenGL::addFLAT(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("FLAT: Geometrical restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *esdline = new QLineEdit(dia);
  QLabel *esd = new QLabel("ESD:", dia);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,2,0);
    grid->addWidget(resiCom,2,1);
  }
  // Set default value for esd
  esdline->setText("0.1");

  grid->addWidget(esd,0,0);
  grid->addWidget(esdline,0,1);

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,1,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertFLAT(esdline->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }

}

void PlayOpenGL::addISOR(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("ISOR: ADP restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *esd1line = new QLineEdit(dia);
  QLineEdit *esd2line = new QLineEdit(dia);
  QLabel *esd1 = new QLabel("ESD1:", dia);
  QLabel *esd2 = new QLabel("ESD2:", dia);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,3,0);
    grid->addWidget(resiCom,3,1);
  }
  // Set default values for s and st
  esd1line->setText("0.1");
  esd2line->setText("0.2");

  grid->addWidget(esd1,0,0);
  grid->addWidget(esd1line,0,1);
  grid->addWidget(esd2,1,0);
  grid->addWidget(esd2line,1,1);

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,4,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertISOR(esd1line->text().toDouble(),esd2line->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }
}

void PlayOpenGL::addRIGU(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("RIGU: ADP restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *esd1line = new QLineEdit(dia);
  QLineEdit *esd2line = new QLineEdit(dia);
  QLabel *esd1 = new QLabel("ESD1:", dia);
  QLabel *esd2 = new QLabel("ESD2:", dia);
  // Set default value for distance and esd
  esd1line->setText("0.004");
  esd2line->setText("0.004");
  int i(0); // index for position of widgets in the container
  grid->addWidget(esd1,i,0);
  grid->addWidget(esd1line,i++,1);
  grid->addWidget(esd2,i,0);
  grid->addWidget(esd2line,i++,1);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific",dia);
    resiCom = new QComboBox(dia);
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,i,0);
    grid->addWidget(resiCom,i++,1);
  }

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,i,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertRIGU(esd1line->text().toDouble(),esd2line->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }
}

void PlayOpenGL::addSADI() {
  //! Adds SADI restraints between the selected atom and all bonded atoms.
  QAction *action = qobject_cast<QAction *>(sender());
  int index=-1;
  if (action)
    index=action->data().toInt();
  else index=-1;
  if (index>mol->asymm.size()) index=-1;
  if (index!=-1) {
    emit insertSADI(index);
  }
}

QList<int> PlayOpenGL::lastClicked(){
  QList<int> pppp;
  //if ((ppp>=0)&&(ppp<mol->showatoms.size())&&(ppp!=pp)&&(ppp!=p)) pppp.append(ppp);
  if ((  p>=0)&&(  p<mol->showatoms.size())) pppp.append(p);
  if (( pp>=0)&&( pp<mol->showatoms.size())&&(pp!=p)) pppp.append(pp);
  return pppp;
}


void PlayOpenGL::addSIMU(){
  QDialog *dia = new QDialog();
  dia->setWindowTitle(tr("SIMU: ADP restraint"));
  dia->setModal(false);

  QGridLayout *grid = new QGridLayout(dia);
  QLineEdit *esd1line = new QLineEdit(dia);
  QLineEdit *esd2line = new QLineEdit(dia);
  QLineEdit *dmaxline = new QLineEdit(dia);
  QLabel *esd1 = new QLabel("ESD1:", dia);
  QLabel *esd2 = new QLabel("ESD2:", dia);
  QLabel *dmax = new QLabel("dmax:", dia);
  // Set default value for distance and esd
  esd1line->setText("0.04");
  esd2line->setText("0.08");
  dmaxline->setText("2.00");

  grid->addWidget(esd1,0,0);
  grid->addWidget(esd1line,0,1);
  grid->addWidget(esd2,1,0);
  grid->addWidget(esd2line,1,1);
  grid->addWidget(dmax,2,0);
  grid->addWidget(dmaxline,2,1);
  QStringList rl = resis();
  QComboBox *resiCom = NULL;
  if (!rl.isEmpty()){
    QLabel *resi = new QLabel("Residue specific", dia);
    resiCom = new QComboBox(dia);
    // Usually you want to define residue restraints with numbers at the atom:
    resiCom->addItem("");
    resiCom->addItems(rl);
    grid->addWidget(resi,3,0);
    grid->addWidget(resiCom,3,1);
  }

  QDialogButtonBox *buttonBox = new QDialogButtonBox();

  QPushButton *OkButton;
  QPushButton *CancelButton;

  OkButton = buttonBox->addButton(QDialogButtonBox::Ok);
  CancelButton = buttonBox->addButton(QDialogButtonBox::Cancel);

  buttonBox->setOrientation(Qt::Horizontal);

  grid->addWidget(buttonBox,4,0);

  dia->setLayout(grid);
  connect(OkButton, SIGNAL(clicked()), dia, SLOT(accept()));
  connect(CancelButton, SIGNAL(clicked()), dia, SLOT(reject()));

  dia->show();
  int a  = dia->exec();

  if(a == QDialog::Accepted)
  {
    // read value from value line
    QString s=(rl.isEmpty())?"":resiCom->currentText();
    emit insertSIMU(esd1line->text().toDouble(),esd2line->text().toDouble(),dmaxline->text().toDouble(),mol->selectedatoms,s);
    mol->selectedatoms.clear();
  }
  else
  {
    if(a == QDialog::Rejected)
      dia->close();
  }

}

QStringList PlayOpenGL::resis(){
  QStringList l;
  for (int i=0; i<mol->asymm.size(); i++){
      if (mol->asymm.at(i).an<0)continue;
    if (mol->asymm.at(i).resiNr>0){
      if (!l.contains(QString::number(mol->asymm.at(i).resiNr),Qt::CaseInsensitive))
        l.append(QString::number(mol->asymm.at(i).resiNr));
      if (!l.contains(mol->asymm.at(i).ResiClass))
        l.append(mol->asymm.at(i).ResiClass);
    }
  }
  if (!l.isEmpty()) l.insert(0,"*");
  return l;
}

void PlayOpenGL::singleBondDistance(){
    if (mol->selectedatoms.size()!=2)return;
    QWidgetList wl = QApplication::allWidgets();
    for (int i=0; i<wl.size(); i++){
        if (wl.at(i)->objectName()=="DFIX_DIST"){
            QLineEdit *valueline = qobject_cast<QLineEdit *>(wl[i]);
            valueline->setText(QString::number(0.01*(mol->Kovalenz_Radien[mol->selectedatoms.at(0).an]+mol->Kovalenz_Radien[mol->selectedatoms.at(1).an]),10,3));
        }
    }
}

void PlayOpenGL::changeEnviRange(){//! a dialog for a new ChGL.envirange.
  bool ok;
  double r=QInputDialog::getDouble (nullptr,"Envi-Range", "change environment range", envirange,0.7,8.0,2,&ok);
  if (ok) envirange=r;
  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else index=rotze;//return;
  if (index<0) return;
  if (ok) envi();
}

void PlayOpenGL::growEnvi(){
    QAction *action = qobject_cast<QAction *>(sender());
    int index=0;
    if (action)
      index=action->data().toInt();
    else index=rotze;//return;
    if (index<0) return;
    if (!isMainWindow) return;
    if (index>=mol->asymm.size())return;
    mol->enviSDM(envirange);

    QString systr;
    QStringList bs;
    for (int i = 0; i < mol->envi_sdm.size(); i++){
      if ((mol->envi_sdm.at(i).a2==index)&&mol->envi_sdm.at(i).d<envirange) {
        if ((!mol->envi_sdm.at(i).covalent)&&(enviCova->isChecked())) continue;
        if ((mol->asymm[mol->envi_sdm.at(i).a1].an==-1)&&(enviNoQ->isChecked())) continue;
        bool symm=((mol->envi_sdm.at(i).sn)||(!(mol->envi_sdm.at(i).floorD==V3(0,0,0))));
        if (symm) {
          QString sss=QString("%1_%2%3%4:%5,").arg(mol->envi_sdm.at(i).sn+1).arg(5-(int)mol->envi_sdm.at(i).floorD.x).arg(5-(int)mol->envi_sdm.at(i).floorD.y).arg(5-(int)mol->envi_sdm.at(i).floorD.z).arg(mol->asymm[mol->envi_sdm.at(i).a1].molindex);
          if (!bs.contains(sss)){
            bs.append(sss);
          }
          systr=QString("%1%2").arg(QString::fromUtf8("»")).arg(bs.indexOf(sss)+1);
        }
      }
    }
    mol->usedSymmetry+=bs;
    mol->complete();

    fuse->setVisible(true);
    grow->setVisible(false);
    emit bigmessage(mol->HumanSymmetry);
    if (AMD_Hack) {
      makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);
      makeAtoms(mol->showatoms);
    }
    update();
}

void PlayOpenGL::showEnvi(){
    QAction *action = qobject_cast<QAction *>(sender());
    int index=0;
    if (action)
      index=action->data().toInt();
    else index=rotze;//return;
    if (index<0) return;
    if (!isMainWindow) return;
    if (index>=mol->asymm.size())return;
    mol->enviSDM(envirange);

    QString systr;
    QStringList bs;
    for (int i = 0; i < mol->envi_sdm.size(); i++){
      if ((mol->envi_sdm.at(i).a2==index)&&mol->envi_sdm.at(i).d<envirange) {
        if ((!mol->envi_sdm.at(i).covalent)&&(enviCova->isChecked())) continue;
        if ((mol->asymm[mol->envi_sdm.at(i).a1].an==-1)&&(enviNoQ->isChecked())) continue;
        bool symm=((mol->envi_sdm.at(i).sn)||(!(mol->envi_sdm.at(i).floorD==V3(0,0,0))));
        if (symm) {
          QString sss=QString("%1_%2%3%4:%5,").arg(mol->envi_sdm.at(i).sn+1).arg(5-(int)mol->envi_sdm.at(i).floorD.x).arg(5-(int)mol->envi_sdm.at(i).floorD.y).arg(5-(int)mol->envi_sdm.at(i).floorD.z).arg(mol->asymm[mol->envi_sdm.at(i).a1].molindex);
          if (!bs.contains(sss)){
            bs.append(sss);
          }
          systr=QString("%1%2").arg(QString::fromUtf8("»")).arg(bs.indexOf(sss)+1);
        }
      }
    }
    mol->usedSymmetry+=bs;
    mol->complete();
    V3 p0 = mol->asymm.at(index).pos;
    double er2=envirange*envirange;
    for (int i = 0; i < mol->showatoms.size(); i++){
        double d = Distance(p0, mol->showatoms.at(i).pos);

        if (d > er2) mol->showatoms[i].hidden=1;
    }
    hiddenThings = true;
    fuse->setVisible(true);
    grow->setVisible(false);
    emit bigmessage(mol->HumanSymmetry);
    makeAtoms(mol->showatoms);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

    makeSelectedAtoms();
   // murx=-__LINE__;
    updateBondActions(__LINE__);

    update();
}

void PlayOpenGL::envi(){
  /*! Finds neighboring atoms around a ChGL.envirange the spezified atom and passes a html table via ChGL.bigmessage
   * Feeds ChGL.enviPositions, ChGL.labs  and ChGL.enviKat.
   * sets the rotation center to the spezified atom.
   * */

  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else index=rotze;//return;
  if (index<0) return;
  if (!isMainWindow){
      enviP0=mol->showatoms.at(index).pos;

      for (int i=0; i<mol->showatoms.size(); i++){
          double d = sqrt(Distance(mol->showatoms.at(index).pos,mol->showatoms.at(i).pos));
          int an1=mol->showatoms.at(i).an;
          int an2=mol->showatoms.at(index).an;
          double dddd=(mol->Kovalenz_Radien[an1]+ mol->Kovalenz_Radien[an2])*0.012;
          bool cov=(d<dddd);
          if (d<envirange) {
            if ((!cov)&&(enviCova->isChecked())) continue;
            if ((an1<0)&&(enviNoQ->isChecked())) continue;
            enviPositions.append(mol->showatoms.at(i).pos);
            enviKat.append((cov)?1:0);
            labs.append(mol->showatoms.at(i).Label);
          }
      }
      enviButt->show();
      enviSelect->show();
      makeEnvi();
      return;
  }
  if (index>=mol->asymm.size())return;
  mol->enviSDM(envirange);
  enviPositions.clear();
  enviKat.clear();
  labs.clear();
  QList<bool> covs;
  //pause=true;

  V3 ppc,ppf,p0;
  // int ssy=0;
  QString systr;
  QStringList bs;
  QString info;
  enviP0=mol->asymm.at(index).pos;
  info.append(QString("<hr><b>Environment of %1 </b><table border=\"0\" cellspacing=\"0\" cellpadding=\"5\">").arg(mol->asymm[index].Label));
  for (int i = 0; i < mol->envi_sdm.size(); i++){
    if ((mol->envi_sdm.at(i).a2==index)&&mol->envi_sdm.at(i).d<envirange) {
      if ((!mol->envi_sdm.at(i).covalent)&&(enviCova->isChecked())) continue;
      if ((mol->asymm[mol->envi_sdm.at(i).a1].an==-1)&&(enviNoQ->isChecked())) continue;
      ppf=mol->cell.symmops.at(mol->envi_sdm.at(i).sn) *
        mol->asymm[mol->envi_sdm.at(i).a1].frac +
        mol->cell.trans.at(mol->envi_sdm.at(i).sn) -//-
        mol->envi_sdm.at(i).floorD;
      mol->frac2kart(ppf,ppc);
      mol->frac2kart(mol->asymm[index].frac,p0);
      /*
         if (fabs(mol->envi_sdm.at(i).d-sqrt(Distance(p0,ppc)))>0.5){
         ppf=mol->cell.symmops.at(mol->envi_sdm.at(i).sn) *
         mol->asymm[mol->envi_sdm.at(i).a2].frac -
         mol->cell.trans.at(mol->envi_sdm.at(i).sn) -
         mol->envi_sdm.at(i).floorD;
         mol->frac2kart(ppf,ppc);

         }// */
      enviPositions.append(ppc);
      covs.append(mol->envi_sdm.at(i).covalent);
      enviKat.append((mol->envi_sdm.at(i).covalent)?1:0);
      if((abs(mol->asymm[mol->envi_sdm.at(i).a1].an-7)<2)
          &&(abs(mol->asymm[mol->envi_sdm.at(i).a2].an-7)<2)
          &&(fabs(mol->envi_sdm.at(i).d-2.725)<0.275)){enviKat.last()=2;}

      if((mol->asymm[mol->envi_sdm.at(i).a1].an==0)&&(abs(mol->asymm[mol->envi_sdm.at(i).a2].an-7)<2)&&(fabs(mol->envi_sdm.at(i).d-1.875)<0.275)){enviKat.last()=2;}
      if((mol->asymm[mol->envi_sdm.at(i).a2].an==0)&&(abs(mol->asymm[mol->envi_sdm.at(i).a1].an-7)<2)&&(fabs(mol->envi_sdm.at(i).d-1.875)<0.275)){enviKat.last()=2;}
      bool symm=((mol->envi_sdm.at(i).sn)||(!(mol->envi_sdm.at(i).floorD==V3(0,0,0))));
      if (symm) {
        QString sss=QString("%1_%2%3%4:%5,").arg(mol->envi_sdm.at(i).sn+1).arg(5-(int)mol->envi_sdm.at(i).floorD.x).arg(5-(int)mol->envi_sdm.at(i).floorD.y).arg(5-(int)mol->envi_sdm.at(i).floorD.z).arg(mol->asymm[mol->envi_sdm.at(i).a1].molindex);
        if (!bs.contains(sss)){
          bs.append(sss);
        }

        systr=QString("%1%2").arg(QString::fromUtf8("»")).arg(bs.indexOf(sss)+1);
        labs.append(mol->asymm[ mol->envi_sdm.at(i).a1].Label+systr);
      }
      else labs.append(mol->asymm[ mol->envi_sdm.at(i).a1].Label);
      info.append(QString("<tr><th style=\"background:%6\" >%1%4</th><td style=\"background:%5\" >%7<font color=%3> %2&nbsp;&Aring;</font></b></td>")
          .arg(mol->asymm[ mol->envi_sdm.at(i).a1].Label)
          .arg(mol->envi_sdm.at(i).d,8,'f',3)
          .arg((mol->envi_sdm.at(i).covalent)?"green":"black")
          .arg(symm?systr:"").arg((labs.size()%2)?"#eeeeee":"white")
          .arg((labs.size()%2)?"#d4d4e4":"#efefff")
          .arg((mol->envi_sdm.at(i).covalent)?"<b>":""));
      for (int j=0; j<enviPositions.size()-1; j++){
        double w=
          mol->winkel(p0-enviPositions.at(j),
              p0-enviPositions.last());

        info.append(QString("<td style=\"background:%3\" align=right>%4<font color=%2> %1&deg;</font></b></td>")
            .arg(w,8,'f',2)
            .arg(((mol->envi_sdm.at(i).covalent)&&(covs.at(j)))?"green":"black")
            .arg(((j+1)%2)?(labs.size()%2)?"#d4d4d4":"#e4e4e4":(labs.size()%2)?"#eeeeee":"#ffffff")
            .arg(((mol->envi_sdm.at(i).covalent)&&(covs.at(j)))?"<b>":""));

      }
      info.append("</tr>\n");
    }
  }
  info.append("<tr><td style=\"background:#efefff\"></td><td style=\"background:#efefff\"></td>");
  for (int j=0; j<enviPositions.size()-1; j++){
    info.append(QString("<th style=\"background:%2\">%1</th>").arg(labs.at(j)).arg(((j+1)%2)?"#d4d4e4":"#efefff"));
  }
  QString symml="";
  for (int j=0; j<bs.size(); j++){
    symml+=mol->symmcode2human(bs.at(j),j+1);
  }
  info.append(QString("</tr>\n</table>%1<hr>\n").arg(symml));
  emit bigmessage(info);
  enviButt->show();
  enviSelect->show();
  //pause=false;
  makeEnvi();
  setRotationCenter(index);

  update();

}

//*
void PlayOpenGL::clearEnvi(){
  // *! Clears ChGL.enviPositions, ChGL.enviKat. and ChGL.labs * /
    if (enviPositions.isEmpty()) return;
  enviPositions.clear();
  labs.clear();
  enviKat.clear();
  enviButt->hide();
  enviSelect->hide();
  makeEnvi();
  if (isMainWindow) rotCenter();
}

void PlayOpenGL::selectEnvi(){
  mol->selectedatoms.clear();
  if (rotze>=0) {
    mol->selectedatoms.append(mol->showatoms[rotze]);
    mol->selectedatoms.last().style=rotze;
  }
  for (int i=0; i<mol->showatoms.size(); i++){
    for (int j=0; j< enviPositions.size(); j++){
      if (mol->showatoms.at(i).pos==enviPositions.at(j)) {
        mol->selectedatoms.append(mol->showatoms[i]);
        mol->selectedatoms.last().style=i;
      }
    }
  }
  updateBondActions(__LINE__);
  update();

}
// */



void PlayOpenGL::setRotationCenter(){
  //! Sets the rotation center to the spezified atom
  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else return;
  if (index==(int)((GLuint)-1))return;
  rotze=((int)index<mol->showatoms.size())?index:-1;
  if (rotze>-1){
    rCenter->show();
    homeXY();
    updateScreenAtoms();
  }
  update();
}

void PlayOpenGL::setRotationCenter(V3 center){

  //altcenter=center;
  rotze=-76185;//zip code
  rCenter->show();
  update();
}


void PlayOpenGL::setRotationCenter(int rz){
  /*! Sets the rotation center to the spezified atom
   * @param index of the spezified atom.
   */
  rotze=((int)rz<mol->showatoms.size())?rz:-1;
  if (rotze>-1){
      rCenter->show();
      homeXY();
      updateScreenAtoms();
  }
  update();
}


void PlayOpenGL::findPivot(){

    if ((altpivot)||(whyNotSplit==nullptr)) return;
   // for(int i=0; i<mol->showatoms.size(); i++) mol->showatoms[i].hidden=0;
    if (mol->selectedatoms.size()<3) {
        whyNotSplit->setText("<h2>Select at least 3 atoms first.</h2>");
        emit splitable(canSplitRotate());
        return;
    }

    pivot=-1;pivot2=-1;
    emit pivot2Changed(pivot2);
    emit pivot1Changed(pivot);
    int molidx=-1,mi;
    seat.clear();
    for (int i=0; i<mol->selectedatoms.size();i++){//all selectet atoms have to be in the same molecule
        int ii=mol->selectedatoms.at(i).style;
        if ((ii<0)||(ii>=mol->asymm.size()))continue;
        mi=mol->asymm.at(ii).molindex;
        if ((molidx!=-1)&&(molidx!=mi)) {
            whyNotSplit->setText("<h2>Selected atoms have to be in the same molecule.</h2>");
            emit splitable(canSplitRotate());
            return;
        }
        molidx=mi;
        seat.append(ii);
    }
    int atinmol=0;
    for (int i=0; i<mol->asymm.size();i++){
        if (mol->asymm.at(i).molindex==molidx) atinmol++;
    }

    if (mol->selectedatoms.size()>atinmol) {
        whyNotSplit->setText("<h2>Only a part of a molecule may be selected.<br>Or you have to specify atoms definig the rotation axis yourself.</h2>");
        emit splitable(canSplitRotate());
        return;
    } //only a part of the molecule may be selected
    // */
    for (int i=0;i<seat.size();i++){

        int isin=0,ks=mol->knoepfe.at(seat.at(i)).neighbors.size();
        for (int j=0;j<ks; j++){
            int nb=mol->knoepfe.at(seat.at(i)).neighbors.at(j);
            if (seat.contains(nb))isin++;
        }
        if ((pivot==-1)&&(isin<ks)) {pivot=seat.at(i);emit pivot1Changed(pivot);}
        else if (isin<ks) {
            whyNotSplit->setText("<h2>Only one atom is allowed to have bonds to none selected atoms.</h2>");
            emit splitable(canSplitRotate());
            return;
        }
    }

    if (pivot<0) {
        emit splitable(canSplitRotate());
        return;
    }
    int isin=0,ks=mol->knoepfe.at(pivot).neighbors.size();
    for (int j=0;j<ks; j++){
        int nb=mol->knoepfe.at(pivot).neighbors.at(j);
        if (seat.contains(nb)){pivot2=nb;emit pivot2Changed(pivot2);isin++;}
    }
    if (isin>1) {
        whyNotSplit->setText("<h2>pivot atom may be bonded to only one selected atoms.</h2>");
        emit splitable(canSplitRotate());
        return;
    }
    whyNotSplit->setText(QString("<h2>You can rotate around %1=%2</h2>").arg(mol->asymm.at(pivot).Label).arg(mol->asymm.at(pivot2).Label));
    emit splitable(canSplitRotate());
}

void PlayOpenGL::splitRotate(){
    mol->splitLbonds.clear();
    mol->splitRbonds.clear();
    mol->splitLeftAtoms.clear();
    mol->splitRightAtoms.clear();
    if (freeRot->isChecked()){
     //   updateBondActions(__LINE__);
    }

    if (!canSplitRotate()){makeSplitRot();return;}

    if (!freeRot->isChecked()&&((pivot<0)||(pivot2<0))) {makeSplitRot();return;}
    for (int i=0; i<seat.size();i++){
        mol->splitLeftAtoms.append(mol->asymm[seat.at(i)]);
        mol->splitRightAtoms.append(mol->asymm[seat.at(i)]);
       // if (seat.at(i)!=pivot) mol->showatoms[seat.at(i)].hidden=1;
    }
    /*printf("?splitRotate %d %d  %d\n",
           mol->splitLeftAtoms.size(),
           mol->splitRightAtoms.size(),
           seat.size());*/
    //mol->selectedatoms.clear();

    double arr[4];
    arr[0]=angle1->value();
    V3 ax=((manualAx->isChecked())||(freeRot->isChecked()))?
                Normalize(V3(axx->value(),axy->value(),axz->value())):
                Normalize(mol->asymm.at(pivot).pos-mol->asymm.at(pivot2).pos);
    arr[1]=ax.x;
    arr[2]=ax.y;
    arr[3]=ax.z;
    if ((!manualAx->isChecked())&&(!freeRot->isChecked())){
        axx->setValue(ax.x);
        axy->setValue(ax.y);
        axz->setValue(ax.z);
    }
    if (!freeRot->isChecked()){
        screenRotx=ax;
        Matrix M90 = mol->rotarb(90,ax);
        V3 cx = ax % V3(0,1,0);


        V3 cy = Normalize(M90 * cx);
        V3 cz = ax % cy;

        screenRoty = Normalize(cz);
    }
    if ((freeRot->isChecked())&&(pivot1cb->isEnabled())){
        V3 o(0,0,0);
        for (int i=0; i<mol->selectedatoms.size(); i++){
            o+=mol->selectedatoms.at(i).pos;
        }
        o*=1.0/mol->selectedatoms.size();
        orgx->setValue(o.x);
        orgy->setValue(o.y);
        orgz->setValue(o.z);
    }
    if (freeRot->isChecked()){
        V3 o(0,0,0);
        for (int i=0; i<mol->splitLeftAtoms.size(); i++){
            o+=mol->splitLeftAtoms.at(i).pos;
        }
        o*=1.0/mol->splitLeftAtoms.size();
        V3 no(orgx->value(), orgy->value(), orgz->value());
        for (int i=0; i<mol->splitLeftAtoms.size(); i++){
            mol->splitLeftAtoms[i].pos+=no-o;
        }
    }
    pivot1cb->setEnabled(!freeRot->isChecked());
    pivot2cb->setEnabled(!freeRot->isChecked());


    Matrix R =  ((freeRot->isChecked()))?freeRotMat : mol->rotarb(arr);
    /*
    V3 axe = V3((R.m23 - R.m32), (R.m31 - R.m13), (R.m12 - R.m21));
    double _si = sqrt(Norm(axe)) / 2.0;
    double _co = ((R.m11 + R.m22 + R.m33) - 1.0)/ 2.0;
    axe=Normalize(axe);
    double wt = atan2(_si, _co) / M_PI * 180.0;
    printf("%f grad  %f %f %f\n", wt, axe.x, axe.y, axe.z);
    */
    arr[0]=-angle2->value();
    Matrix Rbar=mol->rotarb(arr);
    SRorg=V3(orgx->value(), orgy->value(), orgz->value());
    if (!freeRot->isChecked()){
        SRorg=mol->asymm.at(pivot).pos;
        orgx->setValue(SRorg.x);
        orgy->setValue(SRorg.y);
        orgz->setValue(SRorg.z);
    }
    for (int i=0; i<seat.size();i++){
        mol->splitLeftAtoms[i].pos=R*(mol->splitLeftAtoms.at(i).pos-SRorg)+SRorg;
        mol->splitRightAtoms[i].pos=Rbar*(mol->splitRightAtoms.at(i).pos-SRorg)+SRorg;
    }
    mol->splitRbonds=mol->connecting(mol->splitRightAtoms,true);
    mol->splitLbonds=mol->connecting(mol->splitLeftAtoms,true);
    //rotze=pivot;
    makeSplitRot();
    update();
}

bool PlayOpenGL::canSplitRotate(){

    if (freeRot==nullptr) return false;//!!!
    if ((freeRot!=nullptr)&&(freeRot->isChecked())&&(mol->selectedatoms.size()>2)) return true;
    if ((pivot>-1)&&(pivot2>-1)&&(pivot<mol->asymm.size())&&(pivot2<mol->asymm.size())&&(mol->selectedatoms.size()>2)) return true;
    return false;
}


void PlayOpenGL::highliteQPeak(double co){

    inFocus=-1;

    for (int i=0;i<mol->showatoms.size();i++){
      if (mol->showatoms[i].an==-66) continue;
      if (mol->showatoms[i].an>=0) continue;
      if (mol->showatoms[i].hidden)continue;
      if (mol->showatoms[i].peakHeight<=co) {
          inFocus=i;
          emit message(QString("%1 %2").arg(mol->showatoms.at(inFocus).Label).arg(mol->showatoms.at(inFocus).peakHeight));
          break;
      }
    }
    update();
}

void PlayOpenGL::renderText(const QVector2D &scpos, const QString &text, int an, int pts){
  QPainter painter(this);

  QFont f = myFont;
  f.setPointSize((an<0)? (pts*2)/3: pts);
  QFontMetrics mtr(f);
  QRect rc=mtr.boundingRect(text+"L");
  QImage image(rc.width(),rc.height(), QImage::Format_ARGB32_Premultiplied);
  image.fill(0);
  QPainter p;
  p.begin(&image);
  p.setPen(labelColor);
  p.setBackground(Qt::NoBrush);
  p.setFont(f);
  p.drawText(image.rect(), Qt::AlignCenter, text);
  p.end();
  painter.drawImage(scpos.x(), scpos.y()-rc.height(), image);
  painter.end();
}

void PlayOpenGL::renderText(const QVector2D &scpos, const QString &text, int an){
  QPainter painter(this);
  QFont f = (an<0)?nonAtomFont:myFont;
  QFontMetrics mtr(f);
  QRect rc=mtr.boundingRect(text+"L");
  QImage image(rc.width(),rc.height(), QImage::Format_ARGB32_Premultiplied);
  image.fill(focusedLabelBGColor.rgba());
  QPainter p;
  p.begin(&image);
  p.setPen(focusedLabelColor);
  p.setBackground(Qt::black);
  p.setFont(f);
  p.drawText(image.rect(), Qt::AlignCenter, text);
  p.end();
  painter.drawImage(scpos.x(), scpos.y()-rc.height(), image);
  painter.end();
}


void PlayOpenGL::renderText(const QVector2D &scpos, int i){
    QPainter painter(this);
    QRect rc = altLabels.at(i).rect();
    painter.drawImage(scpos.x(), scpos.y()-rc.height(), altLabels.at(i));
    painter.end();
}

void PlayOpenGL::renderRect(){
    makeCurrent();
    QPainter painter(this);
    QPen pen(labelColor);
    pen.setStyle(Qt::DotLine);
    painter.setPen(pen);
    painter.drawRect(scrx0,scry0,scrx-scrx0,scry-scry0);
    painter.end();
}

void PlayOpenGL::renderText(const V3 &pos, const QString &text, QColor &color, uint bg){
    QVector3D mapped3d = pmv.map(QVector3D(pos.x, pos.y, pos.z));
    int w=width();
    int h=height();
    QVector2D vec2 = QVector2D((1 + mapped3d.x()) * (w / 2), (1 - mapped3d.y()) * (h / 2));
    QPainter painter(this);
    QFontMetrics mtr(myFont);
    QRect rc=mtr.boundingRect(text+"L");
    QImage image(rc.width(),rc.height(), QImage::Format_ARGB32_Premultiplied);
    image.fill(bg);
    QPainter p;
    p.begin(&image);
    p.setPen(color);
    p.setBackground(Qt::NoBrush);
    p.setFont(myFont);
    p.drawText(image.rect(), Qt::AlignCenter, text);
    p.end();
    painter.drawImage(vec2.x(), vec2.y()-rc.height(), image);
    painter.end();
}


void PlayOpenGL::renderText(QMatrix4x4 mat,int w, int h, const V3 &pos, const QString &text, QColor &color, uint bg){
    QVector3D mapped3d = mat.map(QVector3D(pos.x, pos.y, pos.z));
    QVector2D vec2 = QVector2D((1 + mapped3d.x()) * (w / 2), (1 - mapped3d.y()) * (h / 2));
    QPainter painter(this);
    QFontMetrics mtr(nonAtomFont);
    QRect rc=mtr.boundingRect(text+"L");
    QImage image(rc.width(),rc.height(), QImage::Format_ARGB32_Premultiplied);
    image.fill(bg);
    QPainter p;
    p.begin(&image);
    p.setPen(color);
    p.setBackground(Qt::NoBrush);
    p.setFont(nonAtomFont);
    p.drawText(image.rect(), Qt::AlignCenter, text);
    p.end();
    painter.drawImage(vec2.x(), vec2.y()-rc.height(), image);
    painter.end();
}


void PlayOpenGL::makeLabels(){
    altLabels.clear();
    nonAtomFont = myFont;
    QString label;
    nonAtomFont.setPointSize(qMax(((myFont.pointSize()*2)/3),7));
    for (int i = 0; i < mol->showatoms.size(); i++){
        QFont f = (mol->showatoms.at(i).an<0)?nonAtomFont:myFont;
        QFontMetrics mtr(f);
        label = mol->showatoms.at(i).Label;
        if (shortLabels->isChecked()) {
          label=label.section("_",0,0);
          label=label.section(QString::fromUtf8("»"),0,0);
        }
        QRect rc=mtr.boundingRect(label+"L");
        QImage image(rc.width(),rc.height(), QImage::Format_ARGB32_Premultiplied);
        image.fill(labelBGColor.rgba());
        QPainter p;
        p.begin(&image);
        p.setPen(labelColor);
        p.setBackground(Qt::NoBrush);
        p.setFont(f);
        p.drawText(image.rect(), Qt::AlignCenter, label);
        p.end();
        altLabels.append(image);
    }
    update();
    mkF++;

}

void PlayOpenGL::updateBondActions(int callLine){//!<changes the visibility state of QActions in the 'Selection Toolbar'.
  //bool p=pause;
  //pause=true;


  clearSelection->setVisible(!mol->selectedatoms.isEmpty());
    if (!mol->selectedatoms.isEmpty())makeSelectedAtoms();
  /*printf("line:%d  hiddenThings %d\n",__LINE__,hiddenThings , mol->selectedatoms.size());
  if (hiddenThings){
      for (int i=0;i < mol->showatoms.size(); i++)
      if ((mol->showatoms.at(i).an>=0)&&(mol->showatoms.at(i).hidden))printf("#%d %s is hiding from you.\n",i,mol->showatoms.at(i).Label.toStdString().c_str());
  } // */
  invSelection->setVisible(!mol->selectedatoms.isEmpty());
  centroid->setVisible(mol->selectedatoms.size()>1);
  centerSelection->setVisible((!mol->selectedatoms.isEmpty())||(centerSelection->isChecked()));
  delSelAt->setVisible(!mol->selectedatoms.isEmpty());
  delSelAt->setEnabled(!mol->selectedatoms.isEmpty());
  hideNotSelection->setVisible(!mol->selectedatoms.isEmpty());
  if (fVertexes.size()>1)cntrPlot->setVisible(((!fVertexes[0].isEmpty()||!fVertexes[1].isEmpty())&&(mol->selectedatoms.size()==3)));
 // if (mol->selectedatoms.size()==3)
  unhide->setVisible(hiddenThings);
  invhide->setVisible(hiddenThings);
  if (mol->selectedatoms.size()!=2){
    addBond->setVisible(false);
    killBond->setVisible(false);
  }
  else{
    int da=0;
    for (int i=0; i<mol->showbonds.size();i++){
      if (((mol->selectedatoms.at(0).style==mol->showbonds.at(i).a1)||
            (mol->selectedatoms.at(0).style==mol->showbonds.at(i).a2))&&
          ((mol->selectedatoms.at(1).style==mol->showbonds.at(i).a1)||
           (mol->selectedatoms.at(1).style==mol->showbonds.at(i).a2))) da=i;
    }
    if (da){
      addBond->setVisible(false);
      killBond->setVisible(true);
      killBond->setData(da);
    }
    else{
      addBond->setVisible(true);
      killBond->setVisible(false);
    }
  }
  findPivot();
  makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

  mkG++;
}

void PlayOpenGL::updateScreenAtoms(){

    //QList<QVector2D> screenAtoms;
    //CEnvironment mol->showatoms;
    if (mol->showatoms.isEmpty())return;
    V3 org = sumse;
    if ((rotze>-1)&&(rotze<mol->showatoms.size())) org = mol->showatoms.at(rotze).pos;
    if (centerSelection->isChecked()){
      if (mol->selectedatoms.isEmpty()) {
        org = sumse;
      }
      else {
        V3 sums=V3(0,0,0);
        for (int i = 0; i < mol->selectedatoms.size(); i++) sums += mol->selectedatoms[i].pos;
        sums *= 1.0 / mol->selectedatoms.size();
        org = sums;
      }
    }
    if (mol->showatoms.size()<screenAtoms.size()) screenAtoms.clear();
    for (int i = 0; i < mol->showatoms.size(); i++){
        QVector3D mapped3d = pmv.map(QVector3D(mol->showatoms.at(i).pos.x - org.x, mol->showatoms.at(i).pos.y - org.y, mol->showatoms.at(i).pos.z - org.z));
        int w=width();
        int h=height();
        QVector2D vec2 = QVector2D((1 + mapped3d.x()) * (w / 2), (1 - mapped3d.y()) * (h / 2));
        if (screenAtoms.size() < i + 1) screenAtoms.append(vec2);
        else screenAtoms[i] = vec2;
        /*printf("%s %f %f  %f %f %f %d %d %d %d\n",mol->showatoms.at(i).Label.toStdString().c_str(),vec2.x(),vec2.y(), mapped3d.x(), mapped3d.y(), mapped3d.z()
               ,w,h,w/2,h/2);// */
    }
}


void PlayOpenGL::makeEnvi(){

    makeCurrent();
    envi_vao.release();
    if (envi_vao.isCreated()) {
        envi_vao.release();
        envi_vao.destroy();
        envi_vertexBuffer.destroy();
        envi_program->release();
        envi_program->removeAllShaders();
    }
    envi_vao.create();
    envi_vao.bind();
    envi_vertexBuffer.create();
    envi_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!envi_program) envi_program = new QOpenGLShaderProgram(this);
    else envi_program->removeAllShaders();

    bool result = envi_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/envi.vert");
    if (!result) qWarning() << envi_program->log();
   // result = envi_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/envi.geom");
   // if (!result) qWarning() << envi_program->log();
    result = envi_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/envi.frag");
    if (!result) qWarning() << envi_program->log();
    result = envi_program->link();
    if (!result) qWarning() << envi_program->log();
    QVector<float> p;
    for (int i=0; i<enviPositions.size(); i++){
        p.append(enviP0.x);
        p.append(enviP0.y);
        p.append(enviP0.z);
        p.append(static_cast<float>((enviKat.at(i))?enviKat.at(i):5));
        p.append(enviPositions.at(i).x);
        p.append(enviPositions.at(i).y);
        p.append(enviPositions.at(i).z);
        p.append(static_cast<float>((enviKat.at(i))?-enviKat.at(i):-5));
    }
     if (!envi_vertexBuffer.bind()) {printf("I was unable to bind envi_vertexBuffer.\n");}
    envi_vertexBuffer.allocate((p.size()) * sizeof(float));
    envi_vertexBuffer.write(0,p.data(),p.size() * sizeof(float));

    envi_program->enableAttributeArray("L0" );
    envi_program->setAttributeBuffer("L0", GL_FLOAT, 0, 4 );
    bool b=shaderInfo;
    shaderInfo=true;
    programInfo(envi_program, "envi");
    shaderInfo=b;
    envi_program->release();
    envi_vertexBuffer.release();
    envi_vao.release();
    mk9++;

    update();
}

void PlayOpenGL::makeMaps(
        QVector< QVector< float > > &_fVertexes,
        QVector< QVector< float > > &_fNormals){
    fVertexes = _fVertexes;
    fNormals  = _fNormals;

    makeCurrent();
    map_vao.release();
    if (map_vao.isCreated()) {
        map_vao.release();
        map_vao.destroy();
        map_vertexBuffer.destroy();
        map_program->release();
        map_program->removeAllShaders();
    }
    map_vao.create();
    map_vao.bind();
    map_vertexBuffer.create();
    map_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR)
    {
      printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);

    }
    if (!map_program) map_program = new QOpenGLShaderProgram(this);
    else map_program->removeAllShaders();

    bool result = map_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/waldwiesen.vert");

    if ( !result ) qWarning() << map_program->log();
    result = map_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/waldwiesen.frag");
    if ( !result ) qWarning() << map_program->log();
    result = map_program->link();
    if ( !result ) qWarning() << map_program->log();
    int siz = fVertexes[0].size() + sizeof(float);
    size_t mapsiz = 0;//fVertexes[0].size()* sizeof(float);
    for (int i=0; i<28; i++) mapsiz+=fVertexes[i].size()* sizeof(float);
    if (!map_vertexBuffer.bind()) {printf("I was unable to bind map_vertexBuffer.\n");}
    map_vertexBuffer.allocate(mapsiz*2);
    int offs=0;
    V3 ctr=V3(0,0,0);
    const float * fv =  fVertexes[0].data();
    for (int i=0; i<fVertexes[0].size()/3; i++){
        ctr+=V3(fv[i*3], fv[i*3+1], fv[i*3+2]);
    }
    ctr*=3.0/fVertexes[0].size();



    offsets.clear();
    for (int i=0; i<28; i++){
        offsets.append(offs);
        int siz = fVertexes[i].size() * sizeof(float);
        map_vertexBuffer.write(offs, fVertexes[i].data(), siz);
        offs+=siz;
    }//*/
    int normaloffset=offs;
    for (int i=0; i<28; i++){
        //offsets.append(offs);
        int siz = fNormals[i].size() * sizeof(float);
        map_vertexBuffer.write(offs, fNormals[i].data(), siz);
        offs+=siz;
    }//*/
    //qDebug()<<offsets;

    map_program->enableAttributeArray("normal" );// */
    map_program->enableAttributeArray("vertex" );
    map_program->setAttributeBuffer("normal", GL_FLOAT, normaloffset, 3 );
    map_program->setAttributeBuffer("vertex", GL_FLOAT, 0, 3 );


    programInfo(map_program, "waldwiesen");
    err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      printf("map_program: %s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);

    }
    map_vertexBuffer.release();
    map_program->release();
    map_vao.release();
    mkB++;
    //update();
}

void PlayOpenGL::setMapsStuff(){
    if (!mol->XLAB)connect(foact, SIGNAL(toggled(bool)), this,  SLOT(update()));
    connect(fofcact, SIGNAL(toggled(bool)), this,  SLOT(update()));
    connect(niceTrans, SIGNAL(toggled(bool)), this,  SLOT(update()));
    connect(lighting, SIGNAL(toggled(bool)), this,  SLOT(update()));
    connect(fillMap, SIGNAL(toggled(bool)), this,  SLOT(update()));
    makeMaps(fVertexes,fNormals);
}


void PlayOpenGL::makeUnitCell(){

    makeCurrent();
    uc_vao.release();
    if (uc_vao.isCreated()) {
        uc_vao.release();
        uc_vao.destroy();
        uc_vertexBuffer.destroy();
        uc_program->release();
        uc_program->removeAllShaders();
    }
    uc_vao.create();
    uc_vao.bind();
    uc_vertexBuffer.create();
    uc_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!uc_vertexBuffer.bind()) {printf("I was unable to bind uc_vertexBuffer.\n");}
    GLfloat vert[] = {
        0.0, 0.0, 0.0, 1.0,
        1.0, 0.0, 0.0, 1.0,

        0.0, 0.0, 0.0, 2.0,
        0.0, 1.0, 0.0, 2.0,

        0.0, 0.0, 0.0, 3.0,
        0.0, 0.0, 1.0, 3.0,

        1.0, 0.0, 0.0, 0.0,
        1.0, 1.0, 0.0, 0.0,

        1.0, 0.0, 0.0, 0.0,
        1.0, 0.0, 1.0, 0.0,

        0.0, 1.0, 0.0, 0.0,
        1.0, 1.0, 0.0, 0.0,

        0.0, 1.0, 0.0, 0.0,
        0.0, 1.0, 1.0, 0.0,

        0.0, 0.0, 1.0, 0.0,
        1.0, 0.0, 1.0, 0.0,

        0.0, 0.0, 1.0, 0.0,
        0.0, 1.0, 1.0, 0.0,

        0.0, 1.0, 1.0, 0.0,
        1.0, 1.0, 1.0, 0.0,

        1.0, 0.0, 1.0, 0.0,
        1.0, 1.0, 1.0, 0.0,

        1.0, 1.0, 0.0, 0.0,
        1.0, 1.0, 1.0, 0.0,

};
    if (!uc_program) uc_program = new QOpenGLShaderProgram(this);
    else uc_program->removeAllShaders();
    bool result = uc_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/uc2.vert");

    if ( !result ) qWarning() << uc_program->log();
    result = uc_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/uc2.frag");
    if ( !result ) qWarning() << uc_program->log();
    result = uc_program->link();
    if ( !result ) qWarning() << uc_program->log();
    if ( !uc_program->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }
    uc_vertexBuffer.allocate(vert, 24 * 4 * sizeof( float )) ;
    //uc_program->setAttributeBuffer(uc_vertexBuffer.bufferId(), GL_FLOAT, 0, 3 );
    uc_program->enableAttributeArray("vertex" );
    uc_program->setAttributeBuffer("vertex", GL_FLOAT, 0, 4 );
    //shaderInfo=true;
    programInfo(uc_program,"unit  cell");
    //shaderInfo=false;
    uc_program->release();
    uc_vao.release();
    mkE++;
    uc_vertexBuffer.release();
    uc_vertexBuffer.destroy();
    update();
}

void PlayOpenGL::contDestruct(){
   cont_vao.release();
   cont_vao.destroy();
   cont_vertexBuffer.destroy();
   cont_program->release();
   cont_program->removeAllShaders();
}

void PlayOpenGL::voroDestruct(){
   voro_vao.release();
   voro_vao.destroy();
   voro_vertexBuffer.destroy();
   voro_program->release();
   voro_program->removeAllShaders();
}

void PlayOpenGL::makeContours(){

    makeCurrent();
    cont_vao.release();
    if (cont_vao.isCreated()) {
        cont_vao.release();
        cont_vao.destroy();
        cont_vertexBuffer.destroy();
        cont_program->release();
        cont_program->removeAllShaders();
    }
    cont_vao.create();
    cont_vao.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    cont_vertexBuffer.create();
    cont_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!cont_vertexBuffer.bind()) {printf("I was unable to bind cont_vertexBuffer.\n");}
    /////////////
    QVector<float> p;
    float val=0.0f;
    for (int ci=0;ci<cont.size();ci++){
        if (contval.contains(ci)) val = contval.value(ci);
        p.append(cont.at(ci).x);
        p.append(cont.at(ci).y);
        p.append(cont.at(ci).z);
        p.append(val);
    }
    if (!cont_program) cont_program = new QOpenGLShaderProgram(this);
    else cont_program->removeAllShaders();
    bool result = cont_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/contour.vert");

    if ( !result ) qWarning() << cont_program->log();
    result = cont_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/contour.frag");

    if ( !result ) qWarning() << cont_program->log();
    result = cont_program->link();
    if ( !result ) qWarning() << cont_program->log();

    if ( !cont_program->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }

    cont_vertexBuffer.allocate(p.data(),p.size()*sizeof(float));
    cont_program->setAttributeBuffer("vertex", GL_FLOAT, 0, 4 );
    cont_program->enableAttributeArray("vertex" );

    programInfo(cont_program,"Contour");

    cont_program->release();
    cont_vao.release();
    mk7++;
    cont_vertexBuffer.release();
    cont_vertexBuffer.destroy();
    update();
}

void PlayOpenGL::makeBackGround(){

    makeCurrent();
    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR)
    {
      printf("before makeBackGround %s error %X\n", parent()->objectName().toStdString().c_str(),err);

    }

    bg_vao.release();
    if (bg_vao.isCreated()) {
        bg_vao.release();
        bg_vao.destroy();
        backGround_vertexBuffer.destroy();
        bg_program->release();
        bg_program->removeAllShaders();
    }
    bg_vao.create();
    bg_vao.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    backGround_vertexBuffer.create();
    backGround_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!backGround_vertexBuffer.bind()) {printf("I was unable to bind backGround_vertexBuffer.\n");}
    GLfloat vert[] ={-1.0,-1.0,0.0,  -1.0, 1.0,0.0,  1.0,-1.0,0.0,  1.0, 1.0,0.0};
    if (!bg_program) bg_program = new QOpenGLShaderProgram(this);
    else bg_program->removeAllShaders();



    bool result = bg_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/bg.vert");

    if ( !result ) qWarning() << bg_program->log();
    result = bg_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/bg.frag");

    if ( !result ) qWarning() << bg_program->log();
    result = bg_program->link();
    if ( !result ) qWarning() << bg_program->log();

    if ( !bg_program->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }
    backGround_vertexBuffer.allocate(vert, 4 *3 * sizeof( float )) ;
    //bg_program->setAttributeBuffer(backGround_vertexBuffer.bufferId(), GL_FLOAT, 0, 3 );
    bg_program->setAttributeBuffer("vertex", GL_FLOAT, 0, 3 );
    bg_program->enableAttributeArray("vertex" );
    //bg_program->enableAttributeArray(backGround_vertexBuffer.bufferId());

    bg_program->setUniformValue("bgcolor", backGroundColor);
    bg_program->setUniformValue("bggradient", bggradient);
    programInfo(bg_program,"bg");

    bg_program->release();
    bg_vao.release();
    mk3++;
    backGround_vertexBuffer.release();
    backGround_vertexBuffer.destroy();
    while((err = glGetError()) != GL_NO_ERROR)
    {
      printf("after makeBackGround %s error %X\n", parent()->objectName().toStdString().c_str(),err);

    }


}

void PlayOpenGL::makeSelectedAtoms(){


    makeCurrent();
    slct_vao.release();
    if (slct_vao.isCreated()) {
        slct_vao.release();
        slct_vao.destroy();
        satom_vertexBuffer.destroy();
        //satom_indexBuffer.destroy();
        satom_program->release();
        satom_program->removeAllShaders();
    }
    //if (mol->selectedatoms.isEmpty()) return;
    slct_vao.create();
    slct_vao.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    satom_vertexBuffer.create();
    satom_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!satom_vertexBuffer.bind()) {printf("I was unable to bind atom_vertexBuffer.\n");}
    //glDisable(GL_BLEND);


    int na = mol->selectedatoms.size();


    //*
    if (!satom_program) satom_program = new QOpenGLShaderProgram(this);
    else satom_program->removeAllShaders();

    bool result = satom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/std.vert");
    if ( !result ) qWarning() << satom_program->log();
    result = satom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/std.geom" );
    if ( !result ) qWarning() << satom_program->log();
    result = satom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/std.frag" );
    if ( !result ) qWarning() << satom_program->log();
    // ...and finally we link them to resolve any references.
    result = satom_program->link();
    if ( !result ) qWarning() << "Could not link shader satom_program:" << satom_program->log();

    QVector<float> pos;
    for (int i=0; i<mol->selectedatoms.size(); i++){
        pos.append(mol->selectedatoms.at(i).pos.x);
        pos.append(mol->selectedatoms.at(i).pos.y);
        pos.append(mol->selectedatoms.at(i).pos.z);
        /*printf("select? %f %f %f    %f %f %f \n",
               mol->selectedatoms.at(i).pos.x-sumse.x,
               mol->selectedatoms.at(i).pos.y-sumse.y,
               mol->selectedatoms.at(i).pos.z-sumse.z,
               sumse.x,sumse.y,sumse.z); // */
    }
    for (int i=0; i<mol->duplicateAtoms.size(); i++){
        pos.append (mol->duplicateAtoms.at(i).pos.x);
        pos.append (mol->duplicateAtoms.at(i).pos.y);
        pos.append (mol->duplicateAtoms.at(i).pos.z);
        /*printf("select? %f %f %f    %f %f %f \n",
               mol->selectedatoms.at(i).pos.x-sumse.x,
               mol->selectedatoms.at(i).pos.y-sumse.y,
               mol->selectedatoms.at(i).pos.z-sumse.z,
               sumse.x,sumse.y,sumse.z); // */
    }
    pos.append(0.1);//inFocus dummy atom
    pos.append(0.2);
    pos.append(0.3);

    satom_vertexBuffer.allocate( pos.data(), (na+1+mol->duplicateAtoms.size()) * 3 * sizeof( float ) );//na + dummy
    if ( !satom_program->bind() )    {
        qWarning() << "Could not bind shader satom_program to context";
        return;
    }
    satom_program->setAttributeBuffer("vertex", GL_FLOAT, 0, 3 );
    satom_program->enableAttributeArray("vertex" );
    satom_program->setUniformValue("pmv", pmv);
    satom_program->setUniformValue("normalMatrix", nm);
    satom_program->setUniformValue("Color", QColor("yellow"));
    satom_program->setUniformValue("lightPos", QVector3D(100,100,100));
    programInfo(satom_program, "std");
    satom_program->release();
    satom_vertexBuffer.release();
    slct_vao.release();
    mk4++;

    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR)    {
        printf("%s makeSelectedAtoms error %X %d number of selected atoms %d %d %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__, na,na+1+mol->duplicateAtoms.size(),pos.size());

    }

    update();
//    */
}

void PlayOpenGL::removeAllVAO(){

    makeCurrent();
    if (envi_vao.isCreated()) {
        envi_vao.bind();
        envi_vertexBuffer.destroy();
        envi_program->release();
        envi_program->removeAllShaders();
        envi_program->deleteLater();
        envi_vao.destroy();
    }
    if (latom_vao.isCreated()) {
        latom_vao.bind();
        //latom_vao.release();
        latom_program->release();
        latom_program->removeAllShaders();
        latom_vertexBuffer.destroy();
        latom_vao.destroy();
    }
    if (slct_vao.isCreated()) {
        slct_vao.bind();
        satom_vertexBuffer.destroy();
        //satom_indexBuffer.destroy();
        satom_program->release();
        satom_program->removeAllShaders();
        slct_vao.destroy();
    }
    if (atom_vao.isCreated()) {
        atom_vao.bind();
        atom_program->release();
        atom_program->removeAllShaders();
        atom_vertexBuffer.destroy();
        //atom_indexBuffer.destroy();
        atom_vao.destroy();
    }
    if (bond_vao.isCreated()) {
        bond_vao.bind();
        bond_vertexBuffer.destroy();
        //bond_indexBuffer.destroy();
        bond_vao.destroy();
    }
    if (bg_vao.isCreated()) {
        bg_vao.bind();
        backGround_vertexBuffer.destroy();
        bg_program->release();
        bg_program->removeAllShaders();
        bg_vao.destroy();
    }
    if (cont_vao.isCreated()) {
        cont_vao.bind();
        cont_vertexBuffer.destroy();
        cont_program->release();
        cont_program->removeAllShaders();
        cont_vao.destroy();
    }
    if (map_vao.isCreated()) {
        map_vao.bind();
        map_vertexBuffer.destroy();
        map_program->release();
        map_program->removeAllShaders();
        map_vao.destroy();
    }
    if (splitrot_vao.isCreated()) {
        splitrot_vao.bind();
        splitrot_vertexBuffer.destroy();
        //splitrot_indexBuffer.destroy();
        splitrot_vao.destroy();
    }
    if (symm_vao0.isCreated()) {
        symm_vao0.bind();
        symm_vertexBuffer0.destroy();
        symm_program0->release();
        symm_program0->removeAllShaders();
        symm_vao0.destroy();
        symm_vao1.bind();
        symm_program1->release();
        symm_program1->removeAllShaders();
        symm_vertexBuffer1.destroy();
        symm_vao1.destroy();
        symm_vao2.bind();
        symm_vertexBuffer2.destroy();
        symm_program2->release();
        symm_program2->removeAllShaders();
        symm_vao2.destroy();
    }
    if (uc_vao.isCreated()) {
        uc_vao.bind();
        uc_vertexBuffer.destroy();
        uc_program->release();
        uc_program->removeAllShaders();
        uc_vao.destroy();
    }
    update();
}

void PlayOpenGL::makeLegendAtoms(){
    if (mol->legendAtoms.isEmpty()) return;
    makeCurrent();
    latom_vao.release();
    if (latom_vao.isCreated()) {
        latom_vao.bind();
        //latom_vao.release();
        latom_vertexBuffer.destroy();
        latom_vao.destroy();
        latom_program->release();
        latom_program->removeAllShaders();
    }
    latom_vao.create();
    latom_vao.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    latom_vertexBuffer.create();
    latom_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!latom_vertexBuffer.bind()) {printf("I was unable to bind latom_vertexBuffer.\n");}

    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR)
    {
        printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);

    }
    int na = mol->legendAtoms.size();
    const float phi = (1.0f + sqrtf(5.0f)) / 2.0f;
    const float du = 1.0f / sqrtf(phi * phi + 1.0f);
    const float dv = phi * du;

    const float eckn[] ={   0.0f,  dv,  du,
                            0.0f,  dv, -du,
                            0.0f, -dv,  du,
                            0.0f, -dv, -du,
                            du,  0.0f,  dv,
                            -du, 0.0f,  dv,
                            du,  0.0f, -dv,
                            -du, 0.0f, -dv,
                            dv,    du, 0.0f,
                            dv,   -du, 0.0f,
                            -dv,   du, 0.0f,
                            -dv,  -du, 0.0f};
    const GLuint triangle[] = {
          1, 0, 8,
          4, 0, 5,
          5, 0, 10,
          8, 0, 4,
         10, 0, 1,
          6, 1, 8,
          7, 1, 6,
         10, 1, 7,
          3, 2, 11,
          4, 2, 9,
          5, 2, 4,
          9, 2, 3,
         11, 2, 5,
          6, 3, 7,
          7, 3, 11,
          9, 3, 6,
          8, 4, 9,
         11, 5, 10,
          9, 6, 8,
         10, 7, 11};//*/
    /*const GLuint triangle_strips[] = {
          1,  0, 8,  4, 9, 2, 3, 11, 7, 10, 1, 0, 0xFFFFFFFF,
          3,  9, 6,  8, 1, 0xFFFFFFFF,
          2, 11, 5, 10, 0, 0xFFFFFFFF,
          1,  7, 6,  3, 0xFFFFFFFF, //3, 7, 6, 1, 0xFFFF,
          2,  4, 5,  0, 0xFFFFFFFF//0, 5, 4, 2, 0xFFFF,
    };//, 10, 5, 11, 2}; */
    //0xFFFF
    double yofs=-6.7;
    double xofs=0;
    for (int ali=0; ali<na; ali++){
        xofs=qMax(mol->arad[qMax(mol->legendAtoms[ali].an,0)],xofs);
    }
    xofs-=3.0;
    for (int ali=0; ali<na; ali++){
        mol->legendAtoms[ali].pos.y=yofs;
        if ((adp->isChecked())&&(mol->legendAtoms[ali].an>=0)){
            yofs+=0.5;
        }
        else{
            if (tubes->isChecked()) {
                yofs+=2*mol->bondStrength+0.15;
            }else{
                yofs+=mol->arad[qMax(mol->legendAtoms[ali].an,0)]+0.07;
                if (ali+1<na) yofs+=mol->arad[qMax(mol->legendAtoms[ali+1].an,0)]+0.07;
            }
        }
        mol->legendAtoms[ali].pos.x=xofs;

        //if (mol->legendAtoms.at(ali).hidden) continue;
    }
    QVector<float> stuff;
    for (int i = 0; i < na; i++){
        bool isd=((mol->legendAtoms.at(i).an==0)&&(mol->legendAtoms.at(i).Label[0]=='D'));
        int style = (mol->legendAtoms.at(i).an<0)?0:
                (0xFF&mol->AtomStyle[mol->legendAtoms.at(i).an])
                |((mol->legendAtoms.at(i).hidden)?512:0)
                |((isd)?1024:0)
                |(((mol->legendAtoms.at(i).isIso))?2048:0);


        V3 ev=V3(0,0,0);
        double *arr = mol->jacobi(mol->legendAtoms.at(i).uc, ev);

        for (int j = 0; j < 12; j++){
        stuff.append(mol->legendAtoms.at(i).pos.x);//the center
        stuff.append(mol->legendAtoms.at(i).pos.y);
        stuff.append(mol->legendAtoms.at(i).pos.z);

        stuff.append(static_cast<float>(mol->legendAtoms.at(i).an));

        stuff.append(eckn[3*j]);// an icosahedron corner
        stuff.append(eckn[3*j+1]);
        stuff.append(eckn[3*j+2]);
        stuff.append(static_cast<float>(j));

        stuff.append(static_cast<float>(arr[0]));
        stuff.append(static_cast<float>(arr[1]));
        stuff.append(static_cast<float>(arr[2]));
        stuff.append(static_cast<float>(arr[3]));

        stuff.append(ev.x);
        stuff.append(ev.y);
        stuff.append(ev.z);
        stuff.append(static_cast<float>(style));

        stuff.append(static_cast<float>((mol->legendAtoms.at(i).an==-42)?mol->legendAtoms.at(i).sof:mol->legendAtoms.at(i).peakHeight));
        stuff.append(static_cast<float>(qMax(0, mol->legendAtoms.at(i).symmGroup)));
        stuff.append(static_cast<float>(mol->legendAtoms.at(i).part));
        stuff.append(static_cast<float>(mol->legendAtoms.at(i).molindex));

        stuff.append(static_cast<float>(i));
        stuff.append(static_cast<float>(mol->legendAtoms.at(i).auidx));
        stuff.append(static_cast<float>(mol->legendAtoms.at(i).resiNr));
        stuff.append(0.0f);

        }
        //atom_vertexBuffer.allocate(stuff.data(),stuff.size()*sizeof(float));
    }
    /*
    QVector<GLuint> indexes;
    uint una = static_cast<uint>(na);
    for (uint i = 0; i < una; i++){
        for (uint j = 0; j < 60; j++){
            indexes.append(triangle[j]+12*i);//_strips
        }
    }

    latom_indexBuffer = QOpenGLBuffer(QOpenGLBuffer::IndexBuffer); // Mind: use 'IndexBuffer' here
    latom_indexBuffer.create();
    latom_indexBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
    if (!latom_indexBuffer.bind()) {printf("I was unable to bind atom_indexBuffer!\n");}
    latom_indexBuffer.allocate( indexes.size() * sizeof(GLuint) );
    latom_indexBuffer.write(0, indexes.data(), indexes.size()* sizeof(GLuint));
    */
    QVector<float> bigstuff;
    for (int i = 0; i < na; i++){
        for (int j = 0; j < 60; j++){
            for (int k = 0; k < 24; k++){
                bigstuff.append(stuff.at(24*(triangle[j]+12*i)+k));
            }
        }
    }
    latom_vertexBuffer.allocate(bigstuff.data(),bigstuff.size()*sizeof(float));
    latom_array_size = bigstuff.size()/24;//indexes.size();



    if (!latom_program) latom_program = new QOpenGLShaderProgram(this);
    else latom_program->removeAllShaders();



    bool result = latom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/atom.vert");
    //bool result = atom_program->addShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/atom.vert");


    if ( !result ) qWarning() << latom_program->log();
    result = latom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/atom.geom");
    //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/atom.geom");
    //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/amd_atom.geom");


    if ( !result ) qWarning() << latom_program->log();
    result = latom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/atom.frag");
    //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/atom.frag");

    if ( !result ) qWarning() << latom_program->log();
    result = latom_program->link();
    if ( !result ) qWarning() << latom_program->log();
    \
        if ( !latom_program->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }

    programInfo(latom_program,"latom");
    GLfloat acv[110*3];
    for (int i=0;i<109;i++) {
        acv[3*i]=mol->atomColor(i,false).redF();
        acv[3*i+1]=mol->atomColor(i,false).greenF();
        acv[3*i+2]=mol->atomColor(i,false).blueF();
    }
    GLfloat arv[110];
    for (int i=0;i<109;i++) {
        arv[i]=static_cast<GLfloat>(mol->arad[i]);
    }

    latom_program->setUniformValueArray("atomColors",acv, 109,3);
    latom_program->setUniformValueArray("arad",arv, 109,1);



    latom_program->bind();
    latom_program->enableAttributeArray(0);
    latom_program->enableAttributeArray(1);
    latom_program->enableAttributeArray(2);
    latom_program->enableAttributeArray(3);
    latom_program->enableAttributeArray(4);
    latom_program->enableAttributeArray(5);

    int stride = 6 * 4 * sizeof(float);
    latom_program->setAttributeBuffer(0, GL_FLOAT, 0, 4, stride);
    latom_program->setAttributeBuffer(1, GL_FLOAT, sizeof(float)*4  , 4, stride);
    latom_program->setAttributeBuffer(2, GL_FLOAT, sizeof(float)*8  , 4, stride);
    latom_program->setAttributeBuffer(3, GL_FLOAT, sizeof(float)*12 , 4, stride);
    latom_program->setAttributeBuffer(4, GL_FLOAT, sizeof(float)*16 , 4, stride);
    latom_program->setAttributeBuffer(5, GL_FLOAT, sizeof(float)*20 , 4, stride);// 20*sizeof(float)
    latom_program->release();

    latom_vao.release();
    mkA++;
    //latom_indexBuffer.release();
    latom_vertexBuffer.release();
    latom_vertexBuffer.destroy();
    //latom_indexBuffer.destroy();

    err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR)
    {
        printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);

    }
    update();

}


void PlayOpenGL::makeSplitRot(){

    makeCurrent();
    splitrot_vao.release();
    if (splitrot_vao.isCreated()) {
        splitrot_vao.release();
        splitrot_vao.destroy();
        splitrot_vertexBuffer.destroy();
        //splitrot_indexBuffer.destroy();
    }
    splitrot_vao.create();
    splitrot_vao.bind();
    splitrot_vertexBuffer.create();
    splitrot_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    //glDisable(GL_BLEND);
    if (!splitrot_vertexBuffer.bind()) {printf("m_vertexBufferObject2 I was unable to bind!\n");}

    QVector<float> pos;
    //QVector<float> col;
    QVector<GLuint> nbonds;
    for (int i=0; i<mol->splitLbonds.size(); i++){
    //for (int i=0; i<mol->splitLeftAtoms.size(); i++){
        pos.append(mol->splitLeftAtoms.at(mol->splitLbonds.at(i).a1).pos.x);
        pos.append(mol->splitLeftAtoms.at(mol->splitLbonds.at(i).a1).pos.y);
        pos.append(mol->splitLeftAtoms.at(mol->splitLbonds.at(i).a1).pos.z);
        pos.append(1.0);

        pos.append(mol->splitLeftAtoms.at(mol->splitLbonds.at(i).a2).pos.x);
        pos.append(mol->splitLeftAtoms.at(mol->splitLbonds.at(i).a2).pos.y);
        pos.append(mol->splitLeftAtoms.at(mol->splitLbonds.at(i).a2).pos.z);
        pos.append(1.0);
    }
        for (int i=0; i<mol->splitRbonds.size(); i++){
        pos.append(mol->splitRightAtoms.at(mol->splitRbonds.at(i).a1).pos.x);
        pos.append(mol->splitRightAtoms.at(mol->splitRbonds.at(i).a1).pos.y);
        pos.append(mol->splitRightAtoms.at(mol->splitRbonds.at(i).a1).pos.z);
        pos.append(2.0);

        pos.append(mol->splitRightAtoms.at(mol->splitRbonds.at(i).a2).pos.x);
        pos.append(mol->splitRightAtoms.at(mol->splitRbonds.at(i).a2).pos.y);
        pos.append(mol->splitRightAtoms.at(mol->splitRbonds.at(i).a2).pos.z);
        pos.append(2.0);
    }
    splitrot_vertexBuffer.allocate(pos.size() * sizeof(float));
    splitrot_vertexBuffer.write(0, (void*)pos.data(), pos.size() * sizeof(float));
    splitSize = pos.size();

    if (!splitrot_program) splitrot_program = new QOpenGLShaderProgram(this);
    else splitrot_program->removeAllShaders();


    bool result2 = splitrot_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/splitrot.vert");
    if ( !result2 ) qWarning() << splitrot_program->log();
    result2 = splitrot_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/splitrot.geom");
    if ( !result2 ) qWarning() << splitrot_program->log();
    result2 = splitrot_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/splitrot.frag");
    if ( !result2 ) qWarning() << splitrot_program->log();
    result2 = splitrot_program->link();
    programInfo(splitrot_program,"splitrot");
    if ( !result2 ) qWarning() << splitrot_program->log();
    if ( !splitrot_program->bind() )    {
        qWarning() << "2Could not bind shader program to context";
        return;
    }

    splitrot_program->setUniformValue("radius", static_cast<float>(mol->bondStrength));
    splitrot_program->setUniformValue("lightPos", QVector3D(100.0, 100.0, 100.0));

    splitrot_program->setUniformValue("pmv", pmv);
    int stride = 0;//6*sizeof(float);
    splitrot_program->setAttributeBuffer( "VertexPosition", GL_FLOAT, 0, 4, stride );
    splitrot_program->enableAttributeArray( "VertexPosition" );
    splitrot_program->release();
    splitrot_vao.release();
    mkC++;
    //splitrot_indexBuffer.release();
    splitrot_vertexBuffer.release();
    splitrot_vertexBuffer.destroy();
    //bond_indexBuffer.destroy();
    update();
}

void PlayOpenGL::makeSymmetryElements(){

    makeCurrent();
    symm_vao0.release();
    symm_vao1.release();
    symm_vao2.release();
    if (symm_vao0.isCreated()) {
        symm_vao0.release();
        symm_vao0.destroy();
        symm_vao1.release();
        symm_vao1.destroy();
        symm_vao2.release();
        symm_vao2.destroy();
        symm_vertexBuffer0.destroy();
        symm_vertexBuffer1.destroy();
        symm_vertexBuffer2.destroy();
        symm_program0->release();
        symm_program0->removeAllShaders();
        symm_program1->release();
        symm_program1->removeAllShaders();
        symm_program2->release();
        symm_program2->removeAllShaders();
    }
    symm_vao0.create();
    symm_vao0.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    GLfloat scv[18*4];
    for (int i=0;i<18;i++) {
        scv[4*i]=mol->symmColors.at(i).redF();
        scv[4*i+1]=mol->symmColors.at(i).greenF();
        scv[4*i+2]=mol->symmColors.at(i).blueF();
        scv[4*i+3]=mol->symmColors.at(i).alphaF();

    }
    symm_vertexBuffer0.create();
    symm_vertexBuffer0.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!symm_vertexBuffer0.bind()) {printf("I was unable to bind symm_vertexBuffer.\n");}


    QVector<float> p;
    axend = 0;
    invend = 0;
    planend = 0;
    for (int i = 0; i < mol->axvecs.size(); i++){
        p.append(mol->axvecs.at(i).x);
        p.append(mol->axvecs.at(i).y);
        p.append(mol->axvecs.at(i).z);
        p.append(mol->axoprs.at(i));
        p.append(mol->axcols.at(i));


    }
    axend = p.size();
    symm_vertexBuffer0.allocate(p.data(),(p.size()*sizeof(float)));

    if (!symm_program0) symm_program0 = new QOpenGLShaderProgram(this);
    else symm_program0->removeAllShaders();
    bool result = symm_program0->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/symm.vert");
    if ( !result ) qWarning() << symm_program0->log();
    result = symm_program0->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/symmaxe.geom");
    if ( !result ) qWarning() << symm_program0->log();
    result = symm_program0->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/symm.frag");
    if ( !result ) qWarning() << symm_program0->log();
    result = symm_program0->link();
    if ( !result ) qWarning() << symm_program0->log();
    if ( !symm_program0->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }
    bool beforeInfo = shaderInfo;
    //shaderInfo=false;//true;

    symm_program0->enableAttributeArray("vertex" );
    symm_program0->enableAttributeArray("oprf" );// */
    symm_program0->enableAttributeArray("colf" );// */
    int stride =  5 * sizeof(float);

    symm_program0->setAttributeBuffer("vertex", GL_FLOAT, 0, 3 , stride);
    symm_program0->setAttributeBuffer("oprf", GL_FLOAT, sizeof(float)*3, 1, stride);
    symm_program0->setAttributeBuffer("colf", GL_FLOAT, sizeof(float)*4, 1, stride);

    symm_program0->setUniformValueArray("Color",scv,18,4);//we have to this for color changes too

    symm_program0->setUniformValue("lightPos", QVector3D(100.0, 100.0, 100.0));
    programInfo(symm_program0, "symm0");
    symm_program0->release();
    symm_vertexBuffer0.release();
    p.clear();

    symm_vao0.release();
    symm_vao1.create();
    symm_vao1.bind();

    for (int i = 0; i < mol->invvecs.size(); i++){
        p.append(mol->invvecs.at(i).x);
        p.append(mol->invvecs.at(i).y);
        p.append(mol->invvecs.at(i).z);
        p.append(mol->invoprs.at(i));
        p.append(5.0f);
    }
    symm_vertexBuffer1.create();
    symm_vertexBuffer1.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!symm_vertexBuffer1.bind()) {printf("I was unable to bind symm_vertexBuffer.\n");}
    invend = p.size();
    symm_vertexBuffer1.allocate(p.data(),(p.size()*sizeof(float)));

    if (!symm_program1) symm_program1 = new QOpenGLShaderProgram(this);
    else symm_program1->removeAllShaders();
    result = symm_program1->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/symm.vert");

    if ( !result ) qWarning() << symm_program1->log();
    result = symm_program1->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/symm.geom");
    if ( !result ) qWarning() << symm_program1->log();
    result = symm_program1->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/symm.frag");
    if ( !result ) qWarning() << symm_program1->log();
    result = symm_program1->link();
    if ( !result ) qWarning() << symm_program1->log();
    if ( !symm_program1->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }
    symm_program1->setUniformValueArray("Color",scv,18,4);//we have to this for color changes too
    symm_program1->setUniformValue("lightPos", QVector3D(100.0, 100.0, 100.0));
    //*
    symm_program1->enableAttributeArray("vertex" );
    symm_program1->enableAttributeArray("oprf" );
    symm_program1->enableAttributeArray("colf" );
    symm_program1->setAttributeBuffer("vertex", GL_FLOAT, 0, 3 , stride);
    symm_program1->setAttributeBuffer("oprf", GL_FLOAT, sizeof(float)*3, 1, stride);
    symm_program1->setAttributeBuffer("colf", GL_FLOAT, sizeof(float)*4, 1, stride);
// */
    programInfo(symm_program1, "symm1");
    symm_program1->release();
    symm_vao1.release();
    p.clear();
    symm_vao2.create();
    symm_vao2.bind();

    for (int ti = 0; ti < mol->ptriangles.size(); ti++){
        p.append(mol->ptriangles.at(ti).nor.x);
        p.append(mol->ptriangles.at(ti).nor.y);
        p.append(mol->ptriangles.at(ti).nor.z);

        p.append(mol->ptriangles.at(ti).mid.x);
        p.append(mol->ptriangles.at(ti).mid.y);
        p.append(mol->ptriangles.at(ti).mid.z);

        p.append(mol->ptriangles.at(ti).verts0.x);
        p.append(mol->ptriangles.at(ti).verts0.y);
        p.append(mol->ptriangles.at(ti).verts0.z);

        p.append(mol->ptriangles.at(ti).verts1.x);
        p.append(mol->ptriangles.at(ti).verts1.y);
        p.append(mol->ptriangles.at(ti).verts1.z);

        p.append(mol->ptriangles.at(ti).glide.x);
        p.append(mol->ptriangles.at(ti).glide.y);
        p.append(mol->ptriangles.at(ti).glide.z);

        p.append(mol->ptriangles.at(ti).n);
        p.append(mol->ptriangles.at(ti).colr);


        //p.append(ti%18);
        /*
       int pp = p.size();
        printf("n%8.4f %8.4f %8.4f m%8.4f %8.4f %8.4f I:%8.4f %8.4f %8.4f II:%8.4f %8.4f %8.4f g %8.4f %3.0f %f n=%3.0f c=%f\n",
               p[pp-17], p[pp-16], p[pp-15],
                p[pp-14], p[pp-13], p[pp-12],
                p[pp-11], p[pp-10], p[pp-9],
                p[pp-8], p[pp-7], p[pp-6],
                p[pp-5], p[pp-4], p[pp-3],
                p[pp-2],
                p[pp-1]);// * /
        {
            V3 start = mol->ptriangles.at(ti).verts0;
            V3 end = mol->ptriangles.at(ti).verts1;
            V3 direction   = Normalize(end - start);
            double l = sqrt(Norm(end - start));
            V3 lot = (direction*V3(1, 0, 0) > 0.8)?
                        (direction*V3(0, 1, 0) > 0.8)?
                        Normalize(direction% V3(0, 0, 1)):
                        Normalize(direction% V3(0, 1, 0)):
                        Normalize(direction% V3(1, 0, 0));
            V3 lot2 = Normalize(direction % lot);
         printf("%f (%f %f %f) lot (%f %f %f)  lot2 (%f %f %f)\n",l,direction.x,direction.y,direction.z,lot.x,lot.y,lot.z,lot2.x,lot2.y,lot2.z);
        }// */
    }



    glend = p.size();
    symm_vertexBuffer2.create();
    symm_vertexBuffer2.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!symm_vertexBuffer2.bind()) {printf("I was unable to bind symm_vertexBuffer.\n");}
    symm_vertexBuffer2.allocate(p.data(),(p.size()*sizeof(float)));

    if (!symm_program2) symm_program2 = new QOpenGLShaderProgram(this);
    else symm_program2->removeAllShaders();
    result = symm_program2->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/symmplane.vert");

    if ( !result ) qWarning() << symm_program2->log();
    result = symm_program2->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/symmplane.geom");
    if ( !result ) qWarning() << symm_program2->log();
    result = symm_program2->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/symmplane.frag");
    if ( !result ) qWarning() << symm_program2->log();
    result = symm_program2->link();
    if ( !result ) qWarning() << symm_program2->log();
    if ( !symm_program2->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }
    stride =  17 * sizeof(float);
    symm_program2->setUniformValueArray("Color",scv,18,4);//we have to this for color changes too
    symm_program2->setUniformValue("lightPos", QVector3D(100.0, 100.0, 100.0));
    //*
    symm_program2->enableAttributeArray("nor");
    symm_program2->enableAttributeArray("mid");
    symm_program2->enableAttributeArray("verts0");
    symm_program2->enableAttributeArray("verts1");
    symm_program2->enableAttributeArray("glide");
    symm_program2->enableAttributeArray("oprf");
    symm_program2->enableAttributeArray("colf");
    symm_program2->setAttributeBuffer("nor", GL_FLOAT, 0, 3 , stride);
    symm_program2->setAttributeBuffer("mid", GL_FLOAT, sizeof(float)*3, 3, stride);
    symm_program2->setAttributeBuffer("verts0", GL_FLOAT, sizeof(float)*6, 3, stride);
    symm_program2->setAttributeBuffer("verts1", GL_FLOAT, sizeof(float)*9, 3, stride);
    symm_program2->setAttributeBuffer("glide", GL_FLOAT, sizeof(float)*12, 3, stride);
    symm_program2->setAttributeBuffer("oprf", GL_FLOAT, sizeof(float)*15, 1, stride);
    symm_program2->setAttributeBuffer("colf", GL_FLOAT, sizeof(float)*16, 1, stride);

    programInfo(symm_program2, "symm2");
    symm_program2->release();
    symm_vao2.release();
    mkD++;


    shaderInfo=beforeInfo;
}


void PlayOpenGL::programInfo(QOpenGLShaderProgram *p, const QString pname){
    if (!shaderInfo) return;
    GLint count;

    GLint size; // size of the variable
    GLenum type; // type of the variable (float, vec3 or mat4, etc)
    char cnam[100];
    sprintf(cnam, "%s", pname.toStdString().c_str());

    const GLsizei bufSize = 20; // maximum name length
    GLchar name[bufSize]; // variable name in GLSL
    GLsizei length; // name length
    //glGetProgramiv(p->programId(),  GL_ACTIVE_ATTRIBUTE_MAX_LENGTH, &count);


    glGetProgramiv(p->programId(), GL_ACTIVE_ATTRIBUTES, &count);
    printf("\n#######################################################################################################################\n"
           "%-20s Active Attributes: %d\n#######################################################################################################################\n", cnam, count);


    for (GLint i = 0; i < count; i++){
        glGetActiveAttrib(p->programId(), (GLuint)i, bufSize, &length, &size, &type, name);

        printf("%-20s Attribute #%-2d Type: %-23s Name: %-20s stringlength=%2d size=%d (%d)\n"
               , cnam
               , i
               , glTypeInt2String(type).toStdString().c_str()
               , name, length
               , size, p->attributeLocation(name));
    }
    glGetProgramiv(p->programId(), GL_ACTIVE_UNIFORMS, &count);
    printf("%-20s Active Uniforms: %d\n#######################################################################################################################\n", cnam, count);

    for (GLint i = 0; i < count; i++){
        glGetActiveUniform(p->programId(), (GLuint)i, bufSize, &length, &size, &type, name);

        printf("%-20s Uniform #%-2d   Type: %-23s Name: %-20s stringlength=%2d size=%d\n", cnam, i, glTypeInt2String(type).toStdString().c_str(), name,length, size);
    }
}
/*
void PlayOpenGL::makeClassicAtoms(const CEnvironment &atoms){

    printf("makeClassicAtoms\n");
    rotze = (!isMainWindow)?0:-1;
    printf("makeAtoms rotation center set to: %d\n",rotze);
    makeCurrent();
    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR)
    {
      printf("before makeAtoms %s error %X\n", parent()->objectName().toStdString().c_str(),err);

    }
    catom_vao.release();
    if (catom_vao.isCreated()) {
        catom_vao.release();
        catom_vao.destroy();
        catom_vertexBuffer.destroy();
        catom_indexBuffer.destroy();
        catom_program->release();
        catom_program->removeAllShaders();
    }
    catom_vao.create();
    catom_vao.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    catom_vertexBuffer.create();
    catom_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!catom_vertexBuffer.bind()) {printf("I was unable to bind classic atom_vertexBuffer.\n");}
    static const GLuint idx2[960] = {
      1,     42,     44,//  0 0.275904 0.321244 0.275904
      42,     12,     43,//  1 0.275904 0.312869 0.285473
      43,     14,     44,//  2 0.312869 0.275904 0.285473
      42,     43,     44,//  3 0.285473 0.285473 0.321244
      12,     45,     47,//  4 0.275904 0.285473 0.312869
      45,      0,     46,//  5 0.275904 0.275904 0.321244
      46,     13,     47,//  6 0.275904 0.312869 0.285473
      45,     46,     47,//  7 0.321244 0.285473 0.285473
      13,     48,     50,//  8 0.275904 0.285473 0.312869
      48,      5,     49,//  9 0.275904 0.275904 0.321244
      49,     14,     50,// 10 0.275904 0.312869 0.285473
      48,     49,     50,// 11 0.321244 0.285473 0.285473
      12,     47,     43,// 12 0.312869 0.324920 0.312869
      47,     13,     50,// 13 0.312869 0.312869 0.324920
      50,     14,     43,// 14 0.312869 0.312869 0.324920
      47,     50,     43,// 15 0.324920 0.324920 0.324920
      0,     45,     52,// 16 0.275904 0.321244 0.275904
      45,     12,     51,// 17 0.275904 0.312869 0.285473
      51,     16,     52,// 18 0.312869 0.275904 0.285473
      45,     51,     52,// 19 0.285473 0.285473 0.321244
      12,     42,     54,// 20 0.275904 0.285473 0.312869
      42,      1,     53,// 21 0.275904 0.275904 0.321244
      53,     15,     54,// 22 0.275904 0.312869 0.285473
      42,     53,     54,// 23 0.321244 0.285473 0.285473
      15,     55,     57,// 24 0.275904 0.285473 0.312869
      55,      7,     56,// 25 0.275904 0.275904 0.321244
      56,     16,     57,// 26 0.275904 0.312869 0.285473
      55,     56,     57,// 27 0.321244 0.285473 0.285473
      12,     54,     51,// 28 0.312869 0.324920 0.312869
      54,     15,     57,// 29 0.312869 0.312869 0.324920
      57,     16,     51,// 30 0.312869 0.312869 0.324920
      54,     57,     51,// 31 0.324920 0.324920 0.324920
      10,     58,     60,// 32 0.275904 0.321244 0.275904
      58,     17,     59,// 33 0.275904 0.312869 0.285473
      59,     18,     60,// 34 0.312869 0.275904 0.285473
      58,     59,     60,// 35 0.285473 0.285473 0.321244
      17,     61,     62,// 36 0.275904 0.285473 0.312869
      61,      0,     52,// 37 0.275904 0.275904 0.321244
      52,     16,     62,// 38 0.275904 0.312869 0.285473
      61,     52,     62,// 39 0.321244 0.285473 0.285473
      16,     56,     64,// 40 0.275904 0.285473 0.312869
      56,      7,     63,// 41 0.275904 0.275904 0.321244
      63,     18,     64,// 42 0.275904 0.312869 0.285473
      56,     63,     64,// 43 0.321244 0.285473 0.285473
      17,     62,     59,// 44 0.312869 0.324920 0.312869
      62,     16,     64,// 45 0.312869 0.312869 0.324920
      64,     18,     59,// 46 0.312869 0.312869 0.324920
      62,     64,     59,// 47 0.324920 0.324920 0.324920
      0,     61,     66,// 48 0.275904 0.321244 0.275904
      61,     17,     65,// 49 0.275904 0.312869 0.285473
      65,     20,     66,// 50 0.312869 0.275904 0.285473
      61,     65,     66,// 51 0.285473 0.285473 0.321244
      17,     58,     68,// 52 0.275904 0.285473 0.312869
      58,     10,     67,// 53 0.275904 0.275904 0.321244
      67,     19,     68,// 54 0.275904 0.312869 0.285473
      58,     67,     68,// 55 0.321244 0.285473 0.285473
      19,     69,     71,// 56 0.275904 0.285473 0.312869
      69,     11,     70,// 57 0.275904 0.275904 0.321244
      70,     20,     71,// 58 0.275904 0.312869 0.285473
      69,     70,     71,// 59 0.321244 0.285473 0.285473
      17,     68,     65,// 60 0.312869 0.324920 0.312869
      68,     19,     71,// 61 0.312869 0.312869 0.324920
      71,     20,     65,// 62 0.312869 0.312869 0.324920
      68,     71,     65,// 63 0.324920 0.324920 0.324920
      11,     72,     70,// 64 0.275904 0.321244 0.275904
      72,     21,     73,// 65 0.275904 0.312869 0.285473
      73,     20,     70,// 66 0.312869 0.275904 0.285473
      72,     73,     70,// 67 0.285473 0.285473 0.321244
      21,     74,     75,// 68 0.275904 0.285473 0.312869
      74,      5,     48,// 69 0.275904 0.275904 0.321244
      48,     13,     75,// 70 0.275904 0.312869 0.285473
      74,     48,     75,// 71 0.321244 0.285473 0.285473
      13,     46,     76,// 72 0.275904 0.285473 0.312869
      46,      0,     66,// 73 0.275904 0.275904 0.321244
      66,     20,     76,// 74 0.275904 0.312869 0.285473
      46,     66,     76,// 75 0.321244 0.285473 0.285473
      21,     75,     73,// 76 0.312869 0.324920 0.312869
      75,     13,     76,// 77 0.312869 0.312869 0.324920
      76,     20,     73,// 78 0.312869 0.312869 0.324920
      75,     76,     73,// 79 0.324920 0.324920 0.324920
      5,     74,     78,// 80 0.275904 0.321244 0.275904
      74,     21,     77,// 81 0.275904 0.312869 0.285473
      77,     23,     78,// 82 0.312869 0.275904 0.285473
      74,     77,     78,// 83 0.285473 0.285473 0.321244
      21,     72,     80,// 84 0.275904 0.285473 0.312869
      72,     11,     79,// 85 0.275904 0.275904 0.321244
      79,     22,     80,// 86 0.275904 0.312869 0.285473
      72,     79,     80,// 87 0.321244 0.285473 0.285473
      22,     81,     83,// 88 0.275904 0.285473 0.312869
      81,      4,     82,// 89 0.275904 0.275904 0.321244
      82,     23,     83,// 90 0.275904 0.312869 0.285473
      81,     82,     83,// 91 0.321244 0.285473 0.285473
      21,     80,     77,// 92 0.312869 0.324920 0.312869
      80,     22,     83,// 93 0.312869 0.312869 0.324920
      83,     23,     77,// 94 0.312869 0.312869 0.324920
      80,     83,     77,// 95 0.324920 0.324920 0.324920
      2,     84,     86,// 96 0.275904 0.321244 0.275904
      84,     24,     85,// 97 0.275904 0.312869 0.285473
      85,     25,     86,// 98 0.312869 0.275904 0.285473
      84,     85,     86,// 99 0.285473 0.285473 0.321244
      24,     87,     88,//100 0.275904 0.285473 0.312869
      87,     11,     69,//101 0.275904 0.275904 0.321244
      69,     19,     88,//102 0.275904 0.312869 0.285473
      87,     69,     88,//103 0.321244 0.285473 0.285473
      19,     67,     90,//104 0.275904 0.285473 0.312869
      67,     10,     89,//105 0.275904 0.275904 0.321244
      89,     25,     90,//106 0.275904 0.312869 0.285473
      67,     89,     90,//107 0.321244 0.285473 0.285473
      24,     88,     85,//108 0.312869 0.324920 0.312869
      88,     19,     90,//109 0.312869 0.312869 0.324920
      90,     25,     85,//110 0.312869 0.312869 0.324920
      88,     90,     85,//111 0.324920 0.324920 0.324920
      11,     87,     79,//112 0.275904 0.321244 0.275904
      87,     24,     91,//113 0.275904 0.312869 0.285473
      91,     22,     79,//114 0.312869 0.275904 0.285473
      87,     91,     79,//115 0.285473 0.285473 0.321244
      24,     84,     93,//116 0.275904 0.285473 0.312869
      84,      2,     92,//117 0.275904 0.275904 0.321244
      92,     26,     93,//118 0.275904 0.312869 0.285473
      84,     92,     93,//119 0.321244 0.285473 0.285473
      26,     94,     95,//120 0.275904 0.285473 0.312869
      94,      4,     81,//121 0.275904 0.275904 0.321244
      81,     22,     95,//122 0.275904 0.312869 0.285473
      94,     81,     95,//123 0.321244 0.285473 0.285473
      24,     93,     91,//124 0.312869 0.324920 0.312869
      93,     26,     95,//125 0.312869 0.312869 0.324920
      95,     22,     91,//126 0.312869 0.312869 0.324920
      93,     95,     91,//127 0.324920 0.324920 0.324920
      10,     96,     89,//128 0.275904 0.321244 0.275904
      96,     27,     97,//129 0.275904 0.312869 0.285473
      97,     25,     89,//130 0.312869 0.275904 0.285473
      96,     97,     89,//131 0.285473 0.285473 0.321244
      27,     98,    100,//132 0.275904 0.285473 0.312869
      98,      6,     99,//133 0.275904 0.275904 0.321244
      99,     28,    100,//134 0.275904 0.312869 0.285473
      98,     99,    100,//135 0.321244 0.285473 0.285473
      28,    101,    102,//136 0.275904 0.285473 0.312869
      101,      2,     86,//137 0.275904 0.275904 0.321244
      86,     25,    102,//138 0.275904 0.312869 0.285473
      101,     86,    102,//139 0.321244 0.285473 0.285473
      27,    100,     97,//140 0.312869 0.324920 0.312869
      100,     28,    102,//141 0.312869 0.312869 0.324920
      102,     25,     97,//142 0.312869 0.312869 0.324920
      100,    102,     97,//143 0.324920 0.324920 0.324920
      6,     98,    104,//144 0.275904 0.321244 0.275904
      98,     27,    103,//145 0.275904 0.312869 0.285473
      103,     29,    104,//146 0.312869 0.275904 0.285473
      98,    103,    104,//147 0.285473 0.285473 0.321244
      27,     96,    105,//148 0.275904 0.285473 0.312869
      96,     10,     60,//149 0.275904 0.275904 0.321244
      60,     18,    105,//150 0.275904 0.312869 0.285473
      96,     60,    105,//151 0.321244 0.285473 0.285473
      18,     63,    107,//152 0.275904 0.285473 0.312869
      63,      7,    106,//153 0.275904 0.275904 0.321244
      106,     29,    107,//154 0.275904 0.312869 0.285473
      63,    106,    107,//155 0.321244 0.285473 0.285473
      27,    105,    103,//156 0.312869 0.324920 0.312869
      105,     18,    107,//157 0.312869 0.312869 0.324920
      107,     29,    103,//158 0.312869 0.312869 0.324920
      105,    107,    103,//159 0.324920 0.324920 0.324920
      7,    108,    106,//160 0.275904 0.321244 0.275904
      108,     30,    109,//161 0.275904 0.312869 0.285473
      109,     29,    106,//162 0.312869 0.275904 0.285473
      108,    109,    106,//163 0.285473 0.285473 0.321244
      30,    110,    112,//164 0.275904 0.285473 0.312869
      110,      8,    111,//165 0.275904 0.275904 0.321244
      111,     31,    112,//166 0.275904 0.312869 0.285473
      110,    111,    112,//167 0.321244 0.285473 0.285473
      31,    113,    114,//168 0.275904 0.285473 0.312869
      113,      6,    104,//169 0.275904 0.275904 0.321244
      104,     29,    114,//170 0.275904 0.312869 0.285473
      113,    104,    114,//171 0.321244 0.285473 0.285473
      30,    112,    109,//172 0.312869 0.324920 0.312869
      112,     31,    114,//173 0.312869 0.312869 0.324920
      114,     29,    109,//174 0.312869 0.312869 0.324920
      112,    114,    109,//175 0.324920 0.324920 0.324920
      8,    110,    116,//176 0.275904 0.321244 0.275904
      110,     30,    115,//177 0.275904 0.312869 0.285473
      115,     32,    116,//178 0.312869 0.275904 0.285473
      110,    115,    116,//179 0.285473 0.285473 0.321244
      30,    108,    117,//180 0.275904 0.285473 0.312869
      108,      7,     55,//181 0.275904 0.275904 0.321244
      55,     15,    117,//182 0.275904 0.312869 0.285473
      108,     55,    117,//183 0.321244 0.285473 0.285473
      15,     53,    119,//184 0.275904 0.285473 0.312869
      53,      1,    118,//185 0.275904 0.275904 0.321244
      118,     32,    119,//186 0.275904 0.312869 0.285473
      53,    118,    119,//187 0.321244 0.285473 0.285473
      30,    117,    115,//188 0.312869 0.324920 0.312869
      117,     15,    119,//189 0.312869 0.312869 0.324920
      119,     32,    115,//190 0.312869 0.312869 0.324920
      117,    119,    115,//191 0.324920 0.324920 0.324920
      9,    120,    122,//192 0.275904 0.321244 0.275904
      120,     33,    121,//193 0.275904 0.312869 0.285473
      121,     34,    122,//194 0.312869 0.275904 0.285473
      120,    121,    122,//195 0.285473 0.285473 0.321244
      33,    123,    124,//196 0.275904 0.285473 0.312869
      123,      1,     44,//197 0.275904 0.275904 0.321244
      44,     14,    124,//198 0.275904 0.312869 0.285473
      123,     44,    124,//199 0.321244 0.285473 0.285473
      14,     49,    126,//200 0.275904 0.285473 0.312869
      49,      5,    125,//201 0.275904 0.275904 0.321244
      125,     34,    126,//202 0.275904 0.312869 0.285473
      49,    125,    126,//203 0.321244 0.285473 0.285473
      33,    124,    121,//204 0.312869 0.324920 0.312869
      124,     14,    126,//205 0.312869 0.312869 0.324920
      126,     34,    121,//206 0.312869 0.312869 0.324920
      124,    126,    121,//207 0.324920 0.324920 0.324920
      1,    123,    118,//208 0.275904 0.321244 0.275904
      123,     33,    127,//209 0.275904 0.312869 0.285473
      127,     32,    118,//210 0.312869 0.275904 0.285473
      123,    127,    118,//211 0.285473 0.285473 0.321244
      33,    120,    129,//212 0.275904 0.285473 0.312869
      120,      9,    128,//213 0.275904 0.275904 0.321244
      128,     35,    129,//214 0.275904 0.312869 0.285473
      120,    128,    129,//215 0.321244 0.285473 0.285473
      35,    130,    131,//216 0.275904 0.285473 0.312869
      130,      8,    116,//217 0.275904 0.275904 0.321244
      116,     32,    131,//218 0.275904 0.312869 0.285473
      130,    116,    131,//219 0.321244 0.285473 0.285473
      33,    129,    127,//220 0.312869 0.324920 0.312869
      129,     35,    131,//221 0.312869 0.312869 0.324920
      131,     32,    127,//222 0.312869 0.312869 0.324920
      129,    131,    127,//223 0.324920 0.324920 0.324920
      4,    132,     82,//224 0.275904 0.321244 0.275904
      132,     36,    133,//225 0.275904 0.312869 0.285473
      133,     23,     82,//226 0.312869 0.275904 0.285473
      132,    133,     82,//227 0.285473 0.285473 0.321244
      36,    134,    135,//228 0.275904 0.285473 0.312869
      134,      9,    122,//229 0.275904 0.275904 0.321244
      122,     34,    135,//230 0.275904 0.312869 0.285473
      134,    122,    135,//231 0.321244 0.285473 0.285473
      34,    125,    136,//232 0.275904 0.285473 0.312869
      125,      5,     78,//233 0.275904 0.275904 0.321244
      78,     23,    136,//234 0.275904 0.312869 0.285473
      125,     78,    136,//235 0.321244 0.285473 0.285473
      36,    135,    133,//236 0.312869 0.324920 0.312869
      135,     34,    136,//237 0.312869 0.312869 0.324920
      136,     23,    133,//238 0.312869 0.312869 0.324920
      135,    136,    133,//239 0.324920 0.324920 0.324920
      9,    134,    138,//240 0.275904 0.321244 0.275904
      134,     36,    137,//241 0.275904 0.312869 0.285473
      137,     38,    138,//242 0.312869 0.275904 0.285473
      134,    137,    138,//243 0.285473 0.285473 0.321244
      36,    132,    140,//244 0.275904 0.285473 0.312869
      132,      4,    139,//245 0.275904 0.275904 0.321244
      139,     37,    140,//246 0.275904 0.312869 0.285473
      132,    139,    140,//247 0.321244 0.285473 0.285473
      37,    141,    143,//248 0.275904 0.285473 0.312869
      141,      3,    142,//249 0.275904 0.275904 0.321244
      142,     38,    143,//250 0.275904 0.312869 0.285473
      141,    142,    143,//251 0.321244 0.285473 0.285473
      36,    140,    137,//252 0.312869 0.324920 0.312869
      140,     37,    143,//253 0.312869 0.312869 0.324920
      143,     38,    137,//254 0.312869 0.312869 0.324920
      140,    143,    137,//255 0.324920 0.324920 0.324920
      3,    144,    142,//256 0.275904 0.321244 0.275904
      144,     39,    145,//257 0.275904 0.312869 0.285473
      145,     38,    142,//258 0.312869 0.275904 0.285473
      144,    145,    142,//259 0.285473 0.285473 0.321244
      39,    146,    147,//260 0.275904 0.285473 0.312869
      146,      8,    130,//261 0.275904 0.275904 0.321244
      130,     35,    147,//262 0.275904 0.312869 0.285473
      146,    130,    147,//263 0.321244 0.285473 0.285473
      35,    128,    148,//264 0.275904 0.285473 0.312869
      128,      9,    138,//265 0.275904 0.275904 0.321244
      138,     38,    148,//266 0.275904 0.312869 0.285473
      128,    138,    148,//267 0.321244 0.285473 0.285473
      39,    147,    145,//268 0.312869 0.324920 0.312869
      147,     35,    148,//269 0.312869 0.312869 0.324920
      148,     38,    145,//270 0.312869 0.312869 0.324920
      147,    148,    145,//271 0.324920 0.324920 0.324920
      8,    146,    111,//272 0.275904 0.321244 0.275904
      146,     39,    149,//273 0.275904 0.312869 0.285473
      149,     31,    111,//274 0.312869 0.275904 0.285473
      146,    149,    111,//275 0.285473 0.285473 0.321244
      39,    144,    151,//276 0.275904 0.285473 0.312869
      144,      3,    150,//277 0.275904 0.275904 0.321244
      150,     40,    151,//278 0.275904 0.312869 0.285473
      144,    150,    151,//279 0.321244 0.285473 0.285473
      40,    152,    153,//280 0.275904 0.285473 0.312869
      152,      6,    113,//281 0.275904 0.275904 0.321244
      113,     31,    153,//282 0.275904 0.312869 0.285473
      152,    113,    153,//283 0.321244 0.285473 0.285473
      39,    151,    149,//284 0.312869 0.324920 0.312869
      151,     40,    153,//285 0.312869 0.312869 0.324920
      153,     31,    149,//286 0.312869 0.312869 0.324920
      151,    153,    149,//287 0.324920 0.324920 0.324920
      3,    154,    150,//288 0.275904 0.321244 0.275904
      154,     41,    155,//289 0.275904 0.312869 0.285473
      155,     40,    150,//290 0.312869 0.275904 0.285473
      154,    155,    150,//291 0.285473 0.285473 0.321244
      41,    156,    157,//292 0.275904 0.285473 0.312869
      156,      2,    101,//293 0.275904 0.275904 0.321244
      101,     28,    157,//294 0.275904 0.312869 0.285473
      156,    101,    157,//295 0.321244 0.285473 0.285473
      28,     99,    158,//296 0.275904 0.285473 0.312869
      99,      6,    152,//297 0.275904 0.275904 0.321244
      152,     40,    158,//298 0.275904 0.312869 0.285473
      99,    152,    158,//299 0.321244 0.285473 0.285473
      41,    157,    155,//300 0.312869 0.324920 0.312869
      157,     28,    158,//301 0.312869 0.312869 0.324920
      158,     40,    155,//302 0.312869 0.312869 0.324920
      157,    158,    155,//303 0.324920 0.324920 0.324920
      2,    156,     92,//304 0.275904 0.321244 0.275904
      156,     41,    159,//305 0.275904 0.312869 0.285473
      159,     26,     92,//306 0.312869 0.275904 0.285473
      156,    159,     92,//307 0.285473 0.285473 0.321244
      41,    154,    160,//308 0.275904 0.285473 0.312869
      154,      3,    141,//309 0.275904 0.275904 0.321244
      141,     37,    160,//310 0.275904 0.312869 0.285473
      154,    141,    160,//311 0.321244 0.285473 0.285473
      37,    139,    161,//312 0.275904 0.285473 0.312869
      139,      4,     94,//313 0.275904 0.275904 0.321244
      94,     26,    161,//314 0.275904 0.312869 0.285473
      139,     94,    161,//315 0.321244 0.285473 0.285473
      41,    160,    159,//316 0.312869 0.324920 0.312869
      160,     37,    161,//317 0.312869 0.312869 0.324920
      161,     26,    159,//318 0.312869 0.312869 0.324920
      160,    161,    159 //319 0.324920 0.324920 0.324920
    };
    static const GLfloat v[486] = {
      -0.525731f,   0.850651f,   0.000000f,//0
      0.525731f,   0.850651f,   0.000000f,//1
      -0.525731f,  -0.850651f,   0.000000f,//2
      0.525731f,  -0.850651f,   0.000000f,//3
      0.000000f,  -0.525731f,   0.850651f,//4
      0.000000f,   0.525731f,   0.850651f,//5
      0.000000f,  -0.525731f,  -0.850651f,//6
      0.000000f,   0.525731f,  -0.850651f,//7
      0.850651f,   0.000000f,  -0.525731f,//8
      0.850651f,   0.000000f,   0.525731f,//9
      -0.850651f,   0.000000f,  -0.525731f,//10
      -0.850651f,   0.000000f,   0.525731f,//11
      0.000000f,   1.000000f,   0.000000f,//12
      -0.309017f,   0.809017f,   0.500000f,//13
      0.309017f,   0.809017f,   0.500000f,//14
      0.309017f,   0.809017f,  -0.500000f,//15
      -0.309017f,   0.809017f,  -0.500000f,//16
      -0.809017f,   0.500000f,  -0.309017f,//17
      -0.500000f,   0.309017f,  -0.809017f,//18
      -1.000000f,   0.000000f,   0.000000f,//19
      -0.809017f,   0.500000f,   0.309017f,//20
      -0.500000f,   0.309017f,   0.809017f,//21
      -0.500000f,  -0.309017f,   0.809017f,//22
      0.000000f,   0.000000f,   1.000000f,//23
      -0.809017f,  -0.500000f,   0.309017f,//24
      -0.809017f,  -0.500000f,  -0.309017f,//25
      -0.309017f,  -0.809017f,   0.500000f,//26
      -0.500000f,  -0.309017f,  -0.809017f,//27
      -0.309017f,  -0.809017f,  -0.500000f,//28
      0.000000f,   0.000000f,  -1.000000f,//29
      0.500000f,   0.309017f,  -0.809017f,//30
      0.500000f,  -0.309017f,  -0.809017f,//31
      0.809017f,   0.500000f,  -0.309017f,//32
      0.809017f,   0.500000f,   0.309017f,//33
      0.500000f,   0.309017f,   0.809017f,//34
      1.000000f,   0.000000f,   0.000000f,//35
      0.500000f,  -0.309017f,   0.809017f,//36
      0.309017f,  -0.809017f,   0.500000f,//37
      0.809017f,  -0.500000f,   0.309017f,//38
      0.809017f,  -0.500000f,  -0.309017f,//39
      0.309017f,  -0.809017f,  -0.500000f,//40
      0.000000f,  -1.000000f,   0.000000f,//41
      0.273267f,   0.961938f,   0.000000f,//42
      0.162460f,   0.951057f,   0.262866f,//43
      0.433889f,   0.862668f,   0.259892f,//44
      -0.273267f,   0.961938f,   0.000000f,//45
      -0.433889f,   0.862668f,   0.259892f,//46
      -0.162460f,   0.951057f,   0.262866f,//47
      -0.160622f,   0.693780f,   0.702046f,//48
      0.160622f,   0.693780f,   0.702046f,//49
      0.000000f,   0.850651f,   0.525731f,//50
      -0.162460f,   0.951057f,  -0.262866f,//51
      -0.433889f,   0.862668f,  -0.259892f,//52
      0.433889f,   0.862668f,  -0.259892f,//53
      0.162460f,   0.951057f,  -0.262866f,//54
      0.160622f,   0.693780f,  -0.702046f,//55
      -0.160622f,   0.693780f,  -0.702046f,//56
      0.000000f,   0.850651f,  -0.525731f,//57
      -0.862668f,   0.259892f,  -0.433889f,//58
      -0.688191f,   0.425325f,  -0.587785f,//59
      -0.702046f,   0.160622f,  -0.693780f,//60
      -0.693780f,   0.702046f,  -0.160622f,//61
      -0.587785f,   0.688191f,  -0.425325f,//62
      -0.259892f,   0.433889f,  -0.862668f,//63
      -0.425325f,   0.587785f,  -0.688191f,//64
      -0.850651f,   0.525731f,   0.000000f,//65
      -0.693780f,   0.702046f,   0.160622f,//66
      -0.961938f,   0.000000f,  -0.273267f,//67
      -0.951057f,   0.262866f,  -0.162460f,//68
      -0.961938f,   0.000000f,   0.273267f,//69
      -0.862668f,   0.259892f,   0.433889f,//70
      -0.951057f,   0.262866f,   0.162460f,//71
      -0.702046f,   0.160622f,   0.693780f,//72
      -0.688191f,   0.425325f,   0.587785f,//73
      -0.259892f,   0.433889f,   0.862668f,//74
      -0.425325f,   0.587785f,   0.688191f,//75
      -0.587785f,   0.688191f,   0.425325f,//76
      -0.262866f,   0.162460f,   0.951057f,//77
      0.000000f,   0.273267f,   0.961938f,//78
      -0.702046f,  -0.160622f,   0.693780f,//79
      -0.525731f,   0.000000f,   0.850651f,//80
      -0.259892f,  -0.433889f,   0.862668f,//81
      0.000000f,  -0.273267f,   0.961938f,//82
      -0.262866f,  -0.162460f,   0.951057f,//83
      -0.693780f,  -0.702046f,   0.160622f,//84
      -0.850651f,  -0.525731f,   0.000000f,//85
      -0.693780f,  -0.702046f,  -0.160622f,//86
      -0.862668f,  -0.259892f,   0.433889f,//87
      -0.951057f,  -0.262866f,   0.162460f,//88
      -0.862668f,  -0.259892f,  -0.433889f,//89
      -0.951057f,  -0.262866f,  -0.162460f,//90
      -0.688191f,  -0.425325f,   0.587785f,//91
      -0.433889f,  -0.862668f,   0.259892f,//92
      -0.587785f,  -0.688191f,   0.425325f,//93
      -0.160622f,  -0.693780f,   0.702046f,//94
      -0.425325f,  -0.587785f,   0.688191f,//95
      -0.702046f,  -0.160622f,  -0.693780f,//96
      -0.688191f,  -0.425325f,  -0.587785f,//97
      -0.259892f,  -0.433889f,  -0.862668f,//98
      -0.160622f,  -0.693780f,  -0.702046f,//99
      -0.425325f,  -0.587785f,  -0.688191f,//100
      -0.433889f,  -0.862668f,  -0.259892f,//101
      -0.587785f,  -0.688191f,  -0.425325f,//102
      -0.262866f,  -0.162460f,  -0.951057f,//103
      0.000000f,  -0.273267f,  -0.961938f,//104
      -0.525731f,   0.000000f,  -0.850651f,//105
      0.000000f,   0.273267f,  -0.961938f,//106
      -0.262866f,   0.162460f,  -0.951057f,//107
      0.259892f,   0.433889f,  -0.862668f,//108
      0.262866f,   0.162460f,  -0.951057f,//109
      0.702046f,   0.160622f,  -0.693780f,//110
      0.702046f,  -0.160622f,  -0.693780f,//111
      0.525731f,   0.000000f,  -0.850651f,//112
      0.259892f,  -0.433889f,  -0.862668f,//113
      0.262866f,  -0.162460f,  -0.951057f,//114
      0.688191f,   0.425325f,  -0.587785f,//115
      0.862668f,   0.259892f,  -0.433889f,//116
      0.425325f,   0.587785f,  -0.688191f,//117
      0.693780f,   0.702046f,  -0.160622f,//118
      0.587785f,   0.688191f,  -0.425325f,//119
      0.862668f,   0.259892f,   0.433889f,//120
      0.688191f,   0.425325f,   0.587785f,//121
      0.702046f,   0.160622f,   0.693780f,//122
      0.693780f,   0.702046f,   0.160622f,//123
      0.587785f,   0.688191f,   0.425325f,//124
      0.259892f,   0.433889f,   0.862668f,//125
      0.425325f,   0.587785f,   0.688191f,//126
      0.850651f,   0.525731f,   0.000000f,//127
      0.961938f,   0.000000f,   0.273267f,//128
      0.951057f,   0.262866f,   0.162460f,//129
      0.961938f,   0.000000f,  -0.273267f,//130
      0.951057f,   0.262866f,  -0.162460f,//131
      0.259892f,  -0.433889f,   0.862668f,//132
      0.262866f,  -0.162460f,   0.951057f,//133
      0.702046f,  -0.160622f,   0.693780f,//134
      0.525731f,   0.000000f,   0.850651f,//135
      0.262866f,   0.162460f,   0.951057f,//136
      0.688191f,  -0.425325f,   0.587785f,//137
      0.862668f,  -0.259892f,   0.433889f,//138
      0.160622f,  -0.693780f,   0.702046f,//139
      0.425325f,  -0.587785f,   0.688191f,//140
      0.433889f,  -0.862668f,   0.259892f,//141
      0.693780f,  -0.702046f,   0.160622f,//142
      0.587785f,  -0.688191f,   0.425325f,//143
      0.693780f,  -0.702046f,  -0.160622f,//144
      0.850651f,  -0.525731f,   0.000000f,//145
      0.862668f,  -0.259892f,  -0.433889f,//146
      0.951057f,  -0.262866f,  -0.162460f,//147
      0.951057f,  -0.262866f,   0.162460f,//148
      0.688191f,  -0.425325f,  -0.587785f,//149
      0.433889f,  -0.862668f,  -0.259892f,//150
      0.587785f,  -0.688191f,  -0.425325f,//151
      0.160622f,  -0.693780f,  -0.702046f,//152
      0.425325f,  -0.587785f,  -0.688191f,//153
      0.273267f,  -0.961938f,   0.000000f,//154
      0.162460f,  -0.951057f,  -0.262866f,//155
      -0.273267f,  -0.961938f,   0.000000f,//156
      -0.162460f,  -0.951057f,  -0.262866f,//157
      0.000000f,  -0.850651f,  -0.525731f,//158
      -0.162460f,  -0.951057f,   0.262866f,//159
      0.162460f,  -0.951057f,   0.262866f,//160
      0.000000f,  -0.850651f,   0.525731f//161
    };

    catom_vertexBuffer.allocate(v, 486 * sizeof(float));

    catom_indexBuffer = QOpenGLBuffer(QOpenGLBuffer::IndexBuffer); // Mind: use 'IndexBuffer' here
    catom_indexBuffer.create();
    catom_indexBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
    if (!catom_indexBuffer.bind()) {printf("I was unable to bind atom_indexBuffer!\n");}
    catom_indexBuffer.allocate( 960 * sizeof(GLuint) );
    catom_indexBuffer.write(0, idx2, 960* sizeof(GLuint));
    catom_array_size = 960;


    if (!catom_program) catom_program = new QOpenGLShaderProgram(this);
    else catom_program->removeAllShaders();



    bool result = catom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/catom.vert");

    if ( !result ) qWarning() << catom_program->log();
    result = catom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/catom.frag");

    if ( !result ) qWarning() << catom_program->log();
    result = catom_program->link();
    if ( !result ) qWarning() << catom_program->log();

    if ( !catom_program->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }

    programInfo(catom_program,"atom");
    GLfloat acv[109*3];
    for (int i=0;i<109;i++) {
        acv[3*i]=mol->atomColor(i,false).redF();
        acv[3*i+1]=mol->atomColor(i,false).greenF();
        acv[3*i+2]=mol->atomColor(i,false).blueF();
    }
    GLfloat arv[109];
    for (int i=0;i<109;i++) {
        arv[i]=static_cast<GLfloat>(mol->arad[i]);
    }

    atom_program->setUniformValueArray("atomColors",acv, 109,3);
    atom_program->setUniformValueArray("arad",arv, 109,1);

    atom_program->setUniformValue("pmv", pmv);


    atom_program->enableAttributeArray(0);
    atom_program->enableAttributeArray(1);
    atom_program->enableAttributeArray(2);
    atom_program->enableAttributeArray(3);
    atom_program->enableAttributeArray(4);
    atom_program->enableAttributeArray(5);

    int stride = 6 * 4 * sizeof(float);
    atom_program->setAttributeBuffer(0, GL_FLOAT, 0, 4, stride);
    atom_program->setAttributeBuffer(1, GL_FLOAT, sizeof(float)*4  , 4, stride);
    atom_program->setAttributeBuffer(2, GL_FLOAT, sizeof(float)*8  , 4, stride);
    atom_program->setAttributeBuffer(3, GL_FLOAT, sizeof(float)*12 , 4, stride);
    atom_program->setAttributeBuffer(4, GL_FLOAT, sizeof(float)*16 , 4, stride);
    atom_program->setAttributeBuffer(5, GL_FLOAT, sizeof(float)*20 , 4, stride);// 20*sizeof(float)

    atom_program->release();
    atom_vao.release();
    atom_indexBuffer.release();
    atom_vertexBuffer.release();
    atom_vertexBuffer.destroy();
    atom_indexBuffer.destroy();
}
// */

void PlayOpenGL::makeAtoms(const CEnvironment &atoms){
    /**
     * @brief this the initializer of a (crazy) idea of shader program.
     * @param atoms is a QList of MyAtoms
     * the idea is to construct atoms based on an icosahedron geometry and pass atom properties via 6 vec4 attributes.
     * from an icosahedron it creates
     * - an icosahedron
     * - a ball
     * - an ellipsoid
     * - a dodecahedron
     * - a cube
     * - a triacontahedron
     * - intersecting planes for adp
     **/

    if (AMD_Hack) {
        makeAtomsWOGS(atoms);
        return;
    }

    rotze = (!isMainWindow)?0:(rotze>=atoms.size())?-1:rotze;

    makeCurrent();
    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
        printf("before makeAtoms %s error %X\n", parent()->objectName().toStdString().c_str(),err);

    }
    atom_vao.release();
    if (atom_vao.isCreated()) {
        atom_vao.release();
        atom_vao.destroy();
        atom_vertexBuffer.destroy();
        //atom_indexBuffer.destroy();
        atom_program->release();
        atom_program->removeAllShaders();
    }
    makeLabels();

    int zz=0;
    sumse=V3(0,0,0);
    for (int i = 0; i < atoms.size(); i++){
        if ((!atoms.at(i).hidden)) {//(mol->showatoms.at(i).an>-1)&&
            sumse += atoms[i].pos;
            zz++;
        }
    }
    if (zz) sumse*=1.0/zz;
    //return;
        //glDisable(GL_BLEND);


    int na = atoms.size();
    //  the Atoms = atoms;


    const float phi = (1.0f + sqrtf(5.0f)) / 2.0f;
    const float du = 1.0f / sqrtf(phi * phi + 1.0f);
    const float dv = phi * du;

    const float eckn[] ={   0.0f,  dv,  du,
                          0.0f,  dv, -du,
                          0.0f, -dv,  du,
                          0.0f, -dv, -du,
                          du,  0.0f,  dv,
                          -du, 0.0f,  dv,
                          du,  0.0f, -dv,
                          -du, 0.0f, -dv,
                          dv,    du, 0.0f,
                          dv,   -du, 0.0f,
                          -dv,   du, 0.0f,
                          -dv,  -du, 0.0f};
    const GLuint triangle[] = {
                               1, 0, 8,
                               4, 0, 5,
                               5, 0, 10,
                               8, 0, 4,
                               10, 0, 1,
                               6, 1, 8,
                               7, 1, 6,
                               10, 1, 7,
                               3, 2, 11,
                               4, 2, 9,
                               5, 2, 4,
                               9, 2, 3,
                               11, 2, 5,
                               6, 3, 7,
                               7, 3, 11,
                               9, 3, 6,
                               8, 4, 9,
                               11, 5, 10,
                               9, 6, 8,
                               10, 7, 11};
    QVector<float> stuff;
    bool wehaveh=false;
    stuff.reserve(na*12*24);
    for (int i = 0; i < na; i++){
        bool isd=((atoms.at(i).an==0)&&(atoms.at(i).Label[0]=='D'));
        int style = (atoms.at(i).an<0)? (atoms.at(i).hidden)?512:0:
                        mol->AtomStyle[atoms.at(i).an]
                            |((atoms.at(i).hidden)?512:0)
                            |((isd)?1024:0)
                            |(((atoms.at(i).isIso))?2048:0);
        if (atoms.at(i).Label.contains("noADP"))style |=ATOM_STYLE_NOADP;
        if ((atoms.at(i).an==0)&&((style&512)!=512)) {

            wehaveh=true;
        }
        V3 ev=V3(0,0,0);
        double *arr = mol->jacobi(atoms.at(i).uc, ev);
        /*
        if ((atoms.at(i).an==-1)&&(atoms.at(i).peakHeight<-10.0))printf("%s %f %f hidden %d  < cutoff %d\n",
                   atoms.at(i).Label.toStdString().c_str(),
                   atoms.at(i).peakHeight,
                   qcutoff,
                   style,
                   atoms.at(i).peakHeight<qcutoff); // */
        for (int j = 0; j < 12; j++){
            stuff.append(atoms.at(i).pos.x);//the center
            stuff.append(atoms.at(i).pos.y);
            stuff.append(atoms.at(i).pos.z);
            stuff.append(static_cast<float>(atoms.at(i).an));

            stuff.append(eckn[3*j]);// an icosahedron corner
            stuff.append(eckn[3*j+1]);
            stuff.append(eckn[3*j+2]);
            stuff.append(static_cast<float>(j));

            stuff.append(static_cast<float>(arr[0]));
            stuff.append(static_cast<float>(arr[1]));
            stuff.append(static_cast<float>(arr[2]));
            stuff.append(static_cast<float>(arr[3]));

            stuff.append(ev.x);
            stuff.append(ev.y);
            stuff.append(ev.z);
            stuff.append(static_cast<float>(style));

            stuff.append(static_cast<float>((atoms.at(i).an==-42)?atoms.at(i).sof:atoms.at(i).peakHeight));
            stuff.append(static_cast<float>(qMax(0, atoms.at(i).symmGroup)));
            stuff.append(static_cast<float>(atoms.at(i).part));
            stuff.append(static_cast<float>(atoms.at(i).molindex));

            stuff.append(static_cast<float>(i));
            stuff.append(static_cast<float>(atoms.at(i).auidx));
            stuff.append(static_cast<float>(atoms.at(i).resiNr));
            stuff.append(0.0f);

        }
    }
    hideHydrogen->setVisible(wehaveh);

    QVector<float> bigstuff;
    //QVector<GLuint> indexes;
    //uint una = static_cast<uint>(na);
    for (int i = 0; i < na; i++){
        for (int j = 0; j < 60; j++){
            for (int k = 0; k < 24; k++){
                bigstuff.append(stuff.at(24*(triangle[j]+12*i)+k));
            }
        }
    }
    atom_vao.create();
    atom_vao.bind();
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    atom_vertexBuffer = QOpenGLBuffer(QOpenGLBuffer::VertexBuffer);
    atom_vertexBuffer.create();
    atom_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!atom_vertexBuffer.bind()) {printf("I was unable to bind atom_vertexBuffer.\n");}


    atom_vertexBuffer.allocate(bigstuff.data(), bigstuff.size() * sizeof(float));

    atom_array_size = bigstuff.size() / 24; // one chunk has 24 floats
    //indexes.size();


    if (!atom_program) atom_program = new QOpenGLShaderProgram(this);
    else atom_program->removeAllShaders();



    bool result = atom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/atom.vert");
    //bool result = atom_program->addShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/atom.vert");


    if ( !result ) qWarning() << atom_program->log();
    result = atom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/atom.geom");
    //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/atom.geom");
    //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/amd_atom.geom");


    if ( !result ) qWarning() << atom_program->log();
    result = atom_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/atom.frag");
    //result = atom_program->addShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/atom.frag");

    if ( !result ) qWarning() << atom_program->log();
    result = atom_program->link();
    if ( !result ) qWarning() << atom_program->log();
    \
        if ( !atom_program->bind() )    {
        qWarning() << "Could not bind shader program to context";
        return;
    }

    programInfo(atom_program,"atom");
    GLfloat acv[110*3];
    for (int i=0;i<109;i++) {
        acv[3*i]=mol->atomColor(i,false).redF();
        acv[3*i+1]=mol->atomColor(i,false).greenF();
        acv[3*i+2]=mol->atomColor(i,false).blueF();
    }
    GLfloat arv[110];
    for (int i=0;i<109;i++) {
        arv[i]=static_cast<GLfloat>(mol->arad[i]);
    }

    atom_program->setUniformValueArray("atomColors",acv, 109,3);
    atom_program->setUniformValueArray("arad",arv, 109,1);
    /*
    printf("'a236 b239 c235 d232 e237 f238 g233 h234 i240 j227 k0 l226 m3 n224 o229 p1 q2 r231 s230 t228 u225 v242 w[] x241 y114 z4'\n");
    printf("a%d b%d c%d d%d e%d f%d g%d h%d i%d j%d k%d l%d m%d n%d o%d p%d q%d r%d s%d t%d u%d v%d w[] x%d y%d z%d\n",

           atom_program->uniformLocation("origin"),              //a1
           atom_program->uniformLocation("pmv"),                 //b2
           atom_program->uniformLocation("normalMatrix"),        //c3
           atom_program->uniformLocation("lightPos"),            //d4
           atom_program->uniformLocation("pmax"),                //e5
           atom_program->uniformLocation("pmin"),                //f6
           atom_program->uniformLocation("lmax"),                //g7
           atom_program->uniformLocation("lmin"),                //h8
           atom_program->uniformLocation("qcutoff"),             //i9
           atom_program->uniformLocation("fiftypct"),            //j10
           atom_program->uniformLocation("LOD"),                 //k11
           atom_program->uniformLocation("eye"),                 //l12
           atom_program->uniformLocation("adp"),                 //m13
           atom_program->uniformLocation("bede"),                //n14
           atom_program->uniformLocation("hidePM1"),             //o15
           atom_program->uniformLocation("QPeakLegend"),         //p16
           atom_program->uniformLocation("Qolor"),               //q17
           atom_program->uniformLocation("highlightParts"),      //r18
           atom_program->uniformLocation("highlightEquivalents"),//s19
           atom_program->uniformLocation("hideHydrogen"),        //t20
           atom_program->uniformLocation("bondradius"),          //u21
           atom_program->uniformLocation("tubes"),               //v22
           //atom_program->uniformLocation("ballStick"),           //w23
           atom_program->uniformLocation("qpeakrad"),             //x24
           atom_program->uniformLocation("atomColors"),           //y
           atom_program->uniformLocation("arad")                  //z

           );// */

   // atom_program->setUniformValue("pmv", pmv);
    /*for (int i = 0; i < atoms.size(); i++){
        qDebug()<<pmv.map(QVector3D(atoms.at(i).pos.x-sumse.x,atoms.at(i).pos.y-sumse.y,atoms.at(i).pos.z-sumse.z));
    }// */


    /*printf("L0: %d L1: %d L2: %d L3: %d L4: %d L5: %d \n",
           atom_program->attributeLocation("L0"),
           atom_program->attributeLocation("L1"),
           atom_program->attributeLocation("L2"),
           atom_program->attributeLocation("L3"),
           atom_program->attributeLocation("L4"),
           atom_program->attributeLocation("L5"));// */
    atom_program->enableAttributeArray(0);
    atom_program->enableAttributeArray(1);
    atom_program->enableAttributeArray(2);
    atom_program->enableAttributeArray(3);
    atom_program->enableAttributeArray(4);
    atom_program->enableAttributeArray(5);

    int stride = 6 * 4 * sizeof(float);
    atom_program->setAttributeBuffer(0, GL_FLOAT, 0, 4, stride);
    atom_program->setAttributeBuffer(1, GL_FLOAT, sizeof(float)*4  , 4, stride);
    atom_program->setAttributeBuffer(2, GL_FLOAT, sizeof(float)*8  , 4, stride);
    atom_program->setAttributeBuffer(3, GL_FLOAT, sizeof(float)*12 , 4, stride);
    atom_program->setAttributeBuffer(4, GL_FLOAT, sizeof(float)*16 , 4, stride);
    atom_program->setAttributeBuffer(5, GL_FLOAT, sizeof(float)*20 , 4, stride);// 20*sizeof(float)

    atom_vao.release();
    mk2++;
    atom_program->release();
    //atom_indexBuffer.release();
    atom_vertexBuffer.release();
    atom_vertexBuffer.destroy();
   // atom_indexBuffer.destroy();
    while((err = glGetError()) != GL_NO_ERROR){
        printf("after1 makeAtoms %s error %X\n", parent()->objectName().toStdString().c_str(),err);

    }
    //double va = viewAngle;
    //setViewAngle(29.0);

    pushRotStat();
    popRotStat();
    makeSelectedAtoms();
    updateScreenAtoms();
    makeLegendAtoms();
    //setViewAngle(va);
    update();
    while((err = glGetError()) != GL_NO_ERROR){
        printf("after2 makeAtoms %s error %X\n", parent()->objectName().toStdString().c_str(),err);
    }
}

void PlayOpenGL::pushRotStat(){
    double ll=L;
    double dim=mol->dimension(mol->showatoms);
    L = (isOrtho)?2/dim: viewAngle * 100.0 / (dim *29.0);
    Matrix m = Matrix(
                mvmat(0,0),
                mvmat(0,1),
                mvmat(0,2),
                mvmat(1,0),
                mvmat(1,1),
                mvmat(1,2),
                mvmat(2,0),
                mvmat(2,1),
                mvmat(2,2));
    double det = determinant(m);


    ll=L/pow(det,1.0/3.0);
    if ((!inRenameMode)&&(!doNotScale)) mvmat.scale(ll, ll, ll);
    pmv = pmat*mvmat;
    nm = mvmat.normalMatrix();
    eye = mvmat.inverted()*QVector4D(0,0,0,1);
}

void PlayOpenGL::popRotStat(){
    return;
}

void PlayOpenGL::makeDraftBonds(const CEnvironment &atoms, const Connection &bonds, const Connection &lbonds, Connection &hbonds){
    makeCurrent();
    draft_bond_vao.release();
    if (draft_bond_vao.isCreated()){
      draft_bond_vao.bind();
      draft_bond_vertexBuffer.destroy();
      draft_bond_vao.destroy();
    }
    if (bonds.isEmpty()&&lbonds.isEmpty()&&hbonds.isEmpty()) return;

    if (!draft_bond_vao.create()) qDebug()<< "can't create bond vertex array object (vao)!";

    draft_bond_vao.bind();
    draft_bond_vertexBuffer = QOpenGLBuffer(QOpenGLBuffer::VertexBuffer);
    draft_bond_vertexBuffer.create();
    bond_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!draft_bond_vertexBuffer.bind()) {printf("m_vertexBufferObject2 I was unable to bind!\n");}


    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
    }


    QVector<float> pos;
    mol->h_bonds3(bonds,atoms);
    hbonds=mol->theH_Bonds;
    int vertcnt=0;

    for (int j=0; j<bonds.size(); j++){

      pos.append(bonds.at(j).ato1->pos.x);
      pos.append(bonds.at(j).ato1->pos.y);
      pos.append(bonds.at(j).ato1->pos.z);
      vertcnt++;
      pos.append(bonds.at(j).ato2->pos.x);
      pos.append(bonds.at(j).ato2->pos.y);
      pos.append(bonds.at(j).ato2->pos.z);
      vertcnt++;
    }

    for (int j=0; j<lbonds.size(); j++){
      pos.append(lbonds.at(j).ato1->pos.x);
      pos.append(lbonds.at(j).ato1->pos.y);
      pos.append(lbonds.at(j).ato1->pos.z);
      vertcnt++;
      pos.append(lbonds.at(j).ato2->pos.x);
      pos.append(lbonds.at(j).ato2->pos.y);
      pos.append(lbonds.at(j).ato2->pos.z);
      vertcnt++;
    }
    for (int j=0; j<hbonds.size(); j++){
      pos.append(hbonds.at(j).ato1->pos.x);
      pos.append(hbonds.at(j).ato1->pos.y);
      pos.append(hbonds.at(j).ato1->pos.z);
      vertcnt++;
      pos.append(hbonds.at(j).ato2->pos.x);
      pos.append(hbonds.at(j).ato2->pos.y);
      pos.append(hbonds.at(j).ato2->pos.z);
      vertcnt++;
    }


    draft_bond_vertexBuffer.allocate(vertcnt * 3 * sizeof(float));
    draft_bond_vertexBuffer.write(0, (void*)pos.data(), vertcnt * 3 * sizeof(float));
    ndrbondx2=vertcnt;
    if (!bond_program) draft_bond_program = new QOpenGLShaderProgram(this);
    else draft_bond_program->removeAllShaders();


    bool result2 = draft_bond_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/draft.vert");
    if ( !result2 ) qWarning() << draft_bond_program->log();
    result2 = draft_bond_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/draft.frag");
    if ( !result2 ) qWarning() << draft_bond_program->log();
    result2 = draft_bond_program->link();
    //shaderInfo = true;
    programInfo(draft_bond_program,"draft_bond");
    //shaderInfo = false;
    if ( !result2 ) qWarning() << draft_bond_program->log();
    if ( !draft_bond_program->bind() )    {
      qWarning() << "2Could not bind shader program to context";
      return;
    }
    int stride = 0;//7*sizeof(float);
    draft_bond_program->enableAttributeArray( "VertexPosition" );
    draft_bond_program->setAttributeBuffer( "VertexPosition", GL_FLOAT, 0, 3, stride );
    draft_bond_vao.release();
    mk8++;
    draft_bond_program->release();
    //bond_indexBuffer.release();
    draft_bond_vertexBuffer.release();
    draft_bond_vertexBuffer.destroy();
    //bond_indexBuffer.destroy();
}

void PlayOpenGL::makeBonds(const CEnvironment &atoms, const Connection &bonds, const Connection &lbonds, Connection &hbonds){//, float* atomColors
    //return;
    if (bondsFirst) makeDraftBonds(atoms,bonds,lbonds,hbonds);
    makeCurrent();

    bond_vao.release();
    if (bond_vao.isCreated()) {
      bond_vao.bind();
      //bond_indexBuffer.destroy();
      bond_vertexBuffer.destroy();
      //bond_vao.release();

      bond_program->removeAllShaders();
      bond_program->release();
      bond_vao.destroy();

      glFlush();
    }
    if (bonds.isEmpty()&&lbonds.isEmpty()&&hbonds.isEmpty()) return;

    if (!bond_vao.create()) qDebug()<< "can't create bond vertex array object (vao)!";

    bond_vao.bind();
    bond_vertexBuffer = QOpenGLBuffer(QOpenGLBuffer::VertexBuffer);
    bond_vertexBuffer.create();
    bond_vertexBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    if (!bond_vertexBuffer.bind()) {printf("m_vertexBufferObject2 I was unable to bind!\n");}


    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
    }
    int zz=0;
    sumse=V3(0,0,0);
    for (int i=0;i<atoms.size();i++){
      if ((!atoms.at(i).hidden)) {//(mol->showatoms.at(i).an>-1)&&
            sumse+=atoms[i].pos;
            zz++;
      }
    }
    if (zz) sumse*=1.0/zz;
    QVector<float> pos;
    //QVector<float> col;
    QVector<float> an;
    QVector<float> styles;
    QVector<float> peakHeights;
    QVector<float> symmGroups;
    mol->h_bonds3(bonds,atoms);
    hbonds=mol->theH_Bonds;
    int vertcnt=0;

    for (int j=0; j<bonds.size(); j++){
      pos.append(bonds.at(j).ato1->pos.x);
      pos.append(bonds.at(j).ato1->pos.y);
      pos.append(bonds.at(j).ato1->pos.z);
      bool isd=((bonds.at(j).ato1->an==0)&&(bonds.at(j).ato1->Label[0]=='D'));
      an.append(static_cast<float>(bonds.at(j).ato1->an));
      //styles.append(7);
      peakHeights.append((bonds.at(j).ato1->an<0) ? bonds.at(j).ato1->peakHeight : bonds.at(j).ato1->part);
      symmGroups.append(qMax(0, bonds.at(j).ato1->symmGroup));
      if (bonds.at(j).ato1->an<0) styles.append((bonds.at(j).ato1->hidden)?512:0);
      else {
            styles.append(static_cast<float>(mol->AtomStyle[bonds.at(j).ato1->an]
                                             |((bonds.at(j).ato1->hidden)?512:0)
                                             |((isd)?1024:0)
                                             |(((bonds.at(j).ato1->symmGroup)&&(bonds.at(j).ato1->part<0))?2048:0)));
      }
      vertcnt++;
      pos.append(bonds.at(j).ato2->pos.x);
      pos.append(bonds.at(j).ato2->pos.y);
      pos.append(bonds.at(j).ato2->pos.z);
      isd=((bonds.at(j).ato2->an==0)&&(bonds.at(j).ato2->Label[0]=='D'));
      an.append(static_cast<float>(bonds.at(j).ato2->an));
      //styles.append(7);
      peakHeights.append((bonds.at(j).ato2->an<0)?bonds.at(j).ato2->peakHeight:bonds.at(j).ato2->part);
      symmGroups.append(qMax(0, bonds.at(j).ato2->symmGroup));
      if (bonds.at(j).ato2->an<0) styles.append((bonds.at(j).ato2->hidden)?512:0);
      else {
            styles.append(static_cast<float>(mol->AtomStyle[bonds.at(j).ato2->an]
                                             |((bonds.at(j).ato2->hidden)?512:0)
                                             |((isd)?1024:0)
                                             |(((bonds.at(j).ato2->symmGroup)&&(bonds.at(j).ato2->part<0))?2048:0)));
      }
      //if ((bonds.at(j).ato1->an==0)||(bonds.at(j).ato2->an==0)) printf("%s === %s\n"
      //             ,bonds.at(j).ato1->Label.toStdString().c_str()
      //             ,bonds.at(j).ato2->Label.toStdString().c_str());
      vertcnt++;
    }

    for (int j=0; j<lbonds.size(); j++){
      pos.append(lbonds.at(j).ato1->pos.x);
      pos.append(lbonds.at(j).ato1->pos.y);
      pos.append(lbonds.at(j).ato1->pos.z);
      bool isd=((lbonds.at(j).ato1->an==0)&&(lbonds.at(j).ato1->Label[0]=='D'));
      an.append(static_cast<float>(lbonds.at(j).ato1->an));
      //styles.append(7);
      peakHeights.append((lbonds.at(j).ato1->an<0)?lbonds.at(j).ato1->peakHeight:lbonds.at(j).ato1->part);
      symmGroups.append(qMax(0, lbonds.at(j).ato1->symmGroup));
      if (lbonds.at(j).ato1->an<0) styles.append((lbonds.at(j).ato1->hidden)?512:0);
      else {
            styles.append(static_cast<float>(mol->AtomStyle[lbonds.at(j).ato1->an]
                                             |((lbonds.at(j).ato1->hidden)?512:0)
                                             |((isd)?1024:0)
                                             |(((lbonds.at(j).ato1->symmGroup)&&(lbonds.at(j).ato1->part<0))?2048:0)));
      }
      vertcnt++;
      pos.append(lbonds.at(j).ato2->pos.x);
      pos.append(lbonds.at(j).ato2->pos.y);
      pos.append(lbonds.at(j).ato2->pos.z);
      isd=((lbonds.at(j).ato2->an==0)&&(lbonds.at(j).ato2->Label[0]=='D'));
      an.append(static_cast<float>(lbonds.at(j).ato2->an));
      //styles.append(7);
      peakHeights.append((lbonds.at(j).ato2->an<0)?lbonds.at(j).ato2->peakHeight:lbonds.at(j).ato2->part);
      symmGroups.append(qMax(0, lbonds.at(j).ato2->symmGroup));
      if (lbonds.at(j).ato2->an<0) styles.append((lbonds.at(j).ato2->hidden)?512:0);
      else {
            styles.append(static_cast<float>(mol->AtomStyle[lbonds.at(j).ato2->an]
                                             |((lbonds.at(j).ato2->hidden)?512:0)
                                             |((isd)?1024:0)
                                             |(((lbonds.at(j).ato2->symmGroup)&&(lbonds.at(j).ato2->part<0))?2048:0)));
      }
      vertcnt++;
    }
    for (int j=0; j<hbonds.size(); j++){
/*
      printf(">%d<Hbond %s = %s %f %d %d  %d %d %d\n",j
             ,hbonds.at(j).ato1->Label.toStdString().c_str()
             ,hbonds.at(j).ato2->Label.toStdString().c_str()
             ,hbonds.at(j).length
             ,hbonds.at(j).ato1->an
             ,hbonds.at(j).ato2->an
             ,hbonds.at(j).a1
             ,hbonds.at(j).a2
             ,atoms.size()
             );// */

      pos.append(hbonds.at(j).ato1->pos.x);
      pos.append(hbonds.at(j).ato1->pos.y);
      pos.append(hbonds.at(j).ato1->pos.z);
      bool isd=((hbonds.at(j).ato1->an==0)&&(hbonds.at(j).ato1->Label[0]=='D'));
      an.append(static_cast<float>(hbonds.at(j).ato1->an));
      //styles.append(7);
      peakHeights.append((hbonds.at(j).ato1->an<0)?hbonds.at(j).ato1->peakHeight:hbonds.at(j).ato1->part);
      symmGroups.append(qMax(0, hbonds.at(j).ato1->symmGroup));
      if (hbonds.at(j).ato1->an<0) styles.append((hbonds.at(j).ato1->hidden)?512:0);
      else {
            styles.append(static_cast<float>(mol->AtomStyle[hbonds.at(j).ato1->an]
                                             |((hbonds.at(j).ato1->hidden)?512:0)
                                             |((isd)?1024:0)
                                             |(((hbonds.at(j).ato1->symmGroup)&&(hbonds.at(j).ato1->part<0))?2048:0)));
      }
      vertcnt++;
      pos.append(hbonds.at(j).ato2->pos.x);
      pos.append(hbonds.at(j).ato2->pos.y);
      pos.append(hbonds.at(j).ato2->pos.z);
      isd=((hbonds.at(j).ato2->an==0)&&(hbonds.at(j).ato2->Label[0]=='D'));
      an.append(static_cast<float>(hbonds.at(j).ato2->an));
      //styles.append(7);
      peakHeights.append((hbonds.at(j).ato2->an<0)?hbonds.at(j).ato2->peakHeight:hbonds.at(j).ato2->part);
      symmGroups.append(qMax(0, hbonds.at(j).ato2->symmGroup));
      if (hbonds.at(j).ato2->an<0) styles.append((hbonds.at(j).ato2->hidden)?512:0);
      else {
            styles.append(static_cast<float>(mol->AtomStyle[hbonds.at(j).ato2->an]
                                             |((hbonds.at(j).ato2->hidden)?512:0)
                                             |((isd)?1024:0)
                                             |(((hbonds.at(j).ato2->symmGroup)&&(hbonds.at(j).ato2->part<0))?2048:0)));
      }
      vertcnt++;
    }
    // */

    //printf("MAKE BONDS: hbonds %d qbonds %d bonds %d  vercnt %d\n", hbonds.size(), lbonds.size(), bonds.size(), vertcnt);

    bond_vertexBuffer.allocate(vertcnt * 7 * sizeof(float));
    bond_vertexBuffer.write(0, (void*)pos.data(), vertcnt * 3 * sizeof(float));
    bond_vertexBuffer.write(vertcnt * 3 * sizeof(float), (void*)an.data(), vertcnt * sizeof(float));
    bond_vertexBuffer.write(vertcnt * 4 * sizeof(float), (void*)styles.data(), vertcnt * sizeof(float));
    bond_vertexBuffer.write(vertcnt * 5 * sizeof(float), (void*)peakHeights.data(), vertcnt * sizeof(float));
    bond_vertexBuffer.write(vertcnt * 6 * sizeof(float), (void*)symmGroups.data(), vertcnt * sizeof(float));
    nbondx2=vertcnt;
    //nbondx2 = nbonds.size();
    /*printf("nbondx2 = %d\n", nbondx2);
    for (int mm=0; mm<nbondx2; ){
        printf("%s %s\n"
               ,atoms.at(nbonds.at(mm)).Label.toStdString().c_str()
               ,atoms.at(nbonds.at(mm+1)).Label.toStdString().c_str());
        mm+=2;
    }*/
    /*
    // create a new buffer for the indexes
    bond_indexBuffer = QOpenGLBuffer(QOpenGLBuffer::IndexBuffer); // Mind: use 'IndexBuffer' here
    bond_indexBuffer.create();
    bond_indexBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
    if (!bond_indexBuffer.bind()) {printf("bond_indexBuffer I was unable to bind!\n");}
    bond_indexBuffer.allocate(nbonds.data(), nbonds.size() * sizeof(GLushort) );*/
    if (!bond_program) bond_program = new QOpenGLShaderProgram(this);
    else bond_program->removeAllShaders();


    bool result2 = bond_program->addCacheableShaderFromSourceFile( QOpenGLShader::Vertex, ":/shaders/bond.vert");
    if ( !result2 ) qWarning() << bond_program->log();
    result2 = bond_program->addCacheableShaderFromSourceFile( QOpenGLShader::Geometry, ":/shaders/bond.geom");
    if ( !result2 ) qWarning() << bond_program->log();
    result2 = bond_program->addCacheableShaderFromSourceFile( QOpenGLShader::Fragment, ":/shaders/bond.frag");
    if ( !result2 ) qWarning() << bond_program->log();
    result2 = bond_program->link();

    programInfo(bond_program,"bond");
    if ( !result2 ) qWarning() << bond_program->log();
    if ( !bond_program->bind() )    {
      qWarning() << "2Could not bind shader program to context";
      return;
    }
    GLfloat acv[109*3];
    for (int i=0;i<109;i++) {
      acv[3*i]=mol->atomColor(i,false).redF();
      acv[3*i+1]=mol->atomColor(i,false).greenF();
      acv[3*i+2]=mol->atomColor(i,false).blueF();
    }
    bond_program->setUniformValueArray("atomColors",acv, 109,3);

    //bond_program->setUniformValue("pmv", pmv);
    int stride = 0;//7*sizeof(float);
    bond_program->enableAttributeArray( "VertexPosition" );
    bond_program->enableAttributeArray( "atomicnumbers" );
    bond_program->enableAttributeArray( "Styles" );
    bond_program->enableAttributeArray( "peakHeights" );
    bond_program->enableAttributeArray( "symmGroup" );
    bond_program->setAttributeBuffer( "VertexPosition", GL_FLOAT, 0, 3, stride );
    bond_program->setAttributeBuffer( "atomicnumbers", GL_FLOAT, sizeof(float)*vertcnt*3, 1, stride );
    bond_program->setAttributeBuffer( "Styles", GL_FLOAT, sizeof(float)*vertcnt*4, 1, stride );
    bond_program->setAttributeBuffer( "peakHeights", GL_FLOAT, sizeof(float)*vertcnt*5, 1, stride );
    bond_program->setAttributeBuffer( "symmGroup", GL_FLOAT, sizeof(float)*vertcnt*6, 1, stride );
    /*printf("attr %d %d %d %d %d\n",bond_program->attributeLocation("VertexPosition"),
           bond_program->attributeLocation("atomicnumbers"),
           bond_program->attributeLocation("Styles"),
           bond_program->attributeLocation("peakHeights"),
           bond_program->attributeLocation("symmGroup"));// */

    bond_vao.release();
    mk6++;
    bond_program->release();
    //bond_indexBuffer.release();
    bond_vertexBuffer.release();
    bond_vertexBuffer.destroy();
    //bond_indexBuffer.destroy();

}

QString PlayOpenGL::glTypeInt2String(int ityp){
    QString s;
    /*

#define GL_FLOAT_VEC2                     0x8B50
#define GL_FLOAT_VEC3                     0x8B51
#define GL_FLOAT_VEC4                     0x8B52
#define GL_INT_VEC2                       0x8B53
#define GL_INT_VEC3                       0x8B54
#define GL_INT_VEC4                       0x8B55
#define GL_BOOL                           0x8B56
#define GL_BOOL_VEC2                      0x8B57
#define GL_BOOL_VEC3                      0x8B58
#define GL_BOOL_VEC4                      0x8B59
#define GL_FLOAT_MAT2                     0x8B5A
#define GL_FLOAT_MAT3                     0x8B5B
#define GL_FLOAT_MAT4                     0x8B5C

#define GL_BYTE           0x1400
#define GL_UNSIGNED_BYTE  0x1401
#define GL_SHORT          0x1402
#define GL_UNSIGNED_SHORT 0x1403
#define GL_INT            0x1404
#define GL_UNSIGNED_INT   0x1405
#define GL_FLOAT          0x1406
#define GL_2_BYTES        0x1407
#define GL_3_BYTES        0x1408
#define GL_4_BYTES        0x1409
#define GL_DOUBLE         0x140A

*/
    switch (ityp){
    case 0x8B50: s = "GL_FLOAT_VEC2"; break;
    case 0x8B51: s = "GL_FLOAT_VEC3"; break;
    case 0x8B52: s = "GL_FLOAT_VEC4"; break;
    case 0x8B53: s = "GL_INT_VEC2"; break;
    case 0x8B54: s = "GL_INT_VEC3"; break;
    case 0x8B55: s = "GL_INT_VEC4"; break;
    case 0x8B56: s = "GL_BOOL"; break;
    case 0x8B57: s = "GL_BOOL_VEC2"; break;
    case 0x8B58: s = "GL_BOOL_VEC3"; break;
    case 0x8B59: s = "GL_BOOL_VEC4"; break;
    case 0x8B5A: s = "GL_FLOAT_MAT2"; break;
    case 0x8B5B: s = "GL_FLOAT_MAT3"; break;
    case 0x8B5C: s = "GL_FLOAT_MAT4"; break;
    case 0x1400: s = "GL_BYTE"; break;
    case 0x1401: s = "GL_UNSIGNED_BYTE"; break;
    case 0x1402: s = "GL_SHORT"; break;
    case 0x1403: s = "GL_UNSIGNED_SHORT"; break;
    case 0x1404: s = "GL_INT"; break;
    case 0x1405: s = "GL_UNSIGNED_INT"; break;
    case 0x1406: s = "GL_FLOAT"; break;
    case 0x1407: s = "GL_2_BYTES"; break;
    case 0x1408: s = "GL_3_BYTES"; break;
    case 0x1409: s = "GL_4_BYTES"; break;
    case 0x140A: s = "GL_DOUBLE"; break;
    //case : s = ""; break;
    //case : s = ""; break;
    default: s = "no clue?";
    }
    return s;
}

void PlayOpenGL::initializeGL(){
    if (parent()){
    isMainWindow = (!parent()->objectName().startsWith("Edit Atom Parameters Dialog"));

    }
    if (verbosity >2 ) printf("\n\nis main window = %d. initializeGL. My parent is: %s (objectName)\n", isMainWindow, parent()->objectName().toStdString().c_str());
    makeCurrent();
    initializeOpenGLFunctions();
    // During init, enable debug output
    //glEnable              ( GL_DEBUG_OUTPUT );
    //glDebugMessageCallback( MessageCallback, 0 );

    const QOpenGLContext *context = this->context();
    QSurface *surf = this->context()->surface();
    if (verbosity >2 ) printf("Widget OpenGl: %d.%d\n",format().majorVersion(),format().minorVersion());
    if (verbosity >2 ) printf("Context valid: %d\n",context->isValid());
    if (verbosity >2 ) printf("Really used OpenGl: %d.%d\n",context->format().majorVersion(),context->format().minorVersion());
    if (verbosity >2 ) printf("Renderer: %s\n", glGetString(GL_RENDERER));
    if (verbosity >2 ) printf("profile = %d = %d (0 = no profile, 1 = core profile, 2 = compatibility profile)\n",format().profile(), context->format().profile());
    if (verbosity >2 ) printf("swapBehavior %d %d\n",format().swapBehavior(), context->format().swapBehavior());
    if (verbosity >2 ) printf("depthBufferSize %d %d\n",format().depthBufferSize(), context->format().depthBufferSize());
    QString s = QString::fromUtf8((char*)glGetString(GL_RENDERER));
    GPUInfoString = QString("Requested OpenGL: %1.%2<br>Context is %3valid.<br> Really used OpenGL: %4.%5<br>Renderer: %6")
                        .arg(format().majorVersion())
                        .arg(format().minorVersion())
                        .arg((context->isValid())?"":"in")
                        .arg(context->format().majorVersion())
                        .arg(context->format().minorVersion())
                        .arg(s)

        ;

    GPUInfoString2 = QString("Req.OpenGL: %1.%2;Context is %3valid.;Really used OpenGL: %4.%5;Renderer: %6")
                        .arg(format().majorVersion())
                        .arg(format().minorVersion())
                        .arg((context->isValid())?"":"in")
                        .arg(context->format().majorVersion())
                        .arg(context->format().minorVersion())
                        .arg(s);
    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      if (verbosity >2 ) printf("test%d %s error %X\n", __LINE__, parent()->objectName().toStdString().c_str(),err);

    }
    GLint nrAttributes;
    glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &nrAttributes);
    if (verbosity >2 ) printf("GL_MAX_VERTEX_ATTRIBS = %d\n",nrAttributes);
    //GL_MAX_VERTEX_ATTRIBS
    glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS, &nrAttributes);
    if (verbosity >2 ) printf("GL_MAX_VERTEX_UNIFORM_COMPONENTS = %d\n",nrAttributes);
    glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS,  &nrAttributes);
    if (verbosity >2 ) printf("GL_MAX_FRAGMENT_UNIFORM_COMPONENTS = %d\n",nrAttributes);
    glGetIntegerv(GL_MAX_GEOMETRY_UNIFORM_COMPONENTS,  &nrAttributes);
    if (verbosity >2 ) printf("GL_MAX_GEOMETRY_UNIFORM_COMPONENTS = %d\n",nrAttributes);
    glGetIntegerv(GL_MAX_GEOMETRY_INPUT_COMPONENTS,  &nrAttributes);
    if (verbosity >2 ) printf("GL_MAX_GEOMETRY_INPUT_COMPONENTS = %d\n",nrAttributes);
    glGetIntegerv(GL_MAX_GEOMETRY_OUTPUT_COMPONENTS,  &nrAttributes);
    if (verbosity >2 ) printf("GL_MAX_GEOMETRY_OUTPUT_COMPONENTS = %d\n",nrAttributes);
    //glGetIntegerv(GL_MAX_VARYING_FLOATS, &nrAttributes);
    //if (verbosity >2 ) printf("GL_MAX_VARYING_FLOATS=%d\n",nrAttributes);
    //glGetIntegerv(GL_MAX_VARYING_COMPONENTS, &nrAttributes);
    //if (verbosity >2 ) printf("GL_MAX_VARYING_COMPONENTS=%d\n",nrAttributes);
    err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      printf("before gl_blend %s error %X\n", parent()->objectName().toStdString().c_str(),err);
    }
    //glEnable(GL_PRIMITIVE_RESTART_FIXED_INDEX);
    glEnable(GL_BLEND);
    //glDisable(GL_BLEND);
    glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
    //glClearColor(0.4f,0.4f,0.4f,1.0f);
    //glClearColor(0.f,0.f,0.f,1.0f);
    glClearColor(1.f,1.f,1.f,1.0f);
    glEnable(GL_DEPTH_TEST );
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
    //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    //GLenum err = GL_NO_ERROR;
    err=GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      printf("after GL_PERSPECTIVE_CORRECTION_HINT %s error %X\n", parent()->objectName().toStdString().c_str(),err);
    }
#ifndef Q_OS_MAC
    setMouseTracking(true);
#endif
    float ratio = width() / static_cast<float>(height());
    pmat.setToIdentity();

    if (viewAngle<0.1) {

        L=(mol->showatoms.isEmpty())?2.0:2.0/mol->dimension(mol->showatoms);

        mvmat.setToIdentity();
        mvmat.lookAt(QVector3D(0.0, 200, 0) , QVector3D(  0.0, 0.0, 0.0), QVector3D(0.0, -100, 400 ));
        mvmat.scale(L,L,L);
        pmat.ortho(-ratio, ratio, -1.0f, 1.0f, -5.0f, 8000.0f);
        previousViewAngle = qMax(viewAngle, 29.0);
        viewAngle=0.01;
        isOrtho=true;
        if (toggleOrthoView!=nullptr)toggleOrthoView->setChecked(true);
        if (verbosity >2 ) printf("ORTHO\n");
    }else{
        if (verbosity >2 ) printf("PERSPECTIVE %f\n", viewAngle);
        pmat.perspective(viewAngle, ratio, -5.0f, 8000.0f );
        isOrtho=false;
        if (toggleOrthoView!=nullptr) toggleOrthoView->setChecked(false);
        L=100.0;

        mvmat.setToIdentity();
        mvmat.lookAt(QVector3D(0.0, 200, 0) , QVector3D(  0.0, 0.0, 0.0), QVector3D(0.0, -100, 400 ));
        mvmat.scale(L,L,L);
    }
    pmv = pmat*mvmat;
    nm = mvmat.normalMatrix();
    eye = mvmat.inverted()*QVector4D(0,0,0,1);
    makeBackGround();
    bool cu= this->context()->makeCurrent(surf);

    if (verbosity >2 ) printf("Context valid: %d v %d widget %d window %d  current %d\n",context->isValid(), this->context()->isValid(), context->isWidgetType(), context->isWindowType(),cu);
    if (verbosity >2 ) printf("\nGL initialized   \n\n");
    emit initialized();
    update();/*
    printf("ballaballa\n");
    QVector<GLfloat> vrt;
    ball(vrt, 2, V3(0,0,0), 5, 0, 0, 2048);
    ball(vrt, 3, V3(0,0,0), 5, 0, 0, 2048);
    ball(vrt, 4, V3(0,0,0), 5, 0, 0, 2048);
    ball(vrt, 5, V3(0,0,0), 5, 0, 0, 2048);
    ball(vrt, 6, V3(0,0,0), 5, 0, 0, 2048);
    ball(vrt, 7, V3(0,0,0), 5, 0, 0, 2048);
    ball(vrt, 8, V3(0,0,0), 5, 0, 0, 2048);
    vrt.clear();
    //cube(vrt, V3(0,0,0),5,0,0,2048);
    // */
}

void PlayOpenGL::resizeGL(int width, int height){

  retinafktr=devicePixelRatio();
  glViewport(0, 0, width*retinafktr, qMax(height*retinafktr, 1.0));
  pmat.setToIdentity();
  float ratio = width / static_cast<float>(height);
  if (isOrtho) pmat.ortho(-ratio, ratio, -1.0f, 1.0f, -5.0f, 8000.0f);
  else pmat.perspective(viewAngle, ratio, 5.0f, 8000.0f );

  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
}

inline void PlayOpenGL::__RotateCSf( float c, float s, float& X, float& Y ) {
  float T = X;
  X = c*X - s*Y;
  Y = s*T + c*Y;
}

inline void PlayOpenGL::__RotateCS( double c, double s, double& X, double& Y ) {
  double T = X;
  X = c*X - s*Y;
  Y = s*T + c*Y;
}

void PlayOpenGL::glRotateMat( QMatrix4x4 &mat, const float dang, const float x , const float y, const float z ) {
  float s = z;
  s = sinf(dang*M_PIf/180);
  const float c = cosf(dang*M_PIf/180);
  if( x!=0.0f ){
    __RotateCSf( c, s, mat(1,0), mat(2,0) );
    __RotateCSf( c, s, mat(1,1), mat(2,1) );
    __RotateCSf( c, s, mat(1,2), mat(2,2) );

  }else if( y!=0.0f ){
    __RotateCSf( c, s, mat(2,0), mat(0,0) );
    __RotateCSf( c, s, mat(2,1), mat(0,1) );
    __RotateCSf( c, s, mat(2,2), mat(0,2) );

  }else{
    __RotateCSf( c, s, mat(0,0), mat(1,0) );
    __RotateCSf( c, s, mat(0,1), mat(1,1) );
    __RotateCSf( c, s, mat(0,2), mat(1,2) );

  }
}

void PlayOpenGL::glRotateMat( Matrix &mat, const float dang, const double x , const double y, const double z ) {
  double s = z;
  s = sin(dang*M_PI/180);
  const double c = cos(dang*M_PI/180);
  if( x!=0.0f ){
    __RotateCS( c, s, mat.m21, mat.m31 );
    __RotateCS( c, s, mat.m22, mat.m32 );
    __RotateCS( c, s, mat.m23, mat.m33 );

  }else if( y!=0.0f ){
    __RotateCS( c, s, mat.m31, mat.m11 );
    __RotateCS( c, s, mat.m32, mat.m12 );
    __RotateCS( c, s, mat.m33, mat.m13 );

  }else{
    __RotateCS( c, s, mat.m11, mat.m21 );
    __RotateCS( c, s, mat.m12, mat.m22 );
    __RotateCS( c, s, mat.m13, mat.m23 );

  }
}

void PlayOpenGL::setLabelSize(int s){
myFont.setPointSize(s);
makeLabels();
update();
}

void PlayOpenGL::wheelEvent(QWheelEvent *event){

#if (QT_VERSION>0x050000)
  int numDegrees = (event->angleDelta().y()==0)?event->angleDelta().x():event->angleDelta().y();
#else
  int numDegrees = event->delta();
#endif
  int numSteps = numDegrees / 120;
  if (event->modifiers()==Qt::NoModifier){
    int d = myFont.pointSize();
    d = (d+numSteps>4)?d+numSteps:d;
    d = qMax(d,7);
    myFont.setPointSize(d);
    makeLabels();
    update();
  }
  else
  if (event->modifiers()==Qt::ControlModifier){
    emit diffscroll(numSteps,1);
  }
  else
  if (event->modifiers()==Qt::ShiftModifier){
    emit diffscroll(numSteps,0);
  }
  else
  if (event->modifiers()==(Qt::AltModifier|Qt::ShiftModifier)){
    double va=qMax(viewAngle+0.1*numSteps,0.01);
    if (verbosity >2 ) printf("va %f %f %d event->angleDelta().y() %d\n",va,0.1*numSteps,numSteps,event->angleDelta().x());
    setViewAngle(va);
  }

}

void PlayOpenGL::mouseMoveEvent(QMouseEvent *event){

  double x = event->pos().x();
  double y = event->pos().y();
  double nahda = pickradius, da = 0;
  rectangle = false;
  int nahdai = -1;
  bool changed = false;
  scrx = event->x();
  scry = event->y();
  for (int j = 0; j < mol->showatoms.size(); j++){
    if (mol->showatoms.at(j).hidden) continue;
    if ((mol->showatoms.at(j).an==-1)&&(mol->showatoms.at(j).peakHeight<qcutoff)) continue;
    da=(((screenAtoms.at(j).x()-event->x())*( screenAtoms.at(j).x()-event->x()))+
        ((screenAtoms.at(j).y()-event->y())*( screenAtoms.at(j).y()-event->y())));
    nahdai=(da<nahda)?j:nahdai;
    nahda=qMin(nahda,da);
  }
  if ((!event->buttons())) {//(mouseOverInteraction)&&
      if (inFocus!=nahdai) changed=true;
      inFocus=nahdai;
      if (inFocus==-1) {
        emit qpfoci(-1.0);
        emit message("");
      } else {
        emit message(mol->showatoms.at(inFocus).Label);
        if ((mol->showatoms.at(inFocus).an<0)) emit qpfoci(mol->showatoms.at(inFocus).peakHeight);
        else emit qpfoci(-1.0);
    }

    if (changed) {

      update();
    }
  }
  GLfloat dx = GLfloat( event->pos().x() - moux) / width();
  GLfloat dy = GLfloat( event->pos().y() - mouy) / height();
  if((event->buttons() & Qt::RightButton)) if (event->modifiers()==Qt::ShiftModifier){
      double ddy = ((invertMouseZoom!=nullptr)&&(invertMouseZoom->isChecked()))?(-dy):(dy);
      double ll = 1.0 - ddy;
      legend_sc *= ll;
  }else{
      double ddy = ((invertMouseZoom!=nullptr)&&(invertMouseZoom->isChecked()))?(-dy):(dy);
      double ll=1.0 - ddy;
      L*=ll;
      mvmat.scale(ll,ll,ll);
      pmv = pmat*mvmat;
      nm = mvmat.normalMatrix();
      eye = mvmat.inverted()*QVector4D(0,0,0,1);
      updateScreenAtoms();
  }else if ((event->buttons() & Qt::LeftButton && (event->modifiers()==(Qt::ShiftModifier|Qt::ControlModifier)))) {
      rectangle=true;
      int minxp=qMin(scrx0,scrx);
      int minyp=qMin(scry0,scry);
      int maxxp=qMax(scrx0,scrx);
      int maxyp=qMax(scry0,scry);
      mol->selectedatoms.clear();
      if (verbosity >2 ) printf("show%d screen%d\n",mol->showatoms.size(),screenAtoms.size());
      for (int j=0; j<screenAtoms.size();j++){
        if (mol->showatoms.at(j).hidden) continue;
        if ((screenAtoms.at(j).x()>minxp)&&
            (screenAtoms.at(j).x()<maxxp)&&
            (screenAtoms.at(j).y()>minyp)&&
            (screenAtoms.at(j).y()<maxyp)){
            /*printf("%d < %f < %d  | %d < %f < %d  j = %d %s\n",
                   minxp, screenAtoms.at(j).x(), maxxp,
                   minyp, screenAtoms.at(j).y(), maxyp,
                   j, mol->showatoms[j].Label.toStdString().c_str() );// */
            mol->selectedatoms.append(mol->showatoms[j]);
            mol->selectedatoms.last().style = j;
        }
      }
      updateBondActions(__LINE__);
      makeSelectedAtoms();
      update();
    } else if((event->buttons() & Qt::LeftButton)){
      if (event->modifiers()==Qt::ShiftModifier) {
        float mf = width()/42.0;


        pmat.translate(mf*dx*sqrt(viewAngle),-mf*dy*sqrt(viewAngle));
          //pmat.translate(3*dx*20,-3*dy*20,0);
          pmv = pmat*mvmat;
          nm = mvmat.normalMatrix();
          eye = mvmat.inverted()*QVector4D(0,0,0,1);
          updateScreenAtoms();
      }else if ((manualAx!=nullptr)&&(manualAx->isChecked())&&(event->modifiers()==Qt::AltModifier)){

              makeCurrent();
              GLdouble mm[16];
              glGetDoublev(GL_MODELVIEW_MATRIX,mm);
              double det=
                mm[0]*mm[5]*mm[10] - mm[8]*mm[5]*mm[2]+
                mm[1]*mm[6]*mm[8]  - mm[9]*mm[6]*mm[0]+
                mm[2]*mm[4]*mm[9]  - mm[10]*mm[4]*mm[1];
              Matrix mat=Matrix(mm[0],mm[1],mm[2],mm[4],mm[5],mm[6],mm[8],mm[9],mm[10]);

              mat=mat*pow(det,-1.0/3.0);


              V3 axe=Normalize(V3(axx->value(),axy->value(),axz->value()));
              axe=transponse(mat)*axe;
              double arr[4];
              arr[0]=dx*360.0;
              arr[1]=0.0;
              arr[2]=1.0;
              arr[3]=0.0;
              axe=mol->rotarb(arr)*axe;
              V3 XX = mol->rotarb(arr)*V3(1, 0, 0);
              arr[0]=dy*360.0;
              arr[1]=XX.x;
              arr[2]=XX.y;
              arr[3]=XX.z;
              axe=mol->rotarb(arr)*axe;
              axe=mat*axe;
              axx->setValue(axe.x);
              axy->setValue(axe.y);
              axz->setValue(axe.z);
              splitRotate();
          }else if ((freeRot!=nullptr)&&(freeRot->isChecked())&&(event->modifiers()==(Qt::AltModifier|Qt::ControlModifier))){

              makeCurrent();
              double arr[4];
              /*
              QVector4D v4xy = QVector4D(dx*360, -dy*360, 0, 1);
              v4xy.normalize();
              QMatrix4x4 ii= mvmat;
              ii=ii.inverted().transposed();
              QVector4D v4xys = ii* v4xy;
              v4xys.normalize();
              V3 xx(v4xys.x(), v4xys.y(), v4xys.z());
              xx=Normalize(xx);// */
              fdx+=dx;
              fdy+=dy;
              Matrix worldtoscreen=transponse(Matrix(mvmat));
              double det=1.0/pow(determinant(worldtoscreen), 1.0/3.0);
              worldtoscreen *= det;
              /*
              freeRotMat = freeRotMat * worldtoscreen;
              glRotateMat(freeRotMat,dy*360.0f,1,0,0);
              glRotateMat(freeRotMat,-dx*360.0f,0,1,0);
              freeRotMat =   freeRotMat * transponse(worldtoscreen);
              */
              V3 wox(1,0,0);
              V3 woy(0,1,0);
              screenRotx = worldtoscreen * wox;
              screenRoty = worldtoscreen * woy;
              /*
              QQuaternion p(fdx * 36.0,screenRoty.x,screenRoty.y,screenRoty.z);
              QQuaternion q(fdy * 36.0,screenRotx.x,screenRotx.y,screenRotx.z);
              QQuaternion pq = QQuaternion::fromEulerAngles(0, fdx * 360.0, fdy * 360.0);
              freeRotMat = Matrix(pq.toRotationMatrix());
              //V3 qax=Normalize(V3(pq.x(),pq.y(),pq.z()));

              //freeRotMat = mol->rotarb(pq.scalar(),qax);
              //mol->rotarb(freeRotMat, fdy*360, screenRotx);
              // */
///*
              arr[0] = dx * 360.0;
              arr[1] = screenRoty.x;
              arr[2] = screenRoty.y;
              arr[3] = screenRoty.z;
              freeRotMat = freeRotMat * mol->rotarb(arr);

              arr[0] = dy * 360.0;
              arr[1] = screenRotx.x;
              arr[2] = screenRotx.y;
              arr[3] = screenRotx.z;
              freeRotMat = freeRotMat * mol->rotarb(arr);
//*/
              det=1.0/pow(determinant(freeRotMat), 1.0/3.0);
              freeRotMat *= det;
              QMatrix3x3 frm;

              frm(0,0)=freeRotMat.m11;
              frm(0,1)=freeRotMat.m12;
              frm(0,2)=freeRotMat.m13;

              frm(1,0)=freeRotMat.m21;
              frm(1,1)=freeRotMat.m22;
              frm(1,2)=freeRotMat.m23;

              frm(2,0)=freeRotMat.m31;
              frm(2,1)=freeRotMat.m32;
              frm(2,2)=freeRotMat.m33;
              QQuaternion p=QQuaternion::fromRotationMatrix(frm);
              qDebug()<<p;
              qDebug()<<qRadiansToDegrees(p.scalar())<<p.vector().length();
              splitRotate();
          }else{
          double slow=1.0;
          if (event->modifiers()==Qt::ShiftModifier) slow=0.25;
          if (event->modifiers()==Qt::ControlModifier) slow=0.06;
          glRotateMat(mvmat,slow*dy*360.0f,1,0,0);
          glRotateMat(mvmat,slow*dx*360.0f,0,1,0);
          pmv = pmat*mvmat;
          nm = mvmat.normalMatrix();
          eye = mvmat.inverted()*QVector4D(0,0,0,1);
          if ((freeRot!=nullptr)&&(freeRot->isChecked())){
              Matrix worldtoscreen=transponse(Matrix(mvmat));
              double det=1.0/pow(determinant(worldtoscreen), 1.0/3.0);
              worldtoscreen *= det;
              V3 wox(1,0,0);
              V3 woy(0,1,0);
              screenRotx = worldtoscreen * wox;
              screenRoty = worldtoscreen * woy;
          }
          updateScreenAtoms();
      }
  }else if (event->buttons() & Qt::MiddleButton){
      if ((freeRot!=nullptr)&&(freeRot->isChecked())&&(event->modifiers()==(Qt::AltModifier|Qt::ControlModifier))){
        V3 mv(dx, -dy, 0);
        makeCurrent();
        Matrix worldtoscreen=transponse(Matrix(mvmat));
        double det=1.0/pow(determinant(worldtoscreen), 1.0/3.0);
        worldtoscreen *= det;
        mv = worldtoscreen * mv;
        V3 o(orgx->value(), orgy->value(), orgz->value());
        o+=mv;
        orgx->setValue(o.x);
        orgy->setValue(o.y);
        orgz->setValue(o.z);

        splitRotate();
      }else{
      //pmat.translate(3*dx*viewAngle,-3*dy*viewAngle,0);
        float mf = width()/42.0;

        pmat.translate(mf*dx*sqrt(viewAngle),-mf*dy*sqrt(viewAngle));
        pmv = pmat*mvmat;
        nm = mvmat.normalMatrix();
        eye = mvmat.inverted()*QVector4D(0,0,0,1);
        updateScreenAtoms();
      }
  }
  moux=x;
  mouy=y;  
  update();
}

void PlayOpenGL::toggleOrtho(bool b){
    if (b) setViewAngle(0.05);
    else setViewAngle(previousViewAngle);
}

void PlayOpenGL::setViewAngle(double ang){
    if ((ang>=0.00)&&(ang<160.0)){

      double ll = ang / viewAngle;
      float ratio = static_cast<float>(width()) / static_cast<float>(height());
      pmat.setToIdentity();
      if (ang<0.1) {
          L=2.0/mol->dimension(mol->showatoms);

          mvmat.setToIdentity();
          mvmat.lookAt(QVector3D(0.0, 200, 0) , QVector3D(  0.0, 0.0, 0.0), QVector3D(0.0, -100, 400 ));
          mvmat.scale(L,L,L);
          pmat.ortho(-ratio,ratio,-1.0f,1.0f,-5.0f,8000.0f);

          previousViewAngle = qMax(viewAngle, 29.0);
          viewAngle=0.01;
          isOrtho=true;
          toggleOrthoView->setChecked(true);
      }
      else {
          viewAngle=ang;
          if (isOrtho){
              double dim=mol->dimension(mol->showatoms);
              L=ang * 100.0 / (dim * 29.0);

              mvmat.setToIdentity();
              mvmat.lookAt(QVector3D(0.0, 200, 0) , QVector3D(  0.0, 0.0, 0.0), QVector3D(0.0, -100, 400 ));
              mvmat.scale(L,L,L);

          }else {
              L*=ll;
              mvmat.scale(ll,ll,ll);

          }
          isOrtho = false;
          pmat.perspective(static_cast<float>(viewAngle), ratio, 5.0f, 8000.0f );
          toggleOrthoView->setChecked(false);
      }
      emit viewAngeleChanged(viewAngle);
      homeXY();
    }
}

void PlayOpenGL::homeXY(){//! Resets the molecule to rotate in the middle of the viewport.
  QVector4D t = pmat.column(3);
  t.setX(0);
  t.setY(0);
  pmat.setColumn(3,t);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
}

void PlayOpenGL::rotCenter(){//! reset the rotation center to the center of gravity.
  rotze=-1;
  update();
  rCenter->hide();
  updateScreenAtoms();
}

void PlayOpenGL::toogleWithSymmetry(bool b){
  withsymm=(b)?0:1;
}

void PlayOpenGL::expand(){//! Lets Molecule.expandAt(int index) serach for neighboring molecules around the specified atom.
  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else return;
  if (index==(int)((GLuint)-1))return;
  mol->expandAt(index);
  fuse->setVisible(true);
  grow->setVisible(false);
  emit bigmessage(mol->HumanSymmetry);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);
    makeAtoms(mol->showatoms);
  update();
}


void PlayOpenGL::hideThisFragment(){//! Hides the specified molecular fragment.
  //sdm();
  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else return;
  if (index>mol->showatoms.size()) return;
  int gute=mol->showatoms.at(index).molindex+9999*mol->showatoms.at(index).symmGroup;
  //int gute=mol->showatoms.at(index).molindex;
  for (int i=0;i<mol->showatoms.size();i++)
    if (mol->showatoms.at(i).an>=0)
      mol->showatoms[i].hidden=
        (mol->showatoms.at(i).molindex+9999*mol->showatoms.at(i).symmGroup!=gute)?mol->showatoms.at(i).hidden:1;
  //(mol->showatoms.at(i).molindex!=gute)?mol->showatoms.at(i).hidden:1;
  //
  //frid=gute;
  hiddenThings=true;
  //hideReason|=HIDE_REASON_THIS_FRAGMENT;
  mol->selectedatoms.clear();  
  makeAtoms(mol->showatoms);
  makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

  makeSelectedAtoms();
 // murx=-__LINE__;
  updateBondActions(__LINE__);
  update();
}

void PlayOpenGL::selectThisFragment(){//! Selects the specified molecular fragment
  //sdm();
  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else return;
  if (index>mol->showatoms.size()) return;
  mol->selectedatoms.clear();
  int gute=mol->showatoms.at(index).molindex+9999*mol->showatoms.at(index).symmGroup;
  //int gute=mol->showatoms.at(index).molindex;
  for (int i=0;i<mol->showatoms.size();i++)
    if (mol->showatoms.at(i).an>=0)
      if(mol->showatoms.at(i).molindex+9999*mol->showatoms.at(i).symmGroup==gute){
        mol->selectedatoms.append(mol->showatoms[i]);
        mol->selectedatoms.last().style=i;
      }
  //(mol->showatoms.at(i).molindex!=gute)?mol->showatoms.at(i).hidden:1;
  //
  updateBondActions(__LINE__);
  makeSelectedAtoms();
  update();
}

void PlayOpenGL::hideSelected(){//! Hides the selected atoms
  for(int i=0;i<mol->selectedatoms.size();i++){
    if (mol->selectedatoms.at(i).style < mol->showatoms.size())
      mol->showatoms[mol->selectedatoms.at(i).style].hidden=1;
    //   std::cout<<mol->showatoms.at(mol->selectedatoms.at(i).style).Label.toStdString()<<std::endl;
  }
  hiddenThings=true;
  mol->selectedatoms.clear();
  makeAtoms(mol->showatoms);
  makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);
  updateBondActions(__LINE__);
  update();
}

void PlayOpenGL::hideOtherFragments(){//! Hides all molecular fragments except from the specified one
  //sdm();
  QAction *action = qobject_cast<QAction *>(sender());
  int index=0;
  if (action)
    index=action->data().toInt();
  else return;
  if (index==(int)((GLuint)-1))return;
  if (index>mol->showatoms.size()) return;
  int gute=mol->showatoms.at(index).molindex+withsymm*9999*mol->showatoms.at(index).symmGroup;
  //int gute=mol->showatoms.at(index).molindex;

  for (int i=0;i<mol->showatoms.size();i++)
    if (mol->showatoms.at(i).an>=0){
      mol->showatoms[i].hidden=
        //		(mol->showatoms.at(i).molindex!=gute)?1:mol->showatoms.at(i).hidden;
        (mol->showatoms.at(i).molindex+withsymm*9999*mol->showatoms.at(i).symmGroup!=gute)?1:mol->showatoms.at(i).hidden;

    }
  hiddenThings=true;
  //frid=gute;
  //hideReason|=HIDE_REASON_OTHER_FRAGMENT;
  mol->selectedatoms.clear();
  makeAtoms(mol->showatoms);
  makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);
  updateBondActions(__LINE__);
  update();
}

QColor PlayOpenGL::farbverlaufQC(double wrt, double min, double max){
    if (min>=max) max=min+0.001;
    double rot,gruen,blau,alpha;
    int lauf=0;
    double farbe[6][4]={{1.0,0.0,0.0,1.0},
                       {1.0,1.0,0.0,1.0},
                          {0.0,1.0,0.0,1.0},
                          {0.0,1.0,1.0,1.0},
                          {0.0,0.0,1.0,1.0},
                          {1.0,0.0,1.0,1.0}};
    double nwrt=(wrt-min)/(max-min);
    nwrt=(nwrt>=1.0)?0.99999:nwrt;
    nwrt=(nwrt<=0.0)?0.00001:nwrt;
    lauf=(int (nwrt/0.2));
    nwrt-=(0.2*lauf);
    nwrt/=(0.2);

    rot=(((1.0-nwrt)*farbe[lauf][0]+farbe[lauf+1][0]*nwrt));
    gruen=(((1.0-nwrt)*farbe[lauf][1]+farbe[lauf+1][1]*nwrt));
    blau=(((1.0-nwrt)*farbe[lauf][2]+farbe[lauf+1][2]*nwrt));
    alpha=1.0;
    return QColor(static_cast<int>(255*rot),
           static_cast<int>(255*gruen),
           static_cast<int>(255*blau),
           static_cast<int>(255*alpha));
}


#if defined (Q_OS_MAC)
void PlayOpenGL::contextMenuEvent(QMouseEvent *event){
    printf("??Context Menu event! %d %d\n", event->pos().x(), event->pos().y());
#else
void PlayOpenGL::contextMenuEvent(QContextMenuEvent *event){
#endif
    if (!atomsClickable) return;
    QPoint menupoint = event->globalPos();
  double nahda=pickradius,da=0;
  int nahdai=-1;
  for (int j=0; j<mol->showatoms.size();j++){
    if (mol->showatoms.at(j).hidden) continue;
    da=(((screenAtoms.at(j).x()-event->x())*( screenAtoms.at(j).x()-event->x()))+
        ((screenAtoms.at(j).y()-event->y())*( screenAtoms.at(j).y()-event->y())));
    nahdai=(da<nahda)?j:nahdai;
    nahda=qMin(nahda,da);
  }
  if ((nahdai < 0)||(mol->showatoms.size()<nahdai)) {
    return;
  }
  else {
    int idx=nahdai;
    ImeanThisAtom=idx;
    if (idx>mol->showatoms.size()) return;
    if ((mol->showatoms.at(idx).an==-1)&&(mol->showatoms.at(idx).peakHeight<qcutoff)) return;
    if (menu == nullptr) menu = new QMenu(mol->showatoms.at(idx).Label);
    else {
        menu->setTitle(mol->showatoms.at(idx).Label);
        menu->clear();
    }
    if (!isMainWindow) {
        if (enviMenu == nullptr) enviMenu = new QMenu("ENVI-Settings");
        else enviMenu->clear();
        enviMenu->addAction("Change envi range", this, SLOT(changeEnviRange()));
        enviMenu->addAction(enviNoQ);
        enviMenu->addAction(enviCova);
        menu->addMenu(enviMenu);
        if ((int) idx < mol->showatoms.size()){
            QAction *a=nullptr;
          a=menu->addAction(QString("List ENVIronment of %1").arg(mol->showatoms.at(idx).Label),this,SLOT(envi()));
          a->setData(idx);
        }
        menu->exec(menupoint);
        return;
    }
    if (restrMenu == nullptr) restrMenu = new QMenu("Add restraints");
    else restrMenu->clear();
    if (constrMenu == nullptr) constrMenu = new QMenu("Add constraints");
    else constrMenu->clear();
    if (enviMenu == nullptr) enviMenu = new QMenu("ENVI-Settings");
    else enviMenu->clear();
    if ((selSfacMenu)&&(!mol->selectedatoms.isEmpty())) {
        menu->addMenu(selSfacMenu);
    }
    if (sfacMenu){
    QList<QAction *> sfas = sfacMenu->actions();
    for (int k=0;k<sfas.size();k++){
      sfas[k]->setEnabled(sfas[k]->text()!=mol->pse(mol->showatoms.at(idx).an));
    }
    if (mol->showatoms.at(idx).symmGroup==0) menu->addMenu(sfacMenu);
    }
    QAction *a=nullptr;
    if (isMainWindow) {
    if ((mol->selectedatoms.isEmpty())&&(mol->showatoms.at(idx).symmGroup==0)) {
          a=menu->addAction(QIcon(theIconPath+"killselected.svg"),QString("Delete %1 ").arg(mol->showatoms.at(idx).Label),
          chparent,SLOT(deleteSelectedAtoms()));           
          a->setData(idx);
    }else if ((mol->showatoms.at(idx).symmGroup==0)) {
      a=menu->addAction(QIcon(theIconPath+"killselected.svg"),QString("Delete selected atoms"),
          chparent,SLOT(deleteSelectedAtoms()));
      a->setData(-1);
    }
    if ((!mol->selectedatoms.isEmpty())) a=menu->addAction("Change RESI or PART of selected atoms",chparent,SLOT(changeResiPart()));
    }
    menu->addSeparator();
    a=menu->addAction(QString("Set rotation center to %1").arg(mol->showatoms.at(idx).Label),this,SLOT(setRotationCenter()));
    a->setData(idx);
    a=menu->addAction(QString("Expand contacts around %1").arg(mol->showatoms.at(idx).Label),this,SLOT(expand()));
    a->setData(idx);
    a=menu->addAction(QString("Hide this fragment"),this,SLOT(hideThisFragment()));
    a->setData(idx);
    a=menu->addAction(QString("Select this fragment"),this,SLOT(selectThisFragment()));
    a->setData(idx);
    a=menu->addAction(QString("Hide other fragments"),this,SLOT(hideOtherFragments()));
    a->setData(idx);
    a=menu->addAction(QString("Fragment includes symmetry generated"), this ,SLOT(toogleWithSymmetry(bool)));
    a->setCheckable(true);
    a->setChecked(!(withsymm));
    if (!mol->selectedatoms.isEmpty()) a=menu->addAction("Hide selected Atoms",this,SLOT(hideSelected()));
    bool hasSameBonds=false;//for sadi two neighbor) atoms need to be same atomic number and no hydrogen
    //it warks if they are not the same but gives a warning and does not make much sense
    if (idx<mol->asymm.size()) for (int nb=1; nb<mol->knoepfe.at(idx).neighbors.size(); nb++){
      if ((mol->knoepfe.at(idx).neighbors.at(nb-1)>=mol->asymm.size())||(mol->knoepfe.at(idx).neighbors.at(nb)>=mol->asymm.size())) continue;
      if ((mol->asymm.at(mol->knoepfe.at(idx).neighbors.at(nb-1)).an>0)&&
          (mol->asymm.at(mol->knoepfe.at(idx).neighbors.at(nb-1)).an==mol->asymm.at(mol->knoepfe.at(idx).neighbors.at(nb)).an))hasSameBonds=true;
    }
    if(!mol->selectedatoms.isEmpty()){
      restrMenu->addAction("ISOR", this, SLOT(addISOR()));
      if (hasSameBonds){
        QAction *restr;
        restr=restrMenu->addAction("SADI", this, SLOT(addSADI()));
        restr->setData(idx);
      }
      if(mol->selectedatoms.size() == 2){
        restrMenu->addAction("DFIX", this, SLOT(addDFIX()));
        restrMenu->addAction("DANG", this, SLOT(addDANG()));
      }
      if(mol->selectedatoms.size() >= 4){
        restrMenu->addAction("FLAT", this, SLOT(addFLAT()));
      }
      if(mol->selectedatoms.size() > 1){
        constrMenu->addAction("EXYZ", this, SLOT(addEXYZ()));
        constrMenu->addAction("EADP", this, SLOT(addEADP()));
        restrMenu->addAction("RIGU", this, SLOT(addRIGU()));
        restrMenu->addAction("DELU", this, SLOT(addDELU()));
        restrMenu->addAction("SIMU", this, SLOT(addSIMU()));
        restrMenu->addAction("CHIV", this, SLOT(addCHIV()));
      }
      menu->addMenu(restrMenu);
      menu->addMenu(constrMenu);
    }else if (hasSameBonds) {
      restrMenu = new QMenu("Add restraints");
      QAction *restr;
      restr=restrMenu->addAction("SADI", this, SLOT(addSADI()));
      restr->setData(idx);
      menu->addMenu(restrMenu);
    }
    menu->addSeparator();
    {
        QAction *a = nullptr;
        a = enviMenu->addAction("Change ENVI Range", this, SLOT(changeEnviRange()));
        a->setData(idx);
        enviMenu->addAction(enviNoQ);
        enviNoQ->setData(idx);
        connect(enviNoQ,SIGNAL(triggered(bool)),this,SLOT(envi()));
        enviMenu->addAction(enviCova);
        enviCova->setData(idx);
        connect(enviCova,SIGNAL(triggered(bool)),this,SLOT(envi()));
    }
    /*
    enviMenu->addAction(QString("Change ENVI Range (%1A)").arg(envirange), this, SLOT(changeEnviRange()));
    enviMenu->addAction(enviNoQ);
    enviMenu->addAction(enviCova);
    */
    menu->addMenu(enviMenu);

    if ((int) idx < mol->asymm.size()){
      a=menu->addAction(QString("List ENVIronment of %1").arg(mol->showatoms.at(idx).Label),this,SLOT(envi()));
      a->setData(idx);      
      a=menu->addAction(QString("Grow ENVIronment of %1").arg(mol->showatoms.at(idx).Label),this,SLOT(growEnvi()));
      a->setData(idx);
      a=menu->addAction(QString("Show only ENVIronment of %1").arg(mol->showatoms.at(idx).Label),this,SLOT(showEnvi()));
      a->setData(idx);
    }
    menu->addSeparator();
    int ssgr=0;
    int adpat=0;
    for (int seli=0; seli<mol->selectedatoms.size(); seli++){
      if ((mol->showatoms.at(seli).symmGroup==0)&&(!mol->selectedatoms.at(seli).isIso))adpat++;
    }
    for (int seli=0; seli<mol->selectedatoms.size(); seli++){
      if (mol->selectedatoms.at(seli).symmGroup>0) ssgr+=mol->selectedatoms.at(seli).symmGroup;
      else {
        ssgr=0;
        break;
      }
    }
    if (ssgr) adpat=0;
    if (ssgr){
      a=menu->addAction(QString("Move selected atoms to selected sites"),
          chparent,SLOT(moveSymmMateSel()));
      //////// 28.03.12
      if ((mol->showatoms.at(idx).part<0)){
        a=menu->addAction(QString("Copy %1 here").arg(mol->showatoms.at(idx).Label.section("_",0,0)),
            chparent,SLOT(copySymmMate()));
        a->setData(idx);
      }
      //////// -
    }else{      
        if ((mol->showatoms.at(idx).symmGroup>0)){
        a=menu->addAction(QString("Move %1 here").arg(mol->showatoms.at(idx).Label.section("_",0,0)),
            chparent,SLOT(moveSymmMate()));
        a->setData(idx);
        if ((mol->showatoms.at(idx).part<0)){
          a=menu->addAction(QString("Copy %1 here").arg(mol->showatoms.at(idx).Label.section("_",0,0)),
              chparent,SLOT(copySymmMate()));
          a->setData(idx);
        }
      }
    }
    menu->addSeparator();
    if ((nahdai<mol->asymm.size())&&(mol->knoepfe.at(idx).neighbors.size()==1)&&((mol->showatoms.at(idx).an==6)||(mol->showatoms.at(idx).an==5))) {
      a=menu->addAction(QString("Add dissordered %1 Hydrogen atoms").arg(((mol->showatoms.at(idx).an==5))?"methyl":"-NH3"),chparent,SLOT(addDissorderedMethyl()));
      a->setData(idx);
      a=menu->addAction(QString("Add ordered %1 Hydrogen atoms").arg(((mol->showatoms.at(idx).an==5))?"methyl":"-NH3+"),chparent,SLOT(addMethyl()));
      a->setData(idx);
      a=menu->addAction(QString("Add terminal %1 Hydrogen atoms").arg(((mol->showatoms.at(idx).an==5))?"=CH2 ":"=NH2+"),chparent,SLOT(addH93()));
      a->setData(idx);
      if (mol->showatoms.at(idx).an==5){
        a=menu->addAction(QString("Add acetylenic C-H Hydrogen atom"),chparent,SLOT(addH163()));
        a->setData(idx);
      }
      if (mol->showatoms.at(idx).an==6){
        a=menu->addAction(QString("Add %1 Hydrogen atoms").arg("-NH2 (non planar)"),chparent,SLOT(addHnonplanarAminR1()));
        a->setData(idx);
      }
    }
    if ((nahdai<mol->asymm.size())&&(mol->knoepfe.at(idx).neighbors.size()==1)&&(mol->showatoms.at(idx).an==7)){
        a->setData(idx);
        a=menu->addAction(QString("Add hydroxyl Hydrogen atom to %1").arg(mol->showatoms.at(idx).Label.section("_",0,0)),chparent,SLOT(addHydroxyl()));
        a->setData(idx);
    }
    if ((nahdai<mol->asymm.size())&&(mol->knoepfe.at(idx).neighbors.size()==2)&&((mol->showatoms.at(idx).an==6)||(mol->showatoms.at(idx).an==5))) {
      a=menu->addAction(QString("Add %1 Hydrogen atoms").arg(((mol->showatoms.at(idx).an==5))?">CH2":">NH2+"),chparent,SLOT(addH23()));
      a->setData(idx);
      a=menu->addAction(QString("Add %1 Hydrogen atom").arg(((mol->showatoms.at(idx).an==5))?">CH":">NH (planar)"),chparent,SLOT(addH43()));
      a->setData(idx);
      //addHnonplanarAminR2();
      if ((mol->showatoms.at(idx).an==6)){
        a=menu->addAction(QString("Add %1 Hydrogen atom").arg(">NH (non planar)"),chparent,SLOT(addHnonplanarAminR2()));
        a->setData(idx);
      }
    }
    if ((idx < mol->asymm.size()) && (mol->asymm.at(idx).an>-1)) {
      a=menu->addAction(QString("Use %1 as new label for rename mode").arg(mol->showatoms.at(idx).Label.section("_",0,0)),chparent,SLOT(renameThisAtom()));
      a->setData(idx);
    }

    if(!mol->selectedatoms.isEmpty()){
      menu->addAction(QString("Make selected atoms ANIS"),this,SLOT(addANIS()));
    }
    if ((adpat)) {
      if (mol->selectedatoms.isEmpty()){
        a=menu->addAction(QIcon(theIconPath+"sina.svg"),QString("make %1 isotropic (ISOT)")
            .arg(mol->showatoms.at(idx).Label),chparent,SLOT(sina()));
        a->setData(idx);
      }else{
        a=menu->addAction(QIcon(theIconPath+"sina.svg"),QString("make selected atoms isotropic (ISOT)"),chparent,SLOT(sina()));
        a->setData(-1);
      }
    }
    if ((idx<mol->asymm.size())&&(mol->asymm.at(idx).an>-1)){
        QAction *a = menu->addAction("Draw voronoi polyeder of this atom",chparent,SLOT(makeVoro()));
        a->setData(idx);
      }
    if ((mol->showatoms.at(p).molindex!=mol->showatoms.at(idx).molindex)){
      menu->addSeparator();
      menu->addAction(QString("Inherit Labels from molecule around %1 to atoms around %2")
          .arg(mol->showatoms.at(p).Label)
          .arg(mol->showatoms.at(idx).Label)
          ,chparent,SLOT(inheritLabels()));
      ppp = pp;
      pp = p;
      p = idx;
    }
    menu->exec(menupoint);
    //delete enviMenu;
    //delete menu;
  }
}


void PlayOpenGL::mousePressEvent(QMouseEvent *event){
#if defined (Q_OS_MAC)
#include <QApplication>
    if ((event->buttons() & Qt::RightButton)){
        contextMenuEvent(event);
        return;
    }
#endif

  if (menu) menu->close();
  if (rectangle){
    rectangle=false;
    update();
  }
  moux=event->pos().x();
  mouy=event->pos().y();

  p = qMin(p,mol->showatoms.size()-1);
  pp = qMin(pp,mol->showatoms.size()-1);
  ppp = qMin(ppp,mol->showatoms.size()-1);

  p = qMax(p,0);
  pp = qMax(pp,0);
  ppp = qMax(ppp,0);
  double nahda=pickradius,da=0;
  int nahdai=-1;
  for (int j=0; j<mol->showatoms.size();j++){
    if (mol->showatoms.at(j).hidden) continue;
    if ((mol->showatoms.at(j).an==-1)&&(mol->showatoms.at(j).peakHeight<qcutoff)) continue;

    da=(((screenAtoms.at(j).x()-event->x())*( screenAtoms.at(j).x()-event->x()))+
        ((screenAtoms.at(j).y()-event->y())*( screenAtoms.at(j).y()-event->y())));
    nahdai=(da<nahda)?j:nahdai;
    nahda=qMin(nahda,da);
  }
  if (event->buttons() & Qt::LeftButton){
    scrx0=moux;
    scry0=mouy;
    if ((event->buttons() & Qt::LeftButton) && (nahdai >= 0) && (nahdai < mol->showatoms.size())){
        int index=nahdai;
        if (event->modifiers()==Qt::NoModifier) {
        double w=0,dw=0;
        if ((pp!=p)&&(pp!=index)&&(p!=index)) {
          V3 aa1=mol->showatoms.at(p).pos-mol->showatoms.at(pp).pos;
          V3 aa2=mol->showatoms.at(p).pos-mol->showatoms.at(index).pos;
          w=mol->winkel(aa1,aa2);
          if ((ppp!=p)&&(ppp!=pp)&&(ppp!=index))
            dw=mol->dieder(mol->showatoms.at(pp).pos-mol->showatoms.at(ppp).pos,
                mol->showatoms.at(pp).pos-mol->showatoms.at(p).pos,
                mol->showatoms.at(p).pos-mol->showatoms.at(index).pos);
        }
        if (atomsClickable){
          emit jumpit(index);
         /* printf("%s %d %d\n",
              mol->showatoms.at(index).Label.toStdString().c_str(),
              mol->showatoms.at(index).scod,
              mol->showatoms.at(index).an);*/
          if (inRenameMode) return;
          if (!inRenameMode){
            apair.clear();
            if (event->modifiers()==Qt::NoModifier) {
              int isschon=-1;
              for (int i=0; i< mol->selectedatoms.size();i++){
                isschon=(mol->selectedatoms.at(i).style==index)?index:isschon;
              }
              mol->selectedatoms.clear();
              if (isschon==-1){
                mol->selectedatoms.append(mol->showatoms[index]);
                mol->selectedatoms.last().style=index;
              }
            }
            if ((manualAx!=nullptr)&&(!manualAx->isChecked())&&(event->modifiers()==Qt::AltModifier)) {
              int pv1=pivot1cb->findText(mol->showatoms.at(p).Label);
              int pv2=pivot2cb->findText(mol->showatoms.at(index).Label);
              if ((pv1>-1)&&(pv2)){
                  emit pivot1Changed(pv1);
                  emit pivot2Changed(pv2);
                  altpivot=true;

              }

            }
            if (event->modifiers()==Qt::ShiftModifier) {
              int min=mol->showatoms.size(), max=-1;
              for (int i=0; i< mol->selectedatoms.size();i++){
                min=qMin(mol->selectedatoms.at(i).style,min);
                max=qMax(mol->selectedatoms.at(i).style,max);
              }
              min=qMin((int)index,min);
              max=qMax((int)index,max);
              mol->selectedatoms.clear();
              for (int i=min; i<=max; i++){
                mol->selectedatoms.append(mol->showatoms[i]);
                mol->selectedatoms.last().style=i;
              }
            }
            if (event->modifiers()==Qt::ControlModifier) {
              if (mol->selectedatoms.contains(mol->showatoms[index])){
                mol->selectedatoms.removeOne(mol->showatoms[index]);
              } else {
                mol->selectedatoms.append(mol->showatoms[index]);
                mol->selectedatoms.last().style=index;
                //                for (int oo=0; oo<mol->selectedatoms.size(); oo++) qDebug()<<mol->selectedatoms.at(oo).Label<<mol->selectedatoms.at(oo).symmGroup<< mol->selectedatoms.at(oo).style<<mol->selectedatoms.size();
              }
            }
            //	  std::cout<<mol->selectedatoms.size()<<mol->selectedatoms.last().Label.toStdString()<<std::endl;

            emit selectionChanged();
            updateBondActions(__LINE__);
            update();
          }
          //6026

          V3 hin;
          double dmsda=0;
          hin = Normalize(mol->showatoms.at(index).pos-mol->showatoms.at(p).pos);
          if ((mol->showatoms.at(index).an>-1) && (mol->showatoms.at(p).an))
            dmsda = fabs(((hin*mol->showatoms.at(index).uc)*hin)- ((hin*mol->showatoms.at(p).uc)*hin));
          //
          bool syok;
          QString symml;
          int syid=mol->showatoms.at(index).Label.section(QString::fromUtf8("»"),1).toInt(&syok);
          if (mol->usedSymmetry.size()&&syok) symml=mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid);
          syid=mol->showatoms.at(p).Label.section(QString::fromUtf8("»"),1).toInt(&syok);
          if ((mol->usedSymmetry.size())&&(syok)&&(!symml.contains(mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid))))
            symml+=mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid);

          syid=mol->showatoms.at(pp).Label.section(QString::fromUtf8("»"),1).toInt(&syok);
          if ((mol->usedSymmetry.size())&&(syok)&&(!symml.contains(mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid))))
            symml+=mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid);

          syid=mol->showatoms.at(ppp).Label.section(QString::fromUtf8("»"),1).toInt(&syok);
          if ((mol->usedSymmetry.size())&&(syok)&&(!symml.contains(mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid))))
            symml+=mol->symmcode2human(mol->usedSymmetry.at(syid-1),syid);
          /*printf("%s org=%s iso=%d\n a%9.6f b%9.6f c%9.6f  %9.6f %9.6f %9.6f\n d%9.6f e%9.6f f%9.6f  %9.6f %9.6f %9.6f\n g%9.6f h%9.6f i%9.6f  %9.6f %9.6f %9.6f\n\n",
            mol->showatoms.at(index).Label.toStdString().c_str(),
            mol->showatoms.at(index).ufiso_org.toStdString().c_str(),
            mol->showatoms.at(index).isIso,
            mol->showatoms.at(index).uf.m11,mol->showatoms.at(index).uf.m12,mol->showatoms.at(index).uf.m13,  mol->showatoms.at(index).uc.m11,mol->showatoms.at(index).uc.m12,mol->showatoms.at(index).uc.m13,
            mol->showatoms.at(index).uf.m21,mol->showatoms.at(index).uf.m22,mol->showatoms.at(index).uf.m23,  mol->showatoms.at(index).uc.m21,mol->showatoms.at(index).uc.m22,mol->showatoms.at(index).uc.m23,
            mol->showatoms.at(index).uf.m31,mol->showatoms.at(index).uf.m32,mol->showatoms.at(index).uf.m33,  mol->showatoms.at(index).uc.m31,mol->showatoms.at(index).uc.m32,mol->showatoms.at(index).uc.m33
            );// */

          if ((index!=p)&&(index!=pp)&&(mol->showatoms.at(index).Label[0]=='Q')){
            /*printf("scod %d %f %f %f \n",mol->showatoms.at(index).scod,
              mol->showatoms.at(index).frac.x,
              mol->showatoms.at(index).frac.y,
              mol->showatoms.at(index).frac.z);//  **/
            emit bigmessage(QString("<hr><table><tr><td><b>%1</b></td><td align=\"right\"></td></tr><tr><td><b>%1--%2</b>"
                  "</td><td align=\"right\">%5 &Aring;</td></tr><tr><td><b>%1--%2--%3</b></td><td "
                  "align=\"right\">%6&deg;</td></tr><tr><td><b>%1--%2--%3--%4 </b></td><td "
                  "align=\"right\">%7&deg;</td></tr><tr><td><b>%1</b> Peak Height: </td><td "
                  "align=\"right\">%8 e&Aring;<sup>-3</sup></td><td>[%9&hellip;%10]"
                  "</td></tr></table>%11")
                .arg((mol->showatoms.at(index).Label))                     //1
                .arg((mol->showatoms.at(p).Label))				//3
                .arg((mol->showatoms.at(pp).Label))			                //5			                                                //8
                .arg((mol->showatoms.at(ppp).Label))			                        //9
                .arg(sqrt(Distance(mol->showatoms.at(index).pos,mol->showatoms.at(p).pos)),8,'f',3) //4
                .arg(w,8,'f',2)
                .arg(dw,8,'f',2)
                .arg(mol->showatoms.at(index).peakHeight,8,'f',2)
                .arg(mol->pmin,3,'f',2)
                .arg(mol->pmax,3,'f',2)
                .arg(symml)

                );
          }
          else if ((index!=p)&&(index!=pp)&&(mol->showatoms.at(index).Label[0]=='L')&&(mol->showatoms.at(index).an==-42)){
            emit bigmessage(QString("<hr><table><tr><td><b>%1</b></td><td align=\"right\"></td></tr><tr><td><b>%1--%2</b>"
                  "</td><td align=\"right\">%5 &Aring;</td></tr><tr><td><b>%1--%2--%3</b></td><td "
                  "align=\"right\">%6&deg;</td></tr><tr><td><b>%1--%2--%3--%4 </b></td><td "
                  "align=\"right\">%7&deg;</td></tr><tr><td><b>%1</b> a exp(-b x<sup>2</sup>)</td><td "
                  "align=\"right\">a = %8 </td><td> b = %9 "
                  "</td></tr></table>%11")
                .arg((mol->showatoms.at(index).Label))                     //1
                .arg((mol->showatoms.at(p).Label))				//3
                .arg((mol->showatoms.at(pp).Label))			                //5			                                                //8
                .arg((mol->showatoms.at(ppp).Label))			                        //9
                .arg(sqrt(Distance(mol->showatoms.at(index).pos,mol->showatoms.at(p).pos)),8,'f',3) //4
                .arg(w,8,'f',2)
                .arg(dw,8,'f',2)
                .arg(mol->showatoms.at(index).sof,8,'f',2)
                .arg(mol->showatoms.at(index).peakHeight,8,'f',2)
                .arg(symml)

                );
          }
          else if ((index!=p)&&(index!=pp)){
            emit bigmessage(QString("<br><hr><table><tr><td><b>%1</b> part: %8 residue: %9 %10 (fragment: %12)</td><td align=\"right\">"
                  "</td></tr><tr><td><b>%1--%2</b></td><td align=\"right\">%5 &Aring;</td><td>&nbsp;&nbsp;&nbsp;DMSDA: %13 &Aring;<sup>2</sup></td></tr>"
                  "<tr><td><b>%1--%2--%3</b></td><td align=\"right\">%6&deg;</td><td>&nbsp;&nbsp;&nbsp;1,3 Distance( DANG) %14 &Aring;</td></tr><tr><td>"
                  "<b>%1--%2--%3--%4 </b></td><td align=\"right\">%7&deg;</td></tr></table>%11")
                .arg((mol->showatoms.at(index).Label))			       //1
                .arg((mol->showatoms.at(p).Label))			               //2
                .arg((mol->showatoms.at(pp).Label))			       //3
                .arg((mol->showatoms.at(ppp).Label))			       //4
                .arg(sqrt(Distance(mol->showatoms.at(index).pos,mol->showatoms.at(p).pos)),8,'f',3) //5
                .arg(w,8,'f',2)			                            //6
                .arg(dw,8,'f',2)                                               //7
                .arg(mol->showatoms.at(index).part)
                .arg(mol->showatoms.at(index).resiNr)
                .arg(mol->showatoms.at(index).ResiClass)
                .arg(symml)
                .arg(mol->showatoms.at(index).molindex)
                .arg(dmsda,5,'e',2)
                .arg(sqrt(Distance(mol->showatoms.at(index).pos,mol->showatoms.at(pp).pos)),8,'f',3)
                );
          }
          //				   ,QMessageBox::Ok);
          ppp = pp;
          pp = p;
          p = index;
        }

        }else if (event->modifiers()==Qt::ShiftModifier) {
          int min=mol->showatoms.size(), max=-1;
          for (int i=0; i< mol->selectedatoms.size();i++){
            min=qMin(mol->selectedatoms.at(i).style,min);
            max=qMax(mol->selectedatoms.at(i).style,max);
          }
          min=qMin((int)index,min);
          max=qMax((int)index,max);
          mol->selectedatoms.clear();
          for (int i=min; i<=max; i++){
            mol->selectedatoms.append(mol->showatoms[i]);
            mol->selectedatoms.last().style=i;
          }
        }
        else if (event->modifiers()==Qt::ControlModifier) {
          if (mol->selectedatoms.contains(mol->showatoms[index])){
            mol->selectedatoms.removeOne(mol->showatoms[index]);
          } else {
            mol->selectedatoms.append(mol->showatoms[index]);
            mol->selectedatoms.last().style=index;
          }
        }

        emit selectionChanged();
        updateBondActions(__LINE__);
        makeSelectedAtoms();
        updateScreenAtoms();
        update();
    }
  }
  if ((event->buttons() & Qt::MiddleButton) && (nahdai >= 0) && (nahdai < mol->showatoms.size()) && (parent()->objectName()=="The Main Window")){
        int index=nahdai;
        rotze=((int)index<mol->showatoms.size())?index:-1;
        if (verbosity >2 ) printf("rot center %d %d %d\n",rotze, mol->showatoms.size(),index);
        if (rotze>-1){
            if (parent()->objectName()=="The Main Window") rCenter->show();
            homeXY();
            updateScreenAtoms();
            //glGetDoublev(GL_MODELVIEW_MATRIX, MM); MM[12] = MM[13] = 0; glLoadMatrixd(MM);
        }
        update();
      }
  update();
}

void PlayOpenGL::along001(){
    makeCurrent();
  double va=viewAngle;
  setViewAngle(29.0);
  mvmat(0,0)=1.f;
  mvmat(0,1)=0.f;
  mvmat(0,2)=0.f;
  mvmat(1,0)=0.f;
  mvmat(1,1)=1.f;
  mvmat(1,2)=0.f;
  mvmat(2,0)=0.f;
  mvmat(2,1)=0.f;
  mvmat(2,2)=1.f;
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
  setViewAngle(va);
  //rotX(0.7);
  //rotY((90.0-mol->cell.ga)/-20.0);
  printf("along001\n");
}

void PlayOpenGL::zoom(double speed){//!scale about speed @param speed scale factor
  double ll=1.0+speed*0.1;
  L*=ll;
  mvmat.scale(ll,ll,ll);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
}

void PlayOpenGL::rotY(double speed){//!rotate around Y axis @param speed in degrees
  //glRotateL(-20.0*speed,0.0f,1.0f,0.0f);
  glRotateMat(mvmat,-20.0*speed,0,1,0);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);

  if (verbosity>4) printf("Y%f \n%9g %9g %9g\n%9g %9g %9g\n%9g %9g %9g\n\n",-20*speed
         ,mvmat(0,0),mvmat(1,0),mvmat(2,0)
         ,mvmat(0,1),mvmat(1,1),mvmat(2,1)
         ,mvmat(0,2),mvmat(1,2),mvmat(2,2));
  updateScreenAtoms();
  update();
}

void PlayOpenGL::gZoom(double speed){//!scale about speed @param speed scale factor
  //glScaled(1.0+speed*0.02,1.0+speed*0.02,1.0+speed*0.02);
  double ll=1.0+speed*0.02;
  L*=ll;
  mvmat.scale(ll,ll,ll);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
}

void PlayOpenGL::rotZ(double speed){//!rotate around Z axis @param speed in degrees
  glRotateMat(mvmat,-20*speed,0,0,1);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
}

void PlayOpenGL::moveY(double speed){//!translate the molecule in Y direction
  //glTranslateL(0.0,speed,0.0);
  pmat.translate(0,speed*viewAngle,0);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();

  update();
}

void PlayOpenGL::moveX(double speed){//!translate the molecule in X direction
  //glTranslateL(speed,0.0,0.0);
  //pmat.translate(speed*viewAngle,0,0);
  pmat.translate(20.0,0,0);
  if (verbosity >2 ) printf("moveX %f %f\n",speed, viewAngle);
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  //  QMessageBox::information(this,"Px move","xmoved",QMessageBox::Ok);
  update();
}

void PlayOpenGL::rotX(double speed){//!rotate around X axis @param speed in degrees
  //glRotateL(-20.0*speed,1.0f,0.0f,0.0f);
  glRotateMat(mvmat,-20*speed,1,0,0);
  if (verbosity>4) printf("X%f \n%9g %9g %9g\n%9g %9g %9g\n%9g %9g %9g\n\n",-20*speed
         ,mvmat(0,0),mvmat(1,0),mvmat(2,0)
         ,mvmat(0,1),mvmat(1,1),mvmat(2,1)
         ,mvmat(0,2),mvmat(1,2),mvmat(2,2));
  pmv = pmat*mvmat;
  nm = mvmat.normalMatrix();
  eye = mvmat.inverted()*QVector4D(0,0,0,1);
  updateScreenAtoms();
  update();
}


void PlayOpenGL::paintBonds(QVector3D org){
  makeCurrent();
  bond_vao.bind();
  if ( !bond_program->bind() ){ qWarning() << "2Could not bind shader program to context"; return;}
  /*printf("bond %d %d\n  %d %d %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  %d  | %d L%d %d\n",
               bond_program->programId(),
               bond_vao.objectId(),
               bond_program->uniformLocation("atomColors"),
               bond_program->uniformLocation("h_bonds"),             //g
               bond_program->uniformLocation("hideHydrogen"),        //d
               bond_program->uniformLocation("hidePM1"),             //k
               bond_program->uniformLocation("highlightEquivalents"),//h
               bond_program->uniformLocation("highlightParts"),      //i
               bond_program->uniformLocation("lightPos"),//q
               bond_program->uniformLocation("normalMatrix"),
               bond_program->uniformLocation("origin"),
               bond_program->uniformLocation("pmv"),
               bond_program->uniformLocation("qPeakBonds"),          //e
               bond_program->uniformLocation("qcutoff"),             //f
               bond_program->uniformLocation("qradius"),             //b
               bond_program->uniformLocation("radius"),              //a
               bond_program->uniformLocation("segments"),            //c
               bond_program->uniformLocation("uniColor"),           //n
               bond_program->uniformLocation("uniColorBonds"),       //m
               nbondx2,
               atom_program->programId(),
               atom_program->isLinked(),
               atom_vao.objectId()
               );// */
  glEnable(GL_DEPTH_TEST );
  glEnable(GL_BLEND);
  glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
  glDisable(GL_CULL_FACE);
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  bond_program->setUniformValue("lightPos", QVector3D(100.0, 100.0, 100.0));
  bond_program->setUniformValue("pmv", pmv);
  bond_program->setUniformValue("origin", org);
  bond_program->setUniformValue("normalMatrix",nm);
  bond_program->setUniformValue("radius", static_cast<float>(mol->bondStrength));
  bond_program->setUniformValue("qradius", 0.1f*static_cast<float>(mol->q_bondWidth));
  bond_program->setUniformValue("segments", qMin(mol->LOD,8));
  bond_program->setUniformValue("hideHydrogen", !hideHydrogen->isVisible());
  bond_program->setUniformValue("qPeakBonds",qPeakBonds->isChecked());
  bond_program->setUniformValue("qcutoff", qcutoff);
  bond_program->setUniformValue("h_bonds", (h_bonds->isChecked()));
  bond_program->setUniformValue("highlightEquivalents", mol->highlightEquivalents);

  bond_program->setUniformValue("highlightParts", highlightParts->isChecked());
  bond_program->setUniformValue("hidePM1", shpm1->isChecked());
  bond_program->setUniformValue("uniColorBonds",(mol->bondColorStyle!=0));
  bond_program->setUniformValue("uniColor",mol->bondColor);
  bond_program->setUniformValue("deuteriumColor", mol->deuteriumColor);

  bond_program->enableAttributeArray( "VertexPosition" );
  bond_program->enableAttributeArray( "atomicnumbers" );
  bond_program->enableAttributeArray( "Styles" );
  bond_program->enableAttributeArray( "peakHeights" );
  bond_program->enableAttributeArray( "symmGroup" );
  glDrawArrays(GL_LINES, 0, nbondx2);
  //else glDrawElements(GL_LINES, nbondx2, GL_UNSIGNED_SHORT, 0);
  //printf("bondvao %d\n",bond_vao.objectId());
  bond_vao.release();
  bond_program->release();

  glFlush();
}

void PlayOpenGL::paintDraftBonds(QVector3D org){
  makeCurrent();
  draft_bond_vao.bind();
  if ( !draft_bond_program->bind() ){ qWarning() << "2Could not bind shader program to context"; return;}
  glEnable(GL_DEPTH_TEST );
  glEnable(GL_BLEND);
  glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
  glDisable(GL_CULL_FACE);
  draft_bond_program->setUniformValue("prmovi", pmv);
  draft_bond_program->setUniformValue("Color", mol->bondColor);
  draft_bond_program->setUniformValue("origin1", org);
  glDrawArrays(GL_LINES, 0, ndrbondx2);

  draft_bond_program->release();
  draft_bond_vao.release();
}

void PlayOpenGL::paintAtoms(QVector3D org){
  makeCurrent();
  GLenum err = GL_NO_ERROR;
  while((err = glGetError()) != GL_NO_ERROR)
  {
        printf("atom_program %s error %X\n", parent()->objectName().toStdString().c_str(),err);

  }
  //if (parent()->objectName()!="The Main Window") printf("paintGL -> %d atom_array_size %d\n",__LINE__,atom_array_size);
  atom_vao.bind();
  if ( !atom_program->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
  //if ( !bond_program->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
  float fiftypct =1.54;
  switch (mol->proba ) {
  case 10 :{ fiftypct = 0.76; break;}   //Hauptachsen der Eliipsoide 10% Wahrscheinlichkeit
  case 30 :{ fiftypct = 1.19; break;}   //Hauptachsen der Eliipsoide 30% Wahrscheinlichkeit
  case 50 :{ fiftypct = 1.54; break;}   //Hauptachsen der Eliipsoide 50% Wahrscheinlichkeit
  case 70 :{ fiftypct = 1.91; break;}   //Hauptachsen der Eliipsoide 70% Wahrscheinlichkeit
  case 90 :{ fiftypct = 2.50; break;}   //Hauptachsen der Eliipsoide 90% Wahrscheinlichkeit
  default: ;
  }
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
  glDisable(GL_CULL_FACE);
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  //a234 b237 c233 d230 e235 f236 g231 h232 i238 j225 k0 l224 m3 n222 o227 p1 q2 r229 s228 t226 u223 v240 w[] x239 y113 z4
  atom_program->setUniformValue("LOD", mol->LOD);//k
  atom_program->setUniformValue("QPeakLegend",!mol->monoQrom);//p
  atom_program->setUniformValue("Qolor", mol->mQolor);//q
  atom_program->setUniformValue("deuteriumColor", mol->deuteriumColor);
  atom_program->setUniformValue("adp", adp->isChecked());//m
  atom_program->setUniformValue("bede", hideBeLo->isChecked());//n
  atom_program->setUniformValue("bondradius", static_cast<float>(mol->bondStrength));//u
  atom_program->setUniformValue("eye", eye);//l
  atom_program->setUniformValue("fiftypct", fiftypct);//j
  atom_program->setUniformValue("hideHydrogen", !hideHydrogen->isVisible());//t
  atom_program->setUniformValue("hidePM1", shpm1->isChecked());//o
  atom_program->setUniformValue("highlightEquivalents", mol->highlightEquivalents);//s
  atom_program->setUniformValue("highlightParts", highlightParts->isChecked());//r
  atom_program->setUniformValue("lightPos", QVector3D(-100.0, 100.0, 800.0));//d
  atom_program->setUniformValue("lmax", static_cast<GLfloat>(mol->lmax));//g
  atom_program->setUniformValue("lmin", static_cast<GLfloat>(mol->lmin));//h
  atom_program->setUniformValue("normalMatrix", nm);//c
  atom_program->setUniformValue("origina", org);//a
  atom_program->setUniformValue("pmax", static_cast<GLfloat>(mol->pmax));//e
  atom_program->setUniformValue("pmin", static_cast<GLfloat>(mol->pmin));//f
  atom_program->setUniformValue("pmva", pmv);//b
  atom_program->setUniformValue("qcutoff", qcutoff);//i
  atom_program->setUniformValue("qpeakrad", static_cast<GLfloat>(mol->qPeakRad*2));//x
  atom_program->setUniformValue("tubes", tubes->isChecked());//v
  atom_program->setUniformValue("highlightResidues", highlightResidues->isChecked());

  atom_program->enableAttributeArray(0);
  atom_program->enableAttributeArray(1);
  atom_program->enableAttributeArray(2);
  atom_program->enableAttributeArray(3);


  //atom_program->setUniformValue("ballStick", ballStick->isChecked());//w
  /*

        printf("a%d b%d c%d d%d e%d f%d g%d h%d i%d j%d k%d l%d m%d n%d o%d p%d q%d r%d s%d t%d u%d v%d w[] x%d y%d z%d\n",

               atom_program->uniformLocation("origin"),              //a1
               atom_program->uniformLocation("pmv"),                 //b2
               atom_program->uniformLocation("normalMatrix"),        //c3
               atom_program->uniformLocation("lightPos"),            //d4
               atom_program->uniformLocation("pmax"),                //e5
               atom_program->uniformLocation("pmin"),                //f6
               atom_program->uniformLocation("lmax"),                //g7
               atom_program->uniformLocation("lmin"),                //h8
               atom_program->uniformLocation("qcutoff"),             //i9
               atom_program->uniformLocation("fiftypct"),            //j10
               atom_program->uniformLocation("LOD"),                 //k11
               atom_program->uniformLocation("eye"),                 //l12
               atom_program->uniformLocation("adp"),                 //m13
               atom_program->uniformLocation("bede"),                //n14
               atom_program->uniformLocation("hidePM1"),             //o15
               atom_program->uniformLocation("QPeakLegend"),         //p16
               atom_program->uniformLocation("Qolor"),               //q17
               atom_program->uniformLocation("highlightParts"),      //r18
               atom_program->uniformLocation("highlightEquivalents"),//s19
               atom_program->uniformLocation("hideHydrogen"),        //t20
               atom_program->uniformLocation("bondradius"),          //u21
               atom_program->uniformLocation("tubes"),               //v22
               //atom_program->uniformLocation("ballStick"),           //w23
               atom_program->uniformLocation("qpeakrad"),             //x24
               atom_program->uniformLocation("atomColors"),           //y
               atom_program->uniformLocation("arad")                  //z

               );// */

  glDrawArrays(GL_TRIANGLES, 0, atom_array_size);
  //else glDrawElements(GL_TRIANGLES, atom_array_size, GL_UNSIGNED_INT, nullptr);

  //printf("atomvao %d\n",atom_vao.objectId());
  atom_vao.release();
  atom_program->release();
  glFlush();
  err = GL_NO_ERROR;
  while((err = glGetError()) != GL_NO_ERROR)  {
        printf("atom_program_released %s error %X %d\n", parent()->objectName().toStdString().c_str(),err, __LINE__);

  }
}

void PlayOpenGL::paintGL(){
    GLenum err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
      printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
    }
    glClearColor(backGroundColor.redF(),backGroundColor.greenF(),backGroundColor.blueF(),backGroundColor.alphaF());
    frame_cnt++;
    QVector3D org = QVector3D(sumse.x, sumse.y, sumse.z);
    if ((rotze>-1)&&(rotze<mol->showatoms.size())) org = QVector3D(mol->showatoms.at(rotze).pos.x, mol->showatoms.at(rotze).pos.y, mol->showatoms.at(rotze).pos.z);
    if ((centerSelection)&&(centerSelection->isChecked())){
      if (mol->selectedatoms.isEmpty()) {
        org = QVector3D(sumse.x, sumse.y, sumse.z);
      }
      else {
        V3 sums=V3(0,0,0);
        for (int i = 0; i < mol->selectedatoms.size(); i++) sums += mol->selectedatoms[i].pos;
        sums *= 1.0 / mol->selectedatoms.size();
        org = QVector3D(sums.x, sums.y, sums.z);
      }
    }
    V3 org3=V3(org.x(), org.y(), org.z());
    altemitte=org3;
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    if ((bg_program)&&(bg_vao.isCreated())){
        if ( !bg_program->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
        bg_vao.bind();
        bg_program->setUniformValue("bgcolor", backGroundColor);
        bg_program->setUniformValue("bggradient", bggradient);
        glDisable(GL_DEPTH_TEST);
        glEnable(GL_BLEND);
        glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
        glDisable(GL_CULL_FACE);
        glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        glEnable(GL_DEPTH_TEST);
        bg_vao.release();
        bg_program->release();
    }
    if (bondsFirst){

        if ((drawBonds)&&(draft_bond_program)&&(draft_bond_vao.isCreated())){
            paintDraftBonds(org);
        }
        if ((drawAtoms)&&(atom_program)&&(atom_vao.isCreated())){
            paintAtoms(org);
        }
    }
    err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
        printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
    }
    if (rectangle) renderRect();
    if (((mol->splitLbonds.size()+mol->splitRbonds.size())>0)&&(splitrot_program)){
        paintSplitrotated(org);
    }
    if (!bondsFirst){
        if ((drawAtoms)&&(atom_program)&&(atom_vao.isCreated())){
            paintAtoms(org);
        }
        if ((drawBonds)&&(bond_program)&&(bond_vao.isCreated())){
            paintBonds(org);
        }
    }

    if ((drawLabels)&&(mol->showatoms.size()<=screenAtoms.size())){
        for (int i = 0; i < mol->showatoms.size(); i++){
            if (mol->showatoms.at(i).hidden) continue;
            if ((mol->showatoms.at(i).an>=0)&&((mol->AtomStyle[mol->showatoms.at(i).an]&ATOM_STYLE_NOLABEL)!=0)) continue;
            if ((mol->showatoms.at(i).an==0)&&(!hideHydrogen->isVisible())) continue;
            if ((mol->showatoms.at(i).an==-42)&&(!hideBeLo->isChecked())) continue;
            if ((mol->showatoms.at(i).an==-1)&&(mol->showatoms.at(i).peakHeight<qcutoff)) continue;
            if (i<altLabels.size()){
                renderText(screenAtoms.at(i),i);
            }
        }
    }
    if ((cont_program)&&(cont_vao.isCreated())&&(!cont.isEmpty())){
        paintContours(org);
    }
    if ((map_vao.isCreated())&&(map_program)&&((!fVertexes[0].isEmpty()||!fVertexes[1].isEmpty()))) {
        paintMaps(org);
    }
    err = GL_NO_ERROR;
        while((err = glGetError()) != GL_NO_ERROR){
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
        }
        if (((!mol->selectedatoms.isEmpty())||(!mol->duplicateAtoms.isEmpty())||(inFocus!=-1))&&(satom_program)&&(slct_vao.isCreated())){
          paintSelected(org);
    }
    err = GL_NO_ERROR;
        while((err = glGetError()) != GL_NO_ERROR){
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
        }

    if ((!enviPositions.isEmpty())&&(envi_program)&&(envi_vao.isCreated())){
          paintEnvi(org3, org);
    }
    if ((uc_program)&&(togUnit->isChecked())&&(uc_vao.isCreated())){
        paintCell(org3, org);
    }
    err = GL_NO_ERROR;
        while((err = glGetError()) != GL_NO_ERROR) {
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
        }
    if ((latom_program)&&(atomLegend->isChecked())&&(latom_vao.isCreated())){
          paintLegend();
    }
    if ((symm_program0)&&(symm_program1)&&(symm_program2)&&((axend>0)||(invend>0)||(planend>0)||(glend>0))){
          paintSymmElements(org);
    }
    if ((voro_vao.isCreated())&&(voro_program)){
          paintVoro(org);
    }    
    if ((!mol->selectedatoms.isEmpty())&&(mol->splitLbonds.isEmpty())){
        for (int i=0; i<mol->selectedatoms.size(); i++){
            if (mol->selectedatoms.at(i).hidden) continue;
            QColor c = selectedLabelColor ;
            QString label = mol->selectedatoms.at(i).Label;
            if (shortLabels->isChecked()) {
              label=label.section("_",0,0);
              label=label.section(QString::fromUtf8("»"),0,0);
            }
            renderText(mol->selectedatoms.at(i).pos-org3, label, c, selectedLabelBGColor.rgba());
        }
    }
    err = GL_NO_ERROR;
        while((err = glGetError()) != GL_NO_ERROR){
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
        }
    if ((inFocus!=-1)&&(inFocus<mol->showatoms.size())){
        QString label = mol->showatoms.at(inFocus).Label;
        if (shortLabels->isChecked()) {
          label=label.section("_",0,0);
          label=label.section(QString::fromUtf8("»"),0,0);
        }
        renderText(screenAtoms.at(inFocus), label, mol->showatoms.at(inFocus).an);
    }
    err = GL_NO_ERROR;
        while((err = glGetError()) != GL_NO_ERROR){
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
        }
    if (parent()->objectName()=="Edit Atom Parameters Dialog"){
        QString label = mol->showatoms.at(0).Label;
        if (shortLabels->isChecked()) {
          label=label.section("_",0,0);
          label=label.section(QString::fromUtf8("»"),0,0);
        }
        renderText(screenAtoms.at(0), label, mol->showatoms.at(0).an);
        label = mol->showatoms.at(1).Label;
        if (shortLabels->isChecked()) {
          label=label.section("_",0,0);
          label=label.section(QString::fromUtf8("»"),0,0);
        }
        renderText(screenAtoms.at(1), label, mol->showatoms.at(1).an);
    }
    err = GL_NO_ERROR;
        while((err = glGetError()) != GL_NO_ERROR){
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);
        }
        if (burnGPU) {
          update();
        }
        static int mksum = 0;
        int mksm = mk1 + mk2 + mk3 + mk4 + mk5 +
                   mk6 + mk7 + mk8 + mk9 + mkA +
                   mkB + mkC + mkD + mkE + mkF + mkG;
        //if (mksm != mksum) printf("%s: mk %d %d %d %d %d   %d %d %d %d %d   %d %d %d %d %d   %d\n", parent()->objectName().toStdString().c_str(), mk1,mk2,mk3,mk4,mk5, mk6,mk7,mk8,mk9,mkA, mkB,mkC,mkD,mkE,mkF, mkG);
        mksum = mksm;
}


void PlayOpenGL::getMVmat(GLdouble mv[16]){
    mv[0]  = static_cast<GLdouble>(mvmat(0,0));
    mv[1]  = static_cast<GLdouble>(mvmat(1,0));
    mv[2]  = static_cast<GLdouble>(mvmat(2,0));
    mv[3]  = static_cast<GLdouble>(mvmat(3,0));

    mv[4]  = static_cast<GLdouble>(mvmat(0,1));
    mv[5]  = static_cast<GLdouble>(mvmat(1,1));
    mv[6]  = static_cast<GLdouble>(mvmat(2,1));
    mv[7]  = static_cast<GLdouble>(mvmat(3,1));

    mv[8]  = static_cast<GLdouble>(mvmat(0,2));
    mv[9]  = static_cast<GLdouble>(mvmat(1,2));
    mv[10] = static_cast<GLdouble>(mvmat(2,2));
    mv[11] = static_cast<GLdouble>(mvmat(3,2));


    mv[12] = static_cast<GLdouble>(mvmat(0,3));
    mv[13] = static_cast<GLdouble>(mvmat(1,3));
    mv[14] = static_cast<GLdouble>(mvmat(2,3));
    mv[15] = static_cast<GLdouble>(mvmat(3,3));

}

void PlayOpenGL::changeBColor(){//! a color dialog for chosing a new back ground color
    QColor c= QColorDialog::getColor(backGroundColor);
    if(c.isValid()) backGroundColor=c;
    update();
}

void PlayOpenGL::changeTColor(){//! a color dialog for chosing a new back ground color
    QColor c= QColorDialog::getColor(backGroundColor);
    if(c.isValid()) labelColor=c;
    update();
}

double PlayOpenGL::pickRad(){
    return sqrt(pickradius);
}

void PlayOpenGL::setPickRad(double r){
    pickradius = r*r;
    showPickRad=true;
}

void PlayOpenGL::changePivot1(int i){
    pivot=i;
    splitRotate();
    emit splitable(canSplitRotate());
}

void PlayOpenGL::changePivot2(int i){
    pivot2=i;
    splitRotate();
    emit splitable(canSplitRotate());
}


void PlayOpenGL::invertSelection(){
    CEnvironment tempa=mol->selectedatoms;

    mol->selectedatoms.clear();

    for (int i=0; i<mol->showatoms.size();i++){
          if (!tempa.contains(mol->showatoms.at(i))) {
          mol->selectedatoms.append(mol->showatoms[i]);
          mol->selectedatoms.last().style=i;
          }
    }
    rehide();
    updateBondActions(__LINE__);
    update();
}

void PlayOpenGL::invertHidden(){
    if (!hiddenThings) return; // nothing is hidden => hide everything now? NO!
    hideHydrogen->setVisible(!hideHydrogen->isVisible());
    for (int i=0;i<mol->legendAtoms.size();i++) mol->legendAtoms[i].hidden=1;
    QHash<int,bool> anExists;
    for (int i = 0; i < mol->showatoms.size(); i++){
          mol->showatoms[i].hidden = (mol->showatoms.at(i).hidden!=0)?0:1;
          if (!mol->showatoms.at(i).hidden) anExists[mol->showatoms.at(i).an]=true;
          if (verbosity >2 ) printf("%-4s %d %d\n",mol->showatoms.at(i).Label.toStdString().c_str(),mol->showatoms.at(i).hidden,mol->showatoms.at(i).an);
    }
    for (int i=0;i<mol->legendAtoms.size();i++) if (anExists.contains(mol->legendAtoms.at(i).an)) mol->legendAtoms[i].hidden=0;
    //frid=-10;
    //hideReason=0;
    mol->selectedatoms.clear();
    //murx=__LINE__;
    makeAtoms(mol->showatoms);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

    updateBondActions(__LINE__);
    update();
}

void PlayOpenGL::hideHydrogens(){
    hiddenThings=true;

    hideHydrogen->setVisible(false);
    for (int i=0;i<mol->showatoms.size();i++) mol->showatoms[i].hidden=
              (mol->showatoms.at(i).an==0)?1:mol->showatoms.at(i).hidden;
    for (int i=0;i<mol->legendAtoms.size();i++) mol->legendAtoms[i].hidden=
              (mol->legendAtoms.at(i).an==0)?1:mol->legendAtoms.at(i).hidden;
    mol->selectedatoms.clear();
    makeAtoms(mol->showatoms);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

    updateBondActions(__LINE__);
    update();
}

void PlayOpenGL::hideNonSelected(){
    if (verbosity >2 ) printf("hide non selected\n");
    for (int i=0;i<mol->showatoms.size();i++) mol->showatoms[i].hidden=1;
    for(int i=0;i<mol->selectedatoms.size();i++){
          if (mol->selectedatoms.at(i).style < mol->showatoms.size())
          mol->showatoms[mol->selectedatoms.at(i).style].hidden=0;
          //   std::cout<<mol->showatoms.at(mol->selectedatoms.at(i).style).Label.toStdString()<<std::endl;
    }
    hiddenThings=true;
    //hideReason|=HIDE_REASON_SELECT;
    mol->selectedatoms.clear();
    makeAtoms(mol->showatoms);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);
    updateBondActions(__LINE__);
    update();
}

void PlayOpenGL::selectResiByNr(int nr){
    mol->selectedatoms.clear();
    apair.clear();
    for (int i=0; i<mol->showatoms.size();i++){
          if (mol->showatoms.at(i).resiNr==nr) {
          mol->selectedatoms.append(mol->showatoms[i]);
          mol->selectedatoms.last().style=i;
          }
    }
    //draw();
    if (hiddenThings) hideNonSelected();
    updateBondActions(__LINE__);
    update();
}


void PlayOpenGL::showMatrix(){
    if (verbosity >2 ) printf("showMatrix\n");
    printf("det %g %g %g\n",mvmat.determinant(),pmat.determinant(),pmv.determinant());
    printf("mvmat: \n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n"
           , mvmat(0,0)
           , mvmat(0,1)
           , mvmat(0,2)
           , mvmat(0,3)

           , mvmat(1,0)
           , mvmat(1,1)
           , mvmat(1,2)
           , mvmat(1,3)

           , mvmat(2,0)
           , mvmat(2,1)
           , mvmat(2,2)
           , mvmat(2,3)

           , mvmat(3,0)
           , mvmat(3,1)
           , mvmat(3,2)
           , mvmat(3,3)

           );

    printf("pmat: \n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n"
           , pmat(0,0)
           , pmat(0,1)
           , pmat(0,2)
           , pmat(0,3)

           , pmat(1,0)
           , pmat(1,1)
           , pmat(1,2)
           , pmat(1,3)

           , pmat(2,0)
           , pmat(2,1)
           , pmat(2,2)
           , pmat(2,3)

           , pmat(3,0)
           , pmat(3,1)
           , pmat(3,2)
           , pmat(3,3)

           );

    printf("pmv: \n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n"
           , pmv(0,0)
           , pmv(0,1)
           , pmv(0,2)
           , pmv(0,3)

           , pmv(1,0)
           , pmv(1,1)
           , pmv(1,2)
           , pmv(1,3)

           , pmv(2,0)
           , pmv(2,1)
           , pmv(2,2)
           , pmv(2,3)

           , pmv(3,0)
           , pmv(3,1)
           , pmv(3,2)
           , pmv(3,3)

           );
}

void PlayOpenGL::saveOrientation(){
    if (verbosity >2 ) printf("saveOrientation\n");

    QString fn=QFileDialog::getSaveFileName(nullptr, tr("Save Orientation into file "), "orientation.ini",
                                              "orientation.ini (*.ini);;");

    if (fn.isEmpty()) return;
    QSettings s(fn,QSettings::IniFormat,this);
    s.beginGroup("ModelViewMatrix");
    float *m = mvmat.data();
    s.beginWriteArray("mvmat",16);
    for (int i=0; i<16; i++){
          s.setArrayIndex(i);
          s.setValue("value",m[i]);
    }
    s.endArray();
    s.endGroup();

    s.beginGroup("ProjectionMatrix");
    m = pmat.data();
    s.beginWriteArray("pmat",16);
    for (int i=0; i<16; i++){
          s.setArrayIndex(i);
          s.setValue("value",m[i]);
    }
    s.endArray();
    s.endGroup();
    s.setValue("ViewAngle",viewAngle);
    s.setValue("PreviousViewAngle",previousViewAngle);
    s.deleteLater();
    /*
        QFile mconf(fn);
        mconf.open(QIODevice::WriteOnly|QIODevice::Text);
        GLdouble mm[16];
        makeCurrent();
        glGetDoublev(GL_MODELVIEW_MATRIX,mm);
        mconf.write(QString("%1/%2/%3/%4/%5/%6/%7/%8/%9/%10/%11/%12/%13/%14/%15/%16\n")
                        .arg(mm[0]).arg(mm[1]).arg(mm[2]).arg(mm[3]).arg( mm[4]).arg(mm[5])
                        .arg(mm[6]).arg(mm[7]).arg( mm[8]).arg(mm[9]).arg(mm[10]).arg(mm[11])
                        .arg( mm[12]).arg(mm[13]).arg(mm[14]).arg(mm[15]).toLatin1());
        mconf.close();
        */
}

void PlayOpenGL::loadOrientation(){
    if (verbosity >2 ) printf("loadOrientation\n");
    //showMatrix();
    QString fn=QFileDialog::getOpenFileName(nullptr, tr("Load Orientation from File "), "orientation.ini",
                                              "orientation.ini (*.ini);;");

    if (fn.isEmpty()) return;
    QSettings s(fn,QSettings::IniFormat,this);
    if (!s.contains("ViewAngle")) return;
    s.beginGroup("ModelViewMatrix");
    float *m = mvmat.data();
    int size = s.beginReadArray("mvmat");

    for (int i=0; i<size; ++i){
          s.setArrayIndex(i);

          m[i] = s.value("value").toFloat();

    }
    s.endArray();
    /*printf("mv %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f %10.6f\n"
               ,m[ 0],m[ 1],m[ 2],m[ 3]
               ,m[ 4],m[ 5],m[ 6],m[ 7]
               ,m[ 8],m[ 9],m[10],m[11]
               ,m[12],m[13],m[14],m[15]);
        */

    mvmat = QMatrix4x4(m).transposed();

    s.endGroup();

    s.beginGroup("ProjectionMatrix");
    m = pmat.data();
    s.beginReadArray("pmat");
    for (int i=0; i<16; i++){
          s.setArrayIndex(i);
          m[i] = s.value("value").toFloat();
    }
    s.endArray();

    pmat = QMatrix4x4(m).transposed();
    s.endGroup();
    viewAngle = s.value("ViewAngle").toDouble();
    previousViewAngle = s.value("PreviousViewAngle").toDouble();
    s.deleteLater();


    pmv = pmat*mvmat;

    nm = mvmat.normalMatrix();
    eye = mvmat.inverted()*QVector4D(0,0,0,1);
    /*QMatrix4x4 _pmv = pmat*mvmat;
        printf("pmv: \n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n%9.6f %9.6f %9.6f %9.6f\n"
               , _pmv(0,0)
               , _pmv(0,1)
               , _pmv(0,2)
               , _pmv(0,3)

               , _pmv(1,0)
               , _pmv(1,1)
               , _pmv(1,2)
               , _pmv(1,3)

               , _pmv(2,0)
               , _pmv(2,1)
               , _pmv(2,2)
               , _pmv(2,3)

               , _pmv(3,0)
               , _pmv(3,1)
               , _pmv(3,2)
               , _pmv(3,3)

               );*/
    updateScreenAtoms();
    update();
    if (viewAngle<1.0){
          isOrtho=true;
          toggleOrthoView->setChecked(true);
    }else{
          isOrtho=false;
          toggleOrthoView->setChecked(false);
    }
}

void PlayOpenGL::zoomOut(){
    if (verbosity >2 ) printf("zoomOut\n");
    double v = (viewAngle<1.0)?29.0:0.01;
    double av = viewAngle;
    setViewAngle(v);
    update();
    setViewAngle(av);
    update();
}

void PlayOpenGL::rotateIdle(){
    if (idl!=NULL){
          rotY();
          update();
    }
}

void PlayOpenGL::fertig(){
    update();
}


void PlayOpenGL::setQpeakCutoff(double cut){

    qcutoff = static_cast<float>(cut);
    update();
}



void PlayOpenGL::toggleAutoUpdate(int b){
    burnGPU = (b==Qt::Checked);
    if (burnGPU){
          if (!tmr){
          tmr = new QTimer();
          connect(tmr, &QTimer::timeout, this, &PlayOpenGL::reportFPS);
          }
          tmr->start(1000);

    }else{
          if (tmr) tmr->stop();
    }
    if (verbosity >2 ) printf("burn GPU! %d\n",burnGPU);
    update();
}

void PlayOpenGL::reportFPS(){
    if (!burnGPU) return;
    QString s = QString("FPS: %1").arg(frame_cnt / 1.0,0,'f',1);
    frame_cnt = 0;
    emit message(s);
}

void PlayOpenGL::wuff(){
    bool beloff=hideBeLo->isChecked();
    for (int i=0;i<mol->showatoms.size();i++) if (mol->showatoms.at(i).an==-42) mol->showatoms[i].hidden=(beloff)?0:1;
    for (int i=0;i<mol->legendAtoms.size();i++) if (mol->legendAtoms.at(i).an==-42) mol->legendAtoms[i].hidden=(beloff)?0:1;
    hiddenThings=hiddenThings||(!hideBeLo->isChecked());
    mol->selectedatoms.clear();
    updateBondActions(__LINE__);
    update();
}

void PlayOpenGL::showHidden(){
    int versteckte=0;
    int h=0;
    int belos=0;
    for (int i=0;i<mol->showatoms.size();i++){
          if (mol->showatoms[i].an==-42) belos++;
          if (mol->showatoms[i].hidden) versteckte++;
          if (mol->showatoms[i].an==0)h++;
    }
    hideBeLo->setChecked(true);//toggle state is OK this way
    if (h) hideHydrogen->setVisible(true);
    if (versteckte==0) return; // all are there nothing to do.
    if (verbosity >2 ) printf("show hidden\n");
    for (int i=0;i<mol->showatoms.size();i++){
          mol->showatoms[i].hidden=0;
    }
    for (int i=0;i<mol->legendAtoms.size();i++) mol->legendAtoms[i].hidden=0;
    hiddenThings=false;
    if (belos>0) wuff();
    makeAtoms(mol->showatoms);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

    makeLegendAtoms();
    updateBondActions(__LINE__);
    update();
}

void PlayOpenGL::rehide(){
    int versteckte=0;
    for (int i=0;i<mol->showatoms.size();i++){

          if (mol->showatoms[i].hidden) versteckte++;
          if ((mol->showatoms.at(i).an==-1))mol->showatoms[i].hidden = (mol->noQPeaksPlease)? 1 : 0;
          if (mol->showatoms[i].hidden) versteckte++;
    }
    if (versteckte==0) return;
    if (verbosity >2 ) printf("rehide %d %d\n", versteckte, AMD_Hack);
    makeAtoms(mol->showatoms);
    makeBonds(mol->showatoms, mol->showbonds, mol->lbonds, mol->theH_Bonds);

    updateBondActions(__LINE__);
    //printf("rehide end\n");
}


void PlayOpenGL::paintLegend(){
    if ( !latom_program->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
    latom_vao.bind();
    double fktr = retinafktr * legend_sc;
    glViewport(0, 0, 415 * fktr, qMax(950 * fktr, 1.0));
    QMatrix4x4 tm,tn;
    tm.setToIdentity();
    tn.setToIdentity();
    //tm.ortho(-3.075, 1.075, -7.1, 2.4 ,-10,10);
    tm.ortho(-3.075, 1.075, -7.1, 2.4 ,-10,10);
    glEnable(GL_DEPTH_TEST );
    glEnable(GL_BLEND);
    glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
    float fiftypct =1.54;
    switch (mol->proba ) {
    case 10 :{ fiftypct = 0.76; break;}   //Hauptachsen der Eliipsoide 10% Wahrscheinlichkeit
    case 30 :{ fiftypct = 1.19; break;}   //Hauptachsen der Eliipsoide 30% Wahrscheinlichkeit
    case 50 :{ fiftypct = 1.54; break;}   //Hauptachsen der Eliipsoide 50% Wahrscheinlichkeit
    case 70 :{ fiftypct = 1.91; break;}   //Hauptachsen der Eliipsoide 70% Wahrscheinlichkeit
    case 90 :{ fiftypct = 2.50; break;}   //Hauptachsen der Eliipsoide 90% Wahrscheinlichkeit
    default: ;
    }
    latom_program->setUniformValue("normalMatrix", tn.normalMatrix());
    latom_program->setUniformValue("pmva", tm);
    latom_program->setUniformValue("origina", QVector3D(0.000000, 0.000000, 0.00000));//QVector3D(-2.750000, -6.800000, 0.00000)
    latom_program->setUniformValue("pmax", static_cast<GLfloat>(mol->pmax));
    latom_program->setUniformValue("pmin", static_cast<GLfloat>(mol->pmin));
    latom_program->setUniformValue("lightPos", QVector3D(-100,100,800));
    latom_program->setUniformValue("lmax", static_cast<GLfloat>(mol->lmax));
    latom_program->setUniformValue("lmin", static_cast<GLfloat>(mol->lmin));
    latom_program->setUniformValue("qcutoff", qcutoff);
    latom_program->setUniformValue("fiftypct", fiftypct);
    latom_program->setUniformValue("LOD", mol->LOD);
    latom_program->setUniformValue("eye", QVector4D(0.000000, 31.475019, 7.868755, 1.000000));
    latom_program->setUniformValue("adp", adp->isChecked());
    latom_program->setUniformValue("bede", hideBeLo->isChecked());
    latom_program->setUniformValue("shpm1", shpm1->isChecked());
    latom_program->setUniformValue("QPeakLegend", !mol->monoQrom);
    latom_program->setUniformValue("Qolor", mol->mQolor);
    latom_program->setUniformValue("deuteriumColor", mol->deuteriumColor);
    latom_program->setUniformValue("highlightParts", highlightParts->isChecked());
    latom_program->setUniformValue("hideHydrogen", !hideHydrogen->isVisible());
    latom_program->setUniformValue("bondradius", static_cast<float>(mol->bondStrength));
    latom_program->setUniformValue("tubes", tubes->isChecked());
    latom_program->setUniformValue("ballStick", ballStick->isChecked());
    latom_program->setUniformValue("qpeakrad", static_cast<GLfloat>(mol->qPeakRad*2));
    latom_program->setUniformValue("highlightResidues", false);

    glDrawArrays(GL_TRIANGLES, 0,  latom_array_size);

    latom_program->release();
    latom_vao.release();
    //tm.scale(1.1);
    for (int i=0; i<mol->legendAtoms.size(); i++){
          if (mol->legendAtoms.at(i).hidden) continue;
          if ((mol->legendAtoms.at(i).an==0)&&!hideHydrogen->isVisible()) continue;
          renderText(QVector2D(
                         330*legend_sc + mol->legendAtoms.at(i).pos.x * 100 * legend_sc,
                         -15*legend_sc + height() - (mol->legendAtoms.at(i).pos.y + 6.7) * 104 * legend_sc),
                     mol->legendAtoms.at(i).Label, mol->legendAtoms.at(i).an, 24 * legend_sc);
    }
    glViewport(0, 0, width()*retinafktr, qMax(height()*retinafktr, 1.0));
}

void PlayOpenGL::paintMaps(const QVector3D org){
    double max;
    int Pers=0;
    bool neutrons=false;
    map_vao.bind();
    map_program->bind();
    glEnable(GL_DEPTH_TEST );
    glEnable(GL_BLEND);
    glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
    glDisable(GL_CULL_FACE);
    int err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
          printf("%s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);

    }
    glLineWidth(linwidth);
    if (niceTrans->isChecked()){
          QMatrix4x4 mm(mvmat);
          max= (fabs(mm(2,0))>fabs(mm(2,1)))?mm(2,0):mm(2,1);
          max=(fabs(max)<fabs(mm(2,2)))?mm(2,2):max;
          if ((max==mm(2,0))&&(max>0.0)) Pers=0; else
          if ((max==mm(2,0))&&(max<0.0)) Pers=1; else
              if ((max==mm(2,1))&&(max>0.0)) Pers=2; else
            if ((max==mm(2,1))&&(max<0.0)) Pers=3; else
            if ((max==mm(2,2))&&(max>0.0)) Pers=4; else
              if ((max==mm(2,2))&&(max<0.0)) Pers=5;
    }else Pers=0;
    if (fillMap->isChecked()) glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    else glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    map_program->setUniformValue("lightPos", QVector3D(100.0, 100.0, 100.0));
    map_program->setUniformValue("pmv", pmv);
    map_program->setUniformValue("origin", org);
    err = GL_NO_ERROR;
    while((err = glGetError()) != GL_NO_ERROR){
          printf("paintMaps %s paintGL error %X %d\n", parent()->objectName().toStdString().c_str(), err, __LINE__);

    }
    map_program->setUniformValue("normalMatrix",nm);
    map_program->setUniformValue("lintrans",static_cast<GLfloat>(lintrans));
    map_program->setUniformValue("lighting",lighting->isChecked());
    if (fofcact->isChecked()) {
          map_program->setUniformValue("Color", dipc);
          glDrawArrays(GL_TRIANGLES, offsets.at(0+4*Pers)/12, fVertexes[0+4*Pers].size()/3);
          map_program->setUniformValue("Color", dimc);
          glDrawArrays(GL_TRIANGLES, offsets.at(1+4*Pers)/12 , fVertexes[1+4*Pers].size()/3);
    }
    if (!mol->XLAB)if (foact->isChecked()) {
          map_program->setUniformValue("Color", fopc);
          glDrawArrays(GL_TRIANGLES, offsets.at(2+4*Pers)/12, fVertexes[2+4*Pers].size()/3);
          if (neutrons){
          map_program->setUniformValue("Color", fomc);
          glDrawArrays(GL_TRIANGLES, offsets.at(3+4*Pers)/12, fVertexes[3+4*Pers].size()/3);
          }
    }
    //glEnable(GL_LIGHTING);
    glDisable(GL_BLEND);
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    map_program->release();
    map_vao.release();
}

void PlayOpenGL::paintCell(const V3 org3, const QVector3D org){
    uc_vao.bind();
    if ( !uc_program->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
    glDisable(GL_DEPTH_TEST );
    glEnable(GL_BLEND);
    glLineWidth(1);
    glEnable(GL_LINE_SMOOTH);
    //glDisable(GL_LIGHTING);
    uc_program->setUniformValue("origin", org);
    uc_program->setUniformValue("pmv", pmv);
    uc_program->setUniformValue("cholesky", mol->cell.Cholesky);
    //glDrawArrays(GL_POINTS, 0,  1);
    glDrawArrays(GL_LINES, 0,  24);
    uc_program->release();
    uc_vao.release();
    if (drawLabels){
          V3 aa(1,0,0);
          V3 bb(0,1,0);
          V3 cc(0,0,1);
          QColor c=Qt::red;
          mol->frac2kart(aa,aa);
          mol->frac2kart(bb,bb);
          mol->frac2kart(cc,cc);
          renderText(aa-org3, "a", c, 0u);
          c=Qt::green;
          renderText(bb-org3, "b", c, 0u);
          c=Qt::blue;
          renderText(cc-org3, "c", c, 0u);
          c=Qt::black;
          renderText(V3(0,0,0)-org3, "0", c, 0u);
    }
}

void PlayOpenGL::paintSymmElements(const QVector3D org){

    glEnable(GL_BLEND);
    glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
    glEnable(GL_DEPTH_TEST );
    glDisable(GL_CULL_FACE);

    //glDisable(GL_DEPTH_TEST );
    glEnable(GL_BLEND);
    glLineWidth(1.0);
    glEnable(GL_LINE_SMOOTH);
    //glEnable(GL_LIGHTING);
    GLfloat scv[18*4];
    for (int i=0;i<18;i++) {
          scv[4*i]=mol->symmColors.at(i).redF();
          scv[4*i+1]=mol->symmColors.at(i).greenF();
          scv[4*i+2]=mol->symmColors.at(i).blueF();
          scv[4*i+3]=mol->symmColors.at(i).alphaF();
    }
    if ((symm_vao0.isCreated())&&axend){
          symm_vao0.bind();
          if ( !symm_program0->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
          symm_program0->setUniformValueArray("Color",scv,18,4);//we have to this for color changes too
          symm_program0->setUniformValue("pmv", pmv);
          symm_program0->setUniformValue("origin", org);
          symm_program0->setUniformValue("normalMatrix", nm);
          symm_program0->setUniformValueArray("dontshowSymElement",mol->dontshowSel.toVector().data(),mol->dontshowSel.count());
          symm_program0->setUniformValueArray("dontshowSymOperation",mol->dontshowSop.toVector().data(),mol->dontshowSop.count());
          symm_program0->setUniformValue("dssesize", static_cast<int>(mol->dontshowSel.count()));
          symm_program0->setUniformValue("dssosize", static_cast<int>(mol->dontshowSop.count()));
          symm_program0->setUniformValue("axisThickness",axisThickness);
          glDrawArrays( GL_LINES, 0,  axend/3);
          symm_program0->release();
          symm_vao0.release();
    }
    if ((symm_vao1.isCreated())&&invend){
          symm_vao1.bind();

          if ( !symm_program1->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
          symm_program1->setUniformValueArray("Color",scv,18,4);
          symm_program1->setUniformValue("pmv", pmv);
          symm_program1->setUniformValue("origin", org);
          symm_program1->setUniformValue("normalMatrix", nm);
          symm_program1->setUniformValueArray("dontshowSymElement",mol->dontshowSel.toVector().data(),mol->dontshowSel.count());
          symm_program1->setUniformValueArray("dontshowSymOperation",mol->dontshowSop.toVector().data(),mol->dontshowSop.count());
          symm_program1->setUniformValue("dssesize", static_cast<int>(mol->dontshowSel.count()));
          symm_program1->setUniformValue("dssosize", static_cast<int>(mol->dontshowSop.count()));;
          symm_program1->setUniformValue("InversionCenterRadius", InversionCenterRadius);
          glDrawArrays( GL_POINTS, 0,  invend/5);
          symm_program1->release();
          symm_vao1.release();
    }
    if ((symm_vao2.isCreated())&&glend){
          symm_vao2.bind();
          glDisable(GL_DEPTH_TEST );

          if ( !symm_program2->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
          symm_program2->setUniformValueArray("Color",scv,18,4);
          symm_program2->setUniformValue("pmv", pmv);
          symm_program2->setUniformValue("origin", org);
          symm_program2->setUniformValue("normalMatrix", nm);
          symm_program2->setUniformValueArray("dontshowSymElement",mol->dontshowSel.toVector().data(),mol->dontshowSel.count());
          symm_program2->setUniformValueArray("dontshowSymOperation",mol->dontshowSop.toVector().data(),mol->dontshowSop.count());
          symm_program2->setUniformValue("dssesize", static_cast<int>(mol->dontshowSel.count()));
          symm_program2->setUniformValue("dssosize", static_cast<int>(mol->dontshowSop.count()));;
          symm_program2->setUniformValue("planeTransparence", planeTransparence);
          symm_program2->setUniformValue("glideVectorThickness", glideVectorThickness);
          symm_program2->setUniformValue("planeFrameThinkness", planeFrameThinkness);

          glDrawArrays( GL_POINTS, 0,  glend/17);
          symm_program2->release();
          symm_vao2.release();
    }
}


void PlayOpenGL::paintContours(const QVector3D org){

    cont_vao.bind();
    cont_program->bind();
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glLineWidth(1.0);
    glEnable(GL_LINE_SMOOTH);
    //glDisable(GL_LIGHTING);
    cont_program->setUniformValue("pmv", pmv);
    cont_program->setUniformValue("origin", org);
    cont_program->setUniformValue("rainbow", rainbowPlot->isChecked());
    cont_program->setUniformValue("contMax", static_cast<float>(contMax));
    cont_program->setUniformValue("contMin", static_cast<float>(contMin));
    glDrawArrays(GL_LINES,0,cont.size());
    cont_program->release();
    cont_vao.release();

}

void PlayOpenGL::paintSplitrotated(const QVector3D org){
    if ( !splitrot_program->bind() ){ qWarning() << "2Could not bind shader program to context"; return;}
    splitrot_vao.bind();
    glEnable(GL_DEPTH_TEST );
    //glEnable(GL_BLEND);

    //glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
    splitrot_program->setUniformValue("pmv", pmv);
    splitrot_program->setUniformValue("origin", org);
    splitrot_program->setUniformValue("normalMatrix",nm);
    splitrot_program->setUniformValue("radius", static_cast<float>(mol->bondStrength*1.5f));
    splitrot_program->setUniformValue("dirx", QVector3D(screenRotx.x,screenRotx.y,screenRotx.z));
    splitrot_program->setUniformValue("diry", QVector3D(screenRoty.x,screenRoty.y,screenRoty.z));
    splitrot_program->setUniformValue("SRorg", QVector3D(SRorg.x,SRorg.y,SRorg.z));

    glDrawArrays(GL_LINES, 0, splitSize);
    splitrot_program->release();
    splitrot_vao.release();
}

void PlayOpenGL::paintSelected(const QVector3D org){

    if (!satom_program->bind() ){qWarning() << "Could not bind shader program to context"; return;}
    slct_vao.bind();
    glDisable(GL_DEPTH_TEST);
    //glDisable(GL_BLEND);
    glEnable(GL_BLEND);
    satom_program->setUniformValue("pmv", pmv);
    satom_program->setUniformValue("origin", org);
    satom_program->setUniformValue("fcoord", org);
    satom_program->setUniformValue("normalMatrix",nm);
    satom_program->setUniformValue("selrad",
                                   ((mol->splitLbonds.size()+mol->splitRbonds.size())>0)?0.2f:mol->selrad);
    if (!mol->selectedatoms.isEmpty()) {
          satom_program->setUniformValue("Color",mol->selectedAtomColor);
          satom_program->setUniformValue("inFocus", -1);
          glDrawArrays( GL_POINTS, 0,  mol->selectedatoms.size());
    }
    if (!mol->duplicateAtoms.isEmpty()) {
          satom_program->setUniformValue("Color", mol->dupplicatedAtomColor);
          satom_program->setUniformValue("inFocus", -2);
          glDrawArrays( GL_POINTS, mol->selectedatoms.size(),  mol->duplicateAtoms.size());
    }
    if ((inFocus!=-1)&&(inFocus<mol->showatoms.size())) {
          satom_program->setUniformValue("fcoord", QVector3D(mol->showatoms.at(inFocus).pos.x, mol->showatoms.at(inFocus).pos.y, mol->showatoms.at(inFocus).pos.z));
          satom_program->setUniformValue("origin", org);
          satom_program->setUniformValue("Color", mol->focusedAtomColor);
          satom_program->setUniformValue("inFocus", inFocus);
          glDrawArrays( GL_POINTS, 0,  1);
    }
    satom_program->release();
    slct_vao.release();
}

void PlayOpenGL::paintEnvi(const V3 org3, const QVector3D org){

    glDisable( GL_DEPTH_TEST );
    glEnable(GL_BLEND);
    glLineWidth(1.0);
    glEnable(GL_LINE_SMOOTH);
    //glDisable(GL_LIGHTING);
    //qDebug()<<mol->enviBondColor<<mol->enviBondColor<<mol->enviDefaultColor;
    envi_vao.bind();
    envi_program->bind();
    envi_program->setUniformValue("pmv", pmv);
    envi_program->setUniformValue("origin", org);
    envi_program->setUniformValue("enviBondColor", mol->enviBondColor);
    envi_program->setUniformValue("enviHBColor", mol->enviHBColor);
    envi_program->setUniformValue("enviDefaultColor", mol->enviDefaultColor);
    glDrawArrays(GL_LINES, 0, enviPositions.size()*2);
    //glEnable(GL_LIGHTING);
    glEnable( GL_DEPTH_TEST );
    envi_program->release();
    envi_vao.release();
    for (int j=0; j < enviPositions.size(); j++ ){
          QColor c = mol->enviDefaultColor;
          switch(enviKat.at(j)){
          case 1: c = mol->enviBondColor; break;
          case 2: c = mol->enviHBColor; break;
          default:
          c = mol->enviDefaultColor;
          }
          renderText(enviPositions.at(j)-org3, labs.at(j), c, 0xffd3d3d3);
          V3 p=(enviPositions.at(j)+enviP0)*0.5-org3;
          renderText(p, QString::number(sqrt(Distance(enviP0,enviPositions.at(j))),'f',3), c, 0);
    }

}

void PlayOpenGL::paintVoro(const QVector3D org){
    if ( !voro_program->bind() ){ qWarning() << "Could not bind shader program to context"; return;}
    voro_vao.bind();
    glEnable(GL_BLEND);
    glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;
    glEnable(GL_DEPTH_TEST );
    glDisable(GL_CULL_FACE);
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    voro_program->setUniformValue("pmv", pmv);
    voro_program->setUniformValue("origin", org);
    voro_program->setUniformValue("normalMatrix", nm);
    glDrawArrays( GL_POINTS, 0,  vtriangles.size());
    voro_program->release();
    voro_vao.release();

}

//////
//PlayOpenGL::
